/* PythonSerialMemo.c
   Andrew Dolgert ajd27@cornell.edu 30 August 2001

   This module defines a Python extension class which stores
   a pointer to a Python object (incrementing its reference
   count), and has only a Get and Set access methods.  It 
   replicates basic Python functionality and no more.  It was
   written because Python's pickle module allows special 
   serializations beyond the use of __getstate__ and
   __setstate__ only for extension types, not for Python classes
   or instances.

   This class was created to help convince pickle to store
   and retrieve SWIGged C instances from serialization.  The
   Python shadow class of the C instance returns a PythonSerialMemo
   from its __getstate__ method.  When pickle sees it has to
   save an extension type, it checks with the copy_reg module,
   where it finds it needs to call a function in the DMPickle
   module to save the PythonSerialMemo type.  That function 
   calls a DMWriter to save the C instance.
   */

#include <Python.h>


staticforward PyTypeObject PythonSerialMemo_PythonSerialMemoType;


typedef struct {
    PyObject_HEAD
    PyObject* Owner;
    void* CObject;
} PythonSerialMemo_PythonSerialMemoObject;


static PyObject*
PythonSerialMemo_new_PythonSerialMemo( PyObject* self, PyObject* args )
{
    PythonSerialMemo_PythonSerialMemoObject* pso;

    if ( !PyArg_ParseTuple( args, ":new_PythonSerialMemo") )
        return NULL;

    pso = PyObject_New( PythonSerialMemo_PythonSerialMemoObject,
                        &PythonSerialMemo_PythonSerialMemoType );
    (*pso).Owner = 0;

    return (PyObject*) pso;
}


/* This records a pointer to the PyObject given as an argument but
   does not increment the reference count.  It possibly should increment
   the reference count here and, of course, decrement it during 
   deallocation or if another pointer is set.
   */
static PyObject*
PythonSerialMemo_SetPythonInstance( PythonSerialMemo_PythonSerialMemoObject* self, PyObject* args )
{
    int result;

    PythonSerialMemo_PythonSerialMemoObject* pso =
        (PythonSerialMemo_PythonSerialMemoObject*) self;

    Py_XDECREF( (*pso).Owner );
    result = PyArg_ParseTuple( args, "O:SetPythonInstance", &((*pso).Owner) );
    if ( 0 == result ) return NULL;

    Py_INCREF((*pso).Owner);
    Py_INCREF(Py_None);
    return Py_None;
}


static PyObject*
PythonSerialMemo_GetPythonInstance( PythonSerialMemo_PythonSerialMemoObject* self, PyObject* args )
{
    int result;

    PythonSerialMemo_PythonSerialMemoObject* pso =
        (PythonSerialMemo_PythonSerialMemoObject*) self;

    result = PyArg_ParseTuple( args, ":GetPythonInstance" );

    return Py_BuildValue( "O", (*pso).Owner );
}


static void
PythonSerialMemo_PythonSerialMemo_dealloc( PythonSerialMemo_PythonSerialMemoObject* self )
{
    Py_XDECREF( (*self).Owner );
    PyObject_Del( self );
}



static PyMethodDef PythonSerialMemo_methods[] = {
    { "new_PythonSerialMemo", PythonSerialMemo_new_PythonSerialMemo, METH_VARARGS },
    { "SetPythonInstance", (PyCFunction) PythonSerialMemo_SetPythonInstance, METH_VARARGS },
    { "GetPythonInstance", (PyCFunction) PythonSerialMemo_GetPythonInstance, METH_VARARGS },
    { NULL, NULL }
};



/* Including this method makes our PythonSerialMemoType a real class with
   attributes, so you can say serialMemo.SetPythonInstance(obj) instead of
   PythonSerialMemo.SetPythonInstance( serialMemo, obj ).
   */
static PyObject*
PythonSerialMemo_getattr( PythonSerialMemo_PythonSerialMemoObject* self, char* name )
{
    return Py_FindMethod( PythonSerialMemo_methods, (PyObject*) self, name );
}


static PyTypeObject PythonSerialMemo_PythonSerialMemoType = {
    PyObject_HEAD_INIT( NULL )
    0,
    "PythonSerialMemo",
    sizeof(PythonSerialMemo_PythonSerialMemoObject),
    0,
    (destructor) PythonSerialMemo_PythonSerialMemo_dealloc,
    (printfunc)0, /* tp_print */
    (getattrfunc) PythonSerialMemo_getattr, /* tp_getattr */
	0,			  /*tp_setattr*/
	(cmpfunc)0,			  /*tp_compare*/
	(unaryfunc)0,			  /*tp_repr*/
	0,			  /*tp_as_number*/
	0,			  /*tp_as_sequence*/
	0,	  /*tp_as_mapping*/
};


static char
PythonSerialMemo_docstring[] = "This extension class allows DMReader and "
                             "DMWriter\nto serialize C objects during Python pickling.";


DL_EXPORT( void )
initPythonSerialMemo( void )
{
    PyObject* module;

    PythonSerialMemo_PythonSerialMemoType.ob_type =
        &PyType_Type;

    //__asm { int 3 }

    module = Py_InitModule3( "PythonSerialMemo", PythonSerialMemo_methods, PythonSerialMemo_docstring );

    if ( NULL != module ) {
        Py_INCREF( &PythonSerialMemo_PythonSerialMemoType );
        PyModule_AddObject( module, "PythonSerialMemoType",
            (PyObject*) &PythonSerialMemo_PythonSerialMemoType );
    }
}
