
##     Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
##     Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

##     This library is free software; you can redistribute it and/or modify it
##     under the terms of the GNU Lesser General Public License as published by the
##     Free Software Foundation; either version 2.1 of the License, or (at your
##     option) any later version.

##     This library is distributed in the hope that it will be useful, but WITHOUT
##     ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
##     FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
##     for more details.

##     You should have received a copy of the GNU Lesser General Public License
##     along with this library; if not, write to the Free Software Foundation,
##     Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
##     http://www.gnu.org/licenses/lgpl.html

##     Author Contact Information:
##     Jim Sethna      
##     sethna@lassp.cornell.edu  
##     117 Clark Hall
##     Cornell University
##     Ithaca, NY 14850

# Synchronizer.py

import sys
sys.path.extend(["/home/data/schiotz/DigitalMaterial/Python",
                 "/home/data/schiotz/DigitalMaterial/src/DMSystem/alpha/obj3D",
                 "/home/data/schiotz/DigitalMaterial/src/DMSystem/Parallel/alpha/obj3D"])
sys.path.append("T:/tcretegn/CVS/DigitalMaterial/src/DMSystem/")

from cPickle import Pickler, Unpickler
from NetBuffers import *
import Numeric
#from DMSystem_Parallel_ShadowSynchronizer import *
from DMSystem import *

"""Implements an MPI based synchronizer."""

class Synchronizer(ShadowSynchronizer):
    """Updates atoms near the CPU boundaries using info from the other CPUs.
    """

    def __init__(self, positions, domain, protocol):
        ShadowSynchronizer.__init__(self)
        self._positions = positions
        self._domain = domain
        self._protocol = protocol
        # Now register itself with the shadow synchronizer
        self.Register(self.SynchronizeFromCpp)

    def Synchronize(self, data):
        """Synchronize a DMArray by copying data from the owning CPUs."""

        nch = data.NumberOfBranches()
        assert nch == 2, "DMArray should have exactly two children."
        real = data.GetBranch(0) # First branch?
        gh = data.GetBranch(1)  # Second branch??

        # Prepare the send buffers
        sendbufs = []
        nprocs = protocol.GetNumberOfProcessors()
        procID = protocol.GetProcessorNumber()
        for target in xrange(nprocs):
            if target == procID:
                sendbufs.append(None)
            else:
                sendlist = self._SendToCopySynch[target]
                # sendlist is a list over leaves of lists of indices
                writebuffer = NetWriteBuffer()
                sendbufs.append(writebuffer)
                wbuf = Pickler(writebuffer, bin=1)
                nleaves = real.GetNumberOfLeaves()
                wbuf.dump(nleaves)  # For consistency checks
                assert nleaves == len(sendlist)
                for leaf in xrange(nleaves):
                    l = real.GetLeaf(leaf)[:]
                    indices = sendlist[leaf]
                    wbuf.dump("newleaf")
                    # d = l.Take(indices)  # Get a list with the data
                    d = Numeric.zeros((len(indices),3), Numeric.Float)
                    for i in xrange(len(indices)):
                        d[i] = l[indices[i]]
                    wbuf.dump(d)
    
        # Do the communication
        recvbufs = self._protocol.Butterfly(sendbufs)
    
        # Unpack and place the ghost atoms
        nleaves = gh.GetNumberOfLeaves()
        ghostlen = [0] * nleaves
        for rb in recvbufs:
            if rb is None:
                continue       # We did not receive anything from ourselves.
            rbuf = Unpickler(rb)
            assert nleaves == rbuf.load()
            for leaf in xrange(nleaves):
                assert rbuf.load() == "newleaf"
                l = gh.GetLeaf(leaf)
                d = rbuf.load()
                if len(l) == ghostlen[leaf]:
                    # We are adding ghosts
                    #l.AddAtomsAtEndWithData(len(d), d)
                    oldlen = len(l)
                    l.AddElementsAtEnd(len(d))
                    l[oldlen:] = d
                else:
                    # We are updating ghosts
                    l[ghostlen[leaf]:ghostlen[leaf]+len(d)] = d
                ghostlen[leaf] = ghostlen[leaf] + len(d)
    
            # Sanity check
            for leaf in xrange(nleaves):
                assert len(gh.GetLeaf(leaf)) == ghostlen[leaf], "Wrong number of ghost elements"
    

    def Rebuild(self, domain):
        """Build the send and receive lists."""
    # Sendlist for ordinary syncronize (no receive list necessary).
        realpos = self._positions.GetBranch(0)
        cpuID = self._protocol.GetProcessorNumber()
        self._SendToCopySynch = self._domain.BuildInterestLists(realpos, cpuID)

        # We p.t. do not need to build receivelists (by sending the
        # sendlists), but it will be needed later.
    
    def SynchronizeFromCpp(self):
        """Used to call the Synchronization method from C++."""
        return self.Synchronize(self.GetSyncArg())
    

###########################################################
#
# TEST SUITE
#
###########################################################

if __name__ == "__main__":
    import sys
    import time
    import os
    from DMSystem import *
    from Numeric import *
    from RandomArray import random, seed
    from NetProtocol import *
    from DomainSubdivision import *

    def printpos(proc, pos):
        def _pr(proc, tp, lf, dat):
            print "Processor %d, %s atoms, leaf %d:" % (proc, tp, lf)
            print dat
        _pr(proc, "real", 0, pos.GetBranch(0).GetBranch(0)[:])
        _pr(proc, "real", 1, pos.GetBranch(0).GetBranch(1)[:])
        _pr(proc, "ghost", 0, pos.GetBranch(1).GetBranch(0)[:])
        _pr(proc, "ghost", 1, pos.GetBranch(1).GetBranch(1)[:])

    # Seed the random number generator (different seed for each node!)
    seed(int(time.time()),os.getpid())

    cpu = (1,1,2)
    basis = identity(3) * array((10.0, 10.0, 20.0))
    domain = DomainSubdivision(basis, cpu, 2.0)
    protocol = NetProtocol()
    assert protocol.GetNumberOfProcessors() == 2
    
    # Make the positions
    procID = protocol.GetProcessorNumber()
    if procID == 0:
        a1 = 10.0 * random((10,3))
        a2 = 10.0 * random((5,3))
    else:
        x = array((0.0, 0.0, 10.0))
        a1 = 10.0 * random((10,3)) + x
        a2 = 10.0 * random((15,3)) + x
    rp1 = ArrayDoubleVector(len(a1))
    rp1[:] = a1
    rp2 = ArrayDoubleVector(len(a2))
    rp2[:] = a2
    realpos = ArrayDoubleVector()
    realpos.Attach(rp1)
    realpos.Attach(rp2)
    rp1.thisown = rp2.thisown = 0
    ghp1 = ArrayDoubleVector(1)
    ghp1.Clear()
    ghp2 = ArrayDoubleVector(1)
    ghp2.Clear()
    ghostpos = ArrayDoubleVector()
    ghostpos.Attach(ghp1)
    ghostpos.Attach(ghp2)
    ghp1.thisown = ghp2.thisown = 0
    positions = ArrayDoubleVector()
    positions.Attach(realpos)
    positions.Attach(ghostpos)
    realpos.thisown = ghostpos.thisown = 0
    # printpos(procID, positions)
    
    synchro = Synchronizer(positions, domain, protocol)
    synchro.Rebuild(domain)
    synchro.Synchronize(positions)
    printpos(procID, positions)

    

    
