
##     Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
##     Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

##     This library is free software; you can redistribute it and/or modify it
##     under the terms of the GNU Lesser General Public License as published by the
##     Free Software Foundation; either version 2.1 of the License, or (at your
##     option) any later version.

##     This library is distributed in the hope that it will be useful, but WITHOUT
##     ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
##     FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
##     for more details.

##     You should have received a copy of the GNU Lesser General Public License
##     along with this library; if not, write to the Free Software Foundation,
##     Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
##     http://www.gnu.org/licenses/lgpl.html

##     Author Contact Information:
##     Jim Sethna      
##     sethna@lassp.cornell.edu  
##     117 Clark Hall
##     Cornell University
##     Ithaca, NY 14850


# MD/__init__.py  --  Interface to the Molecular Dynamics code

# import the swigged module into this namespace

from MD3D import *

# Replace the ListOfAtoms classes with ones that reassigns ownership to C++
# when attaching lists etc.  And add various Get/Set methods
class _ListOfAtomsMixin:
    """Mixin class to finalize the definition of list of atoms classes

    It fixes memory management by redefining Attach and SetBoundaryConditions.

    Access methods are redefined to follow the standard. The added methods are:
        GetCartesianPositions
        GetCartesianPositionsDM   (As a DoubleMatrix)
        SetCartesianPositions
        SetCartesianPositionsDM

    The class to be finalized must be redefined with the original class as the
    LAST base class.
    """

    def __init__(self, *args, **kargs):
        self._baseclass = self.__class__.__bases__[-1]
        args = tuple([self] + list(args))
        apply(self._baseclass.__init__, args, kargs)

    def GetSubList(self, sublist):
        sl = self._baseclass.GetSubList(self, sublist)
        # Cast it to this derived class (ugly!!)
        sl.__class__ = self.__class__
        # This is normally done in __init__, but we skipped that one (!)
        sl._baseclass = sl.__class__.__bases__[-1]
        return sl
        
    def SetBoundaryConditions(self, bc):
        retval = self._baseclass.SetBoundaryConditions(self, bc)
        # Now C++ owns the attached object (see Attach above).
        bc.thisown = 0
        bc.attachedto = self
        return retval

    def GetCartesianPositions(self):
        """Get the positions as cartesian coordinates (A Numerical array)."""
        return self._baseclass.GetCartesianPositions(self)[:]

    def GetCartesianPositionsDM(self):
        """Get the positions as cartesian coordinates (A DoubleMatrix)."""
        return self._baseclass.GetCartesianPositions(self)

    def SetCartesianPositions(self, positions):
        """Get the positions as cartesian coordinates (A Numerical array)."""
        self._baseclass.GetCartesianPositions(self)[:] = positions

    def SetCartesianPositionsDM(self, positions):
        """Set the positions as cartesian coordinates (A DoubleMatrix)."""
        return self._baseclass.SetCartesianPositions(self, positions)

    def GetCartesianVelocities(self):
        """Get the velocities as cartesian coordinates (A Numerical array)."""
        return self._baseclass.GetCartesianVelocities(self)[:]

    def GetCartesianVelocitiesDM(self):
        """Get the velocities as cartesian coordinates (A DoubleMatrix)."""
        return self._baseclass.GetCartesianVelocities(self)

    def SetCartesianVelocities(self, velocities):
        """Get the velocities as cartesian coordinates (A Numerical array)."""
        self._baseclass.GetCartesianVelocities(self)[:] = velocities

    def SetCartesianVelocitiesDM(self, velocities):
        """Set the velocities as cartesian coordinates (A DoubleMatrix)."""
        return self._baseclass.SetCartesianVelocities(self, velocities)


_OriginalListOfAtoms = ListOfAtoms
del ListOfAtoms
class ListOfAtoms(_ListOfAtomsMixin, _OriginalListOfAtoms):
    pass

_OriginalPrimitiveListOfAtoms = PrimitiveListOfAtoms
del PrimitiveListOfAtoms
class PrimitiveListOfAtoms(_ListOfAtomsMixin, _OriginalPrimitiveListOfAtoms):
    pass


