import sys
sys.path.append('../../src/QuasiContinuum/SWIGBuild')
sys.path.append('../../src/QuasiContinuum/SWIGBuild/i686/obj3D')
sys.path.append('../MD/Observers')
from Numeric import *
from QC3D import *

# This is a very simple example where we simulate a defect-free crystal
# with periodic boundary conditions. Not very interesting, but validates
# the method by making a direct comparison with standard Molecular Dynamics.

def GenerateMesh(masters, magic, boxSize, latSpacing, filename):
    """ generate a mesh for the master atoms and saves it to file.
    Written as a function, so that if we run this script again
    we do not have to remesh. """
    i = 0
    # the position of the masters atoms.
    qcPos = masters.GetCartesianPositions()[:]
    # we will transform qcPos into an array of int by multiplying it by "magic"
    qcPosInt = zeros((len(masters),3),Int)
    for p in qcPos:
        for j in range(3): qcPosInt[i][j] = int(p[j]*magic)
        i = i + 1
    # import the Advancing front mesher.
    from Meshers import AdvancingFront3D
    # transform the box into ints.
    bigBoxInt = zeros(3,Int)
    for i in range(3): bigBoxInt[i] = magic * boxSize[i]
    # instanciate the mesher.
    mesher = AdvancingFront3D(None, qcPosInt, int(magic*5*latSpacing))
    mesher.SetBoxSize(bigBoxInt)
    # compute the mesh
    mesher.Advance()
    # get the tetrahedra from the mesher
    tetrahedra = mesher.GetTetrahedra()
    # instanciate a MeshData and create the elements.
    meshData = MeshData()
    for tetrahedron in tetrahedra:
        meshData.CreateElement(tetrahedron, qcAtoms, latMD)
    # initialize the slaves (ie decorate the elements and remove the slivers)
    meshData.InitializeSlaves(qcAtoms)
    print "Initialized Slaves"
    # save meshData to disk, so we do not have to recompute it.
    reader = NetCDFWriter(filename).Put(meshData,'simpleMesh')
    return meshData


# here is the MD potential
#mdPot = EMTPotential('Cu')
mdPot = HolianCutLennardJonesPotential()

latSpacing = mdPot.GetLengthScale() * sqrt(2)

# at this time, we must use a primitive lattice because the basis isn't taken into account
# we must have only one atom per unit cell.
latMD = PrimitiveFCCLattice(latSpacing)

# for QC we take a lattice that is 4 time larger.
latQC = PrimitiveFCCLattice(latSpacing * 4)


#here is the instanciation of the qcPotential. We cannot decorate EMTPotential
if(isinstance(mdPot, EMTPotential)):
    qcPot = QCEMTPotential("Cu")
else:
    qcPot = QCPotential(mdPot, latMD)  # note that we pass a reference lattice to QCPotential.

# the atoms we use for MD
mdAtoms     = DynamicListOfAtoms()
# the non local atoms for QC (ie those whose energy is calculated by regular MD)
qcNonLocals = DynamicListOfAtoms()
# the local atoms for QC (their energy comes is related to the energy of their surrounding elements) 
qcLocals    = DynamicListOfAtoms()
# the master atoms (they contain the degrees of freedom)
qcAtoms     = QCListOfAtoms(qcNonLocals, qcLocals)

# initialize the MD atoms as a big cube
bigBoxSize = array([12., 12, 12]) * latSpacing
RectangularClusterInitializer(bigBoxSize, latMD).Create(mdAtoms)
#initialize the local atoms in this big cube...
RectangularClusterInitializer(bigBoxSize, latQC).Create(qcLocals)
# ... and create a small cube of nonlocal atoms in the center.
smallBoxSize = array([3.1, 3.1, 3.1]) * latSpacing
RectangularClusterInitializer(smallBoxSize, latMD).Create(qcNonLocals)
# we must avoid having local atoms overlapping nonlocals. With an OverlapPruner
# we can remove all the atoms that are closer than some distance to
# any given set of points (given by the second argument of the OverlapPruner).  
OverlapPruner(0.2*latSpacing, qcNonLocals.GetCartesianPositions()).Transform(qcLocals)


# now we can set neighbor locators on our lists of atoms.
cutoff = mdPot.GetCutoffDistance()
# for MD ...
mdNL = CellNeighborList(cutoff, 0.05*cutoff)
mdAtoms.SetNeighborLocator(mdNL)
# ... and for QC
qcNL = CellNeighborList(cutoff, 0.05*cutoff)
qcAtoms.SetNeighborLocator(qcNL)
# note that this internally make sure that slaves will be treated correctly

# now let's set the boundary conditions. We use SkewPeriodicBoundaryConditions
# so we can shear the crystal without creating any surface.
superCell = 1.0 * identity(3)
for i in range(3): superCell[i][i] = bigBoxSize[i]  # a rectangular super cell, for now
bc = SkewPeriodicBoundaryConditions()
bc.SetSuperCellVectors(superCell)
# set the boundary conditions on the lists of atoms
mdAtoms.SetBoundaryConditions(bc)
qcAtoms.SetBoundaryConditions(bc)


# we can now mesh or read the mesh
try:
    meshData = NetCDFReader('simpleMesh.cdf').Get('simpleMesh')
    meshData.InitializeSlaves(qcAtoms)
    print "Initialized Slaves"
except:
    meshData = GenerateMesh(qcAtoms, 17, bigBoxSize, latSpacing, 'simpleMesh.cdf')

# let's see where the masters are. 
from RasMolObserver import RasMolObserver3D
def color(masters, i):
    if i < len(qcNonLocals): return 0
    else: return 1
    
rasmol = RasMolObserver3D(colorMethod = color)
qcAtoms.AddObserver(rasmol)
qcAtoms.Notify()


qcPot.SetMeshData(meshData)
print "Potential sent mesh data"

# we will shear and compress the lattice and compare the result.
transfo = array([[1.  , 0.01, 0.0 ],
                 [0.01, 1.  , 0.0 ],
                 [0.  , 0.0 , 1.005 ]])

format = "step of deformation: %i \t %14.7f\t%14.7f"
print "Compare the energies given by QC (", len(qcAtoms), \
      ") atoms, and MD (", len(mdAtoms), " atoms)."
print format % (0, qcPot.CalculateEnergy(qcAtoms), mdPot.CalculateEnergy(mdAtoms))
for i in range(5):
    superCell = matrixmultiply(superCell, transfo)
    bc.SetSuperCellVectors(superCell)
    LinearPositionTransformer(transfo).Transform(mdAtoms)
    LinearPositionTransformer(transfo).Transform(qcAtoms)
    qcAtoms.Notify()
    print format % (i, qcPot.CalculateEnergy(qcAtoms), mdPot.CalculateEnergy(mdAtoms))

#we have checked the consistence of the QC energy. Let's check the
# forces make sense: we randomly shift the master atoms and use
# a QuickMinAtomsMover to see if we recover our previous result.

from random import uniform
for i in range(len(qcAtoms)):
    for j in range(3):
        qcAtoms.IncrementCartesianPosition(i,j,uniform(-0.1,0.1))

#qcAtoms.IncrementCartesianPositions(uniform(-0.1, 0.1, (len(qcAtoms, 3))))

def relax(atoms, potential, tolerance = 1e-2, step = 0.01):
    e = potential.CalculateEnergy(atoms)
    mover = QuickMinAtomsMover(potential, step/5., tolerance)
    mover.Verbose()
    mover.SetMaxIterations(5)
    mover.Move(atoms)
    mover.SetDt(step)
    mover.SetMaxIterations(100)
    while mover.GetConvergence() > tolerance:
        atoms.Notify()
	mover.Move(atoms)
	de = potential.CalculateEnergy(atoms) - e
	e = e + de
	print mover.GetConvergence(), e, de
    return potential.CalculateEnergy(atoms)

relax(qcAtoms, qcPot)
