// DMException.cpp implementation of the DMException class
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"
#include <iostream>
#include "DMException.h"
#include "DMLastHeader.h"

/*!	Convenient way to make an informative exception.

	\param sClass The class throwing the exception.
	\param sFunction The member function which throws it.
	\param sMessage What the exception is.
	*/
DMException::DMException( const std::string& sClass, const std::string& sFunction,
						 const std::string& sMessage ) :
sClass(sClass), sFunction(sFunction), sMessage(sMessage) {}

/*!	Derived classes may prefer this interface, where they can set the
	message themselves.

	\param sClass The class throwing the exception.
	\param sFunction The member function which throws it.
	*/
DMException::DMException( const std::string& sClass, const std::string& sFunction ) :
sClass(sClass), sFunction(sFunction) {}

/*! We have this just so the destructor can be virtual. */
DMException::~DMException() {}

/*! This returns a copy of the string so we don't have to hold
  on to storage inside.
*/
std::string DMException::GetString( ) const
{
	std::string msg(sClass);
	msg.append("::");
	msg.append(sFunction);
	msg.append(":  ");
	msg.append(sMessage);
	return msg;
}

/*! There is no way to make an operator<< virtual, so you cannot override it
	in base classes.  Instead, we make a single ostream<< for all classes
	derived from DMException, but all it does is call our virtual put function.
	That way, subclasses can choose what to print on error.
	*/
std::ostream& DMException::Put(std::ostream& s) const
{
	return s << "Exception in " << this->GetString() << std::endl;
}

//! This operator permits us to write a DMException to a stream.
/*! In order to override the printing behavior of DMException in
	a subclass, don't try to redefine this operator because you can't.
	Instead, supply a new virtual method put(std::ostream& s) in the
	subclass.

	\param s Is any kind of ostream.
	\param r Is anything derived from DMException.
	\return Returns the stream.  Standard signature of operator<<.
	*/
std::ostream& operator<<(std::ostream& s, const DMException& r)
{
	return r.Put(s);
}
