// DigitalMaterial.h
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

/*!
  \mainpage DMSystem

  These pages document the library DMSystem which is used by
  several projects of the Digital Material including
    - <a href="../../../MD/doc/html/index.html">Molecular Dynamics</a>
    - <a href="../../../QuasiContinuum/doc/html/index.html">Quasi-Continuum Dynamics</a>
    - <a href="../../../PhaseField/doc/html/index.html">Phase Fields</a>
    - <a href="../../../KMC/doc/html/index.html">Kinetic Monte-Carlo</a>

  The DMSystem library provides classes of general utility to the
  whole project.  They include
    - Algorithms of broad applicability
      - ConjugateGradientMinimizer
      - Vectorized matrix routines we call SchiotzVectors
    - Architectural classes
      - A common header file - DigitalMaterial.h
      - Serialization interfaces to NetCDF and ASCII formats
      - ProcessFactory, which generalizes architecture-dependent threading.

  <H2>Requirements</H2>
  This library compiles with
	- Linux gcc
	- Windows Microsoft VC++ or Intel compilers
	- Alpha native or gcc

  There is optional support for serialization with NetCDF if you 
  include the NetCDF library.

  */
/* Defines the dimension of space: change it once, everything recompiles */
#ifndef DIGITALMATERIAL_H
#define DIGITALMATERIAL_H

#ifndef DIMENSION
#define DIMENSION 3
#endif

#ifdef DMPARALLEL
#if DMPARALLEL == 0
#undef DMPARALLEL
#endif // DMPARALLEL == 0
#endif

#ifdef M_PI
#define PI M_PI
#else
#define PI 3.1415926535897932
#endif // M_PI

///////////////////////////////////////////////////////////////////
// System capabilities
///////////////////////////////////////////////////////////////////

#ifdef _MSC_VER
#define DMWINDOWS

#define vsnprintf _vsnprintf
#define HAVE_VSNPRINTF

#endif // _MSC_VER

#ifdef linux

#define HAVE_VSNPRINTF

#endif


///////////////////////////////////////////////////////////////////
// Assertions and Trace macros
///////////////////////////////////////////////////////////////////

#ifdef _MSC_VER
#ifdef __ICL  // Intel compiler (_MSC_VER is defined by the Intel compiler)
#pragma warning(disable: 985) // turns off spurious Windows STL warnings
#else
#pragma warning(disable:4786) // turns off spurious Windows STL warnings
#endif // __ICL
#ifndef _DM_FIRST_HEADER_H_
#include "DMFirstHeader.h"
#endif

#ifdef HAVE_BUGSLAYER
#include "BugslayerUtil.h"
#else // Not HAVE_BUGSLAYER

#ifdef _DEBUG
#include <windows.h>

#define ASSERTMACRO(a,x)	_ASSERTE(x)
#define ASSERT(x)			_ASSERTE(x)
#define VERIFY(x)			((void)(x))
#define SUPERASSERT(x)		_ASSERTE(x)
#define SETDIAGASSERTOPTIONS(x)
#define ADDDIAGASSERTMODULE(x)

#ifdef __cplusplus
//inline void TraceOutput(LPCTSTR, ...) { }
#define TRACE				::printf
#endif

#define TRACE0(fmt)					::printf(fmt)
#define TRACE1(fmt,arg1)			::printf(fmt,arg1)
#define TRACE2(fmt,arg1,arg2)		::printf(fmt,arg1,arg2)
#define TRACE3(fmt,arg1,arg2,arg3)	::printf(fmt,arg1,arg2,arg3)
#endif // HAVE_BUGSLAYER
#endif //_DEBUG

#endif // _MSC_VER

#if defined(linux) || defined(__ALPHA)
#ifndef NDEBUG
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#define ASSERTMACRO(a,x)	assert(x)
#define ASSERT(x)			assert(x)
#define VERIFY(x)			((void)(x))
#define SUPERASSERT(x)		assert(x)
#define SETDIAGASSERTOPTIONS(x)
#define ADDDIAGASSERTMODULE(x)

#ifdef __cplusplus
//inline void TraceOutput(LPCTSTR, ...) { }
#define TRACE				::printf
#endif

#define TRACE0(fmt)					::printf(fmt)
#define TRACE1(fmt,arg1)			::printf(fmt,arg1)
#define TRACE2(fmt,arg1,arg2)		::printf(fmt,arg1,arg2)
#define TRACE3(fmt,arg1,arg2,arg3)	::printf(fmt,arg1,arg2)
#endif //NDEBUG
#endif //linux || __ALPHA

// This is applicable for any platform.  Just clear the macros.
#ifdef NDEBUG
/* //////////////////////////////////////////////////////////////////////
                       _DEBUG Is !!NOT!! Defined
////////////////////////////////////////////////////////////////////// */

#define ASSERTMACRO(a,x)
#define ASSERT(x)
#define VERIFY(x)   ((void)(x))
#define SUPERASSERT(x)
#define SETDIAGASSERTOPTIONS(x)
#define ADDDIAGASSERTMODULE(x)

#ifdef __cplusplus
//inline void TraceOutput(LPCTSTR, ...) { }
#define TRACE(x)
#endif

#define TRACE0(fmt)
#define TRACE1(fmt,arg1)
#define TRACE2(fmt,arg1,arg2)
#define TRACE3(fmt,arg1,arg2,arg3)

#endif  // NDEBUG

#endif // ndef DIGITALMATERIAL_H

