// Buffer.cpp: implementation of the Buffer class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DigitalMaterial.h"
#ifdef DMPARALLEL

#include "MPIRecvBuffer.h"

#include "SerialFactory.h"
#include "DMLastHeader.h"
#include "DMException.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////



MPIRecvBuffer::MPIRecvBuffer(const MPI_Comm comm): DMReader(""), comm(comm)
{
    this->buffer	  = 0;
    this->bufferSize  = 0;
    this->currentPos  = 0;
    // lets say 10000 is enough to start with. It will be extended if necessary
    Allocate(10000);
}

MPIRecvBuffer::~MPIRecvBuffer()
{
    if (buffer) delete [] buffer;
}

void MPIRecvBuffer::Allocate(int newSize)
{
    if(this->buffer) delete [] this->buffer;
    this->buffer	 = new char[newSize];
    this->bufferSize = newSize;
    this->currentPos = 0;
}

std::string MPIRecvBuffer::GetType(const string &name)
{
    throw DMException("MPIRecvBuffer", "GetType", "Not Implemented yet");
}

Serializable *MPIRecvBuffer::Get(const string &name)
{
    Serializable *serializable = SerialFactory::Instance()->New(GetString(""));
    serializable->Load(this);
    return serializable;
}



bool MPIRecvBuffer::Fill(Serializable *toFill, const string &name)
{
    throw DMException("MPIRecvBuffer", "Fill", "Not Implemented yet");
}

bool MPIRecvBuffer::GetBool( const string& name )
{
    return ( GetInt(name) == 0 ? 0: 1);
}

int MPIRecvBuffer::GetInt(const string &name)
{
    int myInt;
    MPI_Unpack(buffer, bufferSize, &currentPos, &myInt, 1, MPI_INT, comm);
    return myInt;
}

double MPIRecvBuffer::GetDouble(const string &name)
{
    double myDble;
    MPI_Unpack(buffer, bufferSize, &currentPos, &myDble, 1, MPI_DOUBLE, comm);
    return myDble;
}

string MPIRecvBuffer::GetString(const string &name)
{
    int size = GetInt("");
    char *c = new char[size];
    MPI_Unpack(buffer, bufferSize, &currentPos, c, size, MPI_CHAR, comm);
    string str(c, size);
    delete [] c;
    return str;
}

int *MPIRecvBuffer::GetIntArray(vector<int> &dims, const string &name)
{
    dims.clear();
    int size = GetInt("");
    int *dim = new int[size];
    MPI_Unpack(buffer, bufferSize, &currentPos, dim, size, MPI_INT, comm);
    dims.insert(dims.begin(), dim, dim+size); //  == dims.assign(dim,dim+size);
    int tot = 1;
    for(int i=0; i<size; ++i) tot *= dim[i];
    delete [] dim;
    int *data = new int[tot];
    MPI_Unpack(buffer, bufferSize, &currentPos, data, tot, MPI_INT, comm);
    return data;	
}

double *MPIRecvBuffer::GetDoubleArray(vector<int> &dims, const string &name)
{
    dims.clear();
    int size = GetInt("");
    int *dim = new int[size];
    MPI_Unpack(buffer, bufferSize, &currentPos, dim, size, MPI_INT, comm);
    dims.insert(dims.begin(), dim, dim+size); //  == dims.assign(dim,dim+size);
    int tot = 1;
    for(int i=0; i<size; ++i) tot *= dim[i];
    delete [] dim;
    double *data = new double[tot];
    MPI_Unpack(buffer, bufferSize, &currentPos, data, tot, MPI_DOUBLE, comm);
    return data;	
}


#endif //DMPARALLEL