// RWLock.h interface for the RWLock class
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef _RWLOCK_H_
#define _RWLOCK_H_

#include "DigitalMaterial.h"
#include "RWLockImp.h"

//! Read-write lock
/**
 * This is the class you work with when you make a read-write lock.
 * A read-write lock models a resource which can have multiple
 * readers but only one writer at a time.
 * You create it with a call to ProcessFactory::GetRWLock().
 * \code
 * RWLock *lock = ProcessFactory::Instance()->GetRWLock();
 * \endcode
 *
 * This class is just a wrapper on a system-dependent implementation
 * of a read-write lock.  On construction, it chooses the correct kind
 * for your system.  Then it forwards all calls to that correct
 * implementation.
 * \author Drew Dolgert
 * \date March 2000
 * \sa RWLockImp, WinRWLock, UnixRWLock
 * \ingroup dmsystem
 */
class RWLock : public RWLockImp
{
public:
	//! Initialize the lock.
	/*! You must call InitializeLock before calling any function
		except CleanupLock(). */
	virtual bool InitializeLock();
	/*! If you have called InitializeLock(), you must call CleanupLock()
		before the lock is deleted. */
	virtual void CleanupLock() { _lock->CleanupLock(); delete _lock; }
	//! Obtain a lock to read from the resource.
	/*! There can be multiple clients with read locks at the same time. */
	virtual void ReadLock() { _lock->ReadLock(); }
	//! Obtain a lock to write to the resource.
	/*! Only one client at a time may have write permission. */
	virtual void WriteLock() { _lock->WriteLock(); }
	//! Release the read lock.
	virtual void ReadRelease() { _lock->ReadRelease(); }
	//! Release the write lock.
	virtual void WriteRelease() { _lock->WriteRelease(); }
private:
	RWLockImp* _lock;
};

#endif // _RWLOCK_H_
