/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
#include "DMFirstHeader.h"
#include <algorithm>
#include <cmath>
#include "DigitalMaterial.h"
#include "DynSer.h"
#include "Rand.h"
#include "SerialBuilder.h"
#include "DMWriter.h"
#include "DMReader.h"
#include "DMLastHeader.h"

// ******************** Serialization object ***********

const std::string DynSer::sType = std::string("DynSer");
const SerialBuilder<DynSer> *_creator = new
SerialBuilder<DynSer>( DynSer::sType );

const std::string DynSerA::sType = std::string("DynSerA");
const SerialBuilder<DynSerA> *_acreator = new
SerialBuilder<DynSerA>( DynSerA::sType );

const std::string DynSerB::sType = std::string("DynSerB");
const SerialBuilder<DynSerB> *_bcreator = new
SerialBuilder<DynSerB>( DynSerB::sType );


DynSer::DynSer( int seed )
{
	rand.SetSeed( seed );

	int iBool = rand.discrete( 0, 1 );
	bData = ( iBool == 1 );

	iData = rand.discrete( 0, 1 << 31 );
	dData = rand.uniform();
	sData = "How you doin?";

	int intDimCnt = rand.discrete( 1, 3 );
	int intDimIdx;
	size_t total = 1;
	for ( intDimIdx = 0; intDimIdx < intDimCnt; intDimIdx++ ) {
		intShape.push_back( rand.discrete( 0, 4 ) );
		total *= intShape[ intDimIdx ];
	}

	if ( total > 0 ) {
		piData = new int[ total ];
		int idx;
		for ( idx = 0; idx < total; idx++ ) {
			piData[ idx ] = rand.discrete( 0, 1 << 31 );
		}
	} else {
		piData = 0;
	}

	int dDimCnt = rand.discrete( 1, 3 );
	int dDimIdx;
	size_t dTotal = 1;
	for ( dDimIdx = 0; dDimIdx < dDimCnt; dDimIdx++ ) {
		doubleShape.push_back( rand.discrete( 0, 4 ) );
		dTotal *= doubleShape[ dDimIdx ];
	}

	if ( dTotal > 0 ) {
		pdData = new double[ dTotal ];
		int idx;
		for ( idx = 0; idx < dTotal; idx++ ) {
			pdData[ idx ] = rand.uniform();
		}
	} else {
		pdData = 0;
	}
}


DynSer::~DynSer()
{
	DeleteArrays();
}


bool DynSer::IsEqual( const DynSer& toCompare )
{
	//if ( this->bData != toCompare.bData ) return false;
	if ( this->iData != toCompare.iData ) return false;
	if ( abs(this->dData-toCompare.dData) > 1e-10 ) return false;
	if ( this->sData.compare( toCompare.sData ) != 0 ) return false;

	if ( this->intShape.size() != toCompare.intShape.size() ) return false;
	if ( !std::equal( this->intShape.begin(), this->intShape.end(),
		toCompare.intShape.begin() ) ) return false;
	if ( intShape.size() > 0 ) {
		int iDimIdx;
		size_t iTotal = 1;
		for ( iDimIdx = 0; iDimIdx < intShape.size(); iDimIdx++ ) {
			iTotal *= intShape[ iDimIdx ];
		}
		if ( !std::equal( this->piData, this->piData+iTotal, toCompare.piData ) )
			return false;
	}

	if ( this->doubleShape.size() != toCompare.doubleShape.size() ) return false;
	if ( !std::equal( this->doubleShape.begin(), this->doubleShape.end(),
		toCompare.doubleShape.begin() ) ) return false;
	if ( doubleShape.size() > 0 ) {
		int dDimIdx;
		size_t dTotal = 1;
		for ( dDimIdx = 0; dDimIdx < doubleShape.size(); dDimIdx++ ) {
			dTotal *= doubleShape[ dDimIdx ];
		}
		int mIdx;
		for ( mIdx = 0; mIdx < dTotal; mIdx++ ) {
			if ( abs(this->pdData[mIdx] - toCompare.pdData[mIdx]) > 1e-10 )
				return false;
		}
	}

	return true;
}


void DynSer::Save( DMWriter* writer ) const
{
	writer->PutInt( iData, "iData" );
	writer->PutDouble( dData, "dData" );
	writer->PutString( sData, "sData" );
	writer->PutIntArray( piData, intShape, "piData" );
	writer->PutDoubleArray( pdData, doubleShape, "pdData" );
}

void DynSer::Load( DMReader* reader )
{
	DeleteArrays();

	iData = reader->GetInt( "iData" );
	dData = reader->GetDouble( "dData" );
	sData = reader->GetString( "sData" );
	piData = reader->GetIntArray( intShape, "piData" );
	pdData = reader->GetDoubleArray( doubleShape, "pdData" );
}

void DynSer::DeleteArrays()
{
	if ( pdData ) delete[] pdData;
	if ( piData ) delete[] piData;
	pdData = 0;
	piData = 0;

	intShape.resize(1);
	intShape[0] = 0;
	doubleShape.resize(1);
	doubleShape[0] = 0;
}


DynSerA::DynSerA( int seed ) : DynSer( seed )
{
	if ( rand.discrete( 0, 1 ) == 1 ) {
		this->pDynA = new DynSerA();
	} else {
		this->pDynA = 0;
	}
}


DynSerA::~DynSerA()
{
	if ( pDynA ) delete pDynA;
}

bool DynSerA::IsEqual( const DynSerA& comp )
{
	if ( DynSer::IsEqual( comp ) == false ) return false;

	if ( (pDynA == 0) != (comp.pDynA == 0) ) return false;
	if ( pDynA != 0 ) {
		if ( pDynA->IsEqual( *comp.pDynA ) == false ) return false;
	}
	if ( dynB.IsEqual( comp.dynB ) == false ) return false;

	return true;
}


void DynSerA::Save( DMWriter* writer ) const
{
	DynSer::Save( writer );

	if ( pDynA ) {
		writer->PutBool( true, "haveA" );
		writer->Put( pDynA, "pDynA" );
	} else {
		writer->PutBool( false, "haveA" );
	}
	writer->Put( &dynB, "pDynB" );
}


void DynSerA::Load( DMReader* reader )
{
	if ( pDynA ) delete pDynA;
	pDynA = 0;

	DynSer::Load( reader );

	bool haveA = reader->GetBool( "haveA" );
	if ( haveA ) {
		Serializable* ser = reader->Get( "pDynA" );
		pDynA = dynamic_cast<DynSerA*>( ser );
	} else {
		pDynA = 0;
	}

	reader->Fill( &dynB, "pDynB" );
}
