// SubjectObserver.h: interface for the Subject and Observer classes
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef OBSERVER_H
#define OBSERVER_H

#include <list>
#include "DigitalMaterial.h"

class Subject;
 //! Observers follow receive messages from generic Subjects.
/** 
  * \sa Subject
  * \ingroup dmsystem
  */
class Observer
{
public:
	virtual ~Observer();
	virtual void Update(Subject *theChangedSubject) = 0;

protected:
	Observer();
	
};


//! A Subject sends message to attached Observers.
/** Adapted from the Subject-Observer section of 
 * "Design Patterns, Elements of Reusable Object-Oriented Software"
 * by E. Gamma, R. Helm, R. Johnson, and J. Vlissides, 
 * Addison Wesley, Massachusetts, 1995
 *ISBN 0-201-63361-2
 *
 * The book suggests a type-specific subject-observer pattern which we have
 * abstracted here at the cost of using RTTI and dynamic casts in the 
 * concrete Subjects and Observers.
 * \sa Observer
 * \ingroup dmsystem
 */
class Subject  
{
public:
	virtual ~Subject(); 
	
	virtual void  AttachObserver(Observer* o) 
	{
		if (o)
			_observers->push_back(o);
	}
	virtual void Detach(Observer* o) 
	{
		if (o)
			_observers->remove(o);
	}
	virtual void Notify() 
	{
		std::list<Observer*>::iterator i = _observers->begin();
		while (i != _observers->end()) 
		{
			(*i)->Update(this);
			i++;
		}
	}

protected:
	Subject();
private:
#ifndef SWIG // XXX SWIG doesn't support standard template libraries
	std::list<Observer*> *_observers;
#endif // SWIG
};

#endif // !defined(OBSERVER_H)
