// Threadable.h interface for the Threadable class
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef _THREADABLE_
#define _THREADABLE_

#ifndef SWIG
#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000
#endif
#include "Runnable.h"

class Thread; // need know no more here.

//! A base class for objects which can be run in a thread.
/**
 * \ingroup dmsystem
 *
 * Inherit from this class in order to create a class which can be run in
 * its own thread.  Your class must call IsRunning() every so often in order
 * to ask the owning Thread class whether it has been requested to stop.  Check
 * out this example:
 *
 * \code
 * class MyCalc : public Threadable {
 *     int i;
 *     void Run() {
 *         while (this->IsRunning()) {
 *             i = i + 1; // Do calculation here
 *             // Then notify display you have done a step.
 *         }
 *     }
 * };
 *
 * int main(void) {
 *     MyCalc sim();
 *     Thread lightning(&sim);
 *     lightning.Start();
 *     DoUserInterface();
 * }
 * \endcode
 * 
 * Because we use explicitly cooperative threads, our techniques are a bit different
 * than Java.  This class isn't just an interface.  It provides a point for
 * a thread to break for others.
 * \author ajd27
 * \sa Runnable, Thread
 */

class Threadable : public Runnable
{
public:
	Threadable();
	bool IsRunning();
	//! Tell the threadable who owns it.
	/*! This method is called only by the Thread class.  You should
		not call it directly. */
	void SetThread(Thread *const thread);
	virtual ~Threadable() {}
private:
	Thread * pThread;
};

#endif // _THREADABLE_
