// WinThread.cpp implementation of the WinThread class
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

/* WinThread.cpp - 21 November 1999
 * This is a thread class which owns a Threadable object.
 * It takes care of it.
 * ajd27
 */

#include "DMFirstHeader.h"
#include <stdlib.h>
#include <process.h>
#include "DigitalMaterial.h"
#include "WinThread.h"
#include "DMLastHeader.h"

 /* Notes on implementation:
 * When the thread hasn't been made, hThread=0.  When it is closed, the handle is
 * closed and we set hThread=0, so if we stopped it, hThread=0.
 *
 * When there is no client, the state of the thread is
 * bPleaseListen=false
 * hNotifyMama = unsignaled (automatic reset event)
 * hPauseEvent = signaled if running, unsignaled if paused
 * Each method is responsible for returning the handles to the above state before
 * they complete.
 */
WinThread::WinThread()
{
	this->runnable = 0;
	this->threadable = 0;
	this->hThread = 0;
	this->bPleaseListen = false;
	this->bPleaseStop = false;
	this->bPaused = false;
	this->hPauseEvent = 0;
	::InitializeCriticalSection(&this->csThreadLock);
	this->hClassLock = ::CreateMutex(0,false,0);
	this->hNotifyMama = 0;
}

void WinThread::SetTask(Runnable *thread)
{
	::WaitForSingleObject(hClassLock,INFINITE);
	this->runnable = thread;
	this->threadable = 0;
	::ReleaseMutex(hClassLock);
}

void WinThread::SetTask(Threadable *thread)
{
	::WaitForSingleObject(hClassLock,INFINITE);
	this->runnable = thread;
	this->threadable = thread;
	::ReleaseMutex(hClassLock);
}

WinThread::~WinThread()
{
	this->Stop();
	if (hPauseEvent) {
		::CloseHandle(hPauseEvent);
	}
	if (hThread) ::CloseHandle(hThread);
	if (hNotifyMama) {
		::CloseHandle(hNotifyMama);
	}
	::DeleteCriticalSection(&csThreadLock);
	::CloseHandle(hClassLock);
}

// This is the actual thread routine.
unsigned __stdcall WinThread::Loop(void * pvoid)
{
	WinThread::PPARAMS params = (WinThread::PPARAMS) pvoid;
	::SetThreadPriority(::GetCurrentThread(),THREAD_PRIORITY_BELOW_NORMAL);
	Runnable *runner = params->thread;
	if (runner) runner->Run();

	return 0; // What does the return value mean?
}

// Create the thread only when you want it to run.  This won't start
// paused.
int WinThread::Start()
{
	::WaitForSingleObject(hClassLock,INFINITE);
	if (this->ThreadExists()) {
		::ReleaseMutex(hClassLock);
		return THREAD_EXISTS;
	}
	// Create events with no security attributes, no manual reset, initial
	// state signaled (all waiting may run) and no name.
	if (hPauseEvent==0) {
		hPauseEvent = ::CreateEvent(NULL,true,true,NULL);
	} else {
		// make sure it is signaled
		::SetEvent(hPauseEvent);
	}

	// Create another event, this one for the thread to notify us.
	if (hNotifyMama==0) {
		hNotifyMama = ::CreateEvent(NULL,true,false,NULL);
	} else {
		::ResetEvent(hNotifyMama);
	}
	// initialize the params structure.
	params.thread = runnable;
	// We prefer _beginthreadex to CreatThreadEx because it initializes
	// the stdlib for threads, and we could likely use <math.h> somewhere.
	// You may not begin a new thread inside a mutex or you will have
	// communication problems.  Who knew?

#ifndef _MT
#error Project must be compiled with multithreading in order to use Windows threads.
#endif

	hThread = (HANDLE) ::_beginthreadex(NULL,0,WinThread::Loop,
			&params,0,&dwThreadID);
	bPaused = false;
	if (hThread==0) {
		ReleaseMutex(hClassLock);
		return NO_THREAD;
	}
	// Although this may seem counterintuitive, we set the worker thread priority below
	// normal.  That means that it is a background thread.  On a machine with a user,
	// this gives good user response.  On a machine without a user, this avoids the
	// sleep(0) which always gives up clock cycles.
	// We could try THREAD_PRIORITY_LOWEST if it should be lower.
	// THREAD_PRIORITY_BELOW_NORMAL is a little higher.
	::ReleaseMutex(hClassLock);
	return SUCCESS;
}

int WinThread::Stop()
{
	::WaitForSingleObject(hClassLock,INFINITE);
	if (!this->ThreadExists()) {
		::ReleaseMutex(hClassLock);
		return SUCCESS;
	}
	int ret = SUCCESS;
	if (threadable) {
		if (bPaused) {
			int pret = this->Resume();
			if (ret==NO_THREAD) {
				::ReleaseMutex(hClassLock);
				return NO_THREAD;
			}
		}
		bPleaseStop = true; // This must be reset by the time we leave.
		DWORD dw = ::WaitForSingleObject(hThread,15000); // milliseconds
		if (dw==WAIT_TIMEOUT) { // If the thread never stopped.
			TerminateThread(hThread,1);
			ret = THREAD_TERMINATED;
		}
		::CloseHandle(hThread);
	} else {
		// This thread wasn't meant to be stopped.
		// This is a bad thing to do.  It leaves resources around.
		TerminateThread(hThread,1);
		::CloseHandle(hThread); // release the resource so it can die.
		ret = THREAD_TERMINATED;
	}

	bPleaseStop = false;
	hThread = 0;
	::ReleaseMutex(hClassLock);
	return ret;
}

// When this method exits, the only change should be that the thread
// is waiting on an unsignaled hPauseEvent.  All booleans and other
// handles should be returned to original conditions.
int WinThread::Pause()
{
	::WaitForSingleObject(hClassLock,INFINITE);
	if ((threadable==0) || (!this->ThreadExists())) {
		::ReleaseMutex(hClassLock);
		return NO_THREAD;
	}
	// Then check if it is already paused.
	// The only way the event would be set is if the thread were paused.
//	DWORD dw = ::WaitForSingleObject(hPauseEvent,0);
//	if (dw==WAIT_TIMEOUT) { // then the event is unsignaled.
	if (bPaused) {
		::ReleaseMutex(hClassLock);
		return ALREADY_PAUSED;
	}

	ResetEvent(hNotifyMama);
	ResetEvent(hPauseEvent);
	bPleaseListen = true;
	// Then wait on the event.
	// In case the thread is killed for some reason, we go to the trouble
	// of waiting for both the event and the thread itself.
	// If the thread dies, then it becomes signaled.
	HANDLE lpHandles[2] = { hNotifyMama, hThread };
	// false means we want to wait for either object to signal, not all.
//	DWORD dw = ::WaitForSingleObject(hNotifyMama,10000);
	// now forget about our pause thread.
	::ResetEvent(hNotifyMama); // reset our event
	DWORD dw = ::WaitForMultipleObjects(2, lpHandles,false,10000);
	bPleaseListen=false; // Reset it now so we don't forget.
	if (dw==WAIT_OBJECT_0+1) { // Thread don't mind. Thread's dead.
		::CloseHandle(hThread);
		hThread = 0;
		::ReleaseMutex(hClassLock);
		bPaused = false;
		return NO_THREAD;
	// It may be that the thread never paused.  Be vindictive and kill it.
	} else if ((dw==WAIT_TIMEOUT) || (dw==WAIT_FAILED)) {
		this->Stop();
		::ReleaseMutex(hClassLock);
		bPaused = false;
		return NO_THREAD;
	} // else WAIT_OBJECT_0 is success on the first object.

	bPaused = true;
	::ReleaseMutex(hClassLock);
	return SUCCESS;
}

int WinThread::Resume()
{
	::WaitForSingleObject(hClassLock,INFINITE);
	if ((threadable==0) || (!this->ThreadExists())) {
		::ReleaseMutex(hClassLock);
		return NO_THREAD;
	}
	// Then check if it is already paused.
	// The only way the event would be set is if the thread were paused.
//	DWORD dw = ::WaitForSingleObject(hPauseEvent,0);
//	if (dw!=WAIT_TIMEOUT) { // then the event is running
	if (!bPaused) {
		::ReleaseMutex(hClassLock);
		return SUCCESS;
	}

	bPleaseListen = false;
	// Tell the thread to run free.  Then wait to make sure it did.
	ResetEvent(hNotifyMama);
	SetEvent(hPauseEvent);
	HANDLE lpHandles[2] = { hNotifyMama, hThread };
	// false means we want to wait for either object to signal, not all.
	DWORD dw = ::WaitForMultipleObjects(2, lpHandles,false,10000);
	bPaused = false;
	if (dw==WAIT_OBJECT_0+1) { // Thread don't mind. Thread's dead.
		CloseHandle(hThread);
		hThread = 0;
		::ReleaseMutex(hClassLock);
		return NO_THREAD;
	// It may be that the thread never paused.  Be vindictive and kill it.
	} else if ((dw==WAIT_TIMEOUT) || (dw==WAIT_FAILED)) {
		this->Stop();
		::ReleaseMutex(hClassLock);
		return NO_THREAD;
	} // else WAIT_OBJECT_0 is success on the first object.
	::ReleaseMutex(hClassLock);
	return SUCCESS;
}

// Toggle needs to be atomic.  You can't be outside this class and ask
// if Paused() Resume() else Pause();  The whole sequence should be done
// within the threadsafe csClassLock critical section. (mutex, really).
int WinThread::Toggle()
{
	::WaitForSingleObject(hClassLock,INFINITE);
	if ((threadable==0) || (!this->ThreadExists())) {
		::ReleaseMutex(hClassLock);
		return NO_THREAD;
	}

//	DWORD dw = ::WaitForSingleObject(hPauseEvent,0);
//	if (dw==WAIT_TIMEOUT) { // then the event is unsignaled.
	int ret;
	if (bPaused) {
		ret = this->Resume();
	} else {
		ret = this->Pause();
	}
	::ReleaseMutex(hClassLock);
	return ret;
}

// Interface to threadable
// Runnable will never call this.
bool WinThread::IsRunning()
{
	// This is an interface to the thread, not a client.
	// If there is no controller, then you might as well continue running.
	if (bPleaseStop) return false;
	// We broker the events with a boolean.  It may sometimes mean the
	// threaded object travels through here an extra loop (because it takes
	// us two steps to ask it to stop: a) listen b) stop at my event).
	// There are great speed advantages to avoiding this event stuff in an inner
	// loop.  Petzold said each of these calls can take 500 lines of code.
	if (bPleaseListen) {
		// One command pauses because they asked us to and tells them
		// that we paused.  This only works in NT and 2000, not 95.
#if _WIN32_WINNT >= 0x0400 // includes Win2k
		DWORD dw = ::SignalObjectAndWait(hNotifyMama,hPauseEvent,INFINITE,false);
#else // Win95, I suppose

		if (!::SetEvent(hNotifyMama)) {
			int iError = ::GetLastError();
		}
		DWORD dw = ::WaitForSingleObject(hPauseEvent,INFINITE);

#endif
		// By now someone has allowed us to run again.  Tell them we're fine
		// in a note of thanks.
		::SetEvent(hNotifyMama);
	} //else {
		//::Sleep(0);
	//}
	return true;
}

// PRIVATE BELOW THIS LINE
bool WinThread::ThreadExists()
{
	// A) Did it die because I asked it to?
	// We wouldn't want to wait for the handle in this case, because we
	// already closed it.
	if (hThread==0) return false;
	// Might it be dead because it gave up on life?
	// Here, the handle should still be open.
	DWORD dw = ::WaitForSingleObject(hThread,0);
	// If the thread is signaled, it's done.
	if (dw==WAIT_OBJECT_0) {
		// Take car of that handle.
		::CloseHandle(hThread);
		hThread = 0;
		return false;
	// If unsignaled, it's running like a champ.
	} else if (dw==WAIT_TIMEOUT) {
		return true;
	}
	return false;
}
