#include "DMFirstHeader.h"
#include "DiscreteAvalanche.h"
#include "HysteresisCmdParsers.h"
#include "HysteresisSimulation.h"
#include "AvalancheAlgorithm.h"
#include "DMException.h"
#include "NetCDFReader.h"
#include "SprintString.h"
#include "CmdStreamReader.h"
#include "DMLastHeader.h"

using namespace std;

EmptyCmdParser::EmptyCmdParser( const std::string& cmdString )
: CmdParser( cmdString ) {}

bool EmptyCmdParser::AcceptCmdLine( std::istream& s )
{
	return this->Accept();
}

IntCmdParser::IntCmdParser( const std::string& cmdString )
: CmdParser( cmdString ) {}

bool IntCmdParser::AcceptCmdLine( std::istream& s )
{
	int tryArg, iArg;

	if (s >> tryArg) {
		iArg = tryArg;
		return this->AcceptInt( iArg );

	} else {
		cout << "You need to specify an integer for this command." << endl;
	}
	return false;
}


ToggleCmdParser::ToggleCmdParser( const std::string& cmdString )
: CmdParser( cmdString ) {}

bool ToggleCmdParser::AcceptCmdLine( std::istream& s )
{
	std::string PlusMinus;
	s >> PlusMinus;
	if ( !s ) return false;

	if ( PlusMinus[0] == '+' ) {
		return this->TurnOn();

	} else if ( PlusMinus[0] == '-' ) {
		return this->TurnOff();

	} else {
		cout << "You need to use + or - to turn this on/off." << endl;
	}
	return false;
}


SetupCmdParser::SetupCmdParser( HysteresisSimulation* pSim )
	: EmptyCmdParser("setup"), pSim(pSim)
{
}

bool SetupCmdParser::Accept( )
{
	try {
		pSim->WriteState( std::cout );
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}

void SetupCmdParser::ShortHelp( std::ostream& o ) const
{
	o << "Asks simulation to print its current state.";
}

void SetupCmdParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << std::endl << "\t";
	ShortHelp( o );
}


HelpCmdParser::HelpCmdParser( CmdStreamReader* cmdReader)
: CmdParser("help"), cmdReader( cmdReader )
{
	ASSERT( 0 != cmdReader );
	if ( 0 == cmdReader ) throw DMException("HelpCmdParser",
		"HelpCmdParser","Need to get a command reader.");
}

bool HelpCmdParser::AcceptCmdLine( std::istream& s )
{
	string inquireCommand;

	bool bFoundCommand = false;

	while (s >> inquireCommand) {
		bFoundCommand = true;
		CmdParser* cmdParser = cmdReader->ParserWhichMatches( inquireCommand );
		if ( 0 == cmdParser ) {
			cout << "Could not find help for \"" << inquireCommand << "\"." <<
				endl;
		} else {
			cmdParser->LongHelp( cout );
		}
	}		

	if ( !bFoundCommand ) {
		cout << "HysteresisConsole by Matt Kuntz, Andrew Dolgert" << endl;
		cout << "and Jim Sethna. Copyright 2001" << endl;
		cout << "Issue one command per line.  Each command is identified" << endl;
		cout << "by its unique case-insensitive first characters." << endl << endl;

		CmdStreamReader::ParserVector::const_iterator toPrint;
		const CmdStreamReader::ParserVector& parsers = cmdReader->GetCmdParsers();
		for ( toPrint = parsers.begin(); toPrint < parsers.end(); toPrint++ ) {
			cout << (*toPrint)->GetToken() << ": ";
			(*toPrint)->ShortHelp( cout );
			cout << endl;
		}
	}

	cout << endl;
	return false;
}

void HelpCmdParser::ShortHelp( std::ostream& o ) const
{
	o << "Prints help for any command.";
}

void HelpCmdParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << endl << "\tPrints list of commands." << endl;
	o << "Usage: " << cmdString << " <command>" << endl <<
		"\tPrints help for that command." << endl;
}


DimCmdParser::DimCmdParser(HysteresisSimulation* pSim)
	:  IntCmdParser("dimension"), dim(2), pSim(pSim)
{
}


bool DimCmdParser::AcceptInt( int dimension )
{
	try {
		pSim->SetDimension( dimension );
		std::cout << "Dimension set to " << dimension << std::endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}

void DimCmdParser::ShortHelp( std::ostream& o ) const
{
	o << "Sets number of dimensions.";
}

void DimCmdParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << " <int>" << endl;
	o << "\tSet the number of dimensions for the simulation." << endl;
}


LengthCmdParser::LengthCmdParser( HysteresisSimulation* pSim )
	: CmdParser("lengths"), pSim(pSim)
{
}

/*! You can either specify just one length for all dimensions or the
	length of each side with: "dimension 1000" or "dimension 500 1000 1000".
	*/
bool LengthCmdParser::AcceptCmdLine( std::istream& s )
{
	const int MAX_DIMENSIONS = 40;

	SpinCoord dimension[MAX_DIMENSIONS];

	int dimIdx = 0;
	while (s.good()) {
		s >> dimension[dimIdx];
		dimIdx++;
	}
	int dimCnt = dimIdx;

	try {
		if ( 1 == dimCnt ) {
			pSim->SetSideLength( dimension[0] );
			cout << "Lengths set to " << dimension[0] << "." << endl;
		} else {
			for ( dimIdx=0; dimIdx<dimCnt; dimIdx++) {
				pSim->SetEachSideLength( dimIdx, dimension[dimIdx] );
			}

			cout << "Lengths set to ( ";
			for ( dimIdx=0; dimIdx<dimCnt-1; dimIdx++) {
				cout << dimension[dimIdx] << ", ";
			}
			cout << dimension[dimCnt-1] << " )" << endl;
		}
	} catch (DMException e) {
		std::cout << e;
	}

	return false;
}


void LengthCmdParser::ShortHelp( std::ostream& o ) const
{
	o << "Set the side length of the lattice.";
}

void LengthCmdParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << " <int>" << endl;
	o << "\tSet all sides to the same length." << endl;
	o << "Usage: " << cmdString << " <int> <int> <int> ..." << endl;
	o << "\tSet each side length individually." << endl;
}

RestartCmdParser::RestartCmdParser(HysteresisSimulation* pSim)
: CmdParser("restart"), pSim(pSim) {}


bool RestartCmdParser::AcceptCmdLine( std::istream& s )
{
	std::string sFileName;

	s >> sFileName;
	if (!s) {
		sFileName="";
		sFileName.append("start.dat");
	}

	try {
		NetCDFReader lector(sFileName);
		lector.Fill( pSim, "checkpoint" );
		cout << "Restarted from " << sFileName << "." << endl;
	} catch (DMException& e) {
		std::cout << e;
	}

	return false;
}

void RestartCmdParser::ShortHelp( std::ostream& o ) const
{
	o << "Restart the program from a checkpoint file.";
}

void RestartCmdParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << " <filename>" << endl;
	o << "\tRestart the program from a checkpoint file.  The program will" << endl;
	o << "\tstart in the same state except for batch timers, tracing, and" << endl;
	o << "\tcheckpointing.  Dimensions, observers and such will be the same." << endl;
}


QuitCmdParser::QuitCmdParser() : EmptyCmdParser("quit")
{
}

bool QuitCmdParser::Accept( )
{
	return true;
}


void QuitCmdParser::ShortHelp( std::ostream& o ) const
{
	o << "Quit the program.";
}

void QuitCmdParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << endl;
	o << "\t";
	ShortHelp( o );
	o << endl;
}


SeedCmdParser::SeedCmdParser(HysteresisSimulation* pSim)
	: IntCmdParser("seed"), randSeed(1), pSim(pSim)
{
}


bool SeedCmdParser::AcceptInt( int seed )
{
	try {
		pSim->SetSeed( seed );
		cout << "Seed set to " << seed << "." << endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}


void SeedCmdParser::ShortHelp( std::ostream& o ) const
{
	o << "Set the seed for the random number generator.";
}

void SeedCmdParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << " <int>" << endl;
	o << "\t";
	ShortHelp(o);
	o << endl;
}


EndTimeCmdParser::EndTimeCmdParser(HysteresisSimulation* pSim)
: IntCmdParser("endtime"), pSim(pSim) {}

bool EndTimeCmdParser::AcceptInt( int EndTime )
{
	try {
		pSim->SetEndTime( EndTime );
		cout << "End time set to " << EndTime << " minutes." << endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}


void EndTimeCmdParser::ShortHelp( std::ostream& o ) const
{
	o << "Set the time in minutes this program must quit.";
}

void EndTimeCmdParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << " <int>" << endl;
	o << "\tSet the time in minutes this program must quit.  This time" << endl;
	o << "\twill be used by the checkpointer, if there is one, to" << endl;
	o << "\tdetermine when to checkpoint before quitting." << endl;
}


TraceCmdParser::TraceCmdParser(HysteresisSimulation* pSim)
: IntCmdParser("tracefrequency"), pSim(pSim) {}


bool TraceCmdParser::AcceptInt( int traceFrequency )
{
	try {
		pSim->SetTraceFrequency( traceFrequency );
		cout << "Tracing every " << traceFrequency << " seconds." << endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}

void TraceCmdParser::ShortHelp( std::ostream& o ) const
{
	o << "Set how often, in seconds, to print tracing information.";
}

void TraceCmdParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << " <int>" << endl;
	o << "\tHow often, in seconds to print tracing information.  This turns" << endl;
	o << "\ton a simple tracer in the simulation." << endl;
}


CheckpointCmdParser::CheckpointCmdParser(HysteresisSimulation* pSim)
: IntCmdParser("checkpointfrequency"),pSim(pSim) {}


bool CheckpointCmdParser::AcceptInt( int checkpointFrequency )
{
	try {
		pSim->SetCheckpointFrequency( checkpointFrequency );
		cout << "Checkpointing every " << checkpointFrequency << " minutes." << endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}

void CheckpointCmdParser::ShortHelp( std::ostream& o ) const
{
	o << "Set how often, in minutes, to checkpoint the program.";
}

void CheckpointCmdParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << " <int>" << endl;
	o << "\tSet the time in minutes between checkpoints.  This enables" << endl;
	o << "\ta NetCDF checkpointer in the simulation which will make sure" << endl;
	o << "\tto save not only at the given interval but also two hours" << endl;
	o << "\tbefore the end time and at the end time." << endl;
}


DisorderCmdParser::DisorderCmdParser(HysteresisSimulation* pSim)
	: CmdParser("disorder"), disorder(0.9), pSim(pSim)
{}

bool DisorderCmdParser::AcceptCmdLine( std::istream& s )
{
	double disorder;
	s >> disorder;
	if (s) this->disorder = disorder;
	try {
		pSim->SetDisorder( disorder );
		cout << "Disorder set to " << disorder << "." << endl;
	} catch (DMException& e) {
		std::cout << e;
	}
	return false;
}

void DisorderCmdParser::ShortHelp( std::ostream& o ) const
{
	o << "Set the disorder in the field.";
}

void DisorderCmdParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << " <float>" << endl;
	o << "\tSet the disorder in the field.  Typical values are between 0.6" << endl;
	o << "\tand 1.0." << endl;
}

RunCmdParser::RunCmdParser(HysteresisSimulation* pSim)
	: IntCmdParser("run"), pSim(pSim)
{
}

bool RunCmdParser::AcceptInt( int runs )
{
	return Run( runs );
}

void RunCmdParser::ShortHelp( std::ostream& o ) const
{
	o << "Tell the simulation to run a few times.";
}

void RunCmdParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << " <int>" << endl;
	o << "\tTell the simulation to run the given number of times.  Any " << endl;
	o << "\taveraging will be done over these runs." << endl;
}


bool RunCmdParser::Run( int runCnt )
{
	if ( runCnt > 0 ) {
		cout << "Doing " << runCnt << " runs." << endl;
		try {
			pSim->Initialize();
			pSim->Verify();
			
			int runIndex;
			for (runIndex=0; runIndex < runCnt; runIndex++)
			{
				std::cout << "Starting run " << runIndex << std::endl;
				
				AvalancheAlgorithm::Status status = pSim->Run();
				if ( AvalancheAlgorithm::ALLFLIPPED != status ) {
					std::cout << "Stopped before finished." << std::endl;
					return true;
				}
				
				pSim->SaveResults();
				pSim->IncrementSeed();
			}
			
			pSim->SaveAverageResults();
			std::cout << "Done runs" << std::endl;
			
		} catch (DMException e) {
			std::cout << e;
		}
	} else {
		cout << "You should specify some number of runs to do." << endl;
	}

	return false;
}


OnSpanningParser::OnSpanningParser( HysteresisSimulation* pSim )
	: ToggleCmdParser("SpanningWriterObserver"), pSim(pSim)
{ }

bool OnSpanningParser::TurnOn()
{
	try {
		pSim->EnableSpanningWriter(  );
		std::cout << "Spanning writer on." << std::endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}

bool OnSpanningParser::TurnOff()
{
	try {
		pSim->DisableSpanningWriter();
		std::cout << "Spanning writer off." << std::endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}


void OnSpanningParser::ShortHelp( std::ostream& o ) const
{
	o << "Turn on or off the spanning writer.";
}

void OnSpanningParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << " +/-" << endl;
	o << "\tTurn on or off the observer which will write the current " << endl;
	o << "\tspins to a file after the first spanning avalanche." << endl;
}


CorrelationParser::CorrelationParser( HysteresisSimulation* pSim )
	: ToggleCmdParser("CorrelationFunctionObserver"), pSim(pSim)
{ }

bool CorrelationParser::TurnOn()
{
	try {
		pSim->EnableCorrelationFunction( false );
		std::cout << "CorrelationFunction on." << std::endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}

bool CorrelationParser::TurnOff()
{
	try {
		pSim->DisableCorrelationFunction();
		std::cout << "CorrelationFunction off." << std::endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}


void CorrelationParser::ShortHelp( std::ostream& o ) const
{
	o << "Turn on or off the correlation length observer.";
}

void CorrelationParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << " +/-" << endl;
	o << "\tTurn the correlation length observer on or off." << endl;
}


AvgCorrelationParser::AvgCorrelationParser( HysteresisSimulation* pSim )
	: ToggleCmdParser("AvgCorrelationFunctionObserver"), pSim(pSim)
{ }

bool AvgCorrelationParser::TurnOn()
{
	try {
		pSim->EnableCorrelationFunction( true );
		std::cout << "Average CorrelationFunction on." << std::endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}

bool AvgCorrelationParser::TurnOff()
{
	try {
		pSim->EnableCorrelationFunction( false );
		std::cout << "CorrelationFunction off." << std::endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}

void AvgCorrelationParser::ShortHelp( std::ostream& o ) const
{
	o << "Turn on or off the average correlation length observer.";
}

void AvgCorrelationParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << " +/-" << endl;
	o << "\tTurn the average correlation length observer on or off." << endl;
	o << "\tTurning on or off the average observer turns the regular" << endl;
	o << "\tobserver on or off, too." << endl;
}


SizeDistributionParser::SizeDistributionParser( HysteresisSimulation* pSim )
	: ToggleCmdParser("SizeDistributionObserver"), pSim(pSim)
{ }

bool SizeDistributionParser::TurnOn()
{
	try {
		pSim->EnableSizeDistribution( false );
		cout << "Size distribution on." << endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}

bool SizeDistributionParser::TurnOff()
{
	try {
		pSim->DisableSizeDistribution();
		cout << "Size distribution off." << endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}

void SizeDistributionParser::ShortHelp( std::ostream& o ) const
{
	o << "Turn on or off the size distribution observer.";
}

void SizeDistributionParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << " +/-" << endl;
	o << "\tTurn the size distribution observer on or off." << endl;
}

 
AvgSizeDistributionParser::AvgSizeDistributionParser( HysteresisSimulation* pSim )
	: ToggleCmdParser("AvgSizeDistributionObserver"), pSim(pSim)
{ }

bool AvgSizeDistributionParser::TurnOn()
{
	try {
		pSim->EnableSizeDistribution( true );
		cout << "Average size distribution on." << endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}

bool AvgSizeDistributionParser::TurnOff()
{
	try {
		pSim->EnableSizeDistribution( false );
		cout << "Size distribution off." << endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}


void AvgSizeDistributionParser::ShortHelp( std::ostream& o ) const
{
	o << "Turn on or off the average size distribution observer.";
}

void AvgSizeDistributionParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << " +/-" << endl;
	o << "\tTurn the average size distribution observer on or off." << endl;
	o << "\tTurning on or off the average observer turns the regular" << endl;
	o << "\tobserver on or off, too." << endl;
}


ColoringParser::ColoringParser( HysteresisSimulation* pSim )
	: ToggleCmdParser("ColoringObserver"), pSim(pSim)
{ }

bool ColoringParser::TurnOn()
{
	try {
		pSim->EnableColoringObserver();
		cout << "Coloring observer on." << endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}

bool ColoringParser::TurnOff()
{
	try {
		pSim->DisableColoringObserver();
		cout << "Coloring observer off." << endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}


void ColoringParser::ShortHelp( std::ostream& o ) const
{
	o << "Turn on or off the observer that stores avalanche order.";
}

void ColoringParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << " +/-" << endl;
	o << "\tTurn the observer on or off for avalanche order.  This observer" << endl;
	o << "\tis currently used for display but is suited for n-dimensional" << endl;
	o << "\tavalanches." << endl;
}



MagnetizationParser::MagnetizationParser( HysteresisSimulation* pSim )
	: ToggleCmdParser("MagnetizationObserver"), pSim(pSim)
{ }

bool MagnetizationParser::TurnOn()
{
	try {
		pSim->EnableMagnetizationObserver();
		cout <<"Magnetization observer on." << endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}

bool MagnetizationParser::TurnOff()
{
	try {
		pSim->DisableMagnetizationObserver();
		cout << "Magnetization observer off." << endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}


void MagnetizationParser::ShortHelp( std::ostream& o ) const
{
	o << "Turn on or off the magnetization observer.";
}

void MagnetizationParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << " +/-" << endl;
	o << "\tTurn the observer on or off for magnetization." << endl;
}



TimeSeriesParser::TimeSeriesParser( HysteresisSimulation* pSim )
	: ToggleCmdParser("TimeSeriesObserver"), pSim(pSim)
{ }

bool TimeSeriesParser::TurnOn()
{
	try {
		pSim->EnableTimeSeries();
		cout << "Time series on." << endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}

bool TimeSeriesParser::TurnOff()
{
	try {
		pSim->DisableTimeSeries();
		cout << "Time series off." << endl;
	} catch (DMException e) {
		std::cout << e;
	}
	return false;
}


void TimeSeriesParser::ShortHelp( std::ostream& o ) const
{
	o << "Turn on or off the time series observer.";
}

void TimeSeriesParser::LongHelp( std::ostream& o ) const
{
	o << "Usage: " << cmdString << " +/-" << endl;
	o << "\tTurn the observer on or off for time series.  It tells" << endl;
	o << "\tyou how many spins flipped with each avalanche." << endl;
}
