#ifndef _BATCH_TIMER_H_
#define _BATCH_TIMER_H_

#include "DMFirstHeader.h"
#include "DiscreteAvalanche.h"
#include "WinWallClockTimer.h"
#include "EggTimer.h"

class Checkpointer;
class Tracer;

//! Timer to help run your program in batch conditions.
/*! Provides hooks for
		- Trace statements
		- Checkpointing
		- When to quit

	You call TimeSlice() regularly in your calculation loop,
	at least once a minute.  If you call more than once every ten
	seconds, this class will set up a counter and only do time
	checking every nth time you call TimeSlice.  The idea is that
	you don't always know how long your loop will take, and this
	tries to stay out of the way if it gets called a lot.

	\author ajd27
	\date 4 Sep 2000
	*/
class BatchTimer
{
	WinWallClockTimer WallTimer; //!< In order to decide when to quit.
	EggTimer*		  eggTimer;  //!< Tells us when next to wake.

	Checkpointer*	pCheckpointer;	//!< Pointer to our checkpointing.
	Tracer*			pTracer;		//!< Pointer to tracing object.

	typedef int TimeType;
	TimeType EndTime;			//!< When the simulation must die.
	TimeType NextTrace;			//!< The next time to trace.
	TimeType NextCheckpoint;	//!< The next time to checkpoint.
	TimeType SoonestStopTime;	//!< The soonest we need to do anything.

	bool bShouldQuit;	//!< What you access through IsTimeToQuit().
	bool bFirstCalled;	//!< True until after the first time we get called.
	bool bCheckTimes;	//!< Whether to check breaks.
public:
	//! Constructor.
	BatchTimer();
	//! Destructor.
	~BatchTimer();
	//! Set a time when we have to quit.
	void SetEndTime( int NumberOfSeconds );
	//! Set end time in minutes.
	void SetEndTimeMinutes( int NumberOfMinutes ) {
		this->SetEndTime( 60*NumberOfMinutes );
	}
	//! Give us an object which will checkpoint.
	void SetCheckpointer( Checkpointer* pCheckpointer );
	//! Give us an object which will trace.
	void SetTracer( Tracer* pTracer );
	//! Call to check whether it is initialized properly.
	void Verify( void );
	//! Call this once per inner loop.
	bool TimeSlice();
	//! This tells you whether it is time to quit.
	bool IsTimeToQuit() { return bShouldQuit; }
private:
	//! Sets the pointer-to-member-function, CalculateStopAndQuit.
	void SetTimeCalculationFunction( void );

	TimeType CheckCheckpoint( TimeType CurrentTime );
	TimeType CheckTrace( TimeType CurrentTime );

	/*! These routines are the entries in a kind of table
		lookup for how this class calculates the next time
		it must stop.  During initialization, we assign the 
		address of one of these routines to the 
		pointer-to-member-function, CalculateStopAndQuit.
		*/
	//@{
	TimeType SoonestStopFromCheckpoint( TimeType CurrentTime );
	TimeType SoonestStopFromTrace( TimeType CurrentTime );
	TimeType SoonestStopFromCheckpointOrTrace( TimeType CurrentTime );
	TimeType SoonestStopFromEndTime( TimeType CurrentTime );
	//@}

	void CheckTimes();
	TimeType (BatchTimer::*CalculateStopAndQuit)( TimeType CurrentTime );
};

#endif //_BATCH_TIMER_H_
