#include "DMFirstHeader.h"
#include <vector>
#include <algorithm>
#include "CheckSpanning.h"
#include "SpinLattice.h"
#include "DMException.h"
#include "DMAllocator.h"
#include "NotEnoughMemory.h"
#include "DMWriter.h"
#include "DMReader.h"
#include "DMLastHeader.h"

const std::string sType =
	std::string("CheckSpanning");

/*! This static object permits the DMReader to create an empty CheckSpanning. */
const SerialBuilder<CheckSpanning> *_creator = new
	SerialBuilder<CheckSpanning>(sType);

CheckSpanning::CheckSpanning()
{
	numSpanning = 0;
	isSpanning = false;
	lattice = 0;
	D = 0;
	ZeroArrays();
	iVerify = 0;

	firstSpin = true;
	bSerialized = false;
}

CheckSpanning::~CheckSpanning()
{
	this->DeleteArrays();
}


/*! You must call this function before calling Initialize.
	*/
void CheckSpanning::SetLattice( const SpinLattice* lattice )
{
	ASSERT( 0 != lattice );
	if ( 0 == lattice ) throw
		DMException("CheckSpanning","SetLattice",
		"Passed a null pointer.");

	this->lattice = lattice;
	iVerify |= VERIFY_LATTICE;
}

/*! We should be concerned here about whether we are initializing
	from serialization.  We check whether our arrays were already
	initialized before creating them.

	You must call this before either FlipCompleted or
	AvalancheCompleted are called.

	*/
void CheckSpanning::Initialize()
{
	if (lattice == 0) throw
		DMException("CheckSpanning","Initialize",
		"You must first call SetLattice.");

	int newD;
	newD = lattice->GetDimension();

	if (bSerialized) {
		if (D!=newD) throw
			DMException("CheckSpanning","Initialize",
				"Lattice and serialized CheckSpanning have different dimensions.");
	} else {
		if ( D != newD ) {
			this->DeleteArrays();
			D = newD;
			this->AllocateArrays();
		}
	}

	const SpinCoord* sideLengths = lattice->GetSideLengths();
	int dimIdx;
	for ( dimIdx=0; dimIdx<D; dimIdx++ ) {
		sideLength[dimIdx] = sideLengths[dimIdx];
	}

	bSerialized = false;
	iVerify = 0;
}

/*! This class always reinitializes for the next sweep because it
	doesn't really take any data.
	*/
void CheckSpanning::BeginSweep( bool bKeep )
{
	ASSERT( 0 == iVerify );
	if ( 0 != iVerify ) {
		throw DMException( "CheckSpanning", "BeginSweep",
			"You need to call Initialize after you change the lattice." );
	}
	
	numSpanning = 0;
	isSpanning = false;
	firstSpin = true;

	int dimIdx;
	for ( dimIdx=0; dimIdx<D; dimIdx++ ) {
		before[dimIdx] = 0;
		after[dimIdx] = 0;
		spanDir[dimIdx] = 0;
		extent[dimIdx] = 0;
	}
}

/*! Add the current spin to the cross-sections for
	use in detecting a spanning avalanche.

	This function is still multi-dimensional friendly.

	*/
inline void CheckSpanning::FlipCompleted( SpinLattice* lattice,
								  const int* coords, int dDimension )
{
	if(firstSpin)
	{
		firstSpin = false;
		for(int i=0;i<D;i++)
		{
			before[i] = coords[i]-1;
			if ( before[i]<0 ) before[i] = sideLength[i]-1;

			after[i] = coords[i]+1;
			if ( after[i]>=sideLength[i] ) after[i] = 0;

			extent[i]=1;
			spanDir[i]=false;
		}
		return;
	}

	int i;
	for(i=0;i<D;i++)
	{
		if(!spanDir[i])
		{
			if(coords[i]==before[i])
			{
				extent[i]++;
				if(before[i]==after[i])
				{
					spanDir[i]=true;
					isSpanning=true;
				}
				before[i]--;
				if(before[i]<0) before[i]=sideLength[i]-1;
			}
			else if(coords[i]==after[i])
			{
				extent[i]++;
				after[i]++;
				if(after[i]>=sideLength[i]) after[i]=0;
			}
		}
	}
	// For dipole forces, a better spanning avalanche check would be to
	// see if the avalanche is spanning in all directions.
	isSpanning=true;
	for(i=0;i<D;i++)
	{
	  if(!spanDir[i]) isSpanning=false;
	}
}

/*! Check the cross-sections to determine
	the extent of the avalanche, and see whether it
	is spanning.
	*/
inline void CheckSpanning::AvalancheCompleted( AvalancheAlgorithm* avalanche,
									   double dField )
{
	if(isSpanning) 
	{
		numSpanning++;
	}
	firstSpin = true;
}

void CheckSpanning::Save( DMWriter *writer ) const
{
	int iTemp, *piTemp;

	std::vector<int> shape(1);
	writer->PutInt( numSpanning, "numSpanning" );
	iTemp = (isSpanning) ? 1 : 0;
	writer->PutInt( isSpanning, "isSpanning" );
	iTemp = (firstSpin) ? 1 : 0;
	writer->PutInt( iTemp, "isFirstSpin" );

	shape[0] = D;
	writer->PutIntArray( before, shape, "before" );
	writer->PutIntArray( after, shape, "after" );
	writer->PutIntArray( extent, shape, "extent" );
	piTemp = new int[D];
	std::copy( spanDir, spanDir+D, piTemp );
	writer->PutIntArray( piTemp, shape, "spanDir");
	delete[] piTemp;
}

void CheckSpanning::Load( DMReader *reader )
{
	int iTemp, *piTemp;
	std::vector<int> shape;
	this->DeleteArrays();

	numSpanning = reader->GetInt( "numSpanning" );
	iTemp = reader->GetInt( "isSpanning" );
	isSpanning = ( iTemp == 1 );
	iTemp = reader->GetInt( "isFirstSpin" );
	firstSpin = ( iTemp == 1 );

	before = reader->GetIntArray( shape, "before" );
	D = shape[0];
	after = reader->GetIntArray( shape, "after" );
	extent = reader->GetIntArray( shape, "extent" );
	piTemp = reader->GetIntArray( shape, "spanDir" );
	spanDir = new bool[shape[0]];

	// cached variable, but should be allocated with others.
	sideLength = DMAllocate<SpinCoord>( D );

	for (iTemp=0; iTemp<D; iTemp++)
		spanDir[iTemp] = (piTemp[iTemp]==1);
	delete[] piTemp;
	bSerialized = true;
}

const std::string CheckSpanning::GetType() const
{
	return sType;
}

void CheckSpanning::ZeroArrays()
{
	before = 0;
	after = 0;
	spanDir = 0;
	extent = 0;
	sideLength = 0;
}

void CheckSpanning::DeleteArrays()
{
	if (before)		delete [] before;
	if (after)		delete [] after;
	if (spanDir)	delete [] spanDir;
	if (extent)		delete [] extent;
	if (sideLength)	delete [] sideLength;

	ZeroArrays();
}

void CheckSpanning::AllocateArrays()
{
	try {
		before = DMAllocate<SpinCoord>(D);
		after = DMAllocate<SpinCoord>(D);
		spanDir = DMAllocate<bool>(D);
		extent = DMAllocate<SpinCoord>(D);
		sideLength = DMAllocate<SpinCoord>(D);

	} catch ( std::bad_alloc ) {
		DeleteArrays();
		throw NotEnoughMemory("CheckSpanning",
		"AllocateArrays","Dimension*13 bytes");
	}
}