#include "DMFirstHeader.h"
#include <assert.h>
#include <math.h>
#include "CorrelationFunctionObserver.h"
#include "CheckSpanning.h"
#include "SpinLattice.h"
#include "DMReader.h"
#include "DMWriter.h"
#include "DMException.h"
#include "NotEnoughMemory.h"
#include "FailedVerify.h"
#include "SerialBuilder.h"
#include "DMAllocator.h"
#include "DMLastHeader.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

const std::string sType =
	std::string("CorrelationFunctionObserver");

const SerialBuilder<CorrelationFunctionObserver> *_creator = new
	SerialBuilder<CorrelationFunctionObserver>(sType);

/*! Construct a new CorrelationFunction action using the
	action spanAction to detect spanning
	avalanches, and storing a maximum of maxCorrel points
	in the correlation function.
	*/
CorrelationFunctionObserver::CorrelationFunctionObserver()
{
	lattice = 0;
	ZeroArrays();

	lastMax = lastUsedBin = lastCalculatedBin = 0;
	this->maxCorrel = 20000;

	firstSpin = true;

	bSerialized = false;
	iVerify = 0;
}

CorrelationFunctionObserver::~CorrelationFunctionObserver()
{
	DeleteArrays();
}

/**
  *  @param iMaxCorrel integer - Error if correlation length is greater
  *  than the value you set here.
  */
void CorrelationFunctionObserver::SetMaxCorrelationLength( SpinIndex iMaxCorrel )
{
	this->maxCorrel = iMaxCorrel;
	iVerify |= CORREL;
}


/*! You must call this before calling Initialize.
	*/
void CorrelationFunctionObserver::SetSpanning( const CheckSpanning* spanning )
{
	ASSERT( 0 != spanning );
	if (0 == spanning) throw
		DMException("CorrelationFunctionObserver","SetSpanning",
		"Passed null pointer.");

	this->spanning = spanning;

	iVerify |= SPANNING;
}


/*! You must call this before calling Initialize.
	*/
void CorrelationFunctionObserver::SetLattice( const SpinLattice* lattice )
{
	ASSERT( 0 != lattice );
	if (0 == lattice) throw
		DMException("CorrelationFunctionObserver","SetLattice",
		"Passed null pointer.");

	this->lattice = lattice;

	iVerify |= LATTICE;
}


/*! Must call this before the actions are called.
	Must call SetSpanning and SetMaxCorrelationLength before this.

	You can call this as many times as you want.  Each time, it 
	re-initializes the CorrelationFunctionObserver.
	*/
void CorrelationFunctionObserver::Initialize()
{
	ASSERT( spanning != 0 );
	if ( spanning == 0) throw
		DMException("CorrelationFunctionObserver","Initialize",
		"Need to call SetSpanning first.");

	ASSERT( lattice != 0 );
	if ( lattice == 0) throw
		DMException("CorrelationFunctionObserver","Initialize",
		"Need to call SetLattice first.");

	iVerify = INITIALIZE;

	int newD = lattice->GetDimension();
	ASSERT(newD > 0);
	N = lattice->GetLength();

	if (bSerialized) {
		// We read the former dimension from disk.
		// May as well use it for something.
		ASSERT( D == newD );
		if (D != newD) throw
			DMException("CorrelationFunctionObserver","Initialize",
			"Dimension of lattice does not match that of file.");
		bSerialized = false;
	} else {
		D = newD;

		DeleteArrays();
		AllocateArrays();
	}

	// Keep local cache of lengths of sides.
	const SpinCoord* sides = lattice->GetSideLengths();
	int dimIdx;
	for ( dimIdx=0; dimIdx < D; dimIdx++ ) {
		sideLength[dimIdx] = sides[dimIdx];
	}

	int i;
	for(i=0;i<maxCorrel;i++)
	{
		binWidth[i] = 0.0;
	}
}


/*! Call this at the start of a new field sweep in order
	to reset the location of the first spin.
	*/
void CorrelationFunctionObserver::BeginSweep( bool bKeepData )
{
	this->Verify();
	

	if ( bKeepData ) {
		bool bHaveData = ( lastUsedBin > 0 );
		if ( bHaveData )
			accumulator.AddDataSet( *this );
	} else {
		accumulator.Clear();
	}

	int i;
	for(i=0;i<maxCorrel;i++)
	{
		correlationFunction[i] = 0.0;
		tempCorrelationFunction[i] = 0.0;
	}
	
	this->lastUsedBin = 0;
	this->lastMax = 0;
	firstSpin = true;
}


void CorrelationFunctionObserver::Verify()
{
	ASSERT( spanning != 0 );
	if ( spanning == 0) throw
		DMException("CorrelationFunctionObserver","Initialize",
		"Need to call SetSpanning first.");

	ASSERT( lattice != 0 );
	if ( lattice == 0) throw
		DMException("CorrelationFunctionObserver","Initialize",
		"Need to call SetLattice first.");

	if ( iVerify == INITIALIZE ) return;

	if ( (iVerify&CORREL) == 1 ) throw
		FailedVerify("CorrelationFunctionObserver",
		"Must call Initialize if you call SetMaxCorrelationLength.");

	if ( (iVerify&SPANNING) == 1 ) throw
		FailedVerify("CorrelationFunctionObserver",
		"Must call Initialize if you call SetSpanning.");
	
	if ( (iVerify&LATTICE) == 1 ) throw
		FailedVerify("CorrelationFunctionObserver",
		"Must call Initialize if you call SetLattice.");

	if ( (iVerify&INITIALIZE) == 0 ) throw
		FailedVerify("CorrelationFunctionObserver",
		"Must call Initialize.");
}



/** Each time a spin flips, record its location for use in
  * the correlation function.
  */
inline void CorrelationFunctionObserver::FlipCompleted( SpinLattice* lattice,
												const int* coords,
												int dDimension )
{
	ASSERT( 0 != lattice );
	ASSERT( 0 != coords );
	ASSERT( dDimension > 0 );

	if(firstSpin)
	{
		firstSpin = false;
		for( int i=0; i<D; i++)
		{
			firstSpinPosition[i] = coords[i];
		}
	}

	else
	{
		int rSquared=0;
		for ( int i=0; i<D; i++ )
		{
			int x1 = coords[i];
			if(x1>spanning->EmptyRow(i)) x1-=sideLength[i];
			int x2 = firstSpinPosition[i];
			if(x2>spanning->EmptyRow(i)) x2-=sideLength[i];
			int dx = x2-x1;
			rSquared += dx*dx;
		}

		int r = (int)(sqrt(rSquared)+0.5);
		if ( r<maxCorrel )
		{
			if(r>lastMax) lastMax = r;
			tempCorrelationFunction[r]+=1;
		}
	}
}

/*!	Each time an avalanche is completed,
	we need to calculate it's correlation function.
	If it wasn't spanning, add the current avalanche to the correlation function.
	*/
inline void
CorrelationFunctionObserver::AvalancheCompleted
	(
	AvalancheAlgorithm* avalanche,
	double H
	)
{
	ASSERT( 0 != spanning );
	if(!spanning->IsSpanning())
	{
		for(int i=1;i<=lastMax;i++)
		{
			correlationFunction[i]+=tempCorrelationFunction[i];
			tempCorrelationFunction[i]=0;
		}
		if(lastMax>lastUsedBin) lastUsedBin = lastMax;
		lastMax = 0;
		firstSpin = true;
	}
	else
	{
		for(int i=1;i<lastMax;i++)
		{
			tempCorrelationFunction[i]=0;
		}
		lastMax = 0;
		firstSpin = true;
	}
}

/*!	Zero the old bin widths, calculate new bin-widths, and record the
	maximum calculated bin-width.
	Recursively calculate the widths of the
	bins.  The bin at distance r has a width given by the number
	of spins in a D dimensional lattice which are a distance r 
	from the origin.
  */
void CorrelationFunctionObserver::CalculateBinWidths(int dim, int rSquaredSum) const
{
	ASSERT( dim >= 0 );
	double cutoff = sqrt((double) lastUsedBin*lastUsedBin - rSquaredSum);

	if(dim == D-1)
	{
		for(int i=-(int)cutoff; i<=(int)cutoff; i++)
		{
			double rSquared = rSquaredSum+i*i;
			int r = (int) (sqrt(rSquared)+0.5);

			binWidth[r]++;
		}
	}
	else
	{
		for(int i=-(int)cutoff; i<=(int)cutoff; i++)
		{
			CalculateBinWidths(dim+1, rSquaredSum+i*i);
		}
	}
}

/** GetCorrelationFunction: Return the correlation function for
  * the given distance.
  */
double CorrelationFunctionObserver::GetCorrelationFunction(int distance) const
{
	ASSERT( distance<=lastUsedBin );

	if(lastCalculatedBin<lastUsedBin)
	{
		UpdateBinWidths();
	}

	return correlationFunction[distance] / 
		(double)binWidth[distance] / ((double) N);
}

void CorrelationFunctionObserver::UpdateBinWidths() const
{
	for(int i=0;i<maxCorrel;i++) binWidth[i] = 0;
	CalculateBinWidths();
	lastCalculatedBin = lastUsedBin;
}

/*! You have to call Initialize before you will be able
	to save this class.  Calling Initialize ensures it has
	sensible values, including dimensions to the arrays.
	*/
void CorrelationFunctionObserver::Save( DMWriter *writer ) const
{
	int iFirstSpin;
	std::vector<int> shape;

	ASSERT( 0 != writer );
	if ( 0 == writer ) {
		throw DMException( "CorrelationFunctionObserver","Save",
			"Passed a null pointer.");
	}

	ASSERT( maxCorrel > 0 );
	if (0 >= maxCorrel) {
		throw DMException("CorrelationFunctionObserver","Save",
			"Trying to save uninitialized class.");
	}
	writer->PutInt( maxCorrel, "maxCorrel" );
	writer->PutInt( lastUsedBin, "lastUsedBin" );
	writer->PutInt( lastCalculatedBin, "lastCalculatedBin" );
	iFirstSpin = (firstSpin) ? 1 : 0;
	writer->PutInt( iFirstSpin, "isFirstSpin" );
	writer->PutInt( lastMax, "lastMax" );

	shape.push_back(D);
	writer->PutIntArray( firstSpinPosition, shape, "firstSpinPosition");
	shape[0] = maxCorrel;
	writer->PutDoubleArray( correlationFunction, shape,
		"correlationFunction");
	writer->PutDoubleArray( tempCorrelationFunction, shape,
		"tempCorrelationFunction" );
	writer->PutIntArray( binWidth, shape, "binWidth" );

	writer->Put( &accumulator, "accumulator" );
}

void CorrelationFunctionObserver::Load( DMReader *reader )
{
	int iFirstSpin;
	std::vector<int> shape;
	ASSERT( 0 != reader );
	if ( 0 == reader ) {
		throw DMException( "CorrelationFunctionObserver","Save",
			"Passed a null pointer.");
	}
	DeleteArrays();
	iVerify = 0;

	maxCorrel = reader->GetInt( "maxCorrel" );
	ASSERT( 0 != maxCorrel );
	if ( 0 == maxCorrel )
		throw DMException("CorrelationFunctionObserver","Load",
		"Class has improper storage.  maxCorrel is zero.");

	lastUsedBin = reader->GetInt( "lastUsedBin" );
	lastCalculatedBin = reader->GetInt( "lastCalculatedBin" );
	iFirstSpin = reader->GetInt( "isFirstSpin" );
	firstSpin = ( iFirstSpin == 1 );
	lastMax = reader->GetInt( "lastMax" );
	
	firstSpinPosition = reader->GetIntArray(shape,"firstSpinPosition");
	D = shape[0];
	SUPERASSERT( 0 == IsBadReadPtr( firstSpinPosition, D*sizeof(int)));
	correlationFunction = reader->GetDoubleArray(shape,"correlationFunction");
	maxCorrel = shape[0];
	SUPERASSERT( 0 == IsBadReadPtr( correlationFunction, maxCorrel*sizeof(double)));
	tempCorrelationFunction = reader->GetDoubleArray(shape,
		"tempCorrelationFunction");
	SUPERASSERT( 0 == IsBadReadPtr( tempCorrelationFunction,
		maxCorrel*sizeof(double)));
	binWidth = reader->GetIntArray( shape, "binWidth" );
	SUPERASSERT( 0 == IsBadReadPtr( binWidth, maxCorrel*sizeof(int)));

	// Cached variable, but should allocate with the rest.
	sideLength = DMAllocate<SpinCoord>(D);

	reader->Fill( &accumulator, "accumulator" );

	this->SetMaxCorrelationLength( maxCorrel );
	bSerialized = true;
}

const std::string CorrelationFunctionObserver::GetType() const
{
	return sType;
}

void CorrelationFunctionObserver::DeleteArrays()
{
	if (correlationFunction)		delete [] correlationFunction;
	if (tempCorrelationFunction)	delete [] tempCorrelationFunction;
	if (firstSpinPosition)			delete [] firstSpinPosition;

	if (binWidth)	delete [] binWidth;
	if (sideLength)	delete [] sideLength;

	ZeroArrays();
}

void CorrelationFunctionObserver::ZeroArrays()
{
	correlationFunction		= 0;
	tempCorrelationFunction = 0;
	binWidth				= 0;

	firstSpinPosition	= 0;
	sideLength			= 0;
}

void CorrelationFunctionObserver::AllocateArrays()
{
	try {
		binWidth				= DMAllocate<int>(maxCorrel);
		correlationFunction		= DMAllocate<double>(maxCorrel);
		tempCorrelationFunction = DMAllocate<double>(maxCorrel);

		firstSpinPosition	= DMAllocate<int>(D);
		sideLength			= DMAllocate<SpinCoord>(D);

	} catch ( std::bad_alloc ) {
		DeleteArrays();
		throw NotEnoughMemory("CorrelationFunctionObserver",
			"AllocateArrays", "maxCorrel* two doubles and an int");
	}
}

std::ostream&
CorrelationFunctionObserver::SaveAverageData( std::ostream& s )
{
	s << accumulator;
	return s;
}

std::ostream& operator<<( std::ostream& s, const CorrelationFunctionObserver& r )
{
	int i;
	for (i=0; i< r.lastUsedBin-1; i++)
		s << i+1 << "\t" << r.GetCorrelationFunction(i+1) << std::endl;
	return s;
}