#ifndef _CORRELATION_FUNCTION_OBSERVER_H_
#define _CORRELATION_FUNCTION_OBSERVER_H_

#include "DMFirstHeader.h"
#include "DiscreteAvalanche.h"
#include "FlipListener.h"
#include "AvalancheListener.h"
#include "SpinLattice.h"
#include "SpinIndex.h"
#include "XYData.h"
#include "XYAverage.h"
#include "Serializable.h"

class CheckSpanning;

//! Measures the correlation function of a lattice.
/*! Uses CheckSpanning to determine if avalanches are spanning.
	Prune avalanches beyond maxAvalSize spins.  Error if correlation
	length is greater than maxCorrel.

	This class can store cumulative data over several runs.

	\author Matt Kuntz
  */
class CorrelationFunctionObserver : public FlipListener,
		public AvalancheListener, public Serializable, public XYData
{
	const CheckSpanning* spanning;
	const SpinLattice* lattice;

	XYAverage accumulator;
public:
	//! Constructor
	CorrelationFunctionObserver();
	
	// free up resources from correlation function
	//! Destructor
	virtual ~CorrelationFunctionObserver();
	//! Set maximum correlation length.
	void SetMaxCorrelationLength( SpinIndex iMaxCorrel = 20000 );
	//! Give this class a pointer to its CheckSpaning.
	void SetSpanning( const CheckSpanning* spanning );
	//! Give this class a pointer to the SpinLattice
	void SetLattice( const SpinLattice* SpinLattice );
	//! Initialize internal variables.
	void Initialize();

	//! Call at the start of each sweep.
	void BeginSweep( bool bKeepData = false );
	//! Check whether everything was called in turn.
	void Verify();
	
	//ACTION FUNCTIONS

	//! flip a spin.  Update the correlation function to include this spin.
	void FlipCompleted( SpinLattice* lattice, const int* coords,
		int dDimension );
	
	//! end the avalanche.
	void AvalancheCompleted( AvalancheAlgorithm* avalanche, double H);

	//ACCESSOR FUNCTIONS

	//! Get the largest distance in the correlation function
	int GetLargestDistance() const { return lastUsedBin; }

	//! Get the value of the correlation function at the given distance.
	double GetCorrelationFunction(int distance) const;

	//! Get the width of the bin.  How many spins are at this distance?
	int GetBinWidth(int distance) const { return binWidth[distance]; }

	/*! XYData interface */
	//@{
	//! How many points are there in the correlation function?
	int GetPointCnt() const { return lastUsedBin-1; }
	
	//! What is the distance of the ith entry in the correlation function
	double GetX(int i) const { return (double) (i+1); }

	//! What is the value of the ith element of the correlation function
	double GetY(int i) const
	{
		return GetCorrelationFunction(i+1);
	}
	//@}

	/*! Serializable interface */
	//@{
	void Save( DMWriter *writer ) const;
	void Load( DMReader *reader );
	const std::string GetType() const;
	//@}

	std::ostream& SaveAverageData( std::ostream& s );
	friend std::ostream& operator<<( std::ostream& s,
		const CorrelationFunctionObserver& r );

private:
	//IMPLEMENTATION
		
	// Calculate the bin widths for normalizing the histograms
	void CalculateBinWidths(int dim=0, int rSquaredSum=0) const;

	void UpdateBinWidths() const;

	void DeleteArrays();
	void ZeroArrays();
	void AllocateArrays();

	//PRIVATE DATA

	int maxCorrel;	// The maximum correlation length we store

	double* correlationFunction;			// Correlation function data
	// This is temporary in the sense that it lasts for a single avalanche
	// before being copied into the correlationFunction.
	// Temporary storage for the correlation function of the current avalanche
	double* tempCorrelationFunction;		
	// The position of the first spin.
	// Used to calculate the trigger correlation function.
	int* firstSpinPosition;					
	// The widths of the bins
	mutable int* binWidth;					
	// The number of dimensions of the simulation
	int D;									
	SpinCoord* sideLength;
	int N;// total # of spins in the lattice
	// What is the highest bin we have used?
	int lastUsedBin;						
	// What is the last bin for which we have calculated the width.
	mutable int lastCalculatedBin;			
	// Is this the first spin in the avalanche?
	bool firstSpin;							
	// What was the largest distance in the last avalanche?
	int lastMax;							

	// Constants to help with verification of interface.
	enum {
		CORREL = 0x1,
		SPANNING = 0x2,
		LATTICE = 0x4,
		INITIALIZE = 0x8
	};
	int iVerify; // Keeps track of whether everything was called correctly.
	bool bSerialized; //!< Whether we were just loaded from storage.
};

#endif //_CORRELATION_FUNCTION_OBSERVER_H_
