#include "DMFirstHeader.h"
#include "FixedFieldAvalanche.h"
#include "SpinLattice.h"
#include "FailedVerify.h"
#include "GaussianInitializer.h"
#include "FixedFieldAvalancheStart.h"
#include "DMException.h"
#include "NotEnoughMemory.h"
#include "DMAllocator.h"
#include "DMLastHeader.h"

FixedFieldAvalanche::FixedFieldAvalanche()
{
	starter = 0;
	lattice = 0;
	iVerify = 0;
	ZeroArrays();
	M = 0;
}

FixedFieldAvalanche::~FixedFieldAvalanche()
{
	DeleteArrays();
}

/** You have to set the lattice before you can call Initialize.
  */
void FixedFieldAvalanche::SetLattice( SpinLattice * lattice )
{
	ASSERT( 0 != lattice);
	if (lattice == 0) throw DMException( "FixedFieldAvalanche","SetLattice", "Null pointer");

	this->lattice = lattice;
}

/** You have to set the initializer before you can call initialize.
  */
void FixedFieldAvalanche::SetInitializer( GaussianInitializer *initializer )
{
	ASSERT( 0 != initializer );
	if ( 0 == initializer ) throw DMException( "FixedFieldAvalanche",
		"SetInitializer","Null pointer passed in.");

	this->initializer = initializer;
}

/*! You have to call this before calling Initialize.
	*/
void FixedFieldAvalanche::SetStarter( FixedFieldAvalancheStart* starter )
{
	ASSERT( 0 != starter );
	if ( 0 == starter ) throw DMException("FixedFieldAvalanche",
		"SetStarter","Passed a null pointer.");

	this->starter = starter;
}

void FixedFieldAvalanche::Initialize()
{
	ASSERT( 0 != starter );
	ASSERT( 0 != lattice );
	ASSERT( 0 != initializer );
	if ( 0 == starter ) throw DMException("FixedFieldAvalanche","Initialize",
		"You have to call SetStarter before calling Initialize.");
	if ( 0 == lattice ) throw DMException("FixedFieldAvalanche","Initialize",
		"You have to call SetLattice before calling Initialize.");
	if ( 0 == initializer ) throw DMException("FixedFieldAvalanche","Initialize",
		"You have to call SetInitializer before calling Initialize.");

	SetCachedVariables();
	DeleteArrays();
	AllocateArrays();
	InitializeRandomField();
	InitializeNUp();
	M = -N;

	starter->SetAlgorithm(this);
	iVerify |= INITIALIZED;
}

void FixedFieldAvalanche::Verify() const
{
	ASSERT( 0 != (iVerify&INITIALIZED) );
	if ( (iVerify&INITIALIZED) == 0)
		throw FailedVerify( "FixedFieldAvalanche","Failed to call Initialize." );
}

const SpinLattice* FixedFieldAvalanche::GetLattice() const
{
	if ((iVerify&INITIALIZED) == 0) return 0;
	return lattice;
}

const double* FixedFieldAvalanche::GetRandomField() const
{
	if ((iVerify&INITIALIZED) == 0) return 0;
	return pdRandomField;
}

const char* FixedFieldAvalanche::GetNUp() const
{
	if ((iVerify&INITIALIZED) == 0) return 0;
	return pcNUp;
}

AvalancheAlgorithm::Status
FixedFieldAvalanche::FindNextStart( double* H, SpinIndex *iFirstSpin )
{
	return this->starter->FindNextStart( H, iFirstSpin );
}


AvalancheAlgorithm::Status
FixedFieldAvalanche::Cascade( double H, SpinIndex iFirstSpin )
{
	static const SpinIndex ENDOFSHELL = -1;
	SpinIndex neighbors[8];
	int time, i;
	SpinIndex curLoc = iFirstSpin;
	SpinIndex nextLoc;

	spinFlipQueue.push(curLoc);
	spinFlipQueue.push(ENDOFSHELL);
	time=0;
	
	while (!spinFlipQueue.empty()) {
		curLoc = spinFlipQueue.front();
		spinFlipQueue.pop();

		if (curLoc != ENDOFSHELL) {
			if (lattice->Get(curLoc) != 1) {
				lattice->flip(curLoc,1);
				M += 2;
				
				lattice->GetNeighbors(curLoc, neighbors);
				for (i=0; i<Z; i++) {
					nextLoc = neighbors[i];
					pcNUp[nextLoc] += 1;
					if ( (lattice->Get(nextLoc)==-1) &&
						(2*pcNUp[nextLoc]-Z+pdRandomField[nextLoc]+H>0.0) ) {
						spinFlipQueue.push(nextLoc);
					}
				}
			}
		} else {
			time++;
			if (!spinFlipQueue.empty()) spinFlipQueue.push(ENDOFSHELL);
		}

	}
	return SUCCESS;
}

// ********** Member functions private below here. ************/

void FixedFieldAvalanche::InitializeRandomField()
{
	ASSERT( 0 != initializer );
	if ( 0 == initializer ) throw DMException("FixedFieldAvalanche",
		"InitializeRandomField","You need to call SetInitializer first.");
	initializer->Initialize( pdRandomField, N );
}

void FixedFieldAvalanche::InitializeNUp()
{
	int i;
	for (i=0; i<N; i++) {
		pcNUp[i] = 0;
	}
}

void FixedFieldAvalanche::ZeroArrays()
{
	pdRandomField = 0;
	pcNUp = 0;
}

void FixedFieldAvalanche::AllocateArrays()
{
	try {
		pdRandomField	= DMAllocate<double>(N);
		pcNUp			= DMAllocate<char>(N);

	} catch (std::bad_alloc e) {
		DeleteArrays();
		throw NotEnoughMemory( "FixedFieldAvalanche",
			"AllocateArrays","Couldn\'t allocate enough memory.");
	}
}

void FixedFieldAvalanche::DeleteArrays()
{
	if (pdRandomField) delete[] pdRandomField;
	if (pcNUp) delete[] pcNUp;
	ZeroArrays();
}

void FixedFieldAvalanche::SetCachedVariables()
{
	iLength = lattice->GetLength();
	Z = lattice->GetCoordinationNumber();
	N = iLength*iLength;
}