#ifndef _GENERAL_SPIN_LATTICE_H_
#define _GENERAL_SPIN_LATTICE_H_

#include "DMFirstHeader.h"
#include <vector>
#include "DiscreteAvalanche.h"
//#include "stl_bvector.h"
#include "SpinIndex.h"
#include "Serializable.h"

// std::_Bvector
// stl_bvector
typedef std::_Bvector vector_bool;
// typedef bit_vector vector_bool;

//! Rectangular lattice in N dimensions of up or down spins.

/*!	Matt Kuntz's code defined an enum representing UP or DOWN, but it
	was clear that the integer values +1 and -1 were very often useful
	in his code, so we're taking that experience and making this class
	specify that you treat the values of the
	spins as +1 or -1.

	The Save and Load are just a hair from being done.  The hard part,
	saving the spins, works.  The easy part, saving other internal variables,
	has changed and needs to be fixed.

	You can switch this class to handle more than 2^32-1 spins by 
	changing the typedef of vector_bool to the stl_vector version.
	I tool the Gnu STL vector<bool> and hacked it so it will
	address as many spins as it can fit in 2^32 bytes.  That's eight
	times as many, and we can use them.

	It's only a useful option if you also switch the SpinIndex
	to a 64 bit integer (ULONGLONG in MSVC++).

	\author ajd27
	\date 15 Aug 2000
*/
class GeneralSpinLattice : public vector_bool, public Serializable
{
	int DimensionCnt;
	SpinCoord* Dimensions;
	SpinIndex SpinCnt;
public:
//	typedef SpinIndex* iterator;
	//! Constructor
	GeneralSpinLattice();

	//! Destructor.
	~GeneralSpinLattice();

	//! Set number of dimensions.
	void SetDimension( int NumberOfDimensions );

	//! Set the length of all sides.
	bool SetLengths( const SpinCoord* Dimensions );

	//! Identical to Get.  This returns the value of the given spin.
	int operator()( const SpinIndex iPos ) const;

	//! Return the value of a given spin.
	int Get( const SpinIndex iPos ) const;

	//! Flip a spin up or down.
	void flip( const SpinIndex iPos, int orientation );

	//! Flip all spins down
	void SetAllDown( );

	//! How many up neighbors does this spin have?
	int NeighborsUp( const SpinIndex iPos ) const;

	//! Who are the neighbors of this spin?
	void GetNeighbors( const SpinIndex iPos, SpinIndex* const neighbors ) const;

	//! Get the coordinates of this index.
	void GetCoords( SpinIndex spinPos, SpinCoord *const coords ) const;

	//! Get the index of these coordinates.
	SpinIndex GetIndex( const SpinCoord *const coords ) const;

	//! Return the length of the whole array.
	SpinIndex GetLength() const;

	const SpinCoord* GetSideLengths() const;

	//! What is the coordination number of this lattice?
	SpinIndex GetCoordinationNumber() const;

	//! Return the spatial dimension of the lattice.
	int GetDimension() const { return this->DimensionCnt; }

	//! How many spins are up?
	SpinIndex GetTotalNumberUp() const;

	//! This saves the GeneralSpinLattice.
	virtual void Save(DMWriter *writer) const;
	//! Read in new values for this GeneralSpinLattice.
	virtual void Load(DMReader *reader);
	//! Return the signature type of this spinlattice.
	virtual const std::string GetType() const;
	//! Consider this private.  Use GetType instead.
private:
	SpinIndex* Strides; // Dimensions+1 in length.
	SpinIndex* _neighbors; // internal cache variable.

	//! Version of SetLengths which doesn't check for errors.
	bool _SetLengths( const SpinCoord* Dimensions );
	void SetStrides( );

	//! Allocates room for zero spins.  Default state.
	void ZeroSpinArray();

	//! Checks that dimension array is well-formed.
	bool CheckDimensionArray( const SpinIndex* Dimensions );
	int* WriteSlice( SpinIndex start, SpinIndex finish, int& dwLen ) const;
	void ReadSlice( SpinIndex start, SpinIndex finish, int*data );
	void AllocateArrays();
	void DeleteArrays();
	void ZeroArrays();
};

inline SpinIndex
GeneralSpinLattice::GetCoordinationNumber() const
{
	return 2*this->DimensionCnt;
}


inline SpinIndex
GeneralSpinLattice::GetLength() const
{
	return this->SpinCnt;
}


inline const SpinCoord*
GeneralSpinLattice::GetSideLengths() const
{
	return Dimensions;
}

/*! The fastest way to turn a bool into an int.
  */
inline int GeneralSpinLattice::operator()( const SpinIndex iPos ) const
{
	return ((*this)[iPos]) ? 1 : -1;
}

inline int GeneralSpinLattice::Get( const SpinIndex iPos ) const
{
	return ((*this)[iPos]) ? 1 : -1;
}

inline void GeneralSpinLattice::flip( const SpinIndex iPos, int orientation )
{
	(*this)[iPos] = ( orientation == 1 );
}

/*	If you unroll this loop, the code actually becomes a little slower
	because GetNeighbors accesses a different set of memory than the
	values of the spins.
	*/
inline int
GeneralSpinLattice::NeighborsUp( const SpinIndex iPos ) const
{
	GetNeighbors( iPos, this->_neighbors );

	int nup = 0;
	int NeighborIdx;
	for ( NeighborIdx=0; NeighborIdx<GetCoordinationNumber(); NeighborIdx++) {
		SpinIndex ANeighbor = _neighbors[NeighborIdx];
		nup += ((*this)[ANeighbor]) ? 1 : 0;
	}

	return nup;
	
}

/*! 
	@param iPos position of the spin in the lattice
	@param neighbors Array in which the neigbors will be returned.
	Must be able to accept Coordination Number elements.
  */
/*	Combining the loop to get the coordinates and that to find the
	neighbors saved 10 % of the time over not having the loops combined.

	Using a lookup-table instead of if-thens in the inner loop increased
	the execution time by 10-50%.  The variation in slowdown is due to
	cache considerations.

	Switching the order of the tests so that the most common was first
	was not measurably different.  Even a little slower.
	*/
inline void
GeneralSpinLattice::GetNeighbors
	(
	const SpinIndex iPos,
	SpinIndex *const neighbors
	) const
{
	SpinIndex DecreasingIndex = iPos;

	int DimIdx; // still not right.  Compare with below.
	for ( DimIdx=0; DimIdx < this->DimensionCnt; DimIdx++ ) {

		// Find the coordinate along this side.
		SpinCoord SideLength = this->Dimensions[DimIdx];
		SpinCoord coord = DecreasingIndex % SideLength;
		DecreasingIndex /= SideLength;

		// Find the upper and lower neighbors of that coordinate.
		if ( coord == SideLength-1 ) {
			neighbors[2*DimIdx] = iPos +
				(this->Strides[DimIdx] - this->Strides[DimIdx+1] );
			neighbors[2*DimIdx+1] = iPos - this->Strides[DimIdx];

		} else if ( coord == 0 ) {
			neighbors[2*DimIdx] = iPos + this->Strides[DimIdx];
			neighbors[2*DimIdx+1] = iPos +
				this->Strides[DimIdx+1] - this->Strides[DimIdx];

		} else {
			neighbors[2*DimIdx] = iPos + this->Strides[DimIdx];
			neighbors[2*DimIdx+1] = iPos - this->Strides[DimIdx];
		}
	}
}

inline void
GeneralSpinLattice::GetCoords( SpinIndex iPos, SpinCoord *const coords ) const
{
	int DimIdx;
	for( DimIdx=0; DimIdx < this->DimensionCnt; DimIdx++ ) {

		SpinCoord SideLength = this->Dimensions[DimIdx];
		coords[DimIdx] = iPos % SideLength;
		iPos /= SideLength;

	}
}

inline SpinIndex
GeneralSpinLattice::GetIndex( const SpinCoord *const coords ) const
{
	SpinIndex loc = 0;

	for ( int dimIdx = 0; dimIdx < this->DimensionCnt; dimIdx++ )
		loc += coords[ dimIdx ] * this->Strides[ dimIdx ];

	return loc;
}

#endif //_GENERAL_SPIN_LATTICE_H_
