#include "DMFirstHeader.h"
#include <cmath>
#include "DigitalMaterial.h"
#include "WholeGridBase.h"
#include "DMLastHeader.h"

WholeGridBase::WholeGridBase()
{
	colors = 0;
	iMaxVal = 0;
}

WholeGridBase::~WholeGridBase()
{
	if (colors) delete[] colors;
}

RGBQUAD* WholeGridBase::GetPalette()
{
	int i = 0;
	if (colors == 0) {
		colors = new RGBQUAD [256];
		if (colors==0) return 0;
		for(unsigned char r=0;r<256;r+=0x33)
		{
			for(unsigned char g=0;g<256;g+=0x33)
			{
				for(unsigned char b=0;b<256;b+=0x33)
				{
					colors[i].rgbBlue = b;
					colors[i].rgbGreen = g;
					colors[i].rgbRed = r;
					i++;
				}
			}
		}
	}
	iColors = i;
	return colors;
}

int WholeGridBase::GetPaletteSize()
{
	return iColors;
}

int WholeGridBase::SetPalette(RGBQUAD* colors, int iColors)
{
	this->iColors = iColors;
	this->colors = new RGBQUAD[iColors];
	if (colors == 0) return 0;
	::memcpy(this->colors,colors,iColors*sizeof(RGBQUAD));
	return 1;
}

int WholeGridBase::CalculateDecimation(int cWidth, int cHeight, int xDim, int yDim)
{
	int iDecimate;
	if (cWidth==0 || cHeight == 0 || xDim==0 || yDim==0) return 0;
	int decX = xDim/cWidth + 1;
	int decY = yDim/cHeight + 1;
	iDecimate = (decX<decY) ? decY : decX;
	if (iDecimate==1) {
		decX = cWidth/xDim;
		decY = cHeight/yDim;
		iDecimate = (decX<decY) ? -decX : -decY;
		if (iDecimate==0) iDecimate = 1;
	}
	return iDecimate;
}

int WholeGridBase::SetData(int *data, int xDim, int yDim, int iMaxVal)
{
	this->data = data;
	this->xDim = xDim;
	this->yDim = yDim;
	this->iMaxVal = iMaxVal;
	return 1;
}

int WholeGridBase::Draw(HDC hDC, RECT& bounds)
{
	// Make sure we have a palette.
	if (colors==0) {
		RGBQUAD* colors = GetPalette();
		if (colors==0) return 0;
	}

	int cx = bounds.right-bounds.left;
	int cy = bounds.bottom-bounds.top;
	int iDecimate = CalculateDecimation(cx, cy, xDim, yDim);

	int dx, dy;
	RECT drawRect;
	::CopyRect(&drawRect,&bounds);
	if (iDecimate>0) {
		dx = (cx-xDim)/2;
		dy = (cy-yDim)/2;
		::OffsetRect(&drawRect,dx,dy);
		PixelFill(hDC, drawRect);
	} else {
		dx = (cx+xDim*iDecimate)/2;
		dy = (cy+yDim*iDecimate)/2;
		::OffsetRect(&drawRect,dx,dy);
		RectFill(hDC, drawRect, -iDecimate);
	}
	return 1;
}



HGLOBAL WholeGridBase::CreatePackedDIB(RECT& bounds)
{
	HGLOBAL pbmi;
	RECT tRect;
	if (colors==0) {
		RGBQUAD* colors = GetPalette();
		if (colors==0) return 0;
	}
	if (iMaxVal==0) iMaxVal = GetPaletteSize();

	int iDecimate = CalculateDecimation(bounds.right-bounds.left,
			bounds.bottom-bounds.top,xDim,yDim);
	if (iDecimate>=1) {
		::SetRect(&tRect,0,0,xDim,yDim);
		pbmi = CreateExactSizeDIB(tRect, 1);
	} else {
		::SetRect(&tRect,0,0,-iDecimate*xDim,-iDecimate*yDim);
		pbmi = CreateExactSizeDIB(tRect, iDecimate);
	}
	return pbmi;
}

int WholeGridBase::GetPreferredSize(RECT& bounds, RECT& preferred)
{
	int iDecimate = CalculateDecimation(bounds.right-bounds.left,
			bounds.bottom-bounds.top,xDim,yDim);
	if (iDecimate>=1) {
		::SetRect(&preferred,0,0,xDim,yDim);
	} else {
		::SetRect(&preferred,0,0,-iDecimate*xDim,-iDecimate*yDim);
	}
	return 1;
}


HGLOBAL WholeGridBase::CreateExactSizeDIB(RECT& bounds, int iDecimate)
{
	int cWidth = bounds.right-bounds.left;
	int cHeight = bounds.bottom-bounds.top;
	int bitCount = 32;
	int pitch = ((cWidth * bitCount + 31) / 32 ); // from Petzold
	int rowLength = 4*pitch ; // from Petzold
	int dataSize = rowLength*cHeight;
	// should be iMaxVal-1?
	int colorTableSize = (iMaxVal-1)*sizeof(RGBQUAD);
	int totalSize = sizeof(BITMAPINFO)+colorTableSize+dataSize;

	HGLOBAL hbmi = ::GlobalAlloc(GHND | GMEM_SHARE, totalSize/4);
	if (hbmi==0) return 0;
	PBITMAPINFO pbmi = (PBITMAPINFO) ::GlobalLock(hbmi);
	if (pbmi==0) return 0;

	::memset((void *)pbmi,0,totalSize);
	PBITMAPINFOHEADER pbmih = &(pbmi->bmiHeader);
	pbmih->biSize = sizeof(BITMAPINFOHEADER);
	pbmih->biWidth = cWidth;
	pbmih->biHeight = cHeight;
	pbmih->biPlanes = 1;
	pbmih->biBitCount = 32;
	pbmih->biCompression = BI_RGB;
	pbmih->biSizeImage = dataSize;
	pbmih->biXPelsPerMeter = 2835;
	pbmih->biYPelsPerMeter = 2835;
	pbmih->biClrUsed = 0;
	pbmih->biClrImportant = 0;

	::memcpy((void *)pbmi->bmiColors,colors,iMaxVal*sizeof(RGBQUAD));

	DrawToVoid(pbmi->bmiColors+iMaxVal, iDecimate, pitch);

	::GlobalUnlock(hbmi);
	return hbmi;
}


void WholeGridBase::DrawToVoid(RGBQUAD* buf, int iDecimate, int pitch)
{
	int cPix = iDecimate;
	RGBQUAD* pcol;
	int i,j;
	int cx, cy; // pixel size of blocks.
	for (j=0; j<yDim; j++) {
		for (cy=0; cy<cPix; cy++) {
			for (i=0; i<xDim; i++) {
				pcol = &colors[data[i+xDim*(yDim-j-1)]];
				for (cx=0; cx<cPix; cx++) {
					::memcpy((void *)&buf[i*cPix+cx],(void*)pcol,sizeof(RGBQUAD));
				}
			}
			buf += pitch;
		}
	}
}

void WholeGridBase::PixelFill(HDC hDC, RECT& rect)
{
	int ret;
	RGBQUAD *color;
	int i, j;
	for (j=0; j<yDim; j++) {
		for (i=0; i<xDim; i++) {
			color = &colors[data[i+(yDim-j-1)*xDim]];
			ret = ::SetPixel(hDC,rect.left+i,rect.top+j,
				RGB(color->rgbRed,color->rgbGreen,color->rgbBlue));
			if (ret==-1) {
				ret = 20;
			}
		}
	}
}

void WholeGridBase::RectFill(HDC hDC, RECT& bounds, int iDecimate)
{
	int i;
	// If we are going to adjust the bounds to do centering,
	// we should do it here.
	HBRUSH *brushes = BrushesFromPalette();
	if (brushes) {
		BrushRectFill(hDC, bounds, iDecimate, brushes);
		for (i=0; i<GetPaletteSize(); i++)
			::DeleteObject(brushes[i]);
		delete[] brushes;
	} else {
		PlainRectFill(hDC, bounds, iDecimate);
	}
}

void WholeGridBase::BrushRectFill(HDC hDC, RECT& bounds, int iDecimate,
								  HBRUSH *brushes)
{
	int cPix = iDecimate;
	int i,j;
	RECT rect;
	for (j=0; j<yDim; j++) {
		for (i=0; i<xDim; i++) {
			::SetRect(&rect,i*cPix+bounds.left,
				j*cPix+bounds.top,(i+1)*cPix+bounds.left,(j+1)*cPix+bounds.top);
			::FillRect(hDC,&rect,brushes[data[i+(yDim-j-1)*xDim]]);
		}
	}
}

void WholeGridBase::PlainRectFill(HDC hDC, RECT& bounds, int iDecimate)
{
	int i,j;
	int cPix = iDecimate;
	RGBQUAD* pcol;
	RECT rect;
	for (j=0; j<yDim; j++) {
		for (i=0; i<xDim; i++) {
			::SetRect(&rect,i*cPix+bounds.left,
				j*cPix+bounds.top,(i+1)*cPix+bounds.left,(j+1)*cPix+bounds.top);
			pcol = &colors[data[i+(yDim-j-1)*xDim]];
			HBRUSH color = ::CreateSolidBrush(
				RGB(pcol->rgbRed,pcol->rgbGreen,pcol->rgbBlue));
			::FillRect(hDC,&rect,color);
			::DeleteObject(color);
		}
	}
}

HBRUSH* WholeGridBase::BrushesFromPalette()
{
	if (colors==0) GetPalette();
	if (colors==0) return 0;
	if (iColors>256) return 0; // Don't even try.

	HBRUSH* brushes = new HBRUSH[iColors];
	if (brushes==0) return 0;

	RGBQUAD* pcol;
	int i, j;
	j = -1;
	for (i=0; i<iColors; i++) {
		pcol = &colors[i];
		brushes[i] = ::CreateSolidBrush(
			RGB(pcol->rgbRed,pcol->rgbGreen,pcol->rgbBlue));
		if (brushes[i]==0) {
			j = i;
			break;
		}
	}
	if (j>=0) {
		for (i=0; i<j; i++)
			::DeleteObject(brushes[i]);
		return 0;
	}
	return brushes;
}

int WholeGridBase::NumToBits(int n)
{
	int i = 0;
	int j=0;
	while (i<32) {
		if (n & 1) j = i;
		n >>= 1;
		i++;
	}
	// Now j is the number of bits.
	// Find 1,2,4,8,16,32.
	i=0;
	while (i<32) {
		if ((1 << i)>j) break;
		i++;
	}
	return (1 << i);
}
