#ifndef _HYSTERESIS_SIMULATION_H_
#define _HYSTERESIS_SIMULATION_H_

#include "DMFirstHeader.h"
#include <iostream>
#include <vector>
#include "DiscreteAvalanche.h"
#include "SpinIndex.h"
#include "FlipListener.h"
#include "SpinLattice.h"
#include "AvalancheListener.h"
#include "Serializable.h"
#include "AvalancheAlgorithm.h"

class AvalancheStart;
class FixedFieldAvalanche;
class FixedFieldAvalancheStart;
class GaussianInitializer;
class SweepLogic;
class Rand;
class OnSpanning;
class CheckSpanning;
class SizeDistribution;
class ProbabilisticAvalanche;
class ProbabilisticStart;
class CorrelationFunctionObserver;
class MagnetizationObserver;
class ColoringObserver;
class TimeSeries;
class BatchTimer;
class NetCDFCheckpointer;
class Tracer;

//! This is the master class which owns all simulation elements.
/** Saving this should save the whole simulation.
  */
class HysteresisSimulation : public Serializable
{
	int dimCnt;
	SpinCoord iLength;
	typedef std::vector<SpinCoord> DimensionsVector;
	DimensionsVector sideLength;

	int iSeed;
	double dDisorder;
	long lNumberOfMinutes; // don't serialize this.
	long traceSecondCnt; // neither this.
	long checkpointMinuteCnt; // nor this.

	FixedFieldAvalanche* avalancheAlgorithm;
	FixedFieldAvalancheStart* avalancheStarter;
	ProbabilisticAvalanche* pAvalanche;
	ProbabilisticStart* pStart;
	GaussianInitializer* initializer;
	SpinLattice* spinLattice;
	SweepLogic* sweepLogic;
	Rand* rand;
	CheckSpanning* spanning;
	OnSpanning* onSpanning;
	SizeDistribution* sizeDistribution;
	CorrelationFunctionObserver *correlationFunction;
	MagnetizationObserver *magnetization;
	ColoringObserver* coloringObserver;
	TimeSeries *timeSeries;
	BatchTimer *batchTimer;
	Tracer* tracer;
	NetCDFCheckpointer* checkpointer;

	bool bAverageSizeDistributionOn;
	bool bAverageCorrelationFunctionOn;

	bool bOwnBatchTimer;
	bool bFirstAvalanche;
public:
	//! Constructor.
	HysteresisSimulation();
	//! Destructor.
	~HysteresisSimulation();

	//! Sets number of dimensions.
	void SetDimension( int dimCnt );

	//! Set the length of all sides to one value.
	void SetSideLength( SpinCoord length );
	
	//! Set the length of a side of the lattice.
	void SetEachSideLength( int dimIdx, SpinCoord length );
	
	//! Set the random seed
	void SetSeed( int iSeed );

	//! Increase seed by one.
	void IncrementSeed();

	//! Set the amount of disorder
	void SetDisorder( double dDisorder );

	void EnableSpanning();
	void DisableSpanning();
	void EnableSpanningWriter();
	void DisableSpanningWriter();
	void EnableCorrelationFunction( bool bCumulative );
	void DisableCorrelationFunction( );
	void EnableSizeDistribution( bool bCumulative );
	void DisableSizeDistribution( );
	void EnableColoringObserver( );
	void DisableColoringObserver( );
	void EnableMagnetizationObserver( );
	void DisableMagnetizationObserver( );
	void EnableTimeSeries( );
	void DisableTimeSeries( );

	//! How long from now do we need to be done?
	void SetEndTime( long lNumberOfMinutes );

	//! How many seconds until between traces?
	void SetTraceFrequency( long traceSecondCnt );

	//! How long between checkpoints
	void SetCheckpointFrequency( long checkpointMinuteCnt );

	//! Initialize the simulation.
	void Initialize();

	//! Check that everything is hooked up.
	bool Verify();

	//! Tell it to run. Returns whether to quit.
	AvalancheAlgorithm::Status Run( int RunCnt = 0 );

	//! Get the coloring observer if there is one.
	ColoringObserver* GetColoringObserver() const {
		return coloringObserver; }

	//! Show results
	void SaveResults() const;
	void SaveAverageResults() const;
	std::ostream& WriteState( std::ostream& s );

	/*! Serializable interface */
	//@{
	virtual void Save( DMWriter *writer ) const;
	virtual void Load( DMReader *reader );
	virtual const std::string GetType() const;
	//@}
	static const std::string sType;
private:
	// These are cached for my use.
	AvalancheAlgorithm* avalanche;
	AvalancheStart* start;

	void EnableBatchTimer();
	void DisableBatchTimer();
	void EnableSimpleTracer();
	void EnableMagnetizationTracer();
	void DisableTracer();
	void EnableCheckpointer();
	void DisableCheckpointer();

	void InitializeObservers();
	void CreateSubjectComponents();
	void SetSubjectResetParameters();
	void SetSubjectNoResetParameters();
	void SetObserverResetParameters();
	void SetObserverNoResetParameters();
	void MakeSubjectConnections();
	void MakeObserverConnections();
	void AttachObservers();
	void DeleteSubjectComponents();
	void DeleteObserverComponents();
	void RemoveFromAvalancheObservers( AvalancheListener* observer);
	void RemoveFromFlipObservers( FlipListener* observer);
	//	void RemoveListeners();
	void BeginSweep( bool bKeepData = false );
	void ZeroSubjectPointers();
	void ZeroObserverPointers();
	bool bSerialized;
	bool bSubjectInitialized;
	bool bObserverInitialized;
};

#endif
