#ifndef _MAGNETIZATION_OBSERVER_H_
#define _MAGNETIZATION_OBSERVER_H_

#include "DMFirstHeader.h"
#include "DiscreteAvalanche.h"
#include "AvalancheListener.h"
#include "Serializable.h"
#include "SpinIndex.h"
#include "SpinLattice.h"
#include "AvalancheAlgorithm.h"

//! Keeps track of the magnetization at each step.
class MagnetizationObserver : public AvalancheListener, public Serializable
{
	double* H;
	double* M;
	
	int numPoints;
	int curPoint;
	
	// Some of these may not be properly named.
	// These are name changes from the original code done
	// without full understanding.
	double MagStepSize;
	double Hint;
	double LastMagnetization;
	double OutputMagStep;

	SpinLattice* lattice;
	AvalancheAlgorithm* avalanche;
public:
	//! Constructor.
	MagnetizationObserver();

	//! Destructor.
	~MagnetizationObserver();

	//! Set how many timesteps we can handle.
	void SetNumMeasurements( int numPoints = 1000 );
	
	//! Give us a pointer to the lattice.
	void SetLattice( SpinLattice* lattice );

	//! Give us a pointer to the algorithm
	void SetAlgorithm( AvalancheAlgorithm* avalanche );

	//! Call this before you try to use this class.
	void Initialize();

	//! Call this at the start of a new sweep.
	void BeginSweep( bool bKeep = false);
		
	//! This is called after every avalanche.
	void AvalancheCompleted(AvalancheAlgorithm* avalanche,
		double h);
	
	double GetLastMagnetization() const { return LastMagnetization; }

	//! Return the field at a point.
	double GetH(int i) const { return H[i]; }
	//! Return the magnetization at a point.
	double GetM(int i) const { return M[i]; }
	
	/*! Serializable interface */
	//@{
	void Save( DMWriter *writer ) const;
	void Load( DMReader *reader );
	const std::string GetType() const;
	//@}
	
	friend std::ostream& operator<<( std::ostream& s,
		const MagnetizationObserver& r );
private:
	void DeleteArrays( void );
	void AllocateArrays( void );
	void ZeroArrays( void );

	SpinIndex N; // cached from lattice.
	bool bSerialized;
	enum { VERIFY_NUMPOINTS=0x1, VERIFY_LATTICE=0x2, VERIFY_AVALANCHE=0x4 };
	int iVerify;
};


/*! This method stores the magnetization at a given field.  If the field has
	jumped significantly, it interpolates from the last point.

	\bugs While the math in this is simple, it is nearly impossible to figure out.
	*/
inline void
MagnetizationObserver::AvalancheCompleted(AvalancheAlgorithm* avalanche,
													  double CurrentField)
{
	double CurrentMagnetization = ((double)avalanche->GetM()) / N;
	
	const double epsilon = 1e-14;
	
	if (CurrentMagnetization-OutputMagStep < MagStepSize)
	{
		Hint += CurrentField * (CurrentMagnetization-LastMagnetization);
	}
	else
	{
		Hint += CurrentField * (OutputMagStep + MagStepSize - LastMagnetization);
		
		H[curPoint] = Hint / MagStepSize;
		M[curPoint] = OutputMagStep + 0.5 * MagStepSize;
		
		OutputMagStep += MagStepSize;
		curPoint++;
		
		while (CurrentMagnetization-OutputMagStep >= MagStepSize-epsilon)
		{
			H[curPoint] = CurrentField;
			M[curPoint] = OutputMagStep + 0.5 * MagStepSize;
			OutputMagStep += MagStepSize;
			curPoint++;
		}
		Hint = CurrentField * (CurrentMagnetization-OutputMagStep);
	}
	LastMagnetization = CurrentMagnetization;
}


#endif //_MAGNETIZATION_OBSERVER_H_
