#ifndef _POW2_SPIN_LATTICE_H_
#define _POW2_SPIN_LATTICE_H_

#include "DMFirstHeader.h"
#include <vector>
#include "DiscreteAvalanche.h"
//#include "stl_bvector.h"
#include "SpinIndex.h"
#include "Serializable.h"

// std::_Bvector
// stl_bvector
typedef std::_Bvector vector_bool;
// typedef bit_vector vector_bool;

//! Rectangular lattice in N dimensions of up or down spins.

/*!	Matt Kuntz's code defined an enum representing UP or DOWN, but it
	was clear that the integer values +1 and -1 were very often useful
	in his code, so we're taking that experience and making this class
	specify that you treat the values of the
	spins as +1 or -1.

	The Save and Load are just a hair from being done.  The hard part,
	saving the spins, works.  The easy part, saving other internal variables,
	has changed and needs to be fixed.

	You can switch this class to handle more than 2^32-1 spins by 
	changing the typedef of vector_bool to the stl_vector version.
	I tool the Gnu STL vector<bool> and hacked it so it will
	address as many spins as it can fit in 2^32 bytes.  That's eight
	times as many, and we can use them.

	It's only a useful option if you also switch the SpinIndex
	to a 64 bit integer (ULONGLONG in MSVC++).

	\author ajd27
	\date 15 Aug 2000
*/
class Pow2SpinLattice : public Serializable, private vector_bool
{
	int DimensionCnt;
	SpinCoord* Dimensions;
	SpinIndex SpinCnt;
public:
//	typedef SpinIndex* iterator;
	//! Constructor
	Pow2SpinLattice();

	//! Destructor.
	~Pow2SpinLattice();

	//! Set number of dimensions.
	void SetDimension( int NumberOfDimensions );

	//! Set the length of all sides.
	bool SetLengths( const SpinCoord* Dimensions );

	//! Identical to Get.  This returns the value of the given spin.
	int operator()( const SpinIndex iPos ) const;

	//! Return the value of a given spin.
	int Get( const SpinIndex iPos ) const;

	//! Flip a spin up or down.
	void flip( const SpinIndex iPos, int orientation );

	//! Flip all spins down
	void SetAllDown( );

	//! How many up neighbors does this spin have?
	int NeighborsUp( const SpinIndex iPos ) const;

	//! Who are the neighbors of this spin?
	void GetNeighbors( const SpinIndex iPos, SpinIndex* const neighbors ) const;

	//! Get the coordinates of this index.
	void GetCoords( SpinIndex spinPos, SpinCoord *const coords ) const;

	//! Get the index of these coordinates.
	SpinIndex GetIndex( const SpinCoord *const coords ) const;

	//! Return the length of the whole array.
	SpinIndex GetLength() const;

	const SpinCoord* GetSideLengths() const;

	//! What is the coordination number of this lattice?
	SpinIndex GetCoordinationNumber() const;

	//! Return the spatial dimension of the lattice.
	int GetDimension() const { return this->DimensionCnt; }

	//! How many spins are up?
	SpinIndex GetTotalNumberUp() const;

	//! This saves the Pow2SpinLattice.
	virtual void Save(DMWriter *writer) const;
	//! Read in new values for this Pow2SpinLattice.
	virtual void Load(DMReader *reader);
	//! Return the signature type of this spinlattice.
	virtual const std::string GetType() const;
	//! Consider this private.  Use GetType instead.
private:
	SpinIndex* _neighbors; // internal cache variable.

	//! Lowest bit adds one to a coordinate in the SpinIndex.
	SpinIndex* LowestBit;
	//! BitMask extracts a coordinate from a SpinIndex, still shifted.
	SpinIndex* BitMask;
	//! BitShift is how many times to shift a bit to the right.
	int* BitShift;

	//! Version of SetLengths which doesn't check for errors.
	bool _SetLengths( const SpinCoord* Dimensions );
	void SetStrides();

	//! Allocates room for zero spins.  Default state.
	void ZeroSpinArray();

	//! Checks that dimension array is well-formed.
	bool CheckDimensionArray( const SpinIndex* Dimensions );
	int* WriteSlice( SpinIndex start, SpinIndex finish, int& dwLen ) const;
	void ReadSlice( SpinIndex start, SpinIndex finish, int*data );
	void AllocateArrays();
	void DeleteArrays();
	void ZeroArrays();
};

inline SpinIndex
Pow2SpinLattice::GetCoordinationNumber() const
{
	return 2*this->DimensionCnt;
}


inline SpinIndex
Pow2SpinLattice::GetLength() const
{
	return this->SpinCnt;
}


inline const SpinCoord*
Pow2SpinLattice::GetSideLengths() const
{
	return Dimensions;
}

/*! The fastest way to turn a bool into an int.
  */
inline int Pow2SpinLattice::operator()( const SpinIndex iPos ) const
{
	return ((*this)[iPos]) ? 1 : -1;
}

inline int Pow2SpinLattice::Get( const SpinIndex iPos ) const
{
	return ((*this)[iPos]) ? 1 : -1;
}

inline void Pow2SpinLattice::flip( const SpinIndex iPos, int orientation )
{
	(*this)[iPos] = ( orientation == 1 );
}

/*	If you unroll this loop, the code actually becomes much slower
	because GetNeighbors accesses a different set of memory than the
	values of the spins.
	*/
inline int
Pow2SpinLattice::NeighborsUp( const SpinIndex iPos ) const
{
	int nup = 0;
	GetNeighbors( iPos, this->_neighbors );

	int NeighborIdx;
	for ( NeighborIdx=0; NeighborIdx<GetCoordinationNumber(); NeighborIdx++) {
		SpinIndex ANeighbor = _neighbors[NeighborIdx];
		nup += ((*this)[ANeighbor]) ? 1 : 0;
	}

	return nup;
	
}

/*! 
	@param iPos position of the spin in the lattice
	@param neighbors Array in which the neigbors will be returned.
	Must be able to accept Coordination Number elements.
  */
/*	Combining the loop to get the coordinates and that to find the
	neighbors saved 10 % of the time over not having the loops combined.

	Using a lookup-table instead of if-thens in the inner loop increased
	the execution time by 10-50%.  The variation in slowdown is due to
	cache considerations.

	Switching the order of the tests so that the most common was first
	was not measurably different.  Even a little slower.

	It may be possible to eliminate the LowestBit array from this
	routine because it is just the BitMask+1.  That would mean fewer
	memory accesses.
	*/
inline void
Pow2SpinLattice::GetNeighbors
	(
	const SpinIndex iPos,
	SpinIndex *const neighbors
	) const
{
	int DimIdx;
	for ( DimIdx=0; DimIdx < this->DimensionCnt; DimIdx++ ) {

		// Find the coordinate along this side.
		SpinIndex masked = (iPos & this->BitMask[DimIdx]);
		SpinIndex maskRemoved = ( iPos & ~this->BitMask[DimIdx]);

		SpinIndex UpperNeighbor = masked + LowestBit[DimIdx];
		UpperNeighbor -= (UpperNeighbor & LowestBit[DimIdx+1]);
		neighbors[2*DimIdx] = maskRemoved | UpperNeighbor;

		SpinIndex LowerNeighbor = masked - LowestBit[DimIdx];
		LowerNeighbor += ( LowerNeighbor & LowestBit[DimIdx+1] );
		neighbors[2*DimIdx+1] = maskRemoved | LowerNeighbor;
	}
}

inline void
Pow2SpinLattice::GetCoords( SpinIndex iPos, SpinCoord *const coords ) const
{
	int DimIdx;
	for( DimIdx=0; DimIdx < this->DimensionCnt; DimIdx++ ) {
		SpinIndex masked = (iPos & this->BitMask[DimIdx]);
		coords[DimIdx] = ( masked >> this->BitShift[DimIdx] );
	}
}

inline SpinIndex
Pow2SpinLattice::GetIndex( const SpinCoord *const coords ) const
{
	SpinIndex loc = 0;

	for ( int dimIdx = 0; dimIdx < this->DimensionCnt; dimIdx++ )
		loc |= ( coords[ dimIdx ] << this->BitShift[ dimIdx ] );

	return loc;
}

#endif //_POW2_SPIN_LATTICE_H_
