#include "DMFirstHeader.h"
#include <vector>
#include <iostream>
#include "ProbabilisticAvalanche.h"
#include "SpinLattice.h"
#include "GaussianInitializer.h"
#include "FailedVerify.h"
#include "DMException.h"
#include "NotEnoughMemory.h"
#include "DMWriter.h"
#include "DMReader.h"
#include "Rand.h"
#include "SerialBuilder.h"
#include "BatchTimer.h"
#include "DMAllocator.h"
#include "DMLastHeader.h"

const std::string sType =
	std::string("ProbabilisticAvalanche");

/*! This static object permits the DMReader to create an empty ProbabilisticAvalanche. */
const SerialBuilder<ProbabilisticAvalanche> *_creator = new
	SerialBuilder<ProbabilisticAvalanche>(sType);

ProbabilisticAvalanche::ProbabilisticAvalanche()
{
	ZeroPointers();

	lattice = 0;
	initializer = 0;
	rand = 0;
	batchTimer = 0;

	iVerify = 0;
	bSerialized = false;
	bInCascade = false;
}

ProbabilisticAvalanche::~ProbabilisticAvalanche()
{
	DeleteArrays();
}

/** You have to set the lattice before you can call Initialize.
  */
void ProbabilisticAvalanche::SetLattice( SpinLattice *const lattice )
{
	ASSERT( 0 != lattice );
	if (lattice == 0) throw 
		DMException( "ProbabilisticAvalanche","SetLattice", "Passed a null pointer.");
	this->lattice = lattice;
}

/** You have to set the initializer before you can call initialize.
  */
void ProbabilisticAvalanche::SetInitializer( GaussianInitializer *initializer )
{
	ASSERT( 0 != initializer );
	if (initializer == 0) throw 
		DMException( "ProbabilisticAvalanche","SetInitializer", "Passed a null pointer.");
	this->initializer = initializer;
}

void ProbabilisticAvalanche::SetRand( Rand* rand )
{
	ASSERT( 0 != rand );
	if (rand == 0) throw 
		DMException( "ProbabilisticAvalanche","SetRand", "Passed a null pointer.");
	this->rand = rand;
}

/*! Only set a serializer if you want this to print trace information
	or checkpoint.  It runs fine without it.
	*/
void ProbabilisticAvalanche::SetBatchTimer( BatchTimer *batchTimer )
{
	ASSERT( 0 != batchTimer );
	if (0 == batchTimer) throw
		DMException( "ProbabilisticAvalanche","SetBatchTimer","Passed a null pointer");

	this->batchTimer = batchTimer;
}

/*! You must call SetLattice(), SetRand() and 
	SetSerializer() before calling this.  You may call Initialize() as
	many times as you like, and it will reset the internal state.
	*/
void ProbabilisticAvalanche::Initialize()
{
	ASSERT(0 != lattice);
	ASSERT(0 != rand);
	ASSERT(0 != initializer);
	if ( 0 == lattice ) throw
		DMException( "ProbabilisticAvalanche","Initialize",
		"Need to call SetLattice before calling Initialize.");
	if ( 0 == initializer ) throw
		DMException( "ProbabilisticAvalanche","Initialize",
		"Need to call SetInitializer before calling Initialize.");
	if ( 0 == rand ) throw
		DMException( "ProbabilisticAvalanche","Initialize",
		"Need to call SetRand before calling Initialize.");

	SetCachedVariables();
	if (bSerialized) {
		// Time-consuming, but smart if we are coming up
		// from serializing.
		SpinIndex totalUp = lattice->GetTotalNumberUp();
		ASSERT( M == -N+2*totalUp );
		if ( M != -N+2*totalUp ) throw
			DMException( "ProbabilisticAvalanche","Initialize",
			"The lattice magnetization doesn\'t agree with the avalanche.");
	} else {
		DeleteArrays(); // Just in case this is called twice.
		AllocateArrays();
	}
	starter.SetLattice( lattice );
	starter.SetInitializer( initializer );
	starter.SetRand( rand );
	starter.SetDownSpins( psNUp );
	if (batchTimer) starter.SetBatchTimer( batchTimer );
	starter.Initialize();

	bSerialized = false;
	iVerify |= INITIALIZED;
}

void ProbabilisticAvalanche::BeginSweep( bool bKeepData )
{
	this->Verify();

	M = -N;
	InitializeNUp();

	starter.BeginSweep();
}

/*! This checks that you called everything you needed to call before 
	you get started.
	*/
void ProbabilisticAvalanche::Verify( void ) const
{
	ASSERT( 0 != (iVerify&INITIALIZED) );
	if ( (iVerify&INITIALIZED) == 0) throw
		FailedVerify( "ProbabilisticAvalanche","Didn\'t call Initialize." );

	starter.Verify();
}

void ProbabilisticAvalanche::Save( DMWriter *writer ) const
{
	std::vector<int> shape;

	writer->PutInt( M, "magnetization" );
	int iPrep = (bInCascade) ? 1 : 0;
	writer->PutInt( iPrep, "IsInCascade" );
	writer->PutInt( spinFlipQueue.size(), "queueSize" );
	// Save these two as a check when we read it in again.
	if (spinFlipQueue.size()>0) {
		writer->PutInt( spinFlipQueue.front(), "queueFront" );
		writer->PutInt( spinFlipQueue.back(), "queueBack" );
	} else {
		writer->PutInt( 0, "queueFront" );
		writer->PutInt( 0, "queueBack" );
	}

	shape.push_back( (Z+1)*sizeof(SpinIndex)/sizeof(int) );
	writer->PutIntArray(psNUp,shape,"psNUp");

	shape[0] = Z;
	writer->PutDoubleArray(dProbabilityToFlip, shape, "dProbabilityToFlip");

	shape[0] = spinFlipQueue.size();
	ASSERT( shape[0] >= 0 );
	if (shape[0]>0) {
		int *piQueue = new int[shape[0]];
		std::copy(spinFlipQueue.begin(),spinFlipQueue.end(), piQueue );
		writer->PutIntArray( piQueue, shape, "spinFlipQueue" );
		delete[] piQueue;
	}
}

void ProbabilisticAvalanche::Load( DMReader *reader )
{
	// If you call this, you have to initialize again.
	iVerify = 0;
	bSerialized = true;
	DeleteArrays();

	M = reader->GetInt( "magnetization" );
	int iPrep = reader->GetInt( "IsInCascade" );
	this->bInCascade = (iPrep == 1);
	int queueSize = reader->GetInt( "queueSize" );
	int iFront = reader->GetInt( "queueFront" );
	int iBack = reader->GetInt( "queueBack" );

	ASSERT( queueSize >= 0 );
	std::vector<int> shape;
	int* piData = reader->GetIntArray( shape, "psNUp");
	// iCoordination is actually the coordination number + 1.
	int iCoordination = shape[0]*sizeof(int)/sizeof(SpinIndex);
	psNUp = (SpinIndex*) piData;
	dProbabilityToFlip = reader->GetDoubleArray(shape, "dProbabilityToFlip");

	if (queueSize>0) {
		spinFlipQueue.clear();
		int *piQueue = reader->GetIntArray( shape, "spinFlipQueue" );
		ASSERT( piQueue != 0 );
		ASSERT( shape[0] != 0 );
		std::back_insert_iterator<std::deque<SpinIndex> > bi(spinFlipQueue);
		std::copy(piQueue,piQueue+shape[0], bi );
		delete[] piQueue;
	}

	// Check it read in correctly.
	if (queueSize > 0) {
		ASSERT( iFront == spinFlipQueue.front() );
		ASSERT( iBack == spinFlipQueue.back() );
		if ( (spinFlipQueue.front() != iFront) ||
			(spinFlipQueue.back() != iBack) ) {
			throw DMException("ProbabilisticAvalanche","Load",
				"The spinFlipQueue did not read properly.");
		}
	}
}

const std::string ProbabilisticAvalanche::GetType() const
{
	return sType;
}

/*! Unless this class has a correct pointer to the lattice,
	you'll get a null pointer.
	*/
const SpinLattice* ProbabilisticAvalanche::GetLattice() const
{
	return lattice;
}

/*! We won't give you our pointer to down spins unless you
	have called Initialize();
	*/
const SpinIndex* ProbabilisticAvalanche::GetDownSpins() const
{
	Verify();
	return psNUp;
}

inline AvalancheAlgorithm::Status
ProbabilisticAvalanche::FindNextStart( double* H, SpinIndex *iFirstSpin )
{
	if ( bInCascade ) return AvalancheAlgorithm::SUCCESS;
	return starter.FindNextStart( H, iFirstSpin );
}


inline AvalancheAlgorithm::Status
ProbabilisticAvalanche::Cascade( double H, SpinIndex iFirstSpin )
{
	// Put the first spin on the queue and set probabilities
	// Skip this if we were serialized in the middle of a cascade.
	// Could just mark this by serializing bInCascade.
	//	TRACE("ProbabilisticAvalanche::Cascade Prepping flip.\n");
	TRACE(".");
	// This boolean is a marker that we are in a cascade in case
	// a function we call during the cascade should try to serialize us.
	// We'll be able to pick up where we left off.
	if ( !bInCascade ) {
		bInCascade = true;
		// Initialize dProbabilityToFlip[nup]
		this->SetProbabilityToFlip( H );
		this->InitializeSpinFlipQueue( iFirstSpin );
	} else {
		bInCascade = true;
	}

	// Actually flip the spins.
	//	TRACE("ProbabilisticAvalanche::Cascade Cascading queue.\n");
	Status CascadeStatus = CascadeQueue();
	bInCascade = false;
	return CascadeStatus;
}


// ***************** functions private below this line ******************* //

inline void
ProbabilisticAvalanche::InitializeSpinFlipQueue( SpinIndex iFirstSpin )
{
	static const SpinIndex ENDOFSHELL = -1;

	spinFlipQueue.push_back( iFirstSpin );
	spinFlipQueue.push_back( ENDOFSHELL );
}

/*! The only new information passed to this function is in the
	spinFlipQueue.  If that's set up, this does the rest.
	*/
AvalancheAlgorithm::Status
ProbabilisticAvalanche::CascadeQueue( )
{
	static const SpinIndex ENDOFSHELL = -1;
	SpinIndex neighbors[10];
	int i;
	SpinIndex curLoc;
	SpinIndex nextLoc;
	SpinIndex nFlipped = 0;
	int coords[3];
	int nup;
	double dRand;

	while (!spinFlipQueue.empty()) {
		curLoc = spinFlipQueue.front();

		if (curLoc == ENDOFSHELL) {
			if (batchTimer) {
				batchTimer->TimeSlice();
				if ( batchTimer->IsTimeToQuit() ) {
					return AvalancheAlgorithm::TIMEDOUT;
				}
			}
			spinFlipQueue.pop_front();
			if (!spinFlipQueue.empty()) spinFlipQueue.push_back(ENDOFSHELL);

		} else {
			spinFlipQueue.pop_front();

			if (lattice->Get(curLoc) != 1) {
				nup = lattice->NeighborsUp( curLoc );
				--(psNUp[nup]);

				lattice->flip(curLoc,1);
				++nFlipped;
				M += 2;

				lattice->GetCoords( curLoc, coords );
				this->Notify( lattice, coords, lattice->GetDimension() );

				lattice->GetNeighbors(curLoc, neighbors);
				for (i=0; i<Z; i++) {
					nextLoc = neighbors[i];
					if (lattice->Get( nextLoc ) == -1 ) {
						nup = lattice->NeighborsUp( nextLoc )-1;
						--(psNUp[nup]);
						++(psNUp[nup+1]);
						
						dRand = rand->uniform();
						if ( dRand <= dProbabilityToFlip[nup] )
							spinFlipQueue.push_back(nextLoc);
					}
				}
			}
		}
	}
	//	TRACE( "ProbabilisticAvalanche::CascadeQueue Flipped %i spins this cascade.\n",
	//		nFlipped );

	return AvalancheAlgorithm::SUCCESS;
}


void ProbabilisticAvalanche::SetCachedVariables()
{
	Z = lattice->GetCoordinationNumber();
	N = lattice->GetLength();
}

/*! This initializes the array holding how many spins have N up neighbors,
	psNUp.  This assumes we are starting with a lattice where all spins 
	point down.
	*/
void ProbabilisticAvalanche::InitializeNUp()
{
	SpinIndex i = 0;
	//std::cout << "Starting to initialize NUp." << std::endl;
	for (i=0; i<=Z; i++) psNUp[i] = 0;

	//	for (i=0; i<N; i++)
	//		++(psNUp[lattice->NeighborsUp( i )]);

	psNUp[0] = N;

	//std::cout << "Finished initializing NUp." << std::endl;
}

void ProbabilisticAvalanche::SetProbabilityToFlip( double dField )
{
	int iUpNeighbors;
	double hLoc;

	for (iUpNeighbors=0; iUpNeighbors<Z; iUpNeighbors++) {
		hLoc = dField+2*iUpNeighbors-Z;
		dProbabilityToFlip[iUpNeighbors] =
			1-initializer->intrhoc(hLoc+2)/initializer->intrhoc(hLoc);
	}
}

void ProbabilisticAvalanche::DeleteArrays()
{
	if (psNUp) delete[] psNUp;
	if (dProbabilityToFlip) delete[] dProbabilityToFlip;
	ZeroPointers();
}

void ProbabilisticAvalanche::AllocateArrays()
{
	ASSERT( Z > 0 );
	try {
		psNUp = DMAllocate<int>( Z+1 );
		dProbabilityToFlip = DMAllocate<double>( Z );
	} catch ( std::bad_alloc ) {
		DeleteArrays();
		throw NotEnoughMemory("ProbabilisticAvalanche","AllocateArrays",
			"Allocating two small arrays.");
	}
}

void ProbabilisticAvalanche::ZeroPointers()
{
	psNUp = 0;
	dProbabilityToFlip = 0;
}