#ifndef _PROBABILISTIC_AVALANCHE_H_
#define _PROBABILISTIC_AVALANCHE_H_

#include "DMFirstHeader.h"
#include "DiscreteAvalanche.h"
#include "SpinIndex.h"
#include "SpinLattice.h"
#include "AvalancheAlgorithm.h"
#include "Serializable.h"
#include "ProbabilisticStart.h"

class GaussianInitializer;
class Rand;
class BatchTimer;

//! Responsible for cascading an avalanche of spins using only the distribution of fixed fields.
class ProbabilisticAvalanche : public AvalancheAlgorithm, public Serializable
{
	ProbabilisticStart		starter;

	SpinLattice*			lattice;
	GaussianInitializer*	initializer;
	Rand*					rand;
	BatchTimer*				batchTimer;

	//! psNUp is the number of down spins that have N up neighbors.
	SpinIndex*	psNUp; // Must be larger than lattice coordination number+1.
	//! Probability a spin flips when number of up neighbors increases from N to N+1;
	double *		dProbabilityToFlip; // Must be from 0..Z-1, so length Z.
	SpinIndex	M;

	//! The spinFlipQueue is the main tool of this class.
	// It used to be a queue, but I need iterator access.
	std::deque<SpinIndex> spinFlipQueue;
public:
	//! Constructor.
	ProbabilisticAvalanche();
	//! Destructor.
	~ProbabilisticAvalanche();
	//! Tell this object what SpinLattice it should use.
	void SetLattice( SpinLattice *const lattice ) throw (BadArgument);
	//! Tell it how to initialize the random field.
	void SetInitializer( GaussianInitializer *initializer );
	//! Tell it how to initialize the random field.
	void SetRand( Rand *rand );
	//! Give the class a pointer to something which might serialize it.
	void SetBatchTimer( BatchTimer *batchTimer );
	//! This call permits the class to set up cached internal variables.
	void Initialize();
	//! Call at the start of a sweep.
	void BeginSweep( bool bKeepData );
	//! Ask the class whether it is all ready to go.
	void Verify( void ) const;
	//! Ask what lattice this class uses.
	const SpinLattice* GetLattice() const;
	//! Get array of down spins that have N up neighbors.
	const SpinIndex* GetDownSpins() const;
	//! Return magnetization
	SpinIndex GetM() const { return M; }
	//! Finds the start of the next avalanche.
	Status FindNextStart( double* H, SpinIndex *iFirstSpin );
	//! Run through a single avalanche.
	Status Cascade( double H, SpinIndex iFirstSpin );

	/*! Serializable interface */
	//@{
	virtual void Save( DMWriter *writer ) const;
	virtual void Load( DMReader *reader );
	virtual const std::string GetType() const;
	//@}

private:
	//! Put the first spin on the spinFlipQueue.
	void InitializeSpinFlipQueue( SpinIndex iFirstSpin );
	//! Read the spinFlipQueue and flip spins therein.
	Status CascadeQueue( );
	//! Set internal variables from dependencies.
	void SetCachedVariables();
	//! Store array of how many spins have N up neighbors.
	void InitializeNUp();
	//! Initialize dProbabilityToFlip for this field
	void SetProbabilityToFlip( double dField );
	//! Delete arrays
	void DeleteArrays();
	//! Allocate arrays
	void AllocateArrays();
	//! Make sure the pointers are zero so we know arrays aren't allocated.
	void ZeroPointers();

	//! Number of spins. Cached from SpinLattice.
	SpinIndex N;
	//! coordination number of lattice. Cached from SpinLattice.
	int Z;
	//! These constants are quick notes about initializations.
	enum { 
		INITIALIZED = 0x1
	};
	//! Keep track of which initializations have been performed.
	int iVerify;
	bool bSerialized; // Whether we are initializing from serialization.
	bool bInCascade; // True during a call to Cacade().
};

#endif //_PROBABILISTIC_AVALANCHE_H_
