#include "DMFirstHeader.h"
#include "SizeDistribution.h"

#include <math.h>
#include <assert.h>
#include "CheckSpanning.h"
#include "DMException.h"
#include "NotEnoughMemory.h"
#include "DMWriter.h"
#include "DMReader.h"
#include "SerialBuilder.h"
#include "DMAllocator.h"
#include "DMLastHeader.h"

using namespace std;

const std::string sType = std::string("SizeDistribution");

const SerialBuilder<SizeDistribution> *_creator = new
	SerialBuilder<SizeDistribution>(sType);

const double SizeDistribution::DefaultBinSize_c = 1.1;
const int SizeDistribution::DefaultNumBins_c = 10000;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
SizeDistribution::SizeDistribution()
{
	// Pointer to a helper class.
	spanning = 0;

	this->binSize = DefaultBinSize_c;
	this->maxBinNum = DefaultNumBins_c;
	
	ZeroArrays();

	bSerialized = false;
}

/*! You have to call this before calling Initialize().
	*/
void SizeDistribution::SetBinSize( double binSize )
{
	ASSERT( binSize > 0 );
	if (binSize <= 0) throw
		DMException("SizeDistribution","SetBinSize",
		"The bin size should be greater than zero." );

	this->binSize = binSize;
	iVerify |= BINSIZE;
}

/*! You have to call this before calling Initialize().
	*/
void SizeDistribution::SetNumBins( int numBins )
{
	ASSERT( numBins > 0 );
	if (numBins <= 0) throw
		DMException("SizeDistribution","SetNumBins",
		"The number of bins should be greater than zero.");

	this->maxBinNum = numBins;
	iVerify |= NUMBINS;
}

SizeDistribution::~SizeDistribution()
{
	this->DeleteArrays();
}

/*! You have to call this before you initialize.
	*/
void SizeDistribution::SetSpanning( const CheckSpanning* spanning )
{
	ASSERT( 0 != spanning );
	if (spanning == 0) throw
		DMException("SizeDistribution","SetSpanning",
		"The pointer passed was null.");

	this->spanning = spanning;
	iVerify |= SPANNING;
}

/*! You must call SetBinSize, SetNumBins, and SetSpanning
	before calling Initialize.  You must call Initialize before
	the first spin flip or avalanche happens.

	You may call this as many times as you like.  It will
	just re-initialize the class.
	*/
void SizeDistribution::Initialize()
{
	ASSERT( 0 != spanning );
	if ( 0 == spanning ) throw DMException("SizeDistribution",
		"Initialize","You must call SetSpanning before initializing.");
	

	if (!bSerialized) {
		this->AllocateArrays();
	}

	// Initialize cached variables.
	logBinSize = log(binSize);
	invLogBinSize = 1/logBinSize;

	bSerialized = false;
	iVerify = INITIALIZED;
}

void SizeDistribution::Verify()
{
	ASSERT( (iVerify&BINSIZE) == 0 );
	if ( (iVerify&BINSIZE) != 0 ) throw
		DMException("SizeDistribution","Initialize",
		"You need to initialize if you set the bin size.");

	ASSERT( (iVerify&NUMBINS) == 0 );
	if ( (iVerify&NUMBINS) != 0 ) throw
		DMException("SizeDistribution","Initialize",
		"You need to initialize if you set the number of bins.");

	ASSERT( (iVerify&SPANNING) == 0 );
	if ( (iVerify&SPANNING) != 0 ) throw
		DMException("SizeDistribution","Initialize",
		"You need to initialize if you set the spanning.");

	ASSERT( 0 != spanning );
	if ( 0 == spanning ) throw
		DMException("SizeDistribution","Initialize",
		"You need to set the spanning.");

	ASSERT( iVerify == INITIALIZED );
	if ( iVerify != INITIALIZED ) throw
		DMException("SizeDistribution","Initialize",
		"You need to initialize before using this class.");
}

/*! If you start a sweep with BeginSweep(true), this routine
	will accumulate data.
	*/
void SizeDistribution::BeginSweep( bool bKeepData )
{
	Verify();

	if (bKeepData) {
		if ( totalFlipped > 0 )
			accumulator.AddDataSet( *this );
	} else {
		accumulator.Clear();
	}

	currentSize = totalFlipped = highestBin =0;
	numUsedBins = maxCalculatedBin = 0;
	for(int i=0;i<maxBinNum;i++) avalancheSizeDistribution[i]=0;
	numAvalanches = 0;
	sumSSquared = 0;
}

void SizeDistribution::Save( DMWriter *writer ) const
{
	vector<int> shape;

	writer->PutInt( highestBin,"highestBin" );
	writer->PutInt( maxBinNum, "maxBinNum" );
	writer->PutInt( currentSize, "currentSize" );
	writer->PutInt( totalFlipped, "totalFlipped" );
	writer->PutInt( numAvalanches, "numAvalanches" );
	writer->PutInt( numUsedBins, "numUsedBins" );
	writer->PutInt( maxCalculatedBin, "maxCalculatedBin" );
	writer->PutInt( iVerify, "verify" );

	writer->PutDouble( binSize, "binSize" );
	writer->PutDouble( sumSSquared, "sumSSquared" );

	shape.push_back( maxBinNum );
	writer->PutIntArray( avalancheSizeDistribution, shape,
		"avalancheSizeDistribution" );
	writer->PutIntArray( fullBins, shape, "fillBins" );
	writer->PutIntArray( binWidths, shape, "binWidths" );
	writer->PutDoubleArray( binPositions, shape, "binPositions" );
	writer->Put( &(this->accumulator), "accumulator" );
}

void SizeDistribution::Load( DMReader *reader )
{
	vector<int> shape;
	this->DeleteArrays();
	iVerify = 0;

	highestBin = reader->GetInt( "highestBin" );
	maxBinNum = reader->GetInt( "maxBinNum" );
	currentSize = reader->GetInt( "currentSize" );
	totalFlipped = reader->GetInt( "totalFlipped" );
	numAvalanches = reader->GetInt( "numAvalanches" );
	numUsedBins = reader->GetInt( "numUsedBins" );
	maxCalculatedBin = reader->GetInt( "maxCalculatedBin" );
	iVerify = reader->GetInt( "verify" );
	iVerify |= SPANNING;

	binSize = reader->GetDouble( "binSize" );
	sumSSquared = reader->GetDouble( "sumSSquared" );

	avalancheSizeDistribution = reader->GetIntArray( shape,
		"avalancheSizeDistribution" );
	fullBins = reader->GetIntArray( shape, "fillBins" );
	binWidths = reader->GetIntArray( shape, "binWidths" );
	binPositions = reader->GetDoubleArray( shape, "binPositions" );

	reader->Fill( &accumulator, "accumulator" );
	// Call the Set functions so we initialize ourselves properly.
	this->SetBinSize( binSize );
	this->SetNumBins( maxBinNum );
	bSerialized = true;
}

const std::string SizeDistribution::GetType() const
{
	return sType;
}


/*! A spin has flipped.  Increment the size of the current avalanche.
	*/
inline void SizeDistribution::FlipCompleted(SpinLattice* lattice,
									 const int* coords,
									 int dDimension)
{
	++currentSize;
	++totalFlipped;
}

/** An avalanche has ended.  If the avalanche was not spanning,
  * Record the avalanche size in the distribution.
  *
  * There is an error here if the avalanche is too large, and it isn't thrown
  * properly.
  */
inline void SizeDistribution::AvalancheCompleted(AvalancheAlgorithm* avalanche,
										  double H)
{
	numAvalanches++;
	sumSSquared += currentSize*currentSize;
	if(spanning->IsSpanning()) {
		currentSize = 0;
		return;
	}

		// which bin does the avalanche fall in?
	int binNum = (int)(log(currentSize) * invLogBinSize);
	
		// if the avalanche was too big, output an error.
	if(binNum >= maxBinNum)
	{
		//	cerr <<
		//	"SizeDistribution: Avalanche too large for
		//  avalanche size distribution." <<
		//	endl;
	}
		// add one avalanche to the appropriate bin.
	else
	{
		if(binNum>highestBin) highestBin = binNum;
		avalancheSizeDistribution[binNum]++;
	}
		// reset the avalanche size for the next avalanche
	currentSize=0;
}

/*! Calculate the widths of all the bins
	for normalizing the size distribution.  Also record the
	locations of the bins, and record which bins are "real"
	bins.
	*/
void SizeDistribution::CalculateBinWidths() const
{
	double binMiddle;		// The center of the bin
	int biggestInBin;		// The largest integer in the bin
	int biggestInLastBin;   // The largest integer in the last bin

	if(maxCalculatedBin==0) biggestInLastBin = 0;
	else
		biggestInLastBin = (int)(exp((maxCalculatedBin)*logBinSize)-1e-10);

	for(int i=maxCalculatedBin; i<=highestBin; i++)
	{
		biggestInBin = (int)(exp((i+1)*logBinSize)-1e-10);
		binMiddle = 0.5*(biggestInBin + biggestInLastBin+1);
		binWidths[i] = biggestInBin - biggestInLastBin;
		binPositions[i] = binMiddle;

		if(biggestInBin!=biggestInLastBin)
		{
			fullBins[numUsedBins] = i;
			numUsedBins++;
		}
		biggestInLastBin = biggestInBin;
	}
	maxCalculatedBin = highestBin;
}

/*!
	Return the number of the last used bin.
	*/
int SizeDistribution::GetLastBin() const
{
	if(maxCalculatedBin<highestBin) CalculateBinWidths();
	return numUsedBins-1;
}

/*! Return the avalanche size which
	corresponds to bin binNum.
	*/
double SizeDistribution::GetAvalancheSize(int binNum) const
{
	if(maxCalculatedBin<highestBin) CalculateBinWidths();

	ASSERT(binNum<numUsedBins);

	return binPositions[fullBins[binNum]];
}

/*! Return the number of avalanches in
	bin binNum normalized by the width of the bin.
	*/
double SizeDistribution::GetAvalancheCount(int binNum) const
{
	if(maxCalculatedBin<highestBin) CalculateBinWidths();

	ASSERT(binNum<numUsedBins);

	int actualBin = fullBins[binNum];
	return (double)avalancheSizeDistribution[actualBin] / 
		(double)binWidths[actualBin] / (double)totalFlipped;
}

void SizeDistribution::DeleteArrays()
{
	if (avalancheSizeDistribution)
		delete [] avalancheSizeDistribution;
	if (fullBins) delete [] fullBins;
	if (binWidths) delete [] binWidths;
	if (binPositions) delete [] binPositions;

	ZeroArrays();
}

void SizeDistribution::ZeroArrays()
{
	avalancheSizeDistribution = 0;
	fullBins = 0;
	binWidths = 0;
	binPositions = 0;
}

void SizeDistribution::AllocateArrays()
{
	DeleteArrays();

	try {
		avalancheSizeDistribution	= DMAllocate<int>( maxBinNum );
		fullBins					= DMAllocate<int>( maxBinNum );
		binWidths					= DMAllocate<int>( maxBinNum );
		binPositions				= DMAllocate<double>( maxBinNum );

	} catch ( std::bad_alloc ) {
		DeleteArrays();
		throw NotEnoughMemory("SizeDistribution","AllocateArrays",
			"Four arrays of size maxBinNum");
	}
}

std::ostream& SizeDistribution::SaveAverageData( std::ostream& s )
{
	s << accumulator;
	return s;
}

std::ostream& operator<<(std::ostream& s, const SizeDistribution& r)
{
	int i;
	
	s << "SizeDistribution" << endl;
	s << "HighestBin:" << "\t" << r.GetLastBin() << endl;
	s << "BinSize:" << "\t" << r.binSize << endl;
	for (i=0; i<r.GetLastBin(); i++)
		s << r.GetAvalancheSize(i) << "\t" << r.GetAvalancheCount(i)
		<< std::endl;
	return s;
}