#ifndef _SIZE_DISTRIBUTION_H_
#define _SIZE_DISTRIBUTION_H_

#include "DMFirstHeader.h"
#include "DiscreteAvalanche.h"
#include "SpinIndex.h"
#include "FlipListener.h"
#include "AvalancheListener.h"
#include "Serializable.h"
#include "XYData.h"
#include "XYAverage.h"

class CheckSpanning;

//! Records how many avalanches there are of each size.
/*! Think of this as a table where column 1 is the size of the avalanche
	and column 2 is the number of avalanches of that size.

	This class can store cumulative data over several runs.

	\author Matt Kuntz
 */
class SizeDistribution : public FlipListener, public AvalancheListener,
	public Serializable, public XYData
{
	const CheckSpanning* spanning;
	XYAverage accumulator;

	SpinIndex currentSize;				 // The size of the current avalanche
	SpinIndex totalFlipped;              // The total number of spins flipped
	int* avalancheSizeDistribution;// The avalanche size distribution

	int highestBin;	// The highest bin that has been used
	int maxBinNum;	// The maximum possible bin
	double binSize;	// The size of a bin

	int numAvalanches;		// Number of avalanches for normalizing output
	double sumSSquared;		// The second moment of the distribution

	mutable int* fullBins;	// A list of the bins that actually have non-zero widths
	mutable int numUsedBins;// The number of bins that actually have non-zero widths

	mutable int* binWidths;		 // The widths of the bins
	mutable double* binPositions;// The avalanche sizes corresponding to the bins
	mutable int maxCalculatedBin;// The last bin for which a width is calculated

	static const double DefaultBinSize_c;
	static const int DefaultNumBins_c;
public:
	//! Constructor.
	SizeDistribution();

	//! Destructor.
	~SizeDistribution();

	//! Set the bin size of size binning.
	void SetBinSize( double binSize );

	//! How many bins will there be?
	void SetNumBins( int numBins );

	//! Give us a pointer to the CheckSpanning observer.
	void SetSpanning( const CheckSpanning* spanning );

	//! Allow this class to initialize.
	void Initialize();

	//! Check we called everything correctly.
	void Verify();

	//! Call before beginning a sweep of magnetic field.
	void BeginSweep( bool bKeepData = false );

	//ACTION FUNCTIONS

	//! A spin has flipped.  Increment the size of the current avalanche
	void FlipCompleted( SpinLattice* lattice, const int* coords,
		int dDimension );

	//! The avalanche has ended.  Record it's size in the histogram.
	void AvalancheCompleted( AvalancheAlgorithm* avalanche,
		double H);
	
	//ACCESSOR FUNCTIONS

	//! Get the bin of the largest avalanche
	int GetLastBin() const;

	//! Get the avalanche size corresponding to the bin binNum
	double GetAvalancheSize(int binNum) const;

	//! Get the number of avalanches per size in the bin binNum
	double GetAvalancheCount(int binNum) const;

	//! Return the second moment of the distribution.
	double GetSecondMoment() const
	{ return (double)sumSSquared / (double)numAvalanches; }

	/*! XYData interface */
	//@{
	double GetX( int binNum ) const { return GetAvalancheSize( binNum ); }
	double GetY( int binNum ) const { return GetAvalancheCount( binNum ); }
	int GetPointCnt( ) const { return GetLastBin(); }
	//@}

	/*! Serializable interface */
	//@{
	void Save( DMWriter *writer ) const;
	void Load( DMReader *reader );
	const std::string GetType() const;
	//@}

	std::ostream& SaveAverageData( std::ostream& s );
	friend std::ostream& operator<<(std::ostream& s, const SizeDistribution& r);
private:
	//IMPLEMENTATION FUNCTIONS

	void CalculateBinWidths() const;
	void DeleteArrays();
	void ZeroArrays();
	void AllocateArrays();

	//PRIVATE DATA
	// These are cached.
	double logBinSize;					// The log of the bin size: log(binSize)
	double invLogBinSize;				// 1/log(binSize)

	// used for interface verification.
	enum {
		BINSIZE = 0x1,
		NUMBINS = 0x2,
		SPANNING = 0x4,
		INITIALIZED = 0x8
	};
	int iVerify;
	bool bSerialized;
};

std::ostream& operator<<(std::ostream& s, const SizeDistribution& r);

#endif //_SIZE_DISTRIBUTION_H_
