#include "DMFirstHeader.h"
#include <windows.h>
#include <string>
#include <sstream>
#include "SpinLattice2D.h"
#include "DMReader.h"
#include "DMWriter.h"
#include "SerialBuilder.h"
#include "DMLastHeader.h"

using std::stringstream;

const std::string sType =
	std::string("SpinLattice2D");

/*! This static object permits the DMReader to create an empty SpinLattice2D. */
const SerialBuilder<SpinLattice2D> *_creator = new
	SerialBuilder<SpinLattice2D>(sType);

SpinLattice2D::SpinLattice2D() {}

SpinLattice2D::~SpinLattice2D() {}

SpinIndex SpinLattice2D::GetLength() const
{
	return iLength;
}

SpinIndex SpinLattice2D::GetCoordinationNumber() const
{
	return 4;
}

SpinIndex SpinLattice2D::GetTotalNumberUp() const
{
	SpinIndex i, total = 0;
	for (i=0; i<this->size(); i++)
		total += ( (*this)[i] == true ) ? 1 : 0;
	return total;
}

bool SpinLattice2D::SetLength( const SpinIndex iLength )
{
	ASSERT( iLength >= 0 );

	try {
		this->resize( iLength*iLength );
	} catch (std::bad_alloc e) {
		// failed to resize;
		return false;
	}

	this->iLength = iLength;
	this->N = iLength*iLength;
	return true;
}

#define CHAR_BITS 8
//typedef unsigned int DWORD;
const int VBITS = CHAR_BITS* sizeof(DWORD);

void SpinLattice2D::Save( DMWriter *writer ) const
{
	static const int nBlocks = 10;
	SpinIndex N = this->size();
	writer->PutInt( iLength, "x" );
	writer->PutInt( iLength, "y" );
	
	int dwLen = (N+VBITS-1)/VBITS;
	int dwBlockLen = dwLen/nBlocks + 1;
	int iBlockLen = dwBlockLen*VBITS;
	int iBlocks = (N+iBlockLen-1)/iBlockLen;
	writer->PutInt( iBlocks, "nBlocks" );

	std::vector<int> shape(1);
	SpinIndex end;
	int i;
	for (i=0; i<iBlocks; i++) {
		end = (i+1)*iBlockLen;
		if (end>N) end = N;
		int dwLen;
		int* data = WriteSlice( i*iBlockLen, end, dwLen );
		stringstream sCharBuffer;
		sCharBuffer << "block" << i;
		shape[0] = dwLen;
		writer->PutIntArray( (const int*) data, shape, sCharBuffer.str() );
		delete[] data;
	}
}


void SpinLattice2D::Load( DMReader *reader )
{
	int iX = reader->GetInt( "x" );
	int iY = reader->GetInt( "y" );
	ASSERT( iX > 0 );
	ASSERT( iY > 0 );
	iLength = iX;
	N = iLength*iLength;
	this->resize( N );

	int iBlocks = reader->GetInt( "nBlocks" );
	std::vector<int> shape(1);
	SpinIndex start, finish;
	start = finish = 0;
	int i;
	for (i=0; i< iBlocks; i++) {
		stringstream sCharBuffer;
		int *data;
		sCharBuffer << "block" << i;
		data = reader->GetIntArray( shape, sCharBuffer.str() );
		ASSERT( shape[0] > 0 );
		finish = start+shape[0]*VBITS;
		if (finish>N) finish = N;
		ReadSlice( start, finish, data );
		start = finish;
		delete[] data;
	}
}

/*
	int iLen = (N+VBITS-1)/VBITS;
	int i, j;
	DWORD *data = new DWORD[iLen];
	for (i=0; i<iLen; i++) {
		DWORD it = 0;
		for (j=0; j<VBITS; j++) {
			DWORD tr = ( (*this)[i*VBITS+j] ) ? 1 : 0;
			it <<= 1;
			it |= tr;
		}
		data[i] = it;
	}
	std::vector<int> shape;
	shape.push_back( iLen );
	writer->PutIntArray( (const int*) data, shape, "lattice" );
	delete[] data;
*/
int* SpinLattice2D::WriteSlice( SpinIndex start, SpinIndex finish, int& dwLen ) const
{
	SpinIndex len = finish-start;
	dwLen = (len+VBITS-1)/VBITS;
	int i,j;
	DWORD *data = new DWORD[dwLen];
	for (i=0; i< dwLen; i++) {
		DWORD it = 0;
		for (j=VBITS-1; j>=0; --j) {
			DWORD tr = ( (*this)[start+i*VBITS+j] ) ? 1 : 0;
			it <<= 1;
			it |= tr;
		}
		data[i] = it;
	}
	return (int*)data;
}

void SpinLattice2D::ReadSlice( SpinIndex start, SpinIndex finish, int*data )
{
	SpinIndex len = finish-start;
	int dwLen = (len+VBITS-1)/VBITS;

	ASSERT( 0 == IsBadReadPtr( (CONST VOID *) data, sizeof(int)*dwLen) );

	DWORD* dwData = (DWORD*) data;
	DWORD tr;
	int i,j;
	for (i=0; i<dwLen; i++) {
		DWORD it = dwData[i];
		for (j=0; j<VBITS; ++j) {
			tr = ( it & 0x1 );
			(*this)[start+i*VBITS+j] = ( tr == 1 );
			it >>= 1;
		}
	}
}

/*
	std::vector<int> shape;
	this->resize( iX*iY );
	int *data = (int*) reader->GetIntArray( shape, "lattice" );
	int iLen = shape[0];
	int i, j;
	DWORD tr;
	for (i=0; i<iLen; i++) {
		DWORD it = ((DWORD*)data)[i];
		for (j=0; j<VBITS; j++) {
			tr = ( it & 0x1 );
			(*this)[i*VBITS+VBITS-j-1] = ( tr == 1 );
			it >>= 1;
		}
	}
	delete[] data;

*/

/** The signature of this method is an error.  It should return a reference.
  */
const std::string SpinLattice2D::GetType() const
{
	return sType;
}