#include "DMFirstHeader.h"
#include "SweepLogic.h"
#include "SpinIndex.h"
#include "AvalancheAlgorithm.h"
#include "AvalancheStart.h"
#include "DMWriter.h"
#include "DMReader.h"
#include "SerialBuilder.h"
#include "DMLastHeader.h"

const std::string sType =
	std::string("SweepLogic");

/*! This static object permits the DMReader to create an empty SweepLogic. */
const SerialBuilder<SweepLogic> *_creator = new
	SerialBuilder<SweepLogic>(sType);

/*! Doesn't allocate any pointers.
	*/
SweepLogic::SweepLogic()
{
	startingField = -3;
	BeginSweep();
}

/*! The starting field already has a default value of -3.

	@param dField double The starting field.
	*/
void SweepLogic::SetStartingField( double dField )
{
	this->startingField = dField;
}

/*! You need to call this before calling Run().
	*/
void SweepLogic::SetAvalanche( AvalancheAlgorithm * avalanche )
{
	ASSERT( 0 != avalanche );
	if ( 0 == avalanche ) throw
		DMException( "SweepLogic","SetAvalanche",
		"Passed a null pointer.");
	this->avalanche = avalanche;
}

void SweepLogic::BeginSweep( bool bKeepData )
{
	dField = startingField;
	iTotalRuns = 0;
}

void SweepLogic::Save( DMWriter *writer ) const
{
	ASSERT( writer != 0 );
	if ( writer == 0 ) throw
		DMException( "SweepLogic","Save",
		"The writer is a null pointer.");

	writer->PutDouble( startingField, "startingField" );
	writer->PutDouble( dField, "field" );
	writer->PutInt( iTotalRuns, "numRuns" );
}

void SweepLogic::Load( DMReader *reader )
{
	ASSERT( reader != 0 );
	if ( reader == 0 ) throw
		DMException( "SweepLogic","Load",
		"The reader is a null pointer.");

	startingField = reader->GetDouble( "startingField" );
	dField = reader->GetDouble( "field" );
	iTotalRuns = reader->GetInt( "numRuns" );
}

const std::string SweepLogic::GetType() const
{
	return sType;
}

/*!	In this function, we call the AvalancheStart and 
	AvalancheAlgorithm to make it all go.

	Call this with the maximum number of runs we should do 
	before returning.  If we return with less than the number
	given, it means we're done.

	If you insist on calling this again after it tells you
	it did all the runs it can do, it will be okay as long
	as the starter still knows it can't do any more.

	\param SweepCount integer Number of avalanches to do.
	\return int Returns the status from AvalancheAlgorithm to
	indicate successful completion of SweepCount avalanches.
	*/
AvalancheAlgorithm::Status
SweepLogic::Run( int SweepCount )
{
	SpinIndex FirstSpinToFlip;

	AvalancheAlgorithm::Status status = AvalancheAlgorithm::SUCCESS;
	int SweepIndex = 0;
	while ( SweepIndex < SweepCount ) {

		// Find the start of the next avalanche.
		status = avalanche->FindNextStart( &dField, &FirstSpinToFlip );
		if (status != AvalancheAlgorithm::SUCCESS) return status;

		// Flip the spins at that starting point.
		status = avalanche->Cascade( dField, FirstSpinToFlip );
		if (status != AvalancheAlgorithm::SUCCESS) return status;

		SweepIndex++;
		this->iTotalRuns++;

		// Tell all AvalancheListeners that an avalanche is done.
		this->Notify( dField );
	}

	return status;
}

/*! Less of a need for error checking on AddListener,
	RemoveListener, and Notify, because the ObserverList
	class takes care of it.

	@return Whether we could add this listener.  False
	means we already have it.
	*/
bool SweepLogic::AddListener( AvalancheListener *listener )
{
	return listeners.AddObserver( listener );
}

/*!
	@return Whether we can remove the listener.  False means
	it wasn't on our list in the first place.
	*/
bool SweepLogic::RemoveListener( AvalancheListener *listener )
{
	return listeners.RemoveObserver( listener );
}

/*! Tell our listeners that an avalanche completed.
	*/
void SweepLogic::Notify( double dField )
{
	int i;
	for (i=0; i< listeners.size(); ++i) {
		listeners.at(i)->AvalancheCompleted( avalanche, dField );
	}
}