#include "DMFirstHeader.h"
#include <iostream>
#include "DigitalMaterial.h"
#include "XYAverage.h"
#include "NotEnoughMemory.h"
#include "SerialBuilder.h"
#include "DMReader.h"
#include "DMWriter.h"
#include "DMLastHeader.h"

const std::string XYAverage::sType("XYAverage");

/*! This static object permits the DMReader to create an empty CheckSpanning. */
const SerialBuilder<XYAverage> *_creator = new
	SerialBuilder<XYAverage>(XYAverage::sType);

XYAverage::XYAverage( bool bAbsentEqualsZero ) :
	bAbsentEqualsZero(bAbsentEqualsZero), SetCnt(0)
{}

XYAverage::~XYAverage()
{ }

void XYAverage::AddDataSet( const XYData& data )
{
	SetCnt++;

	// Allocate any necessary storage
	int oldCnt = points.size();
	if ( data.GetPointCnt() > oldCnt )
	{
		try {
			points.resize( data.GetPointCnt() );
		} catch ( std::bad_alloc ) {
			throw NotEnoughMemory( "XYAverage","AddDataSet",
				"Trying to accumulate a a data set for averaging.");
		}
	}

	// Accumulate the data
	int PtIdx;
	for ( PtIdx=0; PtIdx < points.size(); PtIdx++ ) {
		// If this point is beyond current accumulated data
		if ( PtIdx >= oldCnt )
		{
			points[ PtIdx ].X	= data.GetX( PtIdx );
			double Y			= data.GetY( PtIdx );
			points[ PtIdx ].Y	= Y;
			points[ PtIdx ].YSq = Y*Y;

			if ( bAbsentEqualsZero ) points[ PtIdx ].SampleCnt = SetCnt;
			else points[ PtIdx ].SampleCnt = 1;
		}

		// Else if new set has fewer points than old.
		else if ( PtIdx >= data.GetPointCnt() )
		{
			if ( !bAbsentEqualsZero ) break;
			points[ PtIdx ].SampleCnt++;
		}

		// Regular points
		else
		{
			bool bXAxisChanged = fabs( points[ PtIdx ].X-data.GetX( PtIdx ) )>1e-10;
			ASSERT( !bXAxisChanged );
			if ( bXAxisChanged ) throw DMException("XYAverage","AddDataSet",
					"In order to accumulate data, the X axis must remain the same.");
			
			double Y = data.GetY( PtIdx );
			points[ PtIdx ].SampleCnt++;
			points[ PtIdx ].Y	+= Y;
			points[ PtIdx ].YSq += Y*Y;
		}
	}
}


void XYAverage::Clear()
{
	points.clear();
	SetCnt = 0;
}


std::ostream& operator<<( std::ostream& out, const XYAverage& data )
{
	int i;

	for ( i=0; i < data.GetPointCnt(); i++ ) {
		out << data.GetX( i ) << "\t" << data.GetY( i ) << "\t"
			<< data.GetStandardDeviation( i ) << std::endl;
	}
	return out;
}


void XYAverage::Save( DMWriter *writer ) const
{
	writer->PutInt( SetCnt, "SetCnt" );
	int iAbsentEqualsZero = (bAbsentEqualsZero) ? 1 : 0;
	writer->PutInt( iAbsentEqualsZero, "bAbsentEqualsZero" );

	std::vector<int> pointsShape(1);
	int ptCnt = points.size();
	writer->PutInt( ptCnt, "ptCnt" );
	if ( ptCnt > 0 ) {
		pointsShape[0] = ptCnt;
		try {
			double* dTransfer = new double[ptCnt];
			int ptIdx;
			
			for (ptIdx=0; ptIdx<ptCnt; ptIdx++) dTransfer[ptIdx] = points[ptIdx].X;
			writer->PutDoubleArray( dTransfer, pointsShape, "pointsX" );
			
			for (ptIdx=0; ptIdx<ptCnt; ptIdx++) dTransfer[ptIdx] = points[ptIdx].Y;
			writer->PutDoubleArray( dTransfer, pointsShape, "pointsY" );
			
			for (ptIdx=0; ptIdx<ptCnt; ptIdx++) dTransfer[ptIdx] = points[ptIdx].YSq;
			writer->PutDoubleArray( dTransfer, pointsShape, "pointsYSq" );
			
			delete[] dTransfer;
			
			int* iTransfer = new int[ptCnt];
			for (ptIdx=0; ptIdx<ptCnt; ptIdx++) iTransfer[ptIdx] = points[ptIdx].SampleCnt;
			writer->PutIntArray( iTransfer, pointsShape, "pointsSampleCnt" );
			
			delete[] iTransfer;
			
		} catch (std::bad_alloc) {
			throw NotEnoughMemory("XYAverage","Save",
				"Could not allocate enough memory to save the averages.");
		}
	}
}


void XYAverage::Load( DMReader *reader )
{
	this->Clear();

	SetCnt = reader->GetInt( "SetCnt" );
	ASSERT( SetCnt >= 0 );
	if ( SetCnt < 0 ) {
		throw DMException("XYAverage","Load",
			"The number of data sets loaded improperly.");
	}

	int iAbsentEqualsZero = reader->GetInt( "bAbsentEqualsZero" );
	bool isBoolean = (iAbsentEqualsZero == 1) || (iAbsentEqualsZero == 0);
	ASSERT( isBoolean );
	if ( !isBoolean ) {
		throw DMException("XYAverage","Load",
			"A boolean didn't load properly.");
	}
	
	int ptCnt = reader->GetInt( "ptCnt" );
	if ( ptCnt > 0 ) {
		
		std::vector<int> shape;
		double* dTransfer = reader->GetDoubleArray( shape, "pointsX" );
		ASSERT( shape.size() > 0 );
		if ( shape.size() <= 0 ) {
			throw DMException("XYAverage","Load","Array returned with zero size." );
		}
		ASSERT( shape.size() == ptCnt );
		if ( shape.size() != ptCnt )
			throw DMException("XYAverage","Load","Array size mismatches stored size.");
		
		int ptIdx;
		for (ptIdx=0; ptIdx<ptCnt; ptIdx++) points[ptIdx].X = dTransfer[ptIdx];
		delete[] dTransfer;
		
		dTransfer = reader->GetDoubleArray( shape, "pointsY" );
		for (ptIdx=0; ptIdx<ptCnt; ptIdx++) points[ptIdx].Y = dTransfer[ptIdx];
		delete[] dTransfer;
		
		dTransfer = reader->GetDoubleArray( shape, "pointsYSq" );
		for (ptIdx=0; ptIdx<ptCnt; ptIdx++) points[ptIdx].YSq = dTransfer[ptIdx];
		delete[] dTransfer;
		
		int* iTransfer = reader->GetIntArray( shape, "pointsSampleCnt" );
		for (ptIdx=0; ptIdx<ptCnt; ptIdx++) points[ptIdx].SampleCnt = iTransfer[ptIdx];
		delete[] iTransfer;
	}
}
