/*
 *
 * Copyright (c) 1994
 * Hewlett-Packard Company
 *
 * Permission to use, copy, modify, distribute and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.  Hewlett-Packard Company makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 *
 *
 * Copyright (c) 1996,1997
 * Silicon Graphics Computer Systems, Inc.
 *
 * Permission to use, copy, modify, distribute and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.  Silicon Graphics makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 */

/* Caution to the wind.  We are making this a non-stl vector so we always
 * have something to use.
 */
#ifndef __STOLEN_BVECTOR_H
#define __STOLEN_BVECTOR_H

#include <memory> // for allocator
#include <algorithm> // for copy
#include "stl_iterator.h" // for random_access_iterator

typedef unsigned int dword;
typedef unsigned __int64 bit_index;

#define CHAR_BIT 8

static const int __WORD_BIT = int(CHAR_BIT*sizeof(dword));
//! Manipulates the bits in an atomic unsigned integer type.
/*!	This class is taken from the Gnu STL libraries.
	*/
struct __bit_reference {
  dword* p;
  dword mask;
  __bit_reference(dword* x, dword y) : p(x), mask(y) {}

public:
	//! Constructor.
	__bit_reference() : p(0), mask(0) {}
  operator bool() const { return !(!(*p & mask)); }
  //! Assignment
  __bit_reference& operator=(bool x) {
    if (x)      
      *p |= mask;
    else 
      *p &= ~mask;
    return *this;
  }
  //! Copy
  __bit_reference& operator=(const __bit_reference& x) { return *this = bool(x); }
	//! Equality test.
  bool operator==(const __bit_reference& x) const {
    return bool(*this) == bool(x);
  }
	//! Ordering
  bool operator<(const __bit_reference& x) const {
    return bool(*this) < bool(x);
  }
	//! Switch a bit from 0 to 1 or 1 to zero.
  void flip() { *p ^= mask; }
};

inline void swap(__bit_reference x, __bit_reference y) {
  bool tmp = x;
  x = y;
  y = tmp;
}

struct __bit_iterator : public random_access_iterator<bool, bit_index> {
  typedef __bit_reference  reference;
  typedef __bit_reference* pointer;
  typedef __bit_iterator iterator;

  dword* p;
  dword offset;
  void bump_up() {
    if (offset++ == __WORD_BIT - 1) {
      offset = 0;
      ++p;
    }
  }
  void bump_down() {
    if (offset-- == 0) {
      offset = __WORD_BIT - 1;
      --p;
    }
  }

  __bit_iterator() : p(0), offset(0) {}
  __bit_iterator(dword* x, dword y) : p(x), offset(y) {}
  reference operator*() const { return reference(p, 1U << offset); }
  iterator& operator++() {
    bump_up();
    return *this;
  }
  iterator operator++(int) {
    iterator tmp = *this;
    bump_up();
    return tmp;
  }
  iterator& operator--() {
    bump_down();
    return *this;
  }
  iterator operator--(int) {
    iterator tmp = *this;
    bump_down();
    return tmp;
  }
  iterator& operator+=(difference_type i) {
    difference_type n = i + offset;
    p += (dword) (n / __WORD_BIT);
    dword newOffset = (dword) (n % __WORD_BIT);
    if (newOffset < 0) {
      offset = (dword) newOffset + __WORD_BIT;
      --p;
    } else
      offset = (dword) newOffset;
    return *this;
  }
  iterator& operator-=(difference_type i) {
    *this += -i;
    return *this;
  }
  iterator operator+(difference_type i) const {
    iterator tmp = *this;
    return tmp += i;
  }
  iterator operator-(difference_type i) const {
    iterator tmp = *this;
    return tmp -= i;
  }
  difference_type operator-(iterator x) const {
    return __WORD_BIT * (p - x.p) + offset - x.offset;
  }
  reference operator[](difference_type i) { return *(*this + i); }
  bool operator==(const iterator& x) const {
    return p == x.p && offset == x.offset;
  }
  bool operator!=(const iterator& x) const {
    return p != x.p || offset != x.offset;
  }
  bool operator<(iterator x) const {
    return p < x.p || (p == x.p && offset < x.offset);
  }
};

struct __bit_const_iterator
  : public random_access_iterator<bool, bit_index>
{
  typedef bool                 reference;
  typedef bool                 const_reference;
  typedef const bool*          pointer;
  typedef __bit_const_iterator const_iterator;

  dword* p;
  dword offset;
  void bump_up() {
    if (offset++ == __WORD_BIT - 1) {
      offset = 0;
      ++p;
    }
  }
  void bump_down() {
    if (offset-- == 0) {
      offset = __WORD_BIT - 1;
      --p;
    }
  }

  __bit_const_iterator() : p(0), offset(0) {}
  __bit_const_iterator(dword* x, dword y) : p(x), offset(y) {}
  __bit_const_iterator(const __bit_iterator& x) : p(x.p), offset(x.offset) {}
  const_reference operator*() const {
    return __bit_reference(p, 1U << offset);
  }
  const_iterator& operator++() {
    bump_up();
    return *this;
  }
  const_iterator operator++(int) {
    const_iterator tmp = *this;
    bump_up();
    return tmp;
  }
  const_iterator& operator--() {
    bump_down();
    return *this;
  }
  const_iterator operator--(int) {
    const_iterator tmp = *this;
    bump_down();
    return tmp;
  }
  const_iterator& operator+=(difference_type i) {
    difference_type n = i + offset;
    p += (dword) (n / __WORD_BIT);
    dword newOffset = (dword) (n % __WORD_BIT);
    if (newOffset < 0) {
      offset = (dword) newOffset + __WORD_BIT;
      --p;
    } else
      offset = (dword) newOffset;
    return *this;
  }
  const_iterator& operator-=(difference_type i) {
    *this += -i;
    return *this;
  }
  const_iterator operator+(difference_type i) const {
    const_iterator tmp = *this;
    return tmp += i;
  }
  const_iterator operator-(difference_type i) const {
    const_iterator tmp = *this;
    return tmp -= i;
  }
  difference_type operator-(const_iterator x) const {
    return __WORD_BIT * (p - x.p) + offset - x.offset;
  }
  const_reference operator[](difference_type i) { 
    return *(*this + i); 
  }
  bool operator==(const const_iterator& x) const {
    return p == x.p && offset == x.offset;
  }
  bool operator!=(const const_iterator& x) const {
    return p != x.p || offset != x.offset;
  }
  bool operator<(const_iterator x) const {
    return p < x.p || (p == x.p && offset < x.offset);
  }
};

// The next few lines are confusing.  What we're doing is declaring a
//  partial specialization of vector<T, Alloc> if we have the necessary
//  compiler support.  Otherwise, we define a class bit_vector which uses
//  the default allocator.  In either case, we typedef "data_allocator" 
//  appropriately.

#define __BVECTOR bit_vector

class bit_vector
{
//  typedef simple_alloc<dword, alloc> data_allocator;  
  typedef std::allocator<dword> allocator_type;
	allocator_type allocator;
public:
  typedef bool value_type;
  typedef bit_index size_type; // 64 bit baby
  typedef bit_index difference_type; // 64 bit baby
  typedef __bit_reference reference;
  typedef bool const_reference;
  typedef __bit_reference* pointer;
  typedef const bool* const_pointer;

  typedef __bit_iterator                iterator;
  typedef __bit_const_iterator          const_iterator;

  typedef reverse_iterator<const_iterator, value_type, const_reference, 
                           difference_type> const_reverse_iterator;
  typedef reverse_iterator<iterator, value_type, reference, difference_type>
          reverse_iterator;

protected:
  iterator start;
  iterator finish;
  dword* end_of_storage;
  dword* bit_alloc(size_type n) {
    return allocator.allocate((dword)((n + __WORD_BIT - 1)/__WORD_BIT),(dword*)0);
  }
  void deallocate() {
    if (start.p)
      allocator.deallocate(start.p, end_of_storage - start.p);
  }
  void initialize(size_type n) {
    dword* q = bit_alloc(n);
    end_of_storage = q + (dword) ((n + __WORD_BIT - 1)/__WORD_BIT);
    start = iterator(q, 0);
    finish = start + difference_type(n);
  }
  void insert_aux(iterator position, bool x) {
    if (finish.p != end_of_storage) {
		std::copy_backward(position, finish, finish + 1);
      *position = x;
      ++finish;
    }
    else {
      size_type len = size() ? 2 * size() : __WORD_BIT;
      dword* q = bit_alloc(len);
      iterator i = std::copy(begin(), position, iterator(q, 0));
      *i++ = x;
      finish = std::copy(position, end(), i);
      deallocate();
      end_of_storage = q + (dword) ((len + __WORD_BIT - 1)/__WORD_BIT);
      start = iterator(q, 0);
    }
  }


public:
  iterator begin() { return start; }
  const_iterator begin() const { return start; }
  iterator end() { return finish; }
  const_iterator end() const { return finish; }

  reverse_iterator rbegin() { return reverse_iterator(end()); }
  const_reverse_iterator rbegin() const { 
    return const_reverse_iterator(end()); 
  }
  reverse_iterator rend() { return reverse_iterator(begin()); }
  const_reverse_iterator rend() const { 
    return const_reverse_iterator(begin()); 
  }

  size_type size() const { return size_type(end() - begin()); }
  size_type max_size() const { return size_type(-1); }
  size_type capacity() const {
    return size_type(const_iterator(end_of_storage, 0) - begin());
  }
  bool empty() const { return begin() == end(); }
  reference operator[](size_type n) {
    return *(begin() + difference_type(n));
  }
  const_reference operator[](size_type n) const {
    return *(begin() + difference_type(n));
  }
  __BVECTOR() : start(iterator()), finish(iterator()), end_of_storage(0) {}
  __BVECTOR(size_type n, bool value) {
    initialize(n);
	std::fill(start.p, end_of_storage, value ? ~0 : 0);
  }
  __BVECTOR(int n, bool value) {
    initialize(n);
	std::fill(start.p, end_of_storage, value ? ~0 : 0);
  }
  __BVECTOR(long n, bool value) {
    initialize(n);
    std::fill(start.p, end_of_storage, value ? ~0 : 0);
  }
  explicit __BVECTOR(size_type n) {
    initialize(n);
    std::fill(start.p, end_of_storage, 0);
  }
  __BVECTOR(const __BVECTOR& x) {
    initialize(x.size());
    std::copy(x.begin(), x.end(), start);
  }

  __BVECTOR(const_iterator first, const_iterator last) {
    size_type n = 0;
    distance(first, last, n);
    initialize(n);
    std::copy(first, last, start);
  }
  __BVECTOR(const bool* first, const bool* last) {
    size_type n = 0;
    distance(first, last, n);
    initialize(n);
    std::copy(first, last, start);
  }

  ~__BVECTOR() { deallocate(); }
  __BVECTOR& operator=(const __BVECTOR& x) {
    if (&x == this) return *this;
    if (x.size() > capacity()) {
      deallocate();
      initialize(x.size());
    }
    std::copy(x.begin(), x.end(), begin());
    finish = begin() + difference_type(x.size());
    return *this;
  }
  void reserve(size_type n) {
    if (capacity() < n) {
      dword* q = bit_alloc(n);
      finish = std::copy(begin(), end(), iterator(q, 0));
      deallocate();
      start = iterator(q, 0);
      end_of_storage = q + (dword) ((n + __WORD_BIT - 1)/__WORD_BIT);
    }
  }
  reference front() { return *begin(); }
  const_reference front() const { return *begin(); }
  reference back() { return *(end() - 1); }
  const_reference back() const { return *(end() - 1); }
  void push_back(bool x) {
    if (finish.p != end_of_storage)
      *finish++ = x;
    else
      insert_aux(end(), x);
  }
  void swap(__BVECTOR& x) {
    std::swap(start, x.start);
    std::swap(finish, x.finish);
    std::swap(end_of_storage, x.end_of_storage);
  }
  iterator insert(iterator position, bool x = bool()) {
    difference_type n = position - begin();
    if (finish.p != end_of_storage && position == end())
      *finish++ = x;
    else
      insert_aux(position, x);
    return begin() + n;
  }

  void insert(iterator position, const_iterator first, 
              const_iterator last) {
    if (first == last) return;
    size_type n = 0;
    distance(first, last, n);
    if (capacity() - size() >= n) {
      std::copy_backward(position, end(), finish + n);
      std::copy(first, last, position);
      finish += n;
    }
    else {
	  // don't have max(size(),n).  ajd27
		size_type maxsn = (size()>n) ? size() : n;
      size_type len = size() + maxsn;
      dword* q = bit_alloc(len);
      iterator i = std::copy(begin(), position, iterator(q, 0));
      i = std::copy(first, last, i);
      finish = std::copy(position, end(), i);
      deallocate();
      end_of_storage = q + (len + __WORD_BIT - 1)/__WORD_BIT;
      start = iterator(q, 0);
    }
  }

  void insert(iterator position, const bool* first, const bool* last) {
    if (first == last) return;
    size_type n = 0;
    distance(first, last, n);
    if (capacity() - size() >= n) {
      std::copy_backward(position, end(), finish + n);
      std::copy(first, last, position);
      finish += n;
    }
    else {
		// ajd27 no max again
 		size_type maxsn = (size()>n) ? size() : n;
     size_type len = size() + maxsn;
      dword* q = bit_alloc(len);
      iterator i = std::copy(begin(), position, iterator(q, 0));
      i = std::copy(first, last, i);
      finish = std::copy(position, end(), i);
      deallocate();
      end_of_storage = q + (len + __WORD_BIT - 1)/__WORD_BIT;
      start = iterator(q, 0);
    }
  }
  
  void insert(iterator position, size_type n, bool x) {
    if (n == 0) return;
    if (capacity() - size() >= n) {
      std::copy_backward(position, end(), finish + difference_type(n));
      std::fill(position, position + difference_type(n), x);
      finish += difference_type(n);
    }
    else {
		// ajd27 no max again
 		size_type maxsn = (size()>n) ? size() : n;
       size_type len = size() + maxsn;
      dword* q = bit_alloc(len);
      iterator i = std::copy(begin(), position, iterator(q, 0));
      std::fill_n(i, n, x);
      finish = std::copy(position, end(), i + difference_type(n));
      deallocate();
      end_of_storage = q + (len + __WORD_BIT - 1)/__WORD_BIT;
      start = iterator(q, 0);
    }
  }

  void insert(iterator pos, int n, bool x)  { insert(pos, (size_type)n, x); }
  void insert(iterator pos, long n, bool x) { insert(pos, (size_type)n, x); }

  void pop_back() { --finish; }
  iterator erase(iterator position) {
    if (position + 1 != end())
      std::copy(position + 1, end(), position);
    --finish;
    return position;
  }
  iterator erase(iterator first, iterator last) {
    finish = std::copy(last, end(), first);
    return first;
  }
  void resize(size_type new_size, bool x = bool()) {
    if (new_size < size()) 
      erase(begin() + difference_type(new_size), end());
    else
      insert(end(), new_size - size(), x);
  }
  void clear() { erase(begin(), end()); }
};


inline bool operator==(const bit_vector& x, const bit_vector& y) {
  return x.size() == y.size() && std::equal(x.begin(), x.end(), y.begin());
}

inline bool operator<(const bit_vector& x, const bit_vector& y) {
  return std::lexicographical_compare(x.begin(), x.end(), y.begin(), y.end());
}


#undef __BVECTOR

#endif /* __SGI_STL_INTERNAL_BVECTOR_H */

// Local Variables:
// mode:C++
// End:
