
#include "vtkRenderWindow.h"
#include "vtkRenderer.h"
#include "vtkRenderWindowInteractor.h"
#include "vtkGlyph3D.h"
#include "vtkPolyDataMapper.h"
#include "vtkLookupTable.h"
#include "vtkCubeSource.h"
#include "vtkActor.h"
#include "vtkScalars.h"
#include "vtkStructuredPointsReader.h"
#include "vtkExtractVOI.h"
#include "vtkThresholdPoints.h"

vtkRenderWindow *renWin;

void main( int argc, char **argv )
{
  FILE* warnings, *info;
  int minToDisplay = 1;
  int maxToDisplay = 10000;

  warnings = stdout;
  info = stdout;

  if ( argc < 2 ) {
    fprintf( warnings, "Usage: vtkview <filename>\n"
                       "       vtkview <filename> <minavalanche> <maxavalanche>\n" );
    fprintf( warnings, "The file is a VTK structured points input file.\n" );
  }
  if ( argc > 3 ) {
    minToDisplay = atoi( argv[2] );
    maxToDisplay = atoi( argv[3] );
  }

  vtkStructuredPointsReader* reader = vtkStructuredPointsReader::New();
    reader->SetFileName( argv[1] );
    reader->OpenVTKFile();
    if ( !reader->IsFileStructuredPoints() ) {
      fprintf( warnings, "The file %s is not a structured points file.\n", argv[1] );
      return;
    }
    int scalarSetCnt = reader->GetNumberOfScalarsInFile();
    fprintf( info, "There are %d scalar datasets in the file.\n", scalarSetCnt );

  reader->Update();
  int ptCnt = reader->GetOutput()->GetNumberOfPoints();
  fprintf( info, "Read %d points from %s.\n", ptCnt, argv[1] );
  int scalarCnt = reader->GetOutput()->GetPointData()->GetScalars()->GetData()->GetSize();
  fprintf( info, "Read %d scalars from %s.\n", scalarCnt, argv[1] );
  int maxScalar = reader->GetOutput()->GetPointData()->GetScalars()->GetRange()[1];
  fprintf( info, "Max avalanche is %d.\n", maxScalar );


  vtkRenderer* ren = vtkRenderer::New();
  renWin = vtkRenderWindow::New();
    renWin->AddRenderer(ren);
    renWin->SetSize (600,600);

  vtkRenderWindowInteractor* iren = vtkRenderWindowInteractor::New();
    iren->SetRenderWindow(renWin);

  vtkExtractVOI* voi = vtkExtractVOI::New();
  voi->SetInput( reader->GetOutput() );
  //voi->SetVOI( 0,15, 0,15, 2,15 );

  // This command chooses which avalanches to show.
  vtkThresholdPoints* threshold = vtkThresholdPoints::New();
  threshold->ThresholdBetween( minToDisplay, maxToDisplay );
  threshold->SetInput( voi->GetOutput() );

  vtkCubeSource* cubeSource = vtkCubeSource::New();
  // cubeSource->SetXLength(); ...

  vtkGlyph3D* spins = vtkGlyph3D::New();
  spins->SetInput( threshold->GetOutput() );
  //spins->SetInput( reader->GetOutput() );
  spins->SetSource( cubeSource->GetOutput() );
  spins->SetScaleFactor( 1.0 );
  spins->SetScaleModeToDataScalingOff();
  //spins->SetColorModeToColorByScalar();
  // spins->SetRange( MINSCALAR, MAXSCALAR );
  // spins->GetRange( range );

  vtkLookupTable* lut = vtkLookupTable::New();
  lut->SetNumberOfColors( 256 );
  //lut->SetHueRange( 0, 0.667 );
  lut->SetHueRange( 0, 0.667 );
  // Give them color but completely transparent.
  //lut->SetAlphaRange( 0, 0 );
  lut->Build();
  // This is an attempt to show only certain avalanches.
  /*  float lutEntry[4];
  // make the interesting ones opaque.
  for ( int lutIdx = 36; lutIdx<102; lutIdx++ ) {
    lut->GetTableValue( lutIdx, lutEntry );
    lutEntry[3] = 1.0;
    lut->SetTableValue( lutIdx, lutEntry );
  }
*/
  vtkPolyDataMapper* spinsMapper = vtkPolyDataMapper::New();
  spinsMapper->SetInput( spins->GetOutput() );
  spinsMapper->SetScalarRange( minToDisplay, maxToDisplay );
  spinsMapper->SetLookupTable( lut );

  vtkActor* spinsActor = vtkActor::New();
  spinsActor->SetMapper( spinsMapper );

  ren->AddActor( spinsActor );

  ren->SetBackground( 0, 0, 0 );
  
  renWin->Render();

  iren->Start();

  reader->Delete();
  ren->Delete();
  renWin->Delete();
  cubeSource->Delete();
  spins->Delete();
  lut->Delete();
  threshold->Delete();
  spinsMapper->Delete();
  spinsActor->Delete();
  voi->Delete();
}