// KMCLattice.cpp: implementation of the KMCLattice class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "KMC.h"
#include "KMCLattice.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

KMCLattice::KMCLattice(BravaisLatticeWithBasis *bravais, int equiv[DIMENSION][DIMENSION])
{
	// Drew's registration with MOP: allows Get and Set functions to be called
	// from dialog boxes, in serialization, and for copying
	if (NotYetRegistered("KMCLattice",typeid(*this))) {
		AddType("KMCLattice",typeid(*this));
	}

	// Store information
	this->bravais = bravais;
	for (int equivVector=0; equivVector < DIMENSION; equivVector++)
		for (int coord=0; coord < DIMENSION; coord++)
			this->equiv[equivVector][coord] = equiv[equivVector][coord];

	// Find size of rectangular lattice
	FindSize(equiv);	// Sets size of rectangular array, and netSize

	// Allocate storage
	OrderParameterArray = new OrderParameterType* [bravais->GetNAtomsPerCell()];
	for (int atomInCell=0; atomInCell<bravais->GetNAtomsPerCell(); atomInCell++)
	{
		OrderParameterArray[atomInCell] = new OrderParameterType[netSize];
	}
}

KMCLattice::~KMCLattice()
{
	for (int atomInCell=0; atomInCell<bravais->GetNAtomsPerCell(); atomInCell++)
		delete [] OrderParameterArray[atomInCell];
	delete [] OrderParameterArray;
}

/**
  * We do a rather simple job of pushing all the sites into a rectanglar box: we assume
  * (1) that all components of equiv[i] are positive for each i
  * (2) that the various equiv[i] sit on the boundary of the box needed to store them
  *     (i.e., that the equiv[i] are as "small" as possible)
  * (3) that the determinant of the equiv vectors is the same as that of the sides of the box
  * The last condition should be dropped: violating it is natural if you want
  * skew boundary conditions on two simultaneous axes. This requires a test when 
  * setting spins to make sure when a spin has two duplicates in the box they both
  * get set.
  */
void KMCLattice::FindSize(int equiv[DIMENSION][DIMENSION])
{
	netSize = 1;
	for (int coord=0; coord<DIMENSION; coord++)
	{
		size[coord]=0;
		for (int equivVector=0; equivVector<DIMENSION; equivVector++)
		{
			if (size[coord] < equiv[equivVector][coord])
				size[coord] = equiv[equivVector][coord];
		}
		netSize *= size[coord];
	}
	assert(IntegerDeterminant(equiv,DIMENSION) == netSize);	// Condition [3] above
}


int KMCLattice::IntegerDeterminant(const int M[][DIMENSION], const int dim)
{
  if(dim==1)
  {
    return M[0][0];
  }
  else if(dim==2)
  {
    return M[0][0]*M[1][1] - M[1][0]*M[0][1];
  }
  else 
  {
    int temp = 0., matrixtemp[DIMENSION][DIMENSION]; 
    int i, j, k;
    {
      for(i=0; i<dim; i++)
	{
	  for(j=1;j<dim;j++)
	    {
	      for(k=0;k<i;k++) matrixtemp[j-1][k] = M[j][k];
	      for(k=i+1;k<dim;k++) matrixtemp[j-1][k-1] = M[j][k];
	    }
	  temp += M[0][i]*IntegerDeterminant(matrixtemp,dim-1)* (i%2 == 0 ? 1. : -1.); 
	}
      return temp;
    }
  }
}

int KMCLattice::IndexFromPosition(int position[DIMENSION])
{
	// This is a bit ugly, and surely pretty slow!
	int index = 0;
	int stride = 1;
	// Put position into box!
	int coord;	// Stupid Microsoft incompatibility

	bool done = false;
	while (!done)
	{
		for (coord=0; coord < DIMENSION; coord++)
		{
			while (position[coord] < 0)
			{
				for (int coord2 = 0; coord2 < DIMENSION; coord2++)
				{
					position[coord2] += equiv[coord][coord2];	// equiv[i] has max of ith coord
				}
			}
			while (position[coord] > size[coord])
			{
				for (int coord2 = 0; coord2 < DIMENSION; coord2++)
				{
					position[coord2] -= equiv[coord][coord2];	// equiv[i] has max of ith coord
				}
			}
		}
		// Later additions can push earlier ones over the line
		done = true;
		for (coord = 0; coord < DIMENSION; coord++)
		{
			done = done && (position[coord] >= 0) && (position[coord] < size[coord]);
		}
	}

	for (coord=DIMENSION-1; coord >= 0; coord --)
	{
		assert(position[coord] >= 0 && position[coord] < size[coord]);	// XXX
		index += stride * position[coord];
		stride *= size[coord];		// Last multiply could be skipped
	}	
	return index;
}

void KMCLattice::PositionFromIndex(int index, int *position)
{
	// The pointer should be to an array of ints
	// something like (int (*)[DIMENSION]) position
	// but I haven't gotten it working: JPS

	for (int coord = DIMENSION-1; coord > 0; coord --)
	{
		position[coord] = index % size[coord];
		index /= size[coord];
	}
	position[0] = index % size[0];

}

void KMCLattice::SetOrderParameter(int index, OrderParameterType value)
{
	SetOrderParameter(0, index, value);
}

void KMCLattice::SetOrderParameter(int atomInCell, int index, OrderParameterType value)
{
	OrderParameterArray[atomInCell][index] = value;
}

OrderParameterType KMCLattice::GetOrderParameter(int index) const
{
	return GetOrderParameter(0, index);
}

OrderParameterType KMCLattice::GetOrderParameter(int atomInCell, int index) const
{
	return OrderParameterArray[atomInCell][index];
}
