// KMCLattice.h: interface for the KMCLattice class.
//
//////////////////////////////////////////////////////////////////////

#ifndef KMCLATTICE_H
#define KMCLATTICE_H

#ifndef SWIG
#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000
#endif // SWIG

#include "Introspective.h"
#include "BravaisLatticeWithBasis.h"

#define OrderParameterType int	// Change to template later
class KMCLattice  : public Introspective
{
public:
	//! KMCLattice: builds Bravais lattice with periodic BC given by array of equiv sites
	/** KMCLattice: builds lattice for Kinetic Monte Carlo simulations.
	  * 
	  * Given a Bravais lattice, KMCLattice will build a lattice big enough
	  * to hold all the inequivalent atoms
	  * 
	  * equiv gives the periodicity of the boundary conditions in terms of the 
	  * Bravais vectors.
	  * For each i, equiv[i] == {0,0,0}
	  * so equiv[i] * bravais->GetLatticeVector(i) summed over i is the same as the origin
	  * We do a rather simple job of pushing all the sites into a rectanglar box: we assume
	  * (1) that all components of equiv[i] are positive for each i
	  * (2) that the various equiv[i] sit on the boundary of the box needed to store them
	  *     (i.e., that the equiv[i] are as "small" as possible)
	  * (3) that the determinant of the equiv vectors is the same as that of the sides of the box
      * (4) that the equiv vectors are ordered so that the i_th vector has the largest value
	  *     of the i_th coordinate.
      * The third condition should be dropped: violating it is natural if you want
	  * skew boundary conditions on two simultaneous axes. This requires a test when 
	  * setting spins to make sure when a spin has two duplicates in the box they both
	  * get set.
	  * The fourth condition also should be dropped. Ideally we'd have a subroutine that
	  * finds the right linear combination of the input equiv vectors to satisfy (3) and (4),
	  * if that is possible in general.
	  *
	  * We haven't checked that our algorithm works for general lattices, or even if
	  * one can satisfy (1-3) for a general lattice...
	  * 
	  * There should be one internal lattice per basis atom if the lattice has
	  * a basis
	  *
	  * Examples should be discussed here:
	  * Square array
	  * FCC with cubic bravais unit cell: four internal arrays
	  * FCC with primitive unit cell: <111> perpendicular to slowest array index
	  *    (best use of cache for growing <111> surfaces, as in Joachim's KMC/MD simulations)
	  * FCC with primitive unit cell and <100> perpendicular to slowest array index
	  *    (Josh's KMC/MD should use this or cubic bravais unit cell)
	  * 
	  * The KMCLattice can be accessed either as a one-dimensional array
	  * or via three integers: it internally maps the three integers into
	  * 
	  * \ingroup KineticMonteCarlo
	  */

	KMCLattice(BravaisLatticeWithBasis *bravais, int equiv[DIMENSION][DIMENSION]);
	virtual ~KMCLattice();
	int IndexFromPosition(int position[DIMENSION]);	// Gets 1-d position in OrderParameterArray from vector coord
	void PositionFromIndex(int index, int *position);	// Gets vector position in Bravais vectors from 1-d position 
	void SetOrderParameter(int index, OrderParameterType value);
	void SetOrderParameter(int atomInCell, int index, OrderParameterType value);
	OrderParameterType GetOrderParameter(int index) const;
	OrderParameterType GetOrderParameter(int atomInCell, int index) const;
	int GetNetSize() const {return netSize;}
	BravaisLatticeWithBasis *GetBravais() const {return bravais;}

private:

	//! IntegerDeterminant: gives determinant of integer matrix
	/** IntegerDeterminant: gives determinant of integer matrix
	  * adapted from Thierry's BravaisLatticeWithBasis::Determinant
	  * this routine is very basic, but we'll use it a few times and for small matrices (TC)
	  */
	int IntegerDeterminant(const int M[DIMENSION][DIMENSION], const int dim);
	//! FindSize: finds size of rectangle needed to encompass region defined by equiv
	void FindSize(int equiv[DIMENSION][DIMENSION]);

	BravaisLatticeWithBasis *bravais;
	int equiv[DIMENSION][DIMENSION];
	int size[DIMENSION];	//! size of rectangular lattice
	int netSize;			//! Total number of sites in lattice
	OrderParameterType **OrderParameterArray;	//! First index for basis, second for array given basis

};

#endif // !defined(KMCLATTICE_H)
