/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
#include "DigitalMaterial.h"

#ifndef DMPARALLEL 
#error "This application should be compiled with DMPARALLEL set to 1 in DigitalMaterial.h / makevars.mk"
#endif //DMPARALLEL

#if DIMENSION != 3
#error "This application is designed to be compiled in 3 D"
#endif // DIMENSION != 3

#include "mpi.h"

#ifdef WIN32
#include "WinProcessFactory.h"
#else
#include "UnixProcessFactory.h"
#endif //WIN32

#include "ProtocolFactory.h"
#include "HomogeneousDomainSubdivision.h"
#include "Migrator.h"
#include "Synchronizer.h"
#include "EMTPotential.h"
#include "ParallelPotential.h"
#include "3DLattices.h"
#include "RectangularClusterInitializer.h"
#include "ParallelAtomsInitializer.h"
#include "DynamicListOfAtoms.h"
#include "VerletAtomsMover.h"
#include "ParallelNeighborLocator.h"
#include "CellNeighborList.h"

using namespace std;

//! A simple example showing how to set up a simulation in parallel.

void main(int argc, char **argv)
{
	MPI_Init(&argc, &argv);

	try
	{
#ifdef WIN32
	WinProcessFactory::Create();
#else
	UnixProcessFactory::Create();
#endif
    // Set the default communication protocol to MPI_World
    ProtocolFactory::Instance()->SetProtocol(ProtocolFactory::World);
	const int nProcs = dmProtocol->GetNumberOfProcessors();
	int tmp0 = (int) pow(nProcs,1/3.);
	int tmp1 = (int) sqrt(double(nProcs)/tmp0);
	int coord;

	int cpuLayout[DIMENSION];
	cpuLayout[0] = tmp0;
	cpuLayout[1] = tmp1;
	cpuLayout[2] = nProcs/tmp0/tmp1;

    const int myProc = dmProtocol->GetProcessorNumber();

	if (myProc == 0) cout <<  "cpu layout = [" << cpuLayout[0] << " , " 
		<< cpuLayout[1] << " , " << cpuLayout[2] << " ]" << endl;

	// CutLennardJonesPotential potential;
	EMTPotential potential;
	ParallelPotential parapotential(&potential);

	double cutoff      = potential.GetCutoffDistance();
	double lattSpacing = potential.GetLengthScale()*sqrt(2.0);
	
	FCCLattice lattice(lattSpacing);
	double shift[DIMENSION] = {0.1,0.03,0.14};
	lattice.SetOrigin(shift);

	// Size of the cube (we divide by the cpuLayout)
	double cubeSize[DIMENSION];
	for(coord=0; coord<DIMENSION; ++coord)
		cubeSize[coord] = 30.0 *lattSpacing; // / cpuLayout[coord];

    // The atoms initializer and its decorator
	RectangularClusterInitializer rect(cubeSize, &lattice);
    ParallelAtomsInitializer paraRect(&rect);
	DynamicListOfAtoms atoms;

    paraRect.Create(&atoms);

	cout << "processor " << myProc << " has " << atoms.GetNumber() << " atoms" << endl;

	HomogeneousDomainSubdivision domain(cpuLayout, myProc, cutoff);
	//! We should solve this problem about the const_cast
	DMArray<double, DIMENSION> *pos = 
		const_cast<DMArray<double, DIMENSION>*>(atoms.GetCartesianPositions());
	Synchronizer synchronizer(pos);
	Migrator migrator(&domain, pos);

	CellNeighborList *neighborLoc = new CellNeighborList(cutoff, 0.05*cutoff);
	ParallelNeighborLocator *paranl = new ParallelNeighborLocator(&migrator, neighborLoc);

	atoms.SetNeighborLocator(paranl);   // this will migrate the atoms if it hasn't been done
    
	cout << "On processor " << myProc << ", we have:" << atoms.GetNumber() << " and " << 
	    atoms.GetCartesianPositions()->GetSynchronizationKit()->GetGhosts()->Size() <<
		" ghost atoms after migration" << endl;
    
	double E = parapotential.CalculateEnergy(&atoms);
	if (myProc == 0) cout << "Energy = " << E << endl;

	
	VerletAtomsMover mover(&parapotential, 0.01, 10);

	for (int step = 0; step < 3; ++step)
	{
		mover.Move(&atoms);
	    E = parapotential.CalculateEnergy(&atoms);
		if (myProc == 0) cout << E << endl;
	} 

	delete paranl;
	delete neighborLoc;

	}
	catch (DMException e){
		std::cout<< e;}

	MPI_Finalize();
}

