// AtomicSurfaceInitializer.cpp: implementation of the AtomicSurfaceInitializer class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
#include "DigitalMaterial.h"

#include "AtomicSurfaceInitializer.h"
#include "3DLattices.h"
#include "SkewPeriodicBoundaryConditions.h"
#include "DynamicListOfAtoms.h"
#include "FixedBody.h"
#include "DMException.h"
#include "SimpleMatrixInverter.h"
#include "ParallelepipedClusterInitializer.h"

#include "DMLastHeader.h"


#if DIMENSION == 3

using namespace std;

const double AtomicSurfaceInitializer::TINY = 1.0e-13;


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

AtomicSurfaceInitializer::AtomicSurfaceInitializer(BravaisLatticeWithBasis *lattice, 
		int millerIndices[DIMENSION], int mode, SkewPeriodicBoundaryConditions *spbc)
{
	assert(DIMENSION==3);
	minimalLength = 6;
	depthOfFixedLayer = 6.;  // this should be set to the cutoff distance of the potential
	depthOfDynamicLayer = 24;
	zCenter = 0.;
	this->mode = mode;
	this->spbc = spbc;
	this->millerIndices[0] = 1; // we set a default millerIndices. This is needed because when we call SetBravaisLattice we need an millerIndices, and when we call SetMillerIndices we need a BravaisLattice...
	for(int coord=1; coord<DIMENSION; ++coord) this->millerIndices[coord] = 0;
	this->lattice = lattice;
	SetMillerIndices(millerIndices);
}

AtomicSurfaceInitializer::~AtomicSurfaceInitializer()
{

}

void AtomicSurfaceInitializer::SetMode(int mode)
{
	ASSERT(mode > 0 && mode < 2);
	if ( mode < 0 || mode >= 2)
		throw DMException("AtomicSurfaceInitializer","SetMode","Invalid mode number");
	this->mode = mode;
}


void AtomicSurfaceInitializer::SetBravaisLattice(BravaisLatticeWithBasis *lattice)
{
  this->lattice = lattice;
  ResetPrivateData();
}

BravaisLatticeWithBasis *AtomicSurfaceInitializer::GetBravaisLattice() const
{
  return lattice;
}

void AtomicSurfaceInitializer::SetDepthOfFixedLayer(double depthOfFixedLayer)
{
        this->depthOfFixedLayer = depthOfFixedLayer;
        ResetPrivateData();
}

void AtomicSurfaceInitializer::SetDepthOfDynamicLayer(double depthOfDynamicLayer)
{
        this->depthOfDynamicLayer = depthOfDynamicLayer;
        ResetPrivateData();
}

void AtomicSurfaceInitializer::SetMillerIndices(const int millerIndices[DIMENSION])
{
	int numberZero = 0;
    int coord;
	for(coord = 0; coord < DIMENSION; ++coord )
	{
		if(millerIndices[coord] == 0)
			++numberZero;
	}
	
	ASSERT(numberZero < 3);
	if(numberZero == 3)
		throw DMException("AtomicSurfaceInitializer","SetMillerIndices","Miller indices all zero");

	for(coord=0; coord<DIMENSION; coord++) this->millerIndices[coord] = millerIndices[coord];
	ResetPrivateData();
}

void AtomicSurfaceInitializer::SetMinimalLength(double minimalLength)
{
	this->minimalLength = minimalLength;
	ResetPrivateData();
}

void AtomicSurfaceInitializer::ResetPrivateData()
{
	int coord, vec;
	int k = millerIndices[0];
	int l = millerIndices[1];
	int m = millerIndices[2];	
	
	// Reset lattice orientation so we have the correct conventional basis vectors
	lattice->ResetOrientation();

	const double (&latticeVectors)[DIMENSION][DIMENSION] = lattice->GetLatticeVectors();
	TinyVector<double, DIMENSION> reciprocalVectors[DIMENSION];
	reciprocalVectors[0] = TinyVector<double, DIMENSION>(latticeVectors[1]).Cross(latticeVectors[2]);
	reciprocalVectors[1] = TinyVector<double, DIMENSION>(latticeVectors[2]).Cross(latticeVectors[0]);
	reciprocalVectors[2] = TinyVector<double, DIMENSION>(latticeVectors[0]).Cross(latticeVectors[1]);

	// will need lengths --- to divide by to get unit vectors
	double lengths[DIMENSION];
	for( vec = 0; vec < DIMENSION; ++vec )
		lengths[vec] = 0.;

	// Construct the surface vectors in the original coordinate system.
	// The third surface vector---the normal to the surface---won't be
	// needed for the boundary conditions, but is needed to construct the 
	// rotation matrix. The second one isn't needed for the rotation matrix
	// (for that we need three orthonormal vectors so we take the cross product
	// of the first and third, called yVect) but will be used for the
	// boundary conditions


	// we can construct three lattice vectors which are perpendicular to the 
	// surface normal by taking the cross product of the normal vector with
	// each reciprocal lattice vector in turn, which gives the formulas below.
	// Clearly since these vectors are all in the plane perpendicular to the
	// normal they can't be independent. What's more important is that one may
	// be zero---this will happen if the corresponding miller indices are zero
	// (at least one is non-zero). We must assign two nonzero vectors to be 
	// the surface vectors. Note that the latticeVectors are a basis
 	// so these surface vectors can only be zero when all the coefficients are.
	for( coord = 0; coord < DIMENSION; ++coord )
	{
		vec = 0;
		surfaceVectors[vec][coord] = m * latticeVectors[1][coord] - l * latticeVectors[2][coord]; 
		if(m != 0 || l != 0) ++vec; // got one!		
		surfaceVectors[vec][coord] = k * latticeVectors[2][coord] - m * latticeVectors[0][coord]; // if first one was nonzero use this formula for the second vector; otherwise use it for the first 
		if( (k == 0 && m && 0) || vec == 0 || m == 0)
			// ie repectively: (i) second formula is useless or (ii) first one
			// was or (iii) if m is zero the first two formulas give parallel
			// vectors (parallel to latticeVector[2])
			// In any case use the third formula for the second vector
			surfaceVectors[1][coord] = l * latticeVectors[0][coord] - k * latticeVectors[1][coord];

		surfaceVectors[2][coord] = k * reciprocalVectors[0][coord] + l * reciprocalVectors[1][coord] + m * reciprocalVectors[2][coord];

		for(vec = 0; vec < DIMENSION; ++vec )
			lengths[vec] += surfaceVectors[vec][coord] * surfaceVectors[vec][coord];
	}

	// we need three orthonormal vectors to make the rotation matrix.
	// These will be the first and third surface vectors and their cross
	// product (third with first, to keep a right handed triad)

	TinyVector<double, DIMENSION> yVect = TinyVector<double, DIMENSION>(surfaceVectors[2]).Cross(surfaceVectors[0]);
	// need to redo the length for  yVect
	lengths[1] = 0.;
	for( coord = 0; coord < DIMENSION; ++coord )
		lengths[1] += yVect[coord] * yVect[coord];
	
	for( vec = 0; vec < DIMENSION; ++vec )
		lengths[vec] = sqrt(lengths[vec]);
	
	// the rotation is such as to bring the normal of the given surface
	// parallel to the z-direction, the first surface vector parallel to the 
	// x-direction and yVect parallel to the y-direction (hence the name)

	// the matrix for the inverse of this operation is simply that whose 
	// columns are the specified unit vectors. So we construct the transpose
	// of that to get the matrix we want.
	double rotation[DIMENSION][DIMENSION];
	for( coord = 0; coord < DIMENSION; ++coord )
		{
			rotation[0][coord] = surfaceVectors[0][coord]/lengths[0]; 
			rotation[1][coord] = yVect[coord]/lengths[1]; 
			rotation[2][coord] = surfaceVectors[2][coord]/lengths[2]; 
		}
	
	// Set new orientation for the lattice
	lattice->RotateLatticeVectors(rotation);

	// next rotate the surface vectors. 
	double newSVector[DIMENSION][DIMENSION];
	for( vec = 0; vec < DIMENSION; ++vec )
	{
		for( coord = 0; coord < DIMENSION; ++coord )
			newSVector[vec][coord] = 0.;
		
		for( coord = 0; coord < DIMENSION; ++coord )
			for( int coord2 = 0; coord2 < DIMENSION; ++coord2 )
				newSVector[vec][coord] += rotation[coord][coord2] * surfaceVectors[vec][coord2];
	}
	
	for( vec = 0; vec < DIMENSION; ++vec )
		for( coord = 0; coord < DIMENSION; ++coord )
			surfaceVectors[vec][coord] = newSVector[vec][coord];

	// The first one should now be parallel to the x-axis and the second should
	// now be in the xy plane, and the third parallel to the z-axis. check this
	double check = fabs(surfaceVectors[0][1]) + fabs(surfaceVectors[0][2]) + fabs(surfaceVectors[1][2])+ fabs(surfaceVectors[2][0]) +  fabs(surfaceVectors[2][1]);
	ASSERT(check < TINY);

	// Now make sure the cell is long enough and multiply by an appropriate integer if not. Similarly for the width, but there it's really the part of the second surface vector that's perpendicular to the first that counts
	superCellLength = 0., superCellWidth = 0.; 
	double dotProduct = 0.; 
	for( coord = 0; coord < DIMENSION; ++coord )
	{
		superCellLength += surfaceVectors[0][coord] * surfaceVectors[0][coord];
		superCellWidth += surfaceVectors[1][coord] * surfaceVectors[1][coord];
		dotProduct += surfaceVectors[0][coord] * surfaceVectors[1][coord];
	}
	superCellLength = sqrt(superCellLength);
	superCellWidth = sqrt(superCellWidth);
	double theta = acos(dotProduct/superCellLength/superCellWidth);
	superCellWidth *= sin(theta);


	if(superCellLength < minimalLength)
	{
		int factor = 1 + (int) (minimalLength/superCellLength);
		for( coord = 0; coord < DIMENSION; ++coord )
			surfaceVectors[0][coord] *= factor;
		superCellLength *= factor;
	}

	if(superCellWidth < minimalLength)
	{
		int factor = 1 + (int) (minimalLength/superCellWidth);
		for( coord = 0; coord < DIMENSION; ++coord )
			surfaceVectors[1][coord] *= factor;
		superCellWidth *= factor;
	}
	// now the vertical surface vector should be made very large (better to have smart SPBC to turn off periodicity in this direction)
	surfaceVectors[2][2] = 1.; // other components should be zero
}

/*!
  This function should be called first if using mode 1. The reason it is made
  separate is to allow more than one block to be created in mode 1. We don't 
  want to blindly reset the lengths each time. Rather we set once
  and make sure in between calls to Create that things have not changed to make
  the boundary conditions inconsistent.

 */

void AtomicSurfaceInitializer::ConfigureBoundaryConditions()
{
	double superCellVectors[DIMENSION][DIMENSION];
	for( int coord = 0; coord < DIMENSION; ++coord )
		for( int vec = 0; vec < DIMENSION; ++vec )
			superCellVectors[coord][vec] = surfaceVectors[vec][coord];
	
	superCellVectors[0][2] = superCellVectors[1][2] = 0.; // instead of 10^-16 or whatever the had from the rotation
	spbc->SetSuperCellVectors(superCellVectors);
}


void AtomicSurfaceInitializer::Create(ListOfAtoms *atoms)
{
	// the first thing to do is to set the SPBC (mode 0) or check SPBC (mode 1)
	// because EnforceBoundaryConditions will be called
	if(!spbc) 
		throw DMException("AtomicSurfaceInitializer","Create",
						  "No SkewPeriodic Boundary Conditions has been passed to AtomicSurface");

	ASSERT(!atoms->GetShape().IsLeaf() || atoms->GetNumber() == 0);

	if( 0 == mode )
		CreateModeZero(atoms);
	else
		CreateModeOne(atoms);

	atoms->SetBoundaryConditions(spbc); // is this worth doing for mode 1?	
}


void AtomicSurfaceInitializer::CreateModeZero(ListOfAtoms *atoms)
{
	int coord, vec;
	ConfigureBoundaryConditions();
	DynamicListOfAtoms *freeAtoms = new DynamicListOfAtoms(0, spbc);
	DynamicListOfAtoms *fixedAtoms = new DynamicListOfAtoms(0, spbc);
	
	// create a few layers of fixed atoms	
	double boundaryVectors[DIMENSION][DIMENSION];
	double clusterCenter[DIMENSION];
	for( coord = 0; coord < DIMENSION; ++coord )
	{
		clusterCenter[coord] = 0.;
		for( vec = 0; vec < 2; ++vec )// not DIMENSION!
		{
			boundaryVectors[coord][vec] = surfaceVectors[vec][coord];
			clusterCenter[coord] += 0.5 * boundaryVectors[coord][vec];
		}
	}
	boundaryVectors[0][2] = boundaryVectors[1][2] = 0.;
	boundaryVectors[2][2] = depthOfFixedLayer;
	clusterCenter[2] = 0.5 * depthOfFixedLayer * 1.000000001;
	ParallelepipedClusterInitializer parallelepiped(boundaryVectors, lattice);
	parallelepiped.SetCenter(clusterCenter);
	parallelepiped.Create(fixedAtoms);
	// create a few layers of moving atoms	
	boundaryVectors[2][2] = depthOfDynamicLayer;
	parallelepiped.SetBoundaryVectors(boundaryVectors);
	clusterCenter[2] = depthOfFixedLayer + 0.5 * depthOfDynamicLayer;
	parallelepiped.SetCenter(clusterCenter);
	parallelepiped.Create(freeAtoms);
	// aggregate these two lists atoms
	atoms->Attach(fixedAtoms);
	atoms->Attach(freeAtoms);
	// shift the atoms along z such that the surface lies at z=0
	double maxHeight = 0, tmp;
	int at;
	for(at=0; at<freeAtoms->GetNumber();++at)
	{
		tmp = freeAtoms->GetCartesianPosition(at,2);
		if(tmp > maxHeight) maxHeight = tmp;
	}
	DMArray<double, DIMENSION> *increments = new DMArray<double, DIMENSION>(atoms->GetCartesianPositions()->GetShape());
	increments->Zero();
	for(at=0; at<atoms->GetNumber();++at)
		(*increments)[at][2] = -maxHeight;
	atoms->IncrementCartesianPositions(increments);
	delete increments;
	// attach a constraint on the fixed atoms 
	FixedBody *fix = new FixedBody(fixedAtoms->GetCartesianPositions());
	fixedAtoms->SetConstraint(fix);

	// following must be done after ClusterInitializer because the latter
	// changes typenames (in this casew sets them to nothing)
	std::string name = "Fixed Atoms in the surface";
	fixedAtoms->SetTypeName(name);
	name = "Free Atoms in the surface";
	freeAtoms->SetTypeName(name);	
	
} 

void AtomicSurfaceInitializer::CreateModeOne(ListOfAtoms *atoms)
{
	// check that boundary conditions have right lengths.
	int coord, vec;
	const double (&actualSuperCellVectors)[DIMENSION][DIMENSION] = spbc->GetSuperCellVectors();
	bool ok = true;
	
	for( coord = 0; coord < DIMENSION; ++coord )
		for( vec = 0; vec < 2; ++vec )
			ok &= (fabs(actualSuperCellVectors[coord][vec] - surfaceVectors[vec][coord]) < TINY);
	
	ok &= ( (fabs(actualSuperCellVectors[0][2]) < TINY) && ( fabs(actualSuperCellVectors[1][2]) < TINY) );
	ok &= ( actualSuperCellVectors[2][2] > 0. );
	
	
	ASSERT(ok);
	if(!ok)
		throw DMException("AtomicSurfaceInitializer","Create","Mode 1: inconsistent boundary conditions");
	
	
	// create a layer of atoms	
	double boundaryVectors[DIMENSION][DIMENSION];
	double clusterCenter[DIMENSION];
	for( coord = 0; coord < DIMENSION; ++coord )
	{
		clusterCenter[coord] = 0.;
		for( vec = 0; vec < 2; ++vec )// not DIMENSION!
		{
			boundaryVectors[coord][vec] = surfaceVectors[vec][coord];
			clusterCenter[coord] += 0.5 * boundaryVectors[coord][vec];
		}
	}
	boundaryVectors[0][2] = boundaryVectors[1][2] = 0.;
	boundaryVectors[2][2] = depthOfFixedLayer;
	clusterCenter[2] = zCenter;
	
	ParallelepipedClusterInitializer parallelepiped(boundaryVectors, lattice);
	parallelepiped.SetCenter(clusterCenter);
	parallelepiped.Create(atoms);
	
}



bool AtomicSurfaceInitializer::SimplifiedVector(const int k,const int l, const int m) // return true if the vector (klm) cannot be written as (k'l'm')i. (k > 0)
{
  if((k!=1) && (l%k == 0) && (m%k == 0)) return false;
  for(int i=2;i<=sqrt((double) k);i++)
      if((k%i == 0) && (l%i==0) && (m%i==0)) return false;
  return true;
}

double AtomicSurfaceInitializer::GetMaxSize() const
{
  double max=depthOfFixedLayer + 2.*depthOfDynamicLayer;
  if(max < superCellLength) max = superCellLength;
  if(max < superCellWidth ) max = superCellWidth;
  return max;
}


#endif //DIMENSION == 3
