// ClusterInitializer.h: interface for the ClusterInitializer class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef CLUSTERINITIALIZER_H
#define CLUSTERINITIALIZER_H

#include "DMFirstHeader.h"
#include "DigitalMaterial.h"
#include "AtomsInitializer.h"

class BravaisLatticeWithBasis;
template<class Type, int N> class DMArray;

//! Base class for creation of clusters of atoms of certain shapes.
/*!	This is a Factory Method base class.
Each implementation of ClusterInitializer will need to define an Inside function,
set up appropriate internal variables for the shape and size of the cluster,
and set up the maxSize array.
*/

class ClusterInitializer: public AtomsInitializer
{
    friend class CompositeClusterInitializer;
public:
    //! Returns a rough estimate of the largest dimension of the cluster 
    /*! in real units. This is meant to be useful for graphics */
    double GetMaxSize() const;

    //! Constructor.
    /*! When overloaded, must set up maxSize array in initializer
    */
    ClusterInitializer(const BravaisLatticeWithBasis *bravais);
    //! Destructor.
    virtual ~ClusterInitializer();
    
    //! Factory method to do the actual creation.
    void Create(ListOfAtoms *atoms);
    
    //! Set the center of the cluster.
    void SetCenter(double clusterCenter[DIMENSION]);
    
    //! Returns the center of the cluster (in real units)
    /*! this is meant to be useful for graphics */
    const double * GetCenter() const {return clusterCenter;}
    
    //! Set a lattice
    void SetBravaisLattice(const BravaisLatticeWithBasis *bravais) ;

    //! Get the current lattice
    const BravaisLatticeWithBasis *GetBravaisLattice() const {return bravais;}

protected:

    //! The lattice that is used for the creation of the cluster.
    const BravaisLatticeWithBasis *bravais;
    
    //! Recursive function. Works in any dimension
    /*! The \a positions array must be preallocated and is filled by this function.
    \a nAtomsInArrays containts the current number of atoms in the cluster. There's 
    one number per type of atoms (and \a positions has one leaf per type) */
    void MakeSlab(int center[DIMENSION],
                  int subDimension,
                  DMArray<double, DIMENSION> *positions,
                  int *nAtomsInArrays); 

    // Called at the begining of Create.
    /*!It sets the bounds on the solid in the lattice coordinates and returns an
    upper bounds to the number of atoms contained in the cluster. 
    */
    virtual int SetMaxSize() = 0;  
    
    //! Tells if point X is inside cluster
    virtual bool Inside(double X[DIMENSION]) = 0;

    //! The coordinates of the cluster in real space
    double clusterCenter[DIMENSION];

    //! Cluster has maximum extent minSize[i]...maxSize[i] along basis vector i
    //@{
    int minSize[DIMENSION];
    int maxSize[DIMENSION];
    //@} 
};

#endif // !defined(CLUSTERINITIALIZER_H)
