// CompositeAtomsMover.cpp: implementation of the CompositeAtomsMover class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"

#include <iostream>
#include <vector>
#include "DigitalMaterial.h"
#include "CompositeAtomsMover.h"
#include "ExtendedNeighborLocator.h"
#include "ListOfAtoms.h"
#include "DMArray.h"
#include "SerialBuilder.h"

#include "DMLastHeader.h"


using namespace std;

static SerialBuilder<CompositeAtomsMover> *_Compositecreator = 
    new SerialBuilder<CompositeAtomsMover>("CompositeAtomsMover");

const string CompositeAtomsMover::className = string("CompositeAtomsMover");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////



CompositeAtomsMover::CompositeAtomsMover(AtomsMover *mover1,AtomsMover *mover2,double tolerance)
{
	this->mover1 = mover1;
	this->mover2 = mover2;

	maxSteps = 10;
	verbose = false;

	this->tolerance = tolerance;
	nIterations = 0;
}

CompositeAtomsMover::~CompositeAtomsMover()
{

}

void CompositeAtomsMover::SetInterestingLeaves(int numberOfInterestingLeaves1, int numberOfInterestingLeaves2, int *interestingLeaves1, int *interestingLeaves2)
{
	this->numberOfInterestingLeaves1 = numberOfInterestingLeaves1;
	this->numberOfInterestingLeaves2 = numberOfInterestingLeaves2;
	this->interestingLeaves1 = interestingLeaves1;
	this->interestingLeaves2 = interestingLeaves2;

}

void CompositeAtomsMover::Move(ListOfAtoms *atoms)
{
	nIterations = 0;
	// attach an ExtendedNeighborLocator to each branch
	NeighborLocator *subNeighborLocator = atoms->GetBranch(0)->GetNeighborLocator()->Copy();
	ExtendedNeighborLocator *extendedNeighborLocator1 = new ExtendedNeighborLocator(subNeighborLocator);
	ExtendedNeighborLocator *extendedNeighborLocator2 = dynamic_cast<ExtendedNeighborLocator*>(extendedNeighborLocator1->Copy());

	atoms->GetBranch(0)->SetNeighborLocator(extendedNeighborLocator1);
	atoms->GetBranch(1)->SetNeighborLocator(extendedNeighborLocator2);

	DMArray<double, DIMENSION> *ghostPositions1 = atoms->GetBranch(0)->GetCartesianPositions()->GetGhosts();
	DMArray<double, DIMENSION> *ghostPositions2 = atoms->GetBranch(1)->GetCartesianPositions()->GetGhosts();

	int numberOfLeaves1 = atoms->GetBranch(0)->NumberOfLeaves();
	int numberOfLeaves2 = atoms->GetBranch(1)->NumberOfLeaves();

	int t;
//	int *sizeOfGhostLeaves1 = new int[numberOfLeaves1];
//	int *sizeOfGhostLeaves2 = new int[numberOfLeaves2];
//	for(t=0;t<numberOfLeaves1;++t)
//		sizeOfGhostLeaves1[t] = 0;
//	for(t=0;t<numberOfLeaves2;++t)
//		sizeOfGhostLeaves2[t] = 0;
	vector<int> sizeOfGhostLeaves1(numberOfLeaves1, 0);
	vector<int> sizeOfGhostLeaves2(numberOfLeaves2, 0);

	
   for(t=0;t<numberOfInterestingLeaves1;++t)
	   sizeOfGhostLeaves1[interestingLeaves1[t]] = atoms->GetBranch(0)->GetLeaf(interestingLeaves1[t])->GetNumber();
   
   for(t=0;t<numberOfInterestingLeaves2;++t)
	   sizeOfGhostLeaves2[interestingLeaves2[t]] = atoms->GetBranch(1)->GetLeaf(interestingLeaves2[t])->GetNumber();

	// for now set the ghost shape for each branch to have the tree structure of the other branch. (Will have to figure this out more when we generalize to arbitrary numbers of branches).

	// Need to communicate which leaves of the "other" branch contain interesting atoms. There'll be an array ofleaf sizes.


//	ghostShape1 = atoms->GetBranch(1)->GetShape()->TreeStructure();
//	ghostShape1.Fill(sizeOfGhostLeaves2);
	ghostShape1 = Shape(atoms->GetBranch(1)->GetShape().GetTreeStructure(), sizeOfGhostLeaves2);
	ghostPositions1->SetShape(ghostShape1);

//	ghostShape2 = atoms->GetBranch(0)->GetShape().TreeStructure();
//	ghostShape2.Fill(sizeOfGhostLeaves1);
	ghostShape2 = Shape(atoms->GetBranch(0)->GetShape().GetTreeStructure(), sizeOfGhostLeaves1);
	ghostPositions2->SetShape(ghostShape2);
	
	do
    {
		// move each branch
		// "synchronize"--copy positions of ghost atoms for mover1
		
		for(t=0;t<numberOfInterestingLeaves2;++t)
		{
			int leaf = interestingLeaves2[t];
			ghostPositions1->GetLeaf(leaf)->Copy(atoms->GetBranch(1)->GetCartesianPositions()->GetLeaf(leaf));
		}
		
		mover1->Move(atoms->GetBranch(0));
		
		
		

		// "synchronize"--copy positions of ghost atoms for mover2

		for(t=0;t<numberOfInterestingLeaves1;++t)
		{
			int leaf = interestingLeaves1[t];
			ghostPositions2->GetLeaf(leaf)->Copy(atoms->GetBranch(0)->GetCartesianPositions()->GetLeaf(leaf));
		}
		mover2->Move(atoms->GetBranch(1));


		if(verbose)
		{
			cout << "Iteration: " << nIterations << endl;
			atoms->Notify();
		}
		++nIterations;
    }
    while(nIterations < maxSteps); // what's our criterion for stopping when it's not necessarily minimization?

//	delete [] sizeOfGhostLeaves1;
//	delete [] sizeOfGhostLeaves2;

	delete extendedNeighborLocator1;
	delete extendedNeighborLocator2;
	
}

////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////


void CompositeAtomsMover::Save(DMWriter *writer) const
{
	AtomsMover::Save(writer);	
}

void CompositeAtomsMover::Load(DMReader *reader)
{
	AtomsMover::Load(reader);
}
