// ConstraintVerletAtomsMover.cpp: implementation of the ConstraintVerletAtomsMover class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"

#include <iostream>
#include <stdexcept>
#include <fstream>

#include "DigitalMaterial.h"

#include "ConstraintVerletAtomsMover.h"
#include "ListOfAtoms.h"
#include "Constraint.h"
#include "DMArray.h"
#include "Constraint.h"
#include "Potential.h"
#include "lapack.h"
#include "SerialBuilder.h"
#include <iostream>


#include "DMLastHeader.h"

const int ConstraintVerletAtomsMover::STEPSBETWEENDISPLAY = 10;

static SerialBuilder<ConstraintVerletAtomsMover> *_ConsVerletcreator = 
    new SerialBuilder<ConstraintVerletAtomsMover>("ConstraintVerletAtomsMover");

using namespace std;

const string className = string("ConstraintVerletAtomsMover");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////



ConstraintVerletAtomsMover::ConstraintVerletAtomsMover(Potential *potential,double dt,double tolerance)
{
	this->potential = potential;
	this->dt = dt;
	this->tolerance = tolerance;

	generalizedCoordinates = 0;
	generalizedVelocities  = 0;
	generalizedForces      = 0;

	residualForceSq = 0.;
	nIterations = 0;
	generalizedForceSq = 0.;

	verbose = false;
	minimize = false;
}




ConstraintVerletAtomsMover::~ConstraintVerletAtomsMover()
{
	delete [] generalizedCoordinates;
	delete [] generalizedVelocities;
	delete [] generalizedForces;
	
	delete [] kineticMatrix;
}



void ConstraintVerletAtomsMover::Allocate(int nDOFs)
{

	this->nDOFs = nDOFs;

	if(generalizedCoordinates)
	{
		delete [] generalizedCoordinates;
		delete [] generalizedVelocities;
		delete [] generalizedForces;
		
		delete [] kineticMatrix;		
	}


	generalizedCoordinates = new double[nDOFs];
	generalizedVelocities  = new double[nDOFs];
	generalizedForces      = new double[nDOFs];

	kineticMatrix          = new double[nDOFs * nDOFs];


}


void ConstraintVerletAtomsMover::Move(ListOfAtoms *atoms)
{
	int paramIdx;

	Constraint *constraint = atoms->GetConstraint();
	DMArray<double,DIMENSION> *forces = new DMArray<double,DIMENSION>(atoms->GetCartesianPositions()->GetShape());

	
	if(nDOFs != constraint->NumberGeneralizedCoordinates())
		Allocate(constraint->NumberGeneralizedCoordinates());


	// set the generalizedCoordinates and generalizedVelocities from the constraint eg:

	constraint->GetGeneralizedCoordinates(generalizedCoordinates);	
	for(int paramIdx=0;paramIdx<nDOFs;++paramIdx)
	{
		generalizedVelocities[paramIdx] = 0;// don't see a way to get the correct velocities from the constraint unless the constraint explicitly stores them
		// should probably zero the real velocities for consistency
	}
	

	//atoms->IncrementCartesianPositions(velocities,0.5*dt);
	nIterations = 0;
	do
	{

		// calculate forces
		potential->CalculateForces(atoms, forces);
		CalculateGeneralizedForces(atoms, constraint, forces);
		
		// increment velocities
		IncrementGeneralizedVelocitiesFromForces(atoms);

		// if doing minimization, do quick min stuff here

		if(minimize)
		{
			for(paramIdx=0;paramIdx<nDOFs;++paramIdx)
			{
				if(generalizedForces[paramIdx]*generalizedVelocities[paramIdx]<0)
					generalizedVelocities[paramIdx] = 0.;
			}
		}
		
		
		// increment the parameters according to their velocities
		for(paramIdx=0;paramIdx<nDOFs;++paramIdx)
			generalizedCoordinates[paramIdx] += dt*generalizedVelocities[paramIdx];
		
		constraint->SetGeneralizedCoordinates(generalizedCoordinates);
		

		if(verbose && nIterations%STEPSBETWEENDISPLAY == 0)
		{
			cout << "Iteration: " << nIterations << endl;
			cout << "Convergence: " << generalizedForceSq << endl;
			atoms->Notify();
		}
		++nIterations;

	}
	while(!(minimize && generalizedForceSq < tolerance) && nIterations < maxSteps);
	// so if nIterations > maxSteps, stop anyway
	// otherwise stop if we're minimizing AND have converged

	delete forces;
}


void ConstraintVerletAtomsMover::CalculateGeneralizedForces(ListOfAtoms *atoms, Constraint *constraint, DMArray<double,DIMENSION> *forces)
{
    int paramIdx,paramIdx2,coord;

	generalizedForceSq = 0.;
    residualForceSq = 0.;
	
	double *forcesCurr,*posCurr;	
	double (*gradientWRTGeneralizedCoordinates)[DIMENSION];
	gradientWRTGeneralizedCoordinates = new double[nDOFs][DIMENSION];
	
	
	for(paramIdx=0;paramIdx<nDOFs;++paramIdx)
	{
		generalizedForces[paramIdx] = 0.;
		for(paramIdx2=0;paramIdx2<nDOFs;++paramIdx2)
			kineticMatrix[paramIdx2*(nDOFs) + paramIdx] = 0.;
	}
	
	DMArrayIterator<double,DIMENSION> forcesIter(forces),posIter(atoms->GetCartesianPositions()); 
	int atom;

	for(forcesIter.Init(),posIter.Init(),atom=0; !forcesIter.IsDone(); forcesIter.Next(),posIter.Next(),++atom)
	{ 
		forcesCurr = forcesIter.Current();
		posCurr = posIter.Current();
		
		// ask the constraint for the derivatives

		constraint->DerivativesWRTGeneralizedCoordinates(atom,gradientWRTGeneralizedCoordinates);
		
		// accumulate the forces on the parameters and the kinetic matrix
		for(coord=0;coord<DIMENSION;++coord)
		{
			for(paramIdx=0;paramIdx<nDOFs;++paramIdx)
			{		
				generalizedForces[paramIdx] += forcesCurr[coord] * gradientWRTGeneralizedCoordinates[paramIdx][coord];
				// Compute the kinetic matrix (relates parameter velocities to atomic velocities)
				for(paramIdx2=0;paramIdx2<nDOFs;++paramIdx2)
					kineticMatrix[paramIdx2*(nDOFs)+paramIdx] += gradientWRTGeneralizedCoordinates[paramIdx2][coord] *gradientWRTGeneralizedCoordinates[paramIdx][coord];
			} // end of paramIdx loop
			residualForceSq += forcesCurr[coord] * forcesCurr[coord];
		} // end of coord loop 
		
	} // end of loop over atoms
	
	
	
	for(paramIdx=0;paramIdx<nDOFs;++paramIdx)
		generalizedForceSq += generalizedForces[paramIdx] * generalizedForces[paramIdx];
	
	delete gradientWRTGeneralizedCoordinates;

}

	
	
void ConstraintVerletAtomsMover::IncrementGeneralizedVelocitiesFromForces(ListOfAtoms *atoms)
{

    int paramIdx,stringLength=1;
    char uplo='U';
    int matrixSize = nDOFs,nrhs = 1,info, lwork=nDOFs*10;
    double *work = new double[lwork];
	int *ipiv = new int[nDOFs];
	
	// solve the linear system to get velocity increments
 

	if(nDOFs == 0)
		return;
	
	
#ifdef _WIN32 
	DSYSV(&uplo,&stringLength,&matrixSize,&nrhs,kineticMatrix, &matrixSize,ipiv,generalizedForces, &matrixSize, work, &lwork,&info);
#else
	dsysv_(&uplo, &matrixSize, &nrhs,kineticMatrix, &matrixSize,ipiv,generalizedForces, &matrixSize, work, &lwork,&info);
#endif
	
	
	for(paramIdx=0;paramIdx<nDOFs;++paramIdx)
		generalizedVelocities[paramIdx] += dt*generalizedForces[paramIdx];
	// Here I replaced the integral of the force by multiplication by dt, which is put in the second step.

	delete work;
	delete ipiv;
}



////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////


void ConstraintVerletAtomsMover::Save(DMWriter *writer) const
{
	AtomsMover::Save(writer);	
}

void ConstraintVerletAtomsMover::Load(DMReader *reader)
{
	AtomsMover::Load(reader);
}
