// HooverAtomsMover.cpp: implementation of the HooverAtomsMover class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DigitalMaterial.h"
#include "HooverAtomsMover.h"
#include "ListOfAtoms.h"
#include "DMArray.h"
#include "SerialBuilder.h"
#include "DMLastHeader.h"

static SerialBuilder<HooverAtomsMover> *_Hoovercreator = 
    new SerialBuilder<HooverAtomsMover>("HooverAtomsMover");

const std::string HooverAtomsMover::className = std::string("HooverAtomsMover");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

HooverAtomsMover::HooverAtomsMover(Potential *potential, double temperature, double dt, double tau, int nSteps)
{
	this->potential = potential;
	this->temperature = temperature;
	this->dt = dt;
	this->tau = tau;
	this->nSteps = nSteps;
	this->zeta = 0.0;

	nAtoms=0;
	forces = NULL;
	
}


void HooverAtomsMover::Move(ListOfAtoms *atoms)	// Verlet
{
	if (nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber(),atoms->GetCartesianVelocities());

	double KE_Target = 0.5 * DIMENSION * nAtoms * temperature; // 1/2 kT per translational degree of freedom
	atoms->IncrementCartesianPositions(atoms->GetCartesianVelocities(), 0.5*dt);

	for (int i=0; i<nSteps-1; i++)
	{
		potential->CalculateForces(atoms, forces);
		atoms->AdjustForces(forces);
		atoms->ScaleCartesianVelocities((1-0.5*zeta*dt)/(1+0.5*zeta*dt));
		atoms->IncrementCartesianMomenta(forces,dt/(1+0.5*zeta*dt));
		atoms->AdjustVelocities();		
		atoms->IncrementCartesianPositions(atoms->GetCartesianVelocities(),dt);
		double KE = atoms->GetKineticEnergy();
		if (KE_Target != 0)	zeta += ((KE - KE_Target)/KE_Target) * dt / (tau * tau);
	}
	potential->CalculateForces(atoms, forces);
	atoms->AdjustForces(forces);
	atoms->IncrementCartesianMomenta(forces,dt);
	double KE = atoms->GetKineticEnergy();	
	if (KE_Target != 0)	zeta += ((KE - KE_Target)/KE_Target) * dt / (tau * tau);
	atoms->IncrementCartesianPositions(atoms->GetCartesianVelocities(), 0.5*dt);
}

void HooverAtomsMover::Allocate(int nAtoms,const DMArray<double, DIMENSION> *array)
{
	this->nAtoms = nAtoms;
	if(forces) delete forces;
	// ajd27 shapeChange
	forces = new DMArray<double, DIMENSION>(array);
}


////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////


void HooverAtomsMover::Save(DMWriter *writer) const
{
	AtomsMover::Save(writer);	

	writer->PutDouble(dt, "dt");
	writer->PutInt(nSteps, "nSteps");

	writer->PutDouble(zeta, "zeta");	
	writer->PutDouble(tau, "tau");	
	writer->PutDouble(temperature, "temperature,");

}


void HooverAtomsMover::Load(DMReader *reader)
{
	AtomsMover::Load(reader);

	SetDt(reader->GetDouble("dt"));
	SetNSteps(reader->GetInt("nSteps"));
	
	SetZeta(reader->GetDouble("zeta"));
	SetTau(reader->GetDouble("tau"));
	SetTemperature(reader->GetDouble("temperature"));
}
