// LangevinAtomsMover.cpp: implementation of the LangevinAtomsMover class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "blas.h"

#include "DigitalMaterial.h"
#include "LangevinAtomsMover.h"
#include "ListOfAtoms.h"
#include "SerialBuilder.h"
#include "DMLastHeader.h"

static SerialBuilder<LangevinAtomsMover> *_Langevincreator = 
    new SerialBuilder<LangevinAtomsMover>("LangevinAtomsMover");


const std::string LangevinAtomsMover::className = std::string("LangevinAtomsMover");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

LangevinAtomsMover::LangevinAtomsMover(Potential *potential,double dt,int n,double friction,double temperature):VerletAtomsMover(potential,dt,n)
{  
  this->friction = friction;
  this->temperature = temperature;
  CalculateParameters();
  randomPosInc = NULL;
  randomVelInc = NULL;
}

LangevinAtomsMover::~LangevinAtomsMover()
{
	if(randomPosInc) delete randomPosInc;

	if(randomVelInc) delete randomVelInc;
}


void LangevinAtomsMover::Move(ListOfAtoms *atoms)
{
	
	const DMArray<double, DIMENSION> *velocities = atoms->GetCartesianVelocities();
	const DMArray<double, DIMENSION> *positions  = atoms->GetCartesianPositions();
	if (!(forces && (forces->GetShape() == positions->GetShape())))
		Allocate(positions);
	
	
	// Calculate forces
	potential->CalculateForces(atoms,forces);
	atoms->AdjustForces(forces);
	
	for(int s=0;s<nSteps;s++)
    {
			// for second and subsequent times through the loop, am relying on forces as calculated for previous time step
			
			GenerateRandomNumbers();
		atoms->AdjustForceIncrements(randomPosInc); // because these are like forces, except we can't add the external force to these!
		atoms->AdjustForceIncrements(randomVelInc);
		
		atoms->IncrementCartesianPositions(velocities,c1*dt);
		atoms->IncrementCartesianPositions(forces,c2*dt*dt);
		
		atoms->IncrementCartesianPositions(randomPosInc, stdDevPos);
		atoms->ScaleCartesianVelocities(c0);
		atoms->IncrementCartesianVelocities(forces,c1_m_c2*dt);
		atoms->AdjustVelocities(); 
		
		// Calculate forces
		
		potential->CalculateForces(atoms,forces);
		atoms->AdjustForces(forces);
		
		atoms->IncrementCartesianVelocities(forces,c2*dt);
		atoms->IncrementCartesianVelocities(randomPosInc,stdDevVel*pvCorrelation);
		atoms->IncrementCartesianVelocities(randomVelInc,stdDevVel*corrConstant);
		atoms->AdjustVelocities(); 
    }
	
	

}


void LangevinAtomsMover::GenerateRandomNumbers()
{
	DMArrayIterator<double, DIMENSION> randPosIncIter(randomPosInc), randVelIncIter(randomVelInc);
	double *currentPos, *currentVel;
	for(randPosIncIter.Init(), randVelIncIter.Init(); !randPosIncIter.IsDone();
		randPosIncIter.Next(), randVelIncIter.Next())
	{
		currentPos = randPosIncIter.Current();
		currentVel = randVelIncIter.Current();
		for(int coord=0;coord<DIMENSION;++coord)
		{
			double r1 = 1.0 - rand.uniform();
// to avoid r1 being zero (don't want to take the log of zero!)
			double r2 = 2*PI*rand.uniform();
			double t = sqrt(-2.0 * log(r1));
			currentPos[coord] = t*cos(r2);
			currentVel[coord] = t*sin(r2);
		}
	}
}


////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////


void LangevinAtomsMover::Save(DMWriter *writer) const
{
	VerletAtomsMover::Save(writer);	

	writer->PutDouble(friction, "friction");
	writer->PutDouble(temperature, "temperature");
	writer->Put(&rand, "rand");
}

void LangevinAtomsMover::Load(DMReader *reader)
{
	VerletAtomsMover::Load(reader);

	SetFriction(reader->GetDouble("friction"));
	SetTemperature(reader->GetDouble("temperature"));
	reader->Fill(&rand, "rand");
}
