// LangevinAtomsMover.h: interface for the LangevinAtomsMover class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef LANGEVINATOMSMOVER_H
#define LANGEVINATOMSMOVER_H

#include <math.h>

#include "DigitalMaterial.h"

#include "VerletAtomsMover.h"
#include "Rand.h"
#include "DMArray.h"


//! This class implements an algorithm for Langevin dynamics at a given temperature and friction parameter

/*!
  \ingroup moleculardynamics
 
  The algorithm implemented here can be used to carry out molecular dynamics
  at constant temperature. It solves the Langevin equation for atoms in a  
  viscous medium parameterised by a friction parameter. The actual algorithm
  is a velocity-Verlet-type algorithm taken from from Ch. 9 of Allen and 
  Tildesley ("Computer Simulation of Liquids", Oxford 1986).

  Note that it uses IncrementCartesianVelocities rather than
  IncrementCartesianMomenta. This is because it takes more than this to be
  able to handle subbranches having different masses. One of the increment
  positions calls would have to have a mass factor, and also the random
  position and velocity increments depend on mass. To deal with several masses
  there is another class, CompositeLangevinAtomsMover
  

 */



class LangevinAtomsMover : public VerletAtomsMover  
{
public:
	LangevinAtomsMover(Potential *potential = 0, double dt = 1., int n = 1, double friction = 0, double temperature = 0.0);
	virtual ~LangevinAtomsMover();
	void Move(ListOfAtoms *atoms);
	void SetDt(double dt) {this->dt = dt; CalculateParameters();}
	void SetFriction(double friction) {this->friction = friction; CalculateParameters();}
	double GetFriction() const {return friction;}
	void SetTemperature(double temp) {this->temperature = temp; CalculateParameters();}
	double GetTemperature() const {return temperature;}
	void SetSeed(int seed) {rand.SetSeed(seed);}


	const std::string &GetType() const
    {
        return className;
    }
#ifndef SWIG
    void Save(DMWriter *writer) const;
	void Load(DMReader *reader);
   

protected:
	double friction;
	double c0;
	double c1;
	double c2;
	double c1_m_c2;
	double temperature;
	double stdDevPos;
	double stdDevVel;
	double pvCorrelation;
	double corrConstant;
	Rand rand;
	DMArray<double, DIMENSION> *randomPosInc;
	DMArray<double, DIMENSION> *randomVelInc;
	void CalculateParameters() // Note particle mass will have to be included here sometime.
	  {

		double fricDt = friction * dt;
	    c0 = exp(-fricDt);
	    if(friction == 0)
	      {
			c1 = 1;
			c2 = 0.5;

			stdDevPos = 0.0;

			stdDevVel = 0.0;
	      }
	    else
	      {
			c1 = (1-c0)/(fricDt);
			c2 = (1-c1)/(fricDt);

			stdDevPos = dt*sqrt(temperature *

				((2.0 - (3 - 4*exp(-fricDt) + exp(-2*fricDt))/fricDt)/fricDt));

			stdDevVel = sqrt(temperature * (1 - exp(-2*fricDt)));
	      }
	    c1_m_c2 = c1 - c2;
	    pvCorrelation = sqrt(3.0)/2. * (1. - fricDt / 8.0);
	    corrConstant  = sqrt((1.0 - pvCorrelation) * (1.0 + pvCorrelation));

	  }

	void Allocate(const DMArray<double, DIMENSION> *array)
	{
		VerletAtomsMover::Allocate(array);
		if(randomPosInc)
			delete randomPosInc;
		// ajd27 shapeChange
		randomPosInc = new DMArray<double, DIMENSION>(array);
		
		if(randomVelInc)
			delete randomVelInc;
		// ajd27 shapeChange
		randomVelInc = new DMArray<double, DIMENSION>(array);
		
	}
	void GenerateRandomNumbers();
	static const std::string className;
#endif // SWIG
};

#endif // !defined(LANGEVINATOMSMOVER_H)




