// MinimizerAtomsMover.cpp: implementation of the MinimizerAtomsMover class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DigitalMaterial.h"
#include "MinimizerAtomsMover.h"
#include "BoundaryConditions.h"
#include "NeighborLocator.h"
#include "ExternalWork.h"
#include "SerialBuilder.h"
#include "DMLastHeader.h"


static SerialBuilder<MinimizerAtomsMover> *_Minimizercreator = 
    new SerialBuilder<MinimizerAtomsMover>("MinimizerAtomsMover");

const std::string MinimizerAtomsMover::className = std::string("MinimizerAtomsMover");

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

MinimizerAtomsMover::MinimizerAtomsMover(Minimizer *minimizer, Potential *potential, ExternalWork *work)
{
  this->potential = potential;
  double tolerance = 1.0e-8;
  this->minimizer = minimizer;
  this->work = work;
  this->verbose = false;
}

MinimizerAtomsMover::~MinimizerAtomsMover()
{

}

void MinimizerAtomsMover::Move(ListOfAtoms *atoms)
{
	int nAtoms = atoms->GetNumber();
    int maxIterations = nAtoms*DIMENSION*2;
    
    //minimizer->SetMaxIterations(maxIterations);
    

    double *positions = new double[nAtoms*DIMENSION];
    atomsMinimizable = new AtomsMinimizable(atoms, potential, work, verbose);

    
    for(int i = 0;i< nAtoms;i++)
		for(int coord = 0;coord<DIMENSION;coord++)	
			positions[DIMENSION*i + coord] = atoms->GetCartesianPosition(i,coord);
    minimizer->Minimize(positions,atomsMinimizable);

/*
    try
    {
		minimizer->Minimize(positions,atomsMinimizable);
    }
    catch (ConjugateGradientMinimizer::Error error)
    {
		std::cerr << "Error in ConjugateGradient: " << error.what << std::endl;
		throw;
    }
    
    if(minimizer->GetNIterations() == minimizer->GetMaxIterations())
	std::cerr <<
	    "Reached maximum number of iterations in conjugate gradient" << std::endl;
  

*/
  
    delete atomsMinimizable;
    delete [] positions;
}


// The following functions "mediate" between the CG minimizer function and the
// actual forces and energy functions in the atoms class.


double *MinimizerAtomsMover::AtomsMinimizable::Gradient(double *positions)
{
    int index = 0,coord;
    DMArrayIterator<double, DIMENSION> posIter(atoms->GetCartesianPositions());
    for( posIter.Init(); !posIter.IsDone(); posIter.Next() )
		for( coord = 0; coord < DIMENSION; ++coord)	
		{
			posIter.Current()[coord] = positions[index++];
		}	   
    
    atoms->AdjustPositions();
    bc->EnforceBoundaryConditions(const_cast< DMArray<double,DIMENSION>* > (atoms->GetCartesianPositions()) );
    neighborLocator->UpdatePosition();
    
    potential->CalculateForces(atoms,forces);
    atoms->AdjustForces(forces);

	if(verbose)
	{
		double nrm = forces->Nrm2();
		cout << "MinimizerAtomsMover: Norm of forces = " << nrm << endl;
	}

    index = 0;
    DMArrayIterator<double, DIMENSION> forcesIter(forces);
    for(forcesIter.Init();!forcesIter.IsDone();forcesIter.Next())
		for(coord = 0;coord<DIMENSION;coord++)	
		{
			gradients[index++] = -forcesIter.Current()[coord];
		}
    return gradients;
}


double MinimizerAtomsMover::AtomsMinimizable::ObjectiveFunction(double *positions)
{	
    int index=0,coord;
    DMArrayIterator<double, DIMENSION> posIter(atoms->GetCartesianPositions());
    for(posIter.Init();!posIter.IsDone();posIter.Next())
		for(coord = 0;coord<DIMENSION;coord++)	
		{
			posIter.Current()[coord] = positions[index++];
		}	   

	
    atoms->AdjustPositions();


    bc->EnforceBoundaryConditions(const_cast< DMArray<double,DIMENSION>* > (atoms->GetCartesianPositions()));
    neighborLocator->UpdatePosition();
    
	double value = potential->CalculateEnergy(atoms) - (work ? work->CalculateWork(atoms):0);

	if(verbose)
	{
		cout << value << endl; 
		atoms->Notify();
	}

    return value;
}


////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////


void MinimizerAtomsMover::Save(DMWriter *writer) const
{
	AtomsMover::Save(writer);	
}

void MinimizerAtomsMover::Load(DMReader *reader)
{
	AtomsMover::Load(reader);
}
