// PredictorCorrectorAtomsMover.h: interface for the PredictorCorrectorAtomsMover class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef PREDICTORCORRECTORATOMSMOVER_H
#define PREDICTORCORRECTORATOMSMOVER_H

#include "DigitalMaterial.h"
#include "VerletAtomsMover.h"
#include "Potential.h"

#define N_VALUES 4 // for 4-value predictor-corrector (could also be 5)

//! PredictorCorrectorAtoms mover integrates the equations of motion using the predictor-corrector algorithm

/** \ingroup moleculardynamics
 *
 * For second order ODEs where rhs may or may not have first derivatives (for
 * 4-value) or for secind order ODEs where rhs must not have first derivatives
 * (for 5- or 6-value). Note that it is not hard to change the coefficients to
 * allow the 5-value and 5-value ones to work with first derivatives on the
 * RHS. See Allen and Tildesley Appendix E ("Computer Simulation of Liquids",
 *  Oxford 1986).
 */


class PredictorCorrectorAtomsMover : public VerletAtomsMover  
{
public:
	PredictorCorrectorAtomsMover(Potential *potential = 0, double dt = 0.01, int nSteps = 1);
	virtual ~PredictorCorrectorAtomsMover();
	void Move(ListOfAtoms *atoms);

	void CalculateCoefficients();

#ifndef SWIG
	void Save(DMWriter *writer) const;
	void Load(DMReader *reader);
	const std::string &GetType() const
		{
			return className;
		}

protected:
	void Allocate(const DMArray<double, DIMENSION> *array);
	double s[N_VALUES];
	double coeff[N_VALUES];
	DMArray<double, DIMENSION> *secondDerivs;
	DMArray<double, DIMENSION> *thirdDerivs;
	DMArray<double, DIMENSION> *fourthDerivs;
	DMArray<double, DIMENSION> *fifthDerivs;
	DMArray<double, DIMENSION> *corrections;

	static const std::string className;
#endif // SWIG
};

#endif // !defined(PREDICTORCORRECTORATOMSMOVER_H)
