// QuickMinAtomsMover.cpp: implementation of the QuickMinAtomsMover class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"
#include <iostream>

#include "DigitalMaterial.h"

#include "QuickMinAtomsMover.h"
#include "ListOfAtoms.h"
#include "DMArray.h"
#include <iomanip>
#include "SerialBuilder.h"
#include "Constraint.h"

#include "DMLastHeader.h"

const int QuickMinAtomsMover::DISPLAYPRECISION = 12;

using namespace std;


static SerialBuilder<QuickMinAtomsMover> *_QMcreator = 
    new SerialBuilder<QuickMinAtomsMover>("QuickMinAtomsMover");

const string QuickMinAtomsMover::className = string("QuickMinAtomsMover");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////



QuickMinAtomsMover::QuickMinAtomsMover(Potential *potential,double dt,double tolerance)
{
	maxIterations = 1000;
	stepsBetweenUpdate = 100;
	instabilityThreshold = 1000.;
	this->potential = potential;
	this->dt = dt;
	this->tolerance = tolerance;
	forces = 0;
	adjustedVelocities = 0;
	nAtoms = -1; // need to call Allocate even when nAtoms is 0, so don't start with it being 0
	nIterations = 0;
	totalForceSq = 0.;
	verbose = false;
	outputStream = &std::cout;
}

QuickMinAtomsMover::~QuickMinAtomsMover()
{
	if (forces) delete forces;
	if(adjustedVelocities) delete adjustedVelocities;
}

void QuickMinAtomsMover::Allocate(int nAtoms, const DMArray<double, DIMENSION> *array)
{
	this->nAtoms = nAtoms;
	if(forces) delete forces;
	if(adjustedVelocities) delete adjustedVelocities;
	// Allocate by the array rather than the shape so that siblings
	// are handled properly and atoms are migrated on all arrays.
	forces = new DMArray<double, DIMENSION> (array);
	adjustedVelocities = new DMArray<double, DIMENSION> (array);
}


void QuickMinAtomsMover::Move(ListOfAtoms *atoms)
{
	int coord;
	double tFSFirst = 0.0,tFSq_M=1e99;
	nIterations=0;
	const DMArray<double, DIMENSION> *velocities = atoms->GetCartesianVelocities();
	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber(), atoms->GetCartesianPositions());

	int oldPrecision = outputStream->precision();
	outputStream->precision(DISPLAYPRECISION);


	double initDt = dt;
	int initSBU = stepsBetweenUpdate;	


	// Initial update of positions by dt/2 (in the main loop we combine the two half step increments)
	atoms->IncrementCartesianPositions(velocities, 0.5 * dt);

	potential->CalculateForces(atoms,forces);
	atoms->AdjustForces(forces);
	double forceNrm2 = forces->Nrm2();
	totalForceSq = forceNrm2 * forceNrm2;
	tFSFirst = totalForceSq;

	while(totalForceSq > tolerance && nIterations < maxIterations)
	{
 
		// Update velocities	
		atoms->IncrementCartesianVelocities(forces,dt);

		// Adjust velocities: Remove components that are not in the direction of the forces. The new velocities will then be in adjustedVelocities
		AdjustVelocities(forces, velocities);
		atoms->SetCartesianVelocities(adjustedVelocities);

		// now apply adjustments to the velocities from constraints
		atoms->AdjustVelocities();


		// Update positions by dt
		atoms->IncrementCartesianPositions(velocities,dt);

        // Get the global total of the force array. (necessary for parallel simulations)
        //const double tmp = totalForceSq;
		//dmProtocol->AllReduce(&tmp, &totalForceSq, 1, DMProtocol::SUM);
		//Nrm2 function now sums for all processors

		if(verbose && nIterations%stepsBetweenUpdate == 0)
		{
			(*outputStream) << "QM:Iteration: " << nIterations << endl;
			(*outputStream) << "QM:Convergence: " << totalForceSq << " rms " << sqrt(totalForceSq/atoms->GetNumber()) << endl;
			(*outputStream) << "QM:Potential Energy: " << potential->CalculateEnergy(atoms) << endl;
			(*outputStream) << "Number Of Atoms: " << atoms->GetNumber() << " Processor: " << dmProtocol->GetProcessorNumber() << endl;
			atoms->Notify();
		}

		// some crude adaptive time-stepping here
		if(totalForceSq > tFSFirst * 100 && nIterations > 0)
		{
			dt = initDt / 10;
			stepsBetweenUpdate = 10;
		}
		if(dt < initDt && totalForceSq < tFSFirst)
		{
			dt = initDt;
			stepsBetweenUpdate = initSBU;
		}

		if(totalForceSq > tFSFirst*instabilityThreshold	&& nIterations >0) // check nIterations > 0 because there's no previous value for nIterations == 0
		{
		  double growth = totalForceSq/tFSFirst;
		  cerr << "Unstable! Growth Factor " << growth << endl;
		  cerr << "Iteration: " << nIterations << " TFS: " << totalForceSq << endl;
		  // should allow access to positions by making a copy of them and 
		  // putting it in the UnstableError.
		  throw UnstableError(growth,nIterations);
		}   

		potential->CalculateForces(atoms,forces);
		atoms->AdjustForces(forces);
		forceNrm2 = forces->Nrm2();
		totalForceSq = forceNrm2 * forceNrm2;		
				
		++nIterations;
	   
	}

	// Set all velocities to exactly zero at end of minimization. 
	adjustedVelocities->Zero();
	atoms->SetCartesianVelocities(adjustedVelocities);
	outputStream->precision(oldPrecision);
	stepsBetweenUpdate = initSBU;
	dt = initDt;
}


void QuickMinAtomsMover::AdjustVelocities(const DMArray<double, DIMENSION> *forces, const DMArray<double, DIMENSION> *velocities)
{
	int coord;
	double vDotF, fSq;
	double *veloCurr, *forcesCurr, *tVeloCurr;
	// This works atom by atom.
	adjustedVelocities->Copy(velocities);

	DMArrayIterator<double, DIMENSION> forcesIter(forces), veloIter(velocities), tVeloIter(adjustedVelocities);
	for(forcesIter.Init(), veloIter.Init(), tVeloIter.Init(); !forcesIter.IsDone(); 
		forcesIter.Next(), veloIter.Next(), tVeloIter.Next())
	{
		vDotF = 0;
		fSq = 0;
		veloCurr = veloIter.Current();
		tVeloCurr = tVeloIter.Current();
		forcesCurr = forcesIter.Current();
		for (coord=0; coord<DIMENSION; coord++)
		{
 			vDotF += veloCurr[coord]* forcesCurr[coord];
// 			fSq += forcesCurr[coord] * forcesCurr[coord];
 		}
 		if(vDotF < 0)
 		{
			for(coord=0;coord<DIMENSION;coord++) tVeloCurr[coord] = 0.; 
 		}
	}
	
}


////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////

void QuickMinAtomsMover::CheckQMParameter(DMReader *reader, double parameter, const std::string parameterName)
{
	double tmp;
	tmp = reader->GetDouble(parameterName);
	ASSERT(tmp == parameter);
	if (tmp != parameter)
	{
		string errorString("Value of ");
		errorString += parameterName;
		errorString += SprintString(" has changed from %f to %f", tmp, parameter);
		throw DMException(this->GetType(),"Load->CheckQMParameter", errorString);
	}
}


void QuickMinAtomsMover::Save(DMWriter *writer) const
{
	AtomsMover::Save(writer);	

	writer->PutDouble(instabilityThreshold, "instabilityThreshold");
	writer->PutInt(stepsBetweenUpdate, "stepsBetweenUpdate");
	writer->PutInt(DISPLAYPRECISION, "DISPLAYPRECISION");

	writer->PutDouble(dt, "dt");
	writer->PutDouble(tolerance, "tolerance");
	writer->PutInt(maxIterations, "maxIterations");
	writer->PutDouble(totalForceSq, "totalForceSq");
	writer->PutInt(nIterations, "nIterations");
	writer->PutBool(verbose, "verbose");
}

void QuickMinAtomsMover::Load(DMReader *reader)
{
	AtomsMover::Load(reader);

	CheckQMParameter(reader, instabilityThreshold, "instabilityThreshold");
	CheckQMParameter(reader, stepsBetweenUpdate, "stepsBetweenUpdate");
	CheckQMParameter(reader, DISPLAYPRECISION, "DISPLAYPRECISION");

	SetDt(reader->GetDouble("dt"));
	SetTolerance(reader->GetDouble("tolerance"));
	SetMaxIterations(reader->GetInt("maxIterations"));
	totalForceSq = reader->GetDouble("totalForceSq");
	nIterations = reader->GetInt("nIterations");
	verbose = reader->GetBool("verbose");

}
