// VerletAtomsMover.cpp: implementation of the VerletAtomsMover class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DigitalMaterial.h"

#include "VerletAtomsMover.h"
#include "ListOfAtoms.h"
#include "DMArray.h"
#include "SerialBuilder.h"


#include "DMLastHeader.h"

static SerialBuilder<VerletAtomsMover> *_Verletcreator = 
    new SerialBuilder<VerletAtomsMover>("VerletAtomsMover");

const std::string VerletAtomsMover::className = std::string("VerletAtomsMover");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

VerletAtomsMover::VerletAtomsMover(Potential *potential, double dt, int nSteps)
{
	this->potential = potential;
	this->dt = dt;
	this->nSteps = nSteps;

	forces = 0;
	SetTime();
}

VerletAtomsMover::~VerletAtomsMover()
{
	if (forces) delete forces;
}

void VerletAtomsMover::Move(ListOfAtoms *atoms)	// Verlet
{
	double halfDt = 0.5*dt;
	if (!(forces && (forces->GetShape() == atoms->GetShape())))
		Allocate(atoms->GetCartesianPositions());
	atoms->IncrementCartesianPositions(atoms->GetCartesianVelocities(), halfDt);
	
	time += halfDt;

	for (int i=0; i<nSteps-1; i++)
	{
		potential->CalculateForces(atoms, forces);
		atoms->AdjustForces(forces);
		atoms->IncrementCartesianMomenta(forces,dt);
		atoms->AdjustVelocities();
		time += halfDt;
		atoms->IncrementCartesianPositions(atoms->GetCartesianVelocities(),dt);
		time += halfDt;
	}
	potential->CalculateForces(atoms, forces);
	atoms->AdjustForces(forces);
	atoms->IncrementCartesianMomenta(forces,dt);
	atoms->AdjustVelocities();

	time += halfDt;
	atoms->IncrementCartesianPositions(atoms->GetCartesianVelocities(), halfDt);
}

void VerletAtomsMover::Allocate(const DMArray<double, DIMENSION> *array)
{
	if (forces) delete forces;
	forces = new DMArray<double, DIMENSION>(array);
}



////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////


void VerletAtomsMover::Save(DMWriter *writer) const
{
	AtomsMover::Save(writer);	

	writer->PutDouble(time, "time");
	writer->PutDouble(dt, "dt");
	writer->PutInt(nSteps, "nSteps");
}

void VerletAtomsMover::Load(DMReader *reader)
{
	AtomsMover::Load(reader);

	SetTime(reader->GetDouble("time"));
	SetDt(reader->GetDouble("dt"));
	SetNSteps(reader->GetInt("nSteps"));
}
