// NudgedElasticBandBarrierFinder.h: interface for the NudgedElasticBandBarrierFinder class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef NUDGEDELASTICBANDBARRIERFINDER_H__
#define NUDGEDELASTICBANDBARRIERFINDER_H__

#include "DMFirstHeader.h"
#include <math.h>
#include <assert.h>
#include "DigitalMaterial.h"

#include "DMArray.h"
class ListOfAtoms;
class Potential;
class BoundaryConditions;
class NeighborLocator;
class ExternalWork;


//! The Find() function of this class takes two ListOfAtomses (and an optional ExternalWork object) corresponding to two neighboring minima in the potential energy landscape and uses the Nudged Elastic Band method to find the minimum energy path between them and also calculates the energy barrier.

/**
 * After completing the Find() function, various "return-value variables" are
 * set, which can be accessed using get functions. There are no Set functions
 * for these! It doesn't make sense to access these variables without having 
 * called Find() first (you'll get zeros, or worse, the results of a previous
 * calculation). These variables include: 
 *
 * saddleCoord: the distance along the chain corresponding to the saddle point.
 * barrierHeight: the energy at this point, relative to the energy of the
 *                first minimum, ie the barrier height.
 *  barrierError: an estimate of the error in the barrier height. Comes from
 *                the difference between 4 and 5 point interpolation schemes.
 * chainLength: the length of the chain.
 * finalTFS: the total force squared at the end of the chain relaxation. 
 *           Indicate how well converged things were.
 * nIterations: the number of iterations taken to converge, if convegence was
 *         reached. Otherwise the maximum allowed number of iterations.
 *
 * Other query-type functions are:
 * GetReplicas(int a): returns a pointer to the a-th replica.
 * GetBarrierPoints(): returns an array (double *[2]) which contains the pairs
 * {position along chain of replica, energy of replica} which is used to 
 * calculate the barrier height.
 * GetSaddlePointConfiguration(): interpolates all atomic positions to their 
 * value at saddleCoord along the chain and returns a new ListOfAtoms in this
 * configuration. This is the saddle point configuration.
 *
 * 
 * Replicas are created and destroyed each time Find() is called. Arrays are 
 * re-allocated only when nReplicas or NEB_ArrayLength changes. A pointer to a 
 * Potential, a NeighborLocator and a BoundaryConditions are passed in to the 
 * constructor. It is important to create the NEB object after the
 * BoundaryConditions object has been fully initialized (lengths set to their
 * correct values etc) otherwise the copies will be wrong.
 *
 *
 * \author nbailey 
 * \date 5/22/2000
 * \ingroup moleculardynamics
 */

class NudgedElasticBandBarrierFinder  
{
 public:
	NudgedElasticBandBarrierFinder(int nReplicas,double dt,double *springConstant,BoundaryConditions *bc, Potential *potential,NeighborLocator *neighborLocator,double tol, ExternalWork *externalWork = 0, bool wantAbsoluteEnergies = false,bool usingClimbingImage = false);
	virtual ~NudgedElasticBandBarrierFinder();
	virtual void Find(ListOfAtoms *atoms1,ListOfAtoms *atoms2);
	
	// Functions to access the basic NEB parameters
	void SetNReplicas(int n)
		{
			if(n != nReplicas)
				Allocate(NEB_ArrayLength,nReplicas);	
			for( int a = 0; a < nReplicas+1; ++a )
				springConstant[a] = 0.;
			// must now call SetSpringConstant!
		}
	void SetSpringConstant(double *k)
		{
			for( int a = 0; a < nReplicas+1; ++a )
				springConstant[a] = k[a];
		}
	void SetTangentMethod(std::string tangMethStr)
		{
			if(tangMethStr == "original")
				this->TANGENT_METH = ORIG_T;
			else if(tangMethStr == "bisect")
				this->TANGENT_METH = BISECT_T;
			else if(tangMethStr == "energy")
				this->TANGENT_METH = ENERGY_T;
			else throw DMException("NudgedElasticBandBarrierFinder","SetTangentMethod",SprintString("Unknown tangent method \"%s\"",tangMethStr.c_str()));
		}
	

	void SetDt(double t) {dt = t;}
	double Dt() const {return dt;}
	int GetMaxIterations() const {return maxIterations;}
	void SetMaxIterations(int n) {maxIterations = n;}

	int GetItToSetClimbing () const { return itToSetClimbing; }
	//! What NEB iteration to set which image is going to be "climbing". If negative, will have a climbing image from the beginning, and the image will be given by the absolute value of itToSetClimbing
	void SetItToSetClimbing ( int itToSetClimbing) { this->itToSetClimbing = itToSetClimbing ;} 
	
	int GetIncreasedSpringForceRange () const { return increasedSpringForceRange; }
	void SetIncreasedSpringForceRange (int increasedSpringForceRange) { this->increasedSpringForceRange = increasedSpringForceRange ;} 

	double GetIncSpringFactor() const { return incSpringFactor; }
	void SetIncSpringFactor(double incSpringFactor) {this->incSpringFactor = incSpringFactor;}

	void SetUsingClimbingImage( bool useCI) { usingClimbingImage = useCI; }
	int GetClimbingReplica() const { return climbingReplica; }
	
	//! choose which method of computing spring forces (0 for original method)
	void SetSpringMethod( int sm ) {springMethod = sm; }
	
	void SetInitializeFromSaved(std::string **repFileList) { initializeFromSaved = true; this->repFileList = repFileList;}


	// Accessing "return" information	
	//! Calculates distances and energies/forces (unadjusted) along the chain
	void CalculateReturnInfo();

	double GetSaddleCoord() const {return saddleCoord;}
	double GetBarrierHeight() const {return barrierHeight;}
	double GetBarrierError() const {return barrierError;}
	double GetChainLength() const {return chainLength;}
	double GetFinalTFS() const {return totalForceSq;}
	const double *GetReplicaForceSq() const { return replicaForceSq; }
	int GetNIterations() const {return nIterations;}
	double GetBackwardBarrier() const { return backwardBarrier; }
	double GetTotalEnergyDrop() const {return totalEnergyDrop;}
	const ListOfAtoms* GetReplicas(int a) const  {return replicas[a];}

	const double *GetFittingCoefficients() { return cof; }
	const double *GetProfileIntegrand() { return profileIntegrand; }
	
#ifndef SWIG
	typedef double (*DoublePair)[2];
	const DoublePair GetBarrierPoints() const {return barrierPoints;}
	virtual ListOfAtoms *GetSaddlePointConfiguration();
#endif //SWIG
	
	void SetVerbosity(int verbosity) {this->verbosity = verbosity;}
    int GetVerbosity() const {return verbosity;}


protected:
#ifndef SWIG
	// protected member functions
	//! creates the replicas by copying atoms1
	virtual void CreateReplicas();
	virtual void QuickMinStep(ListOfAtoms *repAtoms, DMArray<double, DIMENSION> *repForces, int repIdx);

	virtual double CalculateNudging();
	virtual void CalculateLinkVector( int repIdxSecond, double (*linkVector)[DIMENSION], double &linkLength);
	virtual double CalculateTangent(int repIdx, double (*linkVectorMinus)[DIMENSION], double (*linkVectorPlus)[DIMENSION], double linkMinusLength, double linkPlusLength, double *cosPhi);
	virtual double AddSpringForce(int repIdx, double (*linkVectorMinus)[DIMENSION], double (*linkVectorPlus)[DIMENSION], double linkMinusLength, double linkPlusLength, double f_UDotTangent, double cosPhi);



	//! implements the "cimbing image" algorithm on the chosen replica 
	virtual double FlipTangentialForce(int repIdx, double f_UDotTangent);
	virtual double ReplicaForceEnergy(int chainIdx);
	//! separate from CalculateTangent so can be overloaded in derived classes
	virtual void GetDotProducts(int repIdx, double (*linkVectorMinus)[DIMENSION], double (*linkVectorPlus)[DIMENSION], double *tangentLength, double *f_UDotTangent, double *linkPlusDotMinus);

	double CalculateBarrierHeight(int nPoints, double (*barrierPoints)[2]);
	double GetUnKinkFactor(double cosPhi, int repIdx) const;

	virtual void FreeMemoryNAtoms();
	virtual void FreeMemoryNReplicas();
	virtual void Allocate(int nAtoms,int nReplicas);
	virtual void Initialize();
	void StartClimbingImage(int climbingReplica);

	// basic parameters
	enum {ORIG_T,BISECT_T,ENERGY_T} TANGENT_METH;// determines which method of computing the tangent is used
	int nReplicas;
	int NEB_ArrayLength;               // number of atoms per replica
	double dt;                // time step for quick-min
	double tol;               // tolerance for convergence of quick-min
	int maxIterations;
	double cosPhiLarge;       // limits for determining kinkiness
	double cosPhiSmall;       // Large/Small refer to the angle not the cosine
	//! at which iteration the decision about which replica is to be the climbing one is made. Default is 10th.
    int itToSetClimbing;
	//! says how many images on either side of highest energy one to have higher  spring constant (-1 if none)
	int increasedSpringForceRange;
	//! Factor by which their spring constant is increased
	double incSpringFactor;
	//! minSprIdx/maxSprIdx define which spring forces are increased
	int minSprIdx;
	int maxSprIdx;
	//! for a given iteration, how many extreme kinks exist along the chain
	int numKinks;
	static const int STEPSBETWEENOUTPUT;
	//! the number to be added to the number of replicas, giving the total number of systems in the chain---we need to include the end-points of the chain, ie the two minima, for cases where the barrier height is very close to one of them, eg at critical shear
	static const int END_POINTS;

	double *springConstant;
	bool usingClimbingImage;
	int climbingReplica; // this replica feels the potential force reversed along the tangent line, such that it seeks the top of the barrier. -1 means no replica does this.
	bool wantAbsoluteEnergies; // whether to subtract total energy from profile
	//! controls which method of computing spring forces: 0 for original, 1 for modified
	int springMethod;
	bool initializeFromSaved;
	std::string **repFileList;
	int verbosity;

	// Internal parameters
	//! Energy of "start" end point
	double energy1;
	//! Maximum energy along chain
   	double maxEnergy;
	//! which replica, counting from end point, has the maximum energy
	int maxEnergyIdx;


	
	// NEB arrays: need to be allocated and freed carefully!
	ListOfAtoms **replicas;
	NeighborLocator **repNeighborLocators;
	BoundaryConditions **repBoundaryConditions;
	double (*barrierPoints)[2];
	DMArray<double,DIMENSION> **forces;
	double *cof; // coefficients for fitting to polynomial
	double *profileIntegrand;
	double *replicaForceSq;

	double (*springForce)[DIMENSION];
	double (*tangent)[DIMENSION];
	double (*linkVector1)[DIMENSION];
	double (*linkVector2)[DIMENSION];
	
	
	// pointers passed in from outside: not NEB's responsibility
	ListOfAtoms *atoms1,*atoms2;
	NeighborLocator *nl;
	BoundaryConditions *bc;
	Potential *potential;
	ExternalWork *externalWork;
	
	// Return information:
	double saddleCoord;
	double barrierHeight;
	double barrierError;
	double backwardBarrier;
	double chainLength;
	double totalForceSq;
	double totalEnergyDrop;
	int nIterations;

	
private:
	void ResetMaxEnergy() {maxEnergy = barrierPoints[0][1]; maxEnergyIdx = 0;}
	void CheckMaxEnergy(double energy, int chainIdx)
		{
			if(energy > maxEnergy)
			{ maxEnergy = energy; maxEnergyIdx = chainIdx;}
		}
		

#endif // SWIG
};



#endif // !defined(NUDGEDELASTICBANDBARRIERFINDER_H__)

