// SkewPeriodicBoundaryConditions.cpp: implementation of the SkewPeriodicBoundaryConditions class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
#include <math.h>

#include "DigitalMaterial.h"
#include <vector>
#include <algorithm>
#include "TinyVector.h"

#include "SkewPeriodicBoundaryConditions.h"
#include "DMArray.h"
#include "SerialBuilder.h"
#include "lapack.h"

#include "DMLastHeader.h"

static SerialBuilder<SkewPeriodicBoundaryConditions> *_SkewPBCcreator = 
new SerialBuilder<SkewPeriodicBoundaryConditions>("SkewPeriodicBoundaryConditions");

using namespace std;

const double SkewPeriodicBoundaryConditions::TINY = 1.0e-13;

const string SkewPeriodicBoundaryConditions::className = string("SkewPeriodicBoundaryConditions");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

SkewPeriodicBoundaryConditions::SkewPeriodicBoundaryConditions(const bool wrap[DIMENSION])
{
	SetWrap(wrap);
    double cellVectors[DIMENSION][DIMENSION];
    for(int coord=0;coord<DIMENSION;++coord)
        for(int vec=0;vec<DIMENSION;++vec)
            cellVectors[coord][vec] = (coord == vec ? 1. : 0. ); 
    
    SetSuperCellVectors(cellVectors);
}

SkewPeriodicBoundaryConditions::SkewPeriodicBoundaryConditions(const double superCellVectors[DIMENSION][DIMENSION],const bool wrap[DIMENSION])
{
	SetWrap(wrap);
    SetSuperCellVectors(superCellVectors);
}

SkewPeriodicBoundaryConditions::~SkewPeriodicBoundaryConditions()
{
    
}

void SkewPeriodicBoundaryConditions::SetWrap(const bool wrap[DIMENSION])
{
	nWrapped = 0;
	for(int coord=0; coord<DIMENSION; ++coord)
	{
		this->wrap[coord] = (wrap ? wrap[coord]: true);
		if(this->wrap[coord])
			++nWrapped;
	}
}

void SkewPeriodicBoundaryConditions::SetSuperCellVectors(const double superCellVectors[DIMENSION][DIMENSION])
{
    // first index (coord) is whether x, y or z component of a given vector; 
    // second (vec) is which vector. This makes sense from the POV of the 
    // matrix interpretation.
    
    for(int coord=0;coord<DIMENSION;++coord)
        for(int vec=0;vec<DIMENSION;++vec)
            this->superCellVectors[coord][vec] = superCellVectors[coord][vec];
        
   CalculateReducedVectors();

    /* 
    calculate inverse super-cell vectors and the inverse reduced vectors. We 
    use the main superCellVectors for enforcing the boundary conditions, so the 
    space occupied by the atoms (eg what you see if you plot them) has that shape
    but we use the reduced vectors for calculating differences.
    */
    int vec;
    for(vec=0;vec<DIMENSION;++vec)
    {
        for(int coord=0;coord<DIMENSION;++coord)
            inverseSuperCellVectors[coord][vec] = superCellVectors[coord][vec];
    }
    InvertMatrix(inverseSuperCellVectors);

    for(vec=0;vec<DIMENSION;++vec)
    {
        for(int coord=0;coord<DIMENSION;++coord)
            inverseReducedSuperCellVectors[coord][vec] =
                                        reducedSuperCellVectors[coord][vec];
    }
    InvertMatrix(inverseReducedSuperCellVectors);

    BuildWignerSeitzCell();
}


void SkewPeriodicBoundaryConditions::CalculateReducedVectors()
{
    vector<TinyVector<double> > reducedVectors;
    for(int vec=0; vec<DIMENSION; ++vec)
    {
        TinyVector<double> tmp;
        for(int coord=0; coord<DIMENSION; ++coord) 
            tmp[coord] = superCellVectors[coord][vec];
        reducedVectors.push_back(tmp);
    }
    bool goOn;
    do
    {
        goOn = false;
        for(int vec = 0; vec<DIMENSION; ++vec)
        {
            for(int vec2 = 0; vec2<DIMENSION; ++vec2)
            {
                if(vec == vec2) continue;
                const double dot = reducedVectors[vec] * reducedVectors[vec2];
                const double cmp = reducedVectors[vec2].norm2();
                if(dot > 0.5 * cmp)
                {
                    reducedVectors[vec] -= reducedVectors[vec2];
                    goOn = true;
                }
                else if (dot <= -0.5 * cmp)
                {
                    reducedVectors[vec] += reducedVectors[vec2];
                    goOn = true;
                }
            }
        }
    } while(goOn);

    for(int coord=0; coord<DIMENSION; ++coord)
        for(int vec=0; vec<DIMENSION; ++vec) 
            reducedSuperCellVectors[coord][vec] = reducedVectors[vec][coord];
}


void SkewPeriodicBoundaryConditions::InvertMatrix(double mat[DIMENSION][DIMENSION])
{
	// This function has been copied into a general purpose function in the class 
	// SimpleMatrixInverter which has a template parameter N which here would be 
	// DIMENSION. The class is part of DMSystem. So sometime I will remove this
	// function and replace it with a call to SimpleMatrixInterter::Invert

    int ipiv[ DIMENSION ];
    int size = DIMENSION;
    int info;
    double matrix[ DIMENSION * DIMENSION ];
    int lwork = DIMENSION * DIMENSION;
    double *work = new double[lwork];
    int index = 0;
    for(int vec=0;vec<DIMENSION;++vec)
        for(int coord=0;coord<DIMENSION;++coord)
            matrix[index++] = mat[coord][vec];
        
    // LU factorize the matrix
	dgetrf_( &size, &size, matrix, &size,  ipiv,  &info );
        
    ASSERT(info == 0);
    if(info < 0) throw DMException("SkewPeriodicBoundaryConditions", 
        "InvertMatrix", SprintString("dgetrf: %d-th argument had an illegal value", -info));
    if(info > 0) throw DMException("SkewPeriodicBoundaryConditions", 
        "InvertMatrix", SprintString("dgetrf: U(%d,%d) is exactly zero", info,info));
    
    // compute the inverse of the matrix from the LU factorization
    dgetri_( &size, matrix, &size, ipiv, work, &lwork, &info );

    ASSERT(info == 0);
    if(info < 0) throw DMException("SkewPeriodicBoundaryConditions", 
        "InvertMatrix", SprintString("dgetri: %d-th argument had an illegal value", -info));
    if(info > 0) throw DMException("SkewPeriodicBoundaryConditions", 
        "InvertMatrix", SprintString("dgetri: U(%d,%d) is exactly zero", info,info));

    delete [] work;
    index = 0;
    for(int coord=0;coord<DIMENSION;++coord)
        for(int vec=0;vec<DIMENSION;++vec)
            mat[vec][coord] = matrix[index++];
}

/*! 
Building the Wigner-Seitz cell is one thing we use the reduced basis vectors
for. That way we only need to check indices between plus or minus 1 for
each vector.

*/

void SkewPeriodicBoundaryConditions::BuildWignerSeitzCell()
{
    wignerSeitzVectors.clear();
    wignerSeitzVectorsSq.clear();
    numberOfWSVectors = 0;

    const int cubeSize = 1; 
    class WSIndices
    {
    public:
        static bool Manhattan(const TinyVector<int> &a, const TinyVector<int> &b)
        {
            int tot = 0;
            for(int coord=0; coord<DIMENSION; ++coord)
                tot += abs(a[coord]) - abs(b[coord]);
            return tot < 0;
        };
        WSIndices(int maxi, bool wrap[DIMENSION])
        {
            int indices[DIMENSION];
            for(int coord=0; coord<DIMENSION; ++coord) indices[coord] = 0;
            indices[0] = 1;
            while(indices[DIMENSION-1] <= maxi)
            {
				// don't want to include indices which are non-zero for basis
				// vectors which are not wrapped
				int coord;
				bool indicesValid = true;
				for(coord=0;coord<DIMENSION;++coord)
					if(indices[coord] && !wrap[coord])
						indicesValid = false;
				if(indicesValid)
					allIndices.push_back(indices);
                ++indices[0];
                for(coord = 0; coord < DIMENSION-1; ++coord)
                    if(indices[coord] > maxi)
                    {
                        indices[coord] = -maxi;
                        ++indices[coord+1];
                    }
            }
            sort(allIndices.begin(), allIndices.end(), Manhattan);
        }
        int size() const {return allIndices.size();}
        vector<TinyVector<int> > allIndices;
    } wsIndices(cubeSize,wrap);

    for(int i=0; i<wsIndices.size(); ++i)
    {
        int coord;
        double candidate[DIMENSION], copy[DIMENSION];
        for(coord=0; coord<DIMENSION; ++coord)
        {
            candidate[coord] = 0;
            for(int vec=0; vec<DIMENSION; ++vec)
                candidate[coord] += 0.5 * wsIndices.allIndices[i][vec] * 
					reducedSuperCellVectors[coord][vec];
            copy[coord] = candidate[coord];
        }
        
        WignerSeitzDifferences(copy);
        bool same = true;
        // if you're inside the WS cell, WignerSeitzDifferences won't change you.
        for(coord=0; coord<DIMENSION; ++coord) same &= (copy[coord] == candidate[coord]);
        if(same) AddVectorToWSCell(copy);
    }
    //ASSERT(wignerSeitzVectors.size() >= DIMENSION);
    ASSERT(wignerSeitzVectors.size() >= nWrapped);
    
    sort(wignerSeitzVectors.begin(), wignerSeitzVectors.end(), lesser_norm2);
    sort(wignerSeitzVectorsSq.begin(), wignerSeitzVectorsSq.end());


    /* check which vectors can be removed. These are those for which v * v_j < v_j*v_j.
     In addition, some are redundant. Here is a concrete example: if the first two 
     super cell vectors are orthogonal, then their sum will end up in the list of 
     wigner seitz cell vectors. This is redundant. So a criterion to identify these
     redundant vectors is:
     in 2D: if the vector is on 2 other planes, it is redundant
     in 3D: if the vector is on 3 other planes, it is redundant
        if it is on 2 other planes, which are orthogonal to the vector, it is redundant
    */
    for(int vec = numberOfWSVectors-1; vec >= 0; --vec)
    {
        bool ok = true;
        vector<int> planeIndices; // the index of the planes the current vectors is on.
        for(unsigned int vec2 = 0; vec2 < numberOfWSVectors; ++vec2)
        {
            if(vec2 == vec) continue;
            const double fdot = fabs(wignerSeitzVectors[vec] * wignerSeitzVectors[vec2]); 
			// if(fdot > wignerSeitzVectorsSq[vec2]) ok = false;
            if(fabs( fdot - wignerSeitzVectorsSq[vec2]) < TINY)
                planeIndices.push_back(vec2);
        }
        if(DIMENSION == 3)
        {
            //if(planeIndices.size() == 3) ok = false; // the current vector is redundent
            if(planeIndices.size() >= 2)
            {
                const double *v0 = wignerSeitzVectors[vec];
                const double *v1 = wignerSeitzVectors[planeIndices[0]];
                const double *v2 = wignerSeitzVectors[planeIndices[1]];
                double det = 0.;
                for(int i=0; i<DIMENSION; ++i)
                    det = v0[i] * (v1[(i+1)%3]*v2[(i+2)%3] - v1[(i+2)%3]*v2[(i+1)%3]);
                if(fabs(det) < TINY) ok = false;
            }
        }
        else if(planeIndices.size() == 2) ok = false;
        if(!ok)
            RemoveVectorFromWSCell(vec);        
    }

    //ASSERT(numberOfWSVectors >= DIMENSION);	
    ASSERT(numberOfWSVectors >= nWrapped);
  
    TRACE1("%i vectors in the WS cell\n" ,  numberOfWSVectors);

}

void SkewPeriodicBoundaryConditions::AddVectorToWSCell(const TinyVector<double> &toAdd)
{
    this->wignerSeitzVectors.push_back(toAdd);
    this->wignerSeitzVectorsSq.push_back(toAdd.norm2());
    ++numberOfWSVectors;
    if(numberOfWSVectors == 1) minWSnormSq = toAdd.norm2();
    else if(minWSnormSq > toAdd.norm2()) minWSnormSq = toAdd.norm2();  
}

void SkewPeriodicBoundaryConditions::RemoveVectorFromWSCell(int vec)
{
    wignerSeitzVectors.erase(wignerSeitzVectors.begin()+vec);
    wignerSeitzVectorsSq.erase(wignerSeitzVectorsSq.begin()+vec);

    --numberOfWSVectors;

    minWSnormSq = wignerSeitzVectorsSq[0];
    for(unsigned int i=1; i<numberOfWSVectors; ++i)
        if(minWSnormSq > wignerSeitzVectorsSq[i]) 
            minWSnormSq = wignerSeitzVectorsSq[i];
}


void SkewPeriodicBoundaryConditions::EnforceBoundaryConditions(DMArray<double, DIMENSION> *positions) const
{
    double skewPos[DIMENSION], dummy;
    int coord, vec;
    DMArrayIterator<double, DIMENSION> posIter(positions);
    for(posIter.Init(); !posIter.IsDone(); posIter.Next())
    {
        double *atomPos = posIter.Current();
        for(vec=0;vec<DIMENSION;++vec)
            skewPos[vec] = 0;
        for(vec=0;vec<DIMENSION;++vec)
            for(coord=0;coord<DIMENSION;++coord)
                skewPos[vec] += inverseSuperCellVectors[vec][coord] * atomPos[coord];
            
            for(vec=0;vec<DIMENSION;vec++)
            {
				if(wrap[vec])
				{
					if(skewPos[vec] >= 1.)
					{
						skewPos[vec] -= 1.; // Do this if fmod is more expensive than two "if"s.
						if (skewPos[vec] >= 1.)
						{
							skewPos[vec] = modf( skewPos[vec] , &dummy);
						}
					}
					if(skewPos[vec] < 0.0)
					{
						skewPos[vec] += 1.; // Do this if fmod is more expensive than two "if"s.
						if(skewPos[vec] < 0.0)
						{
							skewPos[vec] = 1. +  modf( skewPos[vec] , &dummy);
						}
					}
				} // end of if(wrap[vec])
				
            } // end loop over vec
            
            for(coord=0;coord<DIMENSION;++coord)
                atomPos[coord] = 0;
            for(vec=0;vec<DIMENSION;++vec)
                for(coord=0;coord<DIMENSION;++coord)
                    atomPos[coord] += superCellVectors[coord][vec] * skewPos[vec];
    } // end loop over DMArray
    
}


void SkewPeriodicBoundaryConditions::EnforceBoundaryConditions(double *position) const
{
    double skewPos[DIMENSION], dummy;
    int coord, vec;  
    for( vec = 0; vec < DIMENSION; ++vec )
        skewPos[vec] = 0;
    for( vec = 0; vec < DIMENSION; ++vec )
        for(coord=0;coord<DIMENSION;++coord)
            skewPos[vec] += inverseSuperCellVectors[vec][coord] * position[coord];
	
        
        for( vec = 0; vec < DIMENSION; ++vec )
        {
			if(wrap[vec])
			{
				if(skewPos[vec] >= 1.)
				{
					skewPos[vec] -= 1.;			// Do this if fmod is more expensive than two "if"s.
					if (skewPos[vec] >= 1.)
					{
						skewPos[vec] = modf( skewPos[vec] , &dummy);
					}
				}
				if(skewPos[vec] < 0.0)
				{
					skewPos[vec] += 1.;			// Do this if fmod is more expensive than two "if"s.
					if(skewPos[vec] < 0.0)
					{
						skewPos[vec] = 1. +  modf( skewPos[vec] , &dummy);
					}
					
				}
			} // end if(wrap[vec])
        } // end loop over vec
        for( coord = 0; coord < DIMENSION; ++coord )
            position[coord] = 0;
        for( vec = 0; vec < DIMENSION; ++vec )
            for( coord = 0; coord < DIMENSION; ++coord )
                position[coord] += superCellVectors[coord][vec] * skewPos[vec];
}


void SkewPeriodicBoundaryConditions::DifferenceBoundaryConditions(
                                     DMArray<double, DIMENSION> *displacements) const
{    
    const int nLeaves = displacements->NumberOfLeaves();
    for(int leaf=0; leaf<nLeaves; ++leaf)
    {
        DMArray<double, DIMENSION> *leafArr = displacements->GetLeaf(leaf);
        const int size = leafArr->Size();
        double (&begin)[DIMENSION] = (*leafArr)[0];
        DifferenceBoundaryConditions(&begin,size);
    }
}


void SkewPeriodicBoundaryConditions::DifferenceBoundaryConditions(
                                     double (*displacements)[DIMENSION],
                                     int nAtoms) const
{
    double *distSq = new double[nAtoms];
    int atom;
    for( atom = 0; atom < nAtoms; ++atom )
    {
        double norm = 0.;
        for(int coord=0; coord<DIMENSION; ++coord) 
            norm += displacements[atom][coord] * displacements[atom][coord];
        distSq[atom] = norm;
    }
    for(atom=0; atom<nAtoms; ++atom)
        if(distSq[atom] > minWSnormSq)
			WignerSeitzDifferences(displacements[atom]);
    delete [] distSq;
}


void SkewPeriodicBoundaryConditions::WignerSeitzDifferences(double (&displacement)[DIMENSION]) const
{
    // XXX probably not the fastest algorithm!
    bool goOn = true;
	int counter = 0;
    while(goOn)
    {
		goOn = false;

        for(int vec = numberOfWSVectors - 1; vec >= 0; --vec)
        {

            const double dot = wignerSeitzVectors[vec] * displacement;
            if(dot > wignerSeitzVectorsSq[vec] * (1 + TINY))  // I had to add this TINY, because round-off errors killed me.
            {
                int n = (int) (dot/(wignerSeitzVectorsSq[vec]*(1+TINY)));
				if (n%2 == 1) n++;
                for(int coord=0; coord<DIMENSION; ++coord)
                    displacement[coord] -= n * wignerSeitzVectors[vec][coord] ;
			}
			// 1/19/2002
			// I had problems with a neverending cycle in 2D which cycled from 
			// intersection to intersection (of bounding lines) because
			// for a vector that starts at the intersection,if it's a tiny bit 
			// past one line, it gets sent back to the other of that pair
            // and now it's just inside it, but also just inside the other
			// line of the pair that intersected the first line, and close
			// enough to get moved, this time to the third intersection
			// and then back to the starting point.
			else if (dot <= - wignerSeitzVectorsSq[vec]*(1 + TINY))// was 1-TINY 
            {
				
                int n = -(int) (-dot/(wignerSeitzVectorsSq[vec]*(1+TINY))); // (careful with negative ints) was 1 - TINY
				              
				if (n%2 == -1) n--;
                for(int coord=0; coord<DIMENSION; ++coord)
                    displacement[coord] -= n * wignerSeitzVectors[vec][coord] ;
			}
            else continue;
            goOn = true;
			// break;
        }
		++counter;
		if(counter > 1000)
			throw DMException("SkewPeriodicBoundaryConditions","WignerSeitzDifferences","While loop exceeded 1000 iterations");
    } // end of while loop
}

BoundaryConditions *SkewPeriodicBoundaryConditions::Copy() const
{
    return new SkewPeriodicBoundaryConditions(superCellVectors);
}

void SkewPeriodicBoundaryConditions::Save(DMWriter *writer) const
{
    vector<int> dims(2,DIMENSION);
    writer->PutDoubleArray((double *)superCellVectors, dims, "superCellVectors");
	vector<int> dimsWrap(1,DIMENSION);
	int intWrap[DIMENSION];
	for (int coord = 0; coord < DIMENSION; ++coord )
		if(wrap[coord])
			intWrap[coord] = 1;
		else
			intWrap[coord] = 0;
	writer->PutIntArray((const int*)&intWrap,dimsWrap,"wrap");
}

void SkewPeriodicBoundaryConditions::Load(DMReader *reader)
{
	double localSuperCellVectors[DIMENSION][DIMENSION];

    vector<int> dims;
    double *tmp = reader->GetDoubleArray(dims, "superCellVectors");
    if(dims.size() != 2 || dims[1] != DIMENSION || dims[0] != DIMENSION)
        throw DMException("SkewPeriodicBoundaryConditions", "Load", "corrupted super cell vectors");
    memcpy(localSuperCellVectors, tmp, DIMENSION*DIMENSION*sizeof(double));
    delete [] tmp;
    

	int intWrap[DIMENSION];
	bool localWrap[DIMENSION];
	int *iTmp = reader->GetIntArray(dims,"wrap");
	if(dims.size() != 1 || dims[0] != DIMENSION)
		throw DMException("SkewPeriodicBoundaryConditions", "Load", "corrupted wrap");
	memcpy(&intWrap,iTmp, DIMENSION * sizeof(int));
	delete [] iTmp;	
	for (int coord = 0; coord < DIMENSION; ++coord )
	{
		if(intWrap[coord])
			localWrap[coord] = true;
		else
			localWrap[coord] = false;
	}
	
	// must set wrap before setting SCVs
	SetWrap(localWrap);
	SetSuperCellVectors(localSuperCellVectors);


}


