// SkewPeriodicBoundaryConditions.h: interface for the SkewPeriodicBoundaryConditions class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef SKEWPERIODICBOUNDARYCONDITIONS_H
#define SKEWPERIODICBOUNDARYCONDITIONS_H

#include "DMFirstHeader.h"
#include <vector>

#include "BoundaryConditions.h"

#include "DigitalMaterial.h"

#include "DMArray.h"

#include "TinyVector.h"

//! A boundary conditions class which is periodic with non-orthogonalperodic-axes

/*!
  Instead of the lengths this class has "super cell vectors", which are 
  vectors along the edges of the supercell. These vectors form a 
  matrix, which with its inverse are used to convert actual positions into
  positions in terms of the basis of the basis vectors in which form standard
  periodic boundary conditions are applied. For difference boundary conditions 
  this procedure does not yield the minimum distance image; one really needs
  the Wigner-Seitz cell. Thus there is a function to construct a representation
  of the WS cell and DifferenceBoundaryConditions uses this to return the
  correct differences
  
  \ingroup moleculardynamics
  \author nbailey
  \date 5/28/2001
*/

class SkewPeriodicBoundaryConditions : public BoundaryConditions  
{
public:
    SkewPeriodicBoundaryConditions(const bool wrap[DIMENSION] = 0);
    virtual ~SkewPeriodicBoundaryConditions();
    
    // interface overridden from BoundaryConditions

    void EnforceBoundaryConditions(DMArray<double, DIMENSION> *positions) const;
    void DifferenceBoundaryConditions(DMArray<double, DIMENSION> *displacements) const;
#ifndef SWIG
    void DifferenceBoundaryConditions(double (*displacements)[DIMENSION],int nAtoms) const;
    void EnforceBoundaryConditions(double *position) const;
    
    void Save(DMWriter *writer) const;
    void Load(DMReader *reader);
#endif // SWIG
	//! Get the string name of the class for saving or loading.
    const std::string &GetType() const
		{
			return className;
		}
	
	BoundaryConditions *Copy() const;
	
	void SetWrap(const bool wrap[DIMENSION] = 0);
	void SetSuperCellVectors(const double superCellVectors[DIMENSION][DIMENSION]);
#ifndef SWIG
	
   SkewPeriodicBoundaryConditions(const double superCellVectors[DIMENSION][DIMENSION],const bool wrap[DIMENSION] = 0);
   const double (&GetSuperCellVectors() const)[DIMENSION][DIMENSION]
        { return superCellVectors; }

	bool GetWrap(int coord) const {return wrap[coord];}

private:
    // the following two functions are called by the respective versions
    // of DifferenceBoundaryConditions after the differences are translated to
    // the conventional unit cell centered on the origin.
    // WignerSeitzDifferences then applies additional translations to put
    // the vector inside the Wigner-Seitz cell. They are separate functions
    // because BuildWignerSeitzCell needs this one separately
    void WignerSeitzDifferences(double (&displacements)[DIMENSION]) const;
    double superCellVectors[DIMENSION][DIMENSION];
    double inverseSuperCellVectors[DIMENSION][DIMENSION];

	// ! If the superCellVectors are too skew we choose a different basis which is not (this can always be done)
    double reducedSuperCellVectors[DIMENSION][DIMENSION];
    double inverseReducedSuperCellVectors[DIMENSION][DIMENSION];
    //! Transform the basis vectors, such that |v_i * v_j| < 0.5 v_i * v_i, i != j.
    /*! We add and substract basis vectors together such that
    |v_i * v_j| < 0.5 v_i * v_i, i != j. If this propery holds, then it is possible 
    to show that if w = \sum \lambda_i v_i, with \lambda_i a set of integer numbers and
    max_i |\lambda_i| > 1, then there exists a k for which |w * v_k| > 0.5 v_k * v_k.
    This means that the reduced vectors define the Wigner Seitz cell, and the other vectors
    defining the Wigner Seitz cell can e written as w = \sum_i \lambda_i v_i with 
    |\lambda_i| <= 1.
    */
    void CalculateReducedVectors();

    //! The vectors whose midpoints planes define the Wigner-Seitz cell of the super cell
    std::vector<TinyVector<double, DIMENSION> > wignerSeitzVectors;
    //! The squared norm of the vectors to test.
    std::vector<double> wignerSeitzVectorsSq;
    //! The norm of the smallest wigner seitz vector
    double minWSnormSq;
    unsigned int numberOfWSVectors;
    void InvertMatrix(double mat[DIMENSION][DIMENSION]);
    void BuildWignerSeitzCell();
    void AddVectorToWSCell(const TinyVector<double> &toAdd);
    void RemoveVectorFromWSCell(int vec);
    static const double TINY;
    static bool lesser_norm2(const TinyVector<double> &a, const TinyVector<double> &b)
    { return (a.norm2() < b.norm2()) ;} 


	bool wrap[DIMENSION];
	unsigned int nWrapped;
	static const std::string className;
#endif // SWIG
};

#endif // !defined(SKEWPERIODICBOUNDARYCONDITIONS_H)
