// 3DLattices.cpp: implementations for the 3DLattices
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/


#include "DMFirstHeader.h"
#include "DigitalMaterial.h"
#include "3DLattices.h"

#if DIMENSION == 3


#include "SerialBuilder.h"
#include "DMReader.h"
#include "DMWriter.h"

#include "DMLastHeader.h"

using namespace std;

static SerialBuilder<FCCLattice> *_FCCLatticecreator = 
    new SerialBuilder<FCCLattice>("FCCLattice");

static SerialBuilder<PrimitiveFCCLattice> *_PrimitiveFCCLatticecreator = 
    new SerialBuilder<PrimitiveFCCLattice>("PrimitiveFCCLattice");

static SerialBuilder<BCCLattice> *_BCCLatticecreator = 
    new SerialBuilder<BCCLattice>("BCCLattice");

static SerialBuilder<DiamondLattice> *_DiamondLatticecreator = 
    new SerialBuilder<DiamondLattice>("DiamondLattice");

static SerialBuilder<TriclinicLattice> *_TriclinicLatticecreator = 
    new SerialBuilder<TriclinicLattice>("TriclinicLattice");

const string CubicLattice::className = string("CubicLattice");

const string FCCLattice::className = string("FCCLattice");

const string PrimitiveFCCLattice::className = string("PrimitiveFCCLattice");

const string BCCLattice::className = string("BCCLattice");

const string DiamondLattice::className = string("DiamondLattice");

const string TriclinicLattice::className = string("TriclinicLattice");




////////////////////////  CUBIC LATTICE //////////////////////////////////////

CubicLattice::CubicLattice(double a)
{
    for(int vec=0; vec<DIMENSION; vec++)
    {
        for(int coord=0; coord<DIMENSION;coord++)
        {
            latticeVectors[vec][coord] = (vec==coord? a : 0.);
        }
    }
    latticeConstant = a;
    UpdateMemberVariables();
}

void CubicLattice::SetLatticeConstant(double a)
{
    double factor = a/latticeConstant;
    for(int vec=0; vec<3; vec++)
    {
        for(int coord=0; coord<DIMENSION; coord++)
            latticeVectors[vec][coord] *= factor;
    }
    latticeConstant = a;
    UpdateMemberVariables();
}



////////////////////////  FCC LATTICE //////////////////////////////////////

FCCLattice::FCCLattice(double a): CubicLattice(a)
{
    vector<TinyVector<double,DIMENSION> > atomsCoord;
    atomsCoord.push_back(TinyVector<double, DIMENSION> (0.));
    SetAtomsCoordinatesInCell(atomsCoord);
    UpdateMemberVariables();
}

void FCCLattice::AddAtomInCell(TinyVector<double, DIMENSION> atomCoord,
                               string name)
{
    double shifts[4][3] = {{0.,0.,0.},{0.,0.5,0.5},{0.5,0.,0.5},{0.5,0.5,0.}};
    for(int at=0; at<4; ++at)
        BravaisLatticeWithBasis::AddAtomInCell(atomCoord+shifts[at], name);
}


////////////////////////  PRIMITIVE FCC LATTICE //////////////////////////////////////

PrimitiveFCCLattice::PrimitiveFCCLattice(double a)
{
    for(int vec = 0; vec < DIMENSION; vec++)
    {
        for(int coord = 0; coord < DIMENSION; coord++)
        {
            latticeVectors[vec][coord] = (vec == coord ? 0.0 : 0.5 * a);
        }
    }
    latticeConstant = a;
    UpdateMemberVariables();
}

void PrimitiveFCCLattice::ResetOrientation()
{
    double newVectors[DIMENSION][DIMENSION];
    for(int vec = 0; vec < DIMENSION; vec++)
        for(int coord = 0; coord < DIMENSION; coord++)
        {
            newVectors[vec][coord] =
                (vec == coord ? 0.0 : 0.5 * latticeConstant);
        }
    SetLatticeVectors(newVectors);
}

void PrimitiveFCCLattice::SetLatticeConstant(double a)
{
    double factor = a/latticeConstant;
    for(int vec=0; vec<3; vec++)
    {
        for(int coord=0; coord<DIMENSION; coord++)
            latticeVectors[vec][coord] *= factor;
    }
    latticeConstant = a;
    UpdateMemberVariables();
}


////////////////////////  BCC LATTICE //////////////////////////////////////

BCCLattice::BCCLattice(double a): CubicLattice(a)
{
    vector<TinyVector<double,DIMENSION> > atomsCoord;
    atomsCoord.push_back(TinyVector<double, DIMENSION> (0.));
    SetAtomsCoordinatesInCell(atomsCoord);
    UpdateMemberVariables();
}

void BCCLattice::AddAtomInCell(TinyVector<double, DIMENSION> atomCoord,
                               string name)
{
    double shifts[2][3] = {{0.,0.,0.},{0.5,0.5,0.5}};
    for(int at=0; at<2; ++at)
        BravaisLatticeWithBasis::AddAtomInCell(atomCoord+shifts[at], name);
}


////////////////////////  DIAMOND LATTICE //////////////////////////////////////

DiamondLattice::DiamondLattice(double a): FCCLattice(a)
{
	vector<TinyVector<double,DIMENSION> > atomsCoord;
	atomsCoord.push_back(TinyVector<double, DIMENSION> (0.));
	SetAtomsCoordinatesInCell(atomsCoord);
	UpdateMemberVariables();
}


void DiamondLattice::AddAtomInCell(TinyVector<double, DIMENSION> atomCoord,
                              string name)
{
    double shifts[2][3] = {{0.,0.,0.},{0.25,0.25,0.25}};
	FCCLattice::AddAtomInCell(atomCoord+shifts[0], name);
    FCCLattice::AddAtomInCell(atomCoord+shifts[1], name);
}


////////////////////////  TRICLINIC LATTICE //////////////////////////////////////


TriclinicLattice::TriclinicLattice(const double latticeConstants[DIMENSION],
                                   const double anglesInUnitCell[DIMENSION])
{	
    for (int vec=0; vec<DIMENSION; vec++)
    {
        this->latticeConstants[vec] = latticeConstants[vec];
        this->anglesInUnitCell[vec] = anglesInUnitCell[vec];
    }
    Set3DLatticeVectors();
    UpdateMemberVariables();
}

TriclinicLattice::TriclinicLattice()
{
    for(int vec=0; vec<DIMENSION; ++vec)
    {
        anglesInUnitCell[vec] = PI/2;
        latticeConstants[vec] = 1.0;
    }
    Set3DLatticeVectors();
    UpdateMemberVariables();
}

void TriclinicLattice::SetLatticeConstants(const double latticeConstants[DIMENSION])
{
    for(int vec=0; vec<DIMENSION; vec++)
    {
        double factor = latticeConstants[vec]/this->latticeConstants[vec];
        for(int coord=0; coord<DIMENSION; coord++) this->latticeVectors[vec][coord] *= factor;
        this->latticeConstants[vec] = latticeConstants[vec];
    }
    UpdateMemberVariables();
}

void TriclinicLattice::SetAnglesInUnitCell(const double anglesInUnitCell[DIMENSION])
{
    for(int coord=0; coord<DIMENSION; coord++)
        this->anglesInUnitCell[coord] = anglesInUnitCell[coord];
    Set3DLatticeVectors();
    UpdateMemberVariables();
}

void TriclinicLattice::Set3DLatticeVectors()
{
    latticeVectors[0][0] = latticeConstants[0]; 
    latticeVectors[0][1] = 0.;
    latticeVectors[0][2] = 0.;
    latticeVectors[1][0] = latticeConstants[1]*cos(anglesInUnitCell[2]);
    latticeVectors[1][1] = latticeConstants[1]*sin(anglesInUnitCell[2]);
    latticeVectors[1][2] = 0.;
    latticeVectors[2][0] = latticeConstants[2]* cos(anglesInUnitCell[1]);
    latticeVectors[2][1] = latticeConstants[2]*(cos(anglesInUnitCell[0])-
                           cos(anglesInUnitCell[1])*cos(anglesInUnitCell[2]))
                           /sin(anglesInUnitCell[2]);
    latticeVectors[2][2] = sqrt(latticeConstants[2]*latticeConstants[2]-
                                latticeVectors[2][0]*latticeVectors[2][0]-
                                latticeVectors[2][1]*latticeVectors[2][1]);
    if(latticeVectors[1][1] < 0)
        latticeVectors[2][2] = -latticeVectors[2][2];
    
    UpdateMemberVariables();
}

void TriclinicLattice::Load(DMReader *reader)
{
    BravaisLatticeWithBasis::Load(reader);
    anglesInUnitCell[0] = DotProduct(latticeVectors[1], latticeVectors[2], 3);
    anglesInUnitCell[1] = DotProduct(latticeVectors[2], latticeVectors[0], 3);
    anglesInUnitCell[2] = DotProduct(latticeVectors[0], latticeVectors[1], 3);
    for(int vec=0; vec<DIMENSION; ++vec)
    {
        TinyVector<double, 3> tmp(latticeVectors[vec]);
        latticeConstants[vec] = sqrt(tmp.norm2());
    }
}


#endif // DIMENSION == 3
