// AndersonDynamics.h: interface for the AndersonDynamics class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef ANDERSONDYNAMICS_H
#define ANDERSONDYNAMICS_H

#include "DigitalMaterial.h"
#include "Constraint.h"
#include "Rand.h"

//! The following constraint can be used to thermalize atoms. The algorithm is described in D. Frenkel, "Understanding Molecular Simulations"
/*!
	\ingroup moleculardynamics
	The idea behind the Andersen dynamics is that some of the atoms undergo random collisions with fictious 
	particles that lead to their thermalization. The occurence of these collisions is given by a Poisson 
	process: the probability that a collision occurs in some interval of time is proportionnal to the length
	of the interval. The proportionnality coefficient (called here \a frequency because it is the frequency of 
	the collisions) is the only parameter. It seems that its actual value is not crucial to get a canonical distribution
	but we shouldn't use this method to measure any dynamical quantity. 

	We can implement the method in the form of a constraint where we adjust the velocities at each time step. So in 
	principle, we can use the constraint only with a Verlet-like algorithm. In fact the important parameter is 
	frequency * dt, so this is the parameter which is passed in the constructor. If we change the time step, we
	should change frequencyDt accordingly.
*/
class AndersenDynamics : public Constraint
{
public: 
	AndersenDynamics(double frequencyDt, double sigma, int seed=1);
	~AndersenDynamics(){}

	void AdjustVelocities(DMArray<double, DIMENSION> *velocities);

	void SetFrequencyDt(double frequencyDt) {this->frequencyDt = frequencyDt;}
	double GetFrequencyDt() const {return frequencyDt;}

	void SetSigma(double sigma) {this->sigma = sigma;}
	double GetSigma() const {return sigma;}


	const std::string &GetType() const
	{ 
		return className; 
	}

private:
	double frequencyDt;
	double sigma;
	Rand rand;
	static const std::string className;
};



#endif // !defined(ANDERSONDYNAMICS_H)
