// SingleDirectionRigidBody.h: interface for the SingleDirectionRigidBody class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef SINGLEDIRECTIONRIGIDBODY_H
#define SINGLEDIRECTIONRIGIDBODY_H

#include "DigitalMaterial.h"
#include "Constraint.h"

//! Like NonRotatingRigidBody, but only coord==0 positions are constrained

/*!

  The single direction can be the coord == 0 direction, coord == 1 direction 
  etc. We store just the constrainedCoord of the center of mass, and average 
  only the corresponding components of the forces etc. So the number
  of degrees of freedom is 1 (the directional COM) plus (DIMENSION-1) 
  times the number of atoms. Having the to deal with all the other positions
  makes the functions dealing with GeneralizedCoordinates a bit messy.

  Perhaps this could be made more general: Different directions other than the
  coordinate axes would be allowed (eg constrain motion along an arbitrary
  straight line to be rigid). Also, it could all be included in the same class
  as NonRotatingRigidBody---just turn on and off the components you want
  contrained,  or pass some kind of matrix to specify them.

  The date below refers to creation of this file. The class previously existed
  in the single Constraint.h file.

  \ingroup MolecularDynamics
  \author Nicholas Bailey
  \date March 1, 2001
  
 */ 


class SingleDirectionRigidBody : public Constraint
{
 public:
	SingleDirectionRigidBody(int constrainedCoord =0,double weight = 1.0); 
	~SingleDirectionRigidBody() ;
	Constraint *Copy() const;

	void AdjustForces(DMArray<double, DIMENSION> *forces);
	void AdjustForceIncrements(DMArray<double, DIMENSION> *forces);
	void AdjustVelocities(DMArray<double, DIMENSION> *velocities);

	int NumberGeneralizedCoordinates();
	void SetGeneralizedCoordinates(double *generalizedCoordinates) ;
    void GetGeneralizedCoordinates(double *generalizedCoordinates); 
	void IncrementGeneralizedCoordinate(int n,double inc);
#ifndef SWIG
    void CalculateGeneralizedForces(DMArray<double, DIMENSION> *forces, double *generalizedForces);
    void DerivativesWRTGeneralizedCoordinates(int atom, double (*gradient)[DIMENSION]);
#endif // SWIG
	const std::string &GetType() const
		{
			return className;
		}


private:
	int constrainedCoord;
	double weightFactor;
	void InitializeCOMAndDisplacements();
	void SetCartesianPositionsFromCOM(double *restOfPositions = 0);
	double constrainedCenterOfMass;
	bool centerOfMassIsDefined;
	DMArray<double, DIMENSION> *displacements;
	static const std::string className;
};


#endif // !defined(SINGLEDIRECTIONRIGIDBODY_H)
