// DMArray_AnyType.h: interface for the base class for DMArrays.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef DMARRAY_ANYTYPE_H
#define DMARRAY_ANYTYPE_H

#include "DMFirstHeader.h"
#include <vector>
#include <set>
#include <assert.h>
#include <iostream>

#include "blas.h"
#include "DigitalMaterial.h"
#include "Shape.h"
#include "Serializable.h"

class DMWriter;
class DMReader;
class SynchronizationToken;
class SynchronizationKit_AnyType;
class SynchronizationKit_AnyType;
class Buffer;

typedef unsigned int LeafIndex;
class Index
{
public:
	unsigned int inLeaf;
	LeafIndex leaf;
	Index(LeafIndex leaf=0, unsigned int inLeaf=0): inLeaf(inLeaf), leaf(leaf) {}
	bool operator==(const Index &index) const {return ((inLeaf == index.inLeaf) && (leaf == index.leaf));}
	bool operator!=(const Index &index) const {return !((inLeaf == index.inLeaf) && (leaf == index.leaf));}
	bool operator<(const Index &index) const 
		{return ((leaf < index.leaf) || ((leaf == index.leaf) && (inLeaf < index.inLeaf)) );}
};


/*!
	This file defines the base class for the arrays which are used in several modules, like 
	molecular dynamics and quasi-continuum. There is a top base class common to all the arrays
	(whatever their type and size) called \a DMArray_AnyType. It allows to pass as an 
	argument an array of 
	any type. We want to allow both types of storage in memory: contiguous and chunk-like because 
	a composite array is not requested to be contiguous in memory. DMARRAY's member variable \a shape
	characterizes the first index of the array: it is only a number for a contiguous array, or a 
	collection of ints reflecting the number of entries in the tree structure of a composite array.
  
	This class is abstract with some implementation of the handling of shapes.

	Deriving from DMARRAY are DMMATRIX, whose template arguments are the type of the data and the
	number N of columns. Indexing a DMMATRIX returns a pointer to the column. Because this is not 
	very intuitive when N=1 (we expect to get a Type instead of a pointer), we have defined the special
	class DMVECTOR which can be viewed as a partial template specialization of DMMATRIX<type, N> for
	N=1. Unfortunately VC++ doesn't accept partial template specialization; this is why DMVECTOR has been 
	introduced.

	Finaly we have declared Iterator classes, DMMATRIXITERATOR and DMVECTORITERATOR, in the way 
	described by the Design Patterns book (see comment below). 

	WARNING: the concrete implementations of these classes is driven by the aim of efficiency instead of
	genericity. For exemple, we use memcpy to copy and we suppose all the element occupy the same amount
	of memory, so do not use too strange Types as template arguments (like std::vector or so).

	\date April 2000
	\ingroup dmsystem
	*/


class DMArray_AnyType : public Serializable
{
	SynchronizationToken* siblings;
public:
	//! Constructor of zero size array.
	DMArray_AnyType();
	//! Construct an array with a given size
	DMArray_AnyType(const Shape& shape);
	//! Constructor of one array as a sibling to another.
	DMArray_AnyType(const DMArray_AnyType *array);
	//! Destructor.
	virtual ~DMArray_AnyType();
	//! Number of elements in the array.
	virtual void SetShape(const Shape &shape) = 0;
	int Size() const;
	//! What is the shape of the array.
	const Shape &GetShape() const;
	//! This is the type returned by GetSiblings.
	typedef std::vector<DMArray_AnyType*> siblingContainer;
	//! Tells you all DMArrays created from each other and still of same shape.
	const siblingContainer& GetSiblings() const;
	//! Remove the elements in the set
	virtual void RemoveElements(const std::set<Index> &indices) = 0;
	//! Add space at the end.
	virtual void AddElementsAtEnd( int n ) = 0;
	virtual void Zero() = 0; // initializes the array with zeroes
	//! These support serialization.
	//!{
	virtual void Save(DMWriter *writer) const = 0;
	virtual void Load(DMReader *reader) = 0;

	//!}
	SynchronizationKit_AnyType *GetSynchronizationKit() const {return synchroKit;}
protected:
	// it's allocated by the concrete arrays
	SynchronizationKit_AnyType *synchroKit;
	Shape shape;
};


#endif // !defined(DMARRAY_ANYTYPE_H)
