// Shape.h: interface for the Shape class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef SHAPE_H
#define SHAPE_H

#include "DMFirstHeader.h"
#include<vector>
#include<iostream>
#include "DigitalMaterial.h"
#include "Serializable.h"

//! Expresses the tree structure of DMArrays or ListsOfAtoms.
/** 
Shape is meant to be a common feature of our composed container-like objects, like DMArrays, 
ListsOfAtoms. Shape characterizes the tree structure of these composite objects. 

  Its member variables are \a total (the total number of elements, which is cached), 
  \a leavesSize (the size of the leaves) and \a treeStructure (a container that codes the structure
  of the tree). A leaf is a Shape whose treeStructure is empty.
  
    The default state of a Shape is leaf with 0 elements (such it's possible to transform 
    it to composite if we attach it an other Shape).
    
    Shape can be streamed (<<) and compared (==).
      
        
    Note that this is a different implementation than what we had before: now the emphasis is on 
    the leaves. So it's faster to request the numbers attached to the leaves than to ask for 
    a subShape (ie. a subtree).
    \ingroup dmsystem
**/

class Shape : public Serializable
{
public:
    //! Check equality.
    friend bool operator==(const Shape &shape1, const Shape &shape2);
    //! Print a shape.
    friend std::ostream &operator<<(std::ostream &out, const Shape &shape);
    typedef bool UpDown;
    //! A vector of bools that characterize the path along the tree (with ups and downs).
    typedef std::vector<UpDown> TreeStructure;
    static const bool Up;
    static const bool Down;
    //! Constructor for leaf	
    Shape(int n=0); 
    //! Copy Constructor
    Shape(const Shape &shape);
    //! Take the tree structure of a shape and fill the leaves with values in \a leafValues
    Shape(const TreeStructure &treeStruct, const std::vector<int> &leavesSize);
    //! assignment operator
    Shape &operator=(const Shape &shape);
    //! Destructor.
    ~Shape();
    //! Tells whether we are allowed to attach something to it.
    bool IsLeaf() const {return treeStruct.empty();}
    //! Tells whether shapes have to same tree structure
    bool IsSimilar(const Shape &shape) const {return (treeStruct == shape.treeStruct);}
    //! Returns the total number of elements
    int GetSize() const {return total;}
    //! Cleans up all the internal data. Leaves Shape in a leaf state.
    void Clear();
    //! Merges shape into \a this. They must have the same tree structure 
    void Merge(const Shape &shape);
    //! Attaches shape as a subShape of \a this (which must be composite or an empty leaf)
    void Attach(const Shape &shape);
    //! How many branches does this Shape have?
    int GetNSubShapes() const;
    //! Get a subShape by index.
    Shape GetSubShape(int i) const ;
    //! How many leaves does this Shape have?
    int NumberOfLeaves() const {return leavesSize.size();}
    //! Return the tree structure: ie a similar shape where all leaves are of length 0
    const TreeStructure &GetTreeStructure() const {return treeStruct;}
    const std::vector<int> &GetLeavesSize() const {return leavesSize;}
    
    //! Save the class.
    void Save(DMWriter *writer) const;
    //! Load the class.
    void Load(DMReader *reader);
    //! Get the string name of the class for saving or loading.
    const std::string &GetType() const 
	{
		return className; //"Shape";
	}

    //! Returns the global shape (merges the corresponding shapes over all the processors)
    /*! We do not cache it at this time: all to all communication involved */
    const Shape Global() const;
private:
    TreeStructure treeStruct;
    std::vector<int> leavesSize;
    int total;
	static const std::string className;
};

bool operator==(const Shape &shape1, const Shape &shape2);
std::ostream &operator<<(std::ostream &out, const Shape &shape);


#endif // !defined(SHAPE_H)

