// SynchronizationKit.h
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/


#ifndef SYNCHRONIZATIONKIT_H
#define SYNCHRONIZATIONKIT_H

#include "DigitalMaterial.h"
#include "SynchronizationKit_AnyType.h"

template<class Type, int N> class DMArray;

//! This won't be a problem in serial.
#include "Parallel/Synchronizer.h"

//! A class that is responsible for allocating ghost arrays and calling the synchronizer

/*!
\sa SynchronizationKit_AnyType

*/

template<class Type, int N=1>
class SynchronizationKit : public SynchronizationKit_AnyType
{
public:
    SynchronizationKit(DMArray<Type, N> *array)
    {
        real          = array;
        extendedArray = array;
        ghosts	      = 0;
        calledInit    = false;
    }
    ~SynchronizationKit();
    void SetSynchronizer(Synchronizer *sync);
    const DMArray<Type, N> *GetExtended() const{return extendedArray;}
    //! returns the extended array
    //! returns our \ghosts
    const DMArray<Type, N> *GetGhosts() const  {return ghosts;}
    //! returns the extended array
    DMArray<Type, N> *GetExtended()            {return extendedArray;}
    //! returns our \ghosts
    DMArray<Type, N> *GetGhosts()              {return ghosts;}
    //! returns our \ghosts as a \a DMArray_AnyType
    DMArray_AnyType *GetUntypedGhosts() const  {return ghosts;}

    const Shape &ExtendedShape() const {return extendedArray->GetShape();}

    void WriteItems(const std::vector<Index> &indices, DMWriter *writer);
    void ReadItems(const std::vector<Index> &indices, DMReader *reader);
    void AddItems(const std::vector<Index> &indices, DMReader *reader);
    //! Allocates memory for the \a ghost array
    /*!(see comment in \a SynchronizationKit_AnyType)
    */
    void Init();
    //! Calls our synchronizer's Synchronize method.
    void Synchronize(); 
    void SynchronizeAndAdd(); 
private:
    //! A flag that stores whether our array will be synchronizable.
    bool calledInit;
    //! Our real array.
    DMArray<Type, N> *real;
    //! The ghost array of our real array.
    DMArray<Type, N> *ghosts;
    //! The tree array that has \a real and \a ghosts as branches. 
    DMArray<Type, N> *extendedArray;
};


template<class Type, int N>
SynchronizationKit<Type, N>::~SynchronizationKit()
{
    if(extendedArray != real)
    {
        extendedArray->RemoveBranch(real);
        delete extendedArray;
    }
    extendedArray = 0;
    ghosts = 0;
}


template<class Type, int N>
void SynchronizationKit<Type, N>::SetSynchronizer(Synchronizer *sync)
{
    this->synchronizer = sync;
    if (calledInit) Init();
}


template<class Type, int N>
void SynchronizationKit<Type, N>::Init()
{
    this->calledInit = true;
#ifdef DMPARALLEL
    if(!synchronizer) return;
    if(ghosts) return;
    
    this->extendedArray = new DMArray<Type, N>();
    extendedArray->Attach(real);
    this->ghosts = new DMArray<Type, N>(this->synchronizer->GetGhostsShape());
    extendedArray->Attach(ghosts);
    
    ASSERT(real->shape.IsSimilar(ghosts->GetShape()));
#endif //DMPARALLEL
}

template<class Type, int N>
void SynchronizationKit<Type, N>::Synchronize()
{
#ifdef DMPARALLEL
    ASSERT(calledInit);
    if (this->synchronizer)
        this->synchronizer->Synchronize(real);
#endif //DMPARALLEL
}

template<class Type, int N>
void SynchronizationKit<Type, N>::SynchronizeAndAdd()
{

#ifdef DMPARALLEL
    ASSERT(calledInit);
    if (this->synchronizer)
        this->synchronizer->SynchronizeAndAdd(real);
#endif //DMPARALLEL
}

template<class Type, int N>
void SynchronizationKit<Type, N>::WriteItems(const std::vector<Index> &indices,
                                             DMWriter *writer)
{
    int itemSize = N*sizeof(Type);
    Type *tmp = 0;
    if(indices.size() != 0) tmp = new Type[N*indices.size()];
    Type *tmpVar = tmp;
    for(std::vector<Index>::const_iterator iter=indices.begin();
    iter != indices.end(); ++iter)
    {
        memcpy(tmpVar, real->cache.ptrs[iter->leaf][iter->inLeaf], itemSize);
        tmpVar += N;
    }
    std::vector<int> dims;
    dims.push_back(indices.size());
    int tmpI = N;
    dims.push_back(tmpI);
    writer->PutArray(tmp, dims, "");
    delete [] tmp;
}

template<class Type, int N>
void SynchronizationKit<Type, N>::ReadItems(const std::vector<Index> &indices,
                                            DMReader *reader)
{
    std::vector<int> dims;
    Type *tmp;
    reader->GetArray(dims, "", tmp);
    ASSERT(dims.size() == 2);
    ASSERT(dims[1] == N);
    Type (*tmpVar)[N] = reinterpret_cast<Type (*)[N]>(tmp);
    for(std::vector<Index>::const_iterator iter=indices.begin(); 
    iter != indices.end(); ++iter)
    {
        memcpy(real->cache.ptrs[iter->leaf][iter->inLeaf],
            tmpVar++, N*sizeof(Type));
    }
    delete [] tmp;
}

template<class Type, int N>
void SynchronizationKit<Type, N>::AddItems(const std::vector<Index> &indices,
	                                           DMReader *reader)
{
    std::vector<int> dims;
    double *tmp;
	int i;
    reader->GetArray(dims, "", tmp);
    ASSERT(dims.size() == 2);
    ASSERT(dims[1] == N);
    double (*tmpVar)[N] = reinterpret_cast<double (*)[N]>(tmp);
    for(std::vector<Index>::const_iterator iter=indices.begin(); 
    iter != indices.end(); ++iter)
	{
		real->cache.add_to_ptr(iter->leaf,iter->inLeaf,real,tmpVar++);
	}
    delete [] tmp;

}

#endif // !defined SYNCHRONIZATIONKIT_H
