// SynchronizationToken.h: interface for the SynchronizationToken class
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef _SYNCHRONIZATION_TOKEN_H_
#define _SYNCHRONIZATION_TOKEN_H_

#include <set>
#include "DigitalMaterial.h"

//! Stores relationship among sibling DMArrays.
/*!	We want arrays to be siblings for the sake of network
	migration of several arrays at the same time.  This token
	stores the arrays which have asked to be joined together.

	This object is guaranteed to have a pointer to at least one
	array at all times.  If you try to remove the last one, it
	will fail, so that it always has at least a pointer to 
	the array that owns it.

	When deleting a DMArray, the destructor needs to see whether
	it should free the memory associated with this object.
	\begincode
	if (token->GetSiblings().size()==1)
		free token;
	else
		token->RemoveSibling(this);
	\endcode

	\author ajd27
	\date 29 August 2000
	*/
class DMArray_AnyType;

class SynchronizationToken
{
	typedef DMArray_AnyType::siblingContainer arrayVector;
	arrayVector siblings; //! Siblings are the arrays joined together.
public:
	//! Constructor to record this single array making the token.
	SynchronizationToken( DMArray_AnyType*const array );
	//! Destructor
	~SynchronizationToken( );
	//! Add an array to the list of arrays.
	bool AddSibling( DMArray_AnyType*const array );
	//! Remove an array from the list
	bool RemoveSibling( DMArray_AnyType*const array );
	//! Find out what siblings this token has.
	const arrayVector& GetSiblings( ) const;
#ifndef NDEBUG
	bool Contains( DMArray_AnyType*const array ) const;
#endif
};

#endif //_SYNCHRONIZATION_TOKEN_H_
