// BondBreakerNeighborLocator.h: interface for the BondBreakerNeighborLocator class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef BONDBREAKERNEIGHBORLOCATOR_H
#define BONDBREAKERNEIGHBORLOCATOR_H

#include "DMFirstHeader.h"
#include "DigitalMaterial.h"

#include "NeighborLocator.h"

/** This is a decorator to any type of neighborlocator. As the name suggests, the BondBreakerNeighborLocator
post-processes the pairs finding of another NeighborLocator by removing from the list those pairs which cross
half a plane. This is meant to be used primarily in the framework of fracture, in order to create a real notch
(ie without removing atoms). The specification of the half plane is given by DIMENSION double[DIMENSION].
The first, called \a tipLocation, is a point on the cutting line. The second, \a cutOrientation, is the 
vector perpendicular to both the cutting line and the normal of the cutting plane, oriented such that
it points to the direction of the cut. The third, \a planeOrientation, is the normal to the cutting plane. 
It is not used in 2D.

Examples:

In 2D, if we want to cut along the line x<0, then \a tipLocation = {0, 0} and \a cutOrientation is {-1, 0}. 

In 3D, if we want to cut along the half plane z=0, y>0, then \a tipLocation = {a, 0, 0} (a being any number),
\a cutOrientation = {0,1,0}, and \a planeOrientation = {0, 0, 1}.

Internally the class builds a right handed coordinate system (e1, e2, e3)  ( (e1, e2) in 2D) centered at 
\a tipLocation, where the first basis vector (e1) is \a cutOrientation. In 3D , e2 = \a planeOrientation
and e3 is orthogonal to e1 and e2. In 2D, e2 is orthogonal to e1.

A bond between atoms A and B is cut if the 2 conditions are true: 1. (x_A - \a tipLocation).e2 and 
(x_B - \a tipLocation).e2 have different signs. 
2. (x_A- \a tipLocation).e1 - (x_A - \a tipLocation).e2 * ( (x_B-x_A).e1 / (x_B-x_A).e2 ).

  \author Thierry Cretegny
  \date 4 November 2000.
*/

class BondBreakerNeighborLocator : public NeighborLocator  
{
public:
	BondBreakerNeighborLocator(NeighborLocator *neighborLocator, const double tip1Location[DIMENSION],
		const double tip2Location[DIMENSION], const double tipOrientation[DIMENSION], bool stripe=true);
//	BondBreakerNeighborLocator(NeighborLocator *neighborLocator, const double tipLocation[DIMENSION],
//		const double cutOrientation[DIMENSION], const double planeOrientation[DIMENSION]);
	virtual ~BondBreakerNeighborLocator();
	void SetPositionArray(const DMArray<double, DIMENSION> *positions);
#ifndef SWIG
	bool Neighbors(int atomNumber, int* neighbors, int& numNeighbors, double (*displacements)[DIMENSION],
			double* displacementsSquared, int& spacesLeftInArrays);

	bool HalfNeighbors(int atomNumber, int* neighbors, int& numNeighbors, double (*displacements)[DIMENSION],
			double* displacementsSquared, int& spacesLeftInArrays);

	bool Neighbors(const double *center, int* neighbors, int& numNeighbors, double (*displacements)[DIMENSION],	
	double* displacementsSquared, int& spacesLeftInArrays, double cutoff=-1.) ;

	bool Neighbors(int atomNumber, int type1, int type2, int* neighbors, int& numNeighbors, 
			double (*displacements)[DIMENSION],	double* displacementsSquared, int& spacesLeftInArrays);

#endif //SWIG
//! all the other functions are forwarded to the neighbor locator
	void UpdatePosition() {neighborLocator->UpdatePosition();}
	void SetBoundaryConditions(const BoundaryConditions *bc);
	void SetCutoffDistance(double cutoff) {neighborLocator->SetCutoffDistance(cutoff);}
	double GetCutoffDistance() const {return neighborLocator->GetCutoffDistance();} 
	void SetDrift(double drift) {neighborLocator->SetDrift(drift);}

	NeighborLocator *Copy() const;
private:
	NeighborLocator *neighborLocator;

#ifndef SWIG
	void FindCutBonds(const double *center, int *neighbors, int &numNeighbors, double (*displacements)[DIMENSION],
		double *displacementsSquared, int &slotsLeft, double cutoff);

	double tip1Location[DIMENSION];
	double tip2Location[DIMENSION];
	double tipOrientation[DIMENSION];
	double cutOrientation[DIMENSION];
	double planeOrientation[DIMENSION];
	bool stripe;
#endif // SWIG
};

#endif // !defined BONDBREAKERNEIGHBORLOCATOR_H
