// CompositeNeighborLocator.h: interface for the CompositeNeighborLocator class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef COMPOSITENEIGHBORLOCATOR_H
#define COMPOSITENEIGHBORLOCATOR_H

#include "DMFirstHeader.h"
#include <vector>
#include "DigitalMaterial.h"

#include "NeighborLocator.h"
#include "DMArray.h"

/*! 
A CompositeNeighborLocator is designed to be used with a CompositeListOfAtoms (tough it's 
not necessary). They give the possibility to set different cutoff distances between different
types of atoms. For example if a CompositeNeighborLocator is used by a CompositeListOfAtoms 
\a atoms, it will ask the neighborlocator of \a atoms 'children for its neighboring pairs
(ie using the cutoff distance of the children). The CompositeNeighborLocator has it's own
cutoff distance for pairs of atoms of different type. 

This is useful if some if the children of \atoms are fixed for example (and the potential 
is a pair potential) because we can use a NoNeighborLocator for the fixed atoms.

CompositeNeighborLocator is a base class and we have two different implementations:
SimpleCompositeNeighborLocator, which doesn't store any neighborlist for pairs of different 
types, and CompositeNeighborList, which has its own value of drift and stores a 
neighborList of atoms of different type. Here is a concrete example where this is useful: 
in a scattering experiment where mostly one atom is moving a lot while the others form a
surface and stay around their equilibrium position it makes sense to have allow a bigger drift 
for the scattered atom.
*/

class ListOfAtoms ;

class CompositeNeighborLocator : public NeighborLocator  
{
protected:
	CompositeNeighborLocator(double cutoffDistance, const ListOfAtoms *atoms);
public:
	virtual ~CompositeNeighborLocator();
	virtual void SetPositionArray(const DMArray<double, DIMENSION> *positions);
#ifndef SWIG
    double GetDrift() const;
//! a list of atoms can request a composite neighborlocator to rebuild its children list
	void SetChildren();
	const std::string &GetType() { return className; }
protected:
	std::vector<NeighborLocator *> children;
	const ListOfAtoms *atoms;
private:
	static const std::string className;
#endif //SWIG
};

class SimpleCompositeNeighborLocator: public CompositeNeighborLocator
{
public:
	SimpleCompositeNeighborLocator(double cutoffDistance, const ListOfAtoms *atoms=NULL);
	virtual ~SimpleCompositeNeighborLocator();

#ifndef SWIG
	bool HalfNeighbors(int atomNumber, int* neighbors, int& numNeighbors,
		       double (*displacements)[DIMENSION], double* displacementsSquared, 
		       int& spacesLeftInArrays);
	bool Neighbors(int atomNumber, int* neighbors, int& numNeighbors,
		       double (*displacements)[DIMENSION], double* displacementsSquared, 
		       int& spacesLeftInArrays);
	bool Neighbors(const double *center, int* neighbors, int& numNeighbors,
				double (*displacements)[DIMENSION],	double* displacementsSquared,
				int& spacesLeftInArrays, double cutoff=-1.);
	bool Neighbors(int atomNumber, int type1, int type2,
				int* neighbors, int& numNeighbors, double (*displacements)[DIMENSION],
				double* displacementsSquared, int& spacesLeftInArrays);
	void UpdatePosition();
#endif //SWIG
	const std::string &GetType() { return className; }

	NeighborLocator *Copy() const;
private:
	static const std::string className;
};

class CompositeNeighborList: public CompositeNeighborLocator
{
public:
	CompositeNeighborList(double cutoffDistance, const ListOfAtoms *atoms);
	virtual ~CompositeNeighborList();
	void UpdatePosition();
	void SetCutoffDistance(double cutoff);

	NeighborLocator *Copy() const;
	const std::string &GetType() { return className; }

#ifndef SWIG
	bool HalfNeighbors(int atomNumber, int* neighbors, int& numNeighbors,
		       double (*displacements)[DIMENSION], double* displacementsSquared, 
		       int& spacesLeftInArrays);
	bool Neighbors(int atomNumber, int* neighbors, int& numNeighbors,
		       double (*displacements)[DIMENSION], double* displacementsSquared, 
		       int& spacesLeftInArrays);
	bool Neighbors(const double *center, int* neighbors, int& numNeighbors,
				double (*displacements)[DIMENSION],	double* displacementsSquared,
				int& spacesLeftInArrays, double cutoff=-1.);
	bool Neighbors(int atomNumber, int type1, int type2,
				int* neighbors, int& numNeighbors, double (*displacements)[DIMENSION],
				double* displacementsSquared, int& spacesLeftInArrays);
private:
	class NeighborList
	{
	public:
		NeighborList(int numberOfTypes);
		virtual ~NeighborList(); 
		std::vector<int> *nbrsOfType;
	};
	std::vector<NeighborList> *neighborList2;
	bool DataIsValid();
	void UpdateMyData();
	Shape atomsShape;
	std::vector<int> **neighborList; // first index is which atom, second index is which branch
	void MakeNeighborList();
	//We may want to cache to information allowing us to find which branch an atom is.
	std::vector<int> partialSum;
	int GetType(int nAtom)
	  {
	    std::vector<int>::const_iterator it=partialSum.begin();
	    int Type = 0;
	    while(*it++ < nAtom) ++Type;
	    return Type;
	  }
	static const std::string className;
#endif //SWIG

};
#endif // !defined(COMPOSITENEIGHBORLOCATOR)
