// NeighborLocator.cpp: implementation of the NeighborLocator class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DigitalMaterial.h"

#include "NeighborLocator.h"
#include "FreeBoundaryConditions.h"
#include "DMArray.h"
#include "DMException.h"
#include "AttributeError.h"
#include "DMLastHeader.h"

using namespace std;

const string NeighborLocator::className = string("NeighborLocator");

const string NoNeighborLocator::className = string("NoNeighborLocator");


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

NeighborLocator::NeighborLocator(double cutoffDistance)
{
	bc = 0;
	positions = 0;
	SetCutoffDistance(cutoffDistance);
#ifdef PRUNESTATS
	rejected = 0;
	okay = 0;
#endif //PRUNESTATS
}

NeighborLocator::~NeighborLocator()
{
}

void NeighborLocator::SetPositionArray(const DMArray<double, DIMENSION> *positions)
{
  	this->positions = positions;
   	UpdatePosition();
}

void NeighborLocator::SetCutoffDistance(double cutoff)
{
	cutoffDistance = cutoff;
	cutoffDistance2 = cutoff*cutoff;
}

void NeighborLocator::PruneNeighborCandidates(const double *center, int numberOfCandidates,
			int *indexOfCandidates, const DMArray<double, DIMENSION> *positionArray, 
			double (*displacements)[DIMENSION], double *displacementsSquared, 
			double sqCutDist, int &numberOfNeighbors)
{
	// now we calculate the displacements of all candidates
	int i, coord;
	
	double *positionOfOtherAtom;

	for(i=0; i<numberOfCandidates; ++i)
	{
		positionOfOtherAtom = (*positionArray)[indexOfCandidates[i]];
		for (coord = 0; coord < DIMENSION; coord++)
			displacements[i][coord] = positionOfOtherAtom[coord] - center[coord];
	}
   		
	// Prune: apply boundary conditions and keep only the atoms which
    // are within cutOffDistance of atom atomNumber
	bc->DifferenceBoundaryConditions(displacements, numberOfCandidates);
	double dr2;
	for(i=0; i<numberOfCandidates; ++i)
	{
		dr2 = 0.0;
		for (coord = 0; coord < DIMENSION; coord++) 
			dr2 += displacements[i][coord] * displacements[i][coord];
		displacementsSquared[i] = dr2;
	}

	double (*tmpDisplacementsPtr)[DIMENSION] = displacements;
	double *tmpDisplacementsSquaredPtr = displacementsSquared;
	int *tmpCandidatesPtr = indexOfCandidates;
	numberOfNeighbors = 0;
	for(i=0; i<numberOfCandidates; ++i)
	{
		if (tmpDisplacementsSquaredPtr[i] <= sqCutDist)
		{
			*indexOfCandidates++ = tmpCandidatesPtr[i];
			*displacementsSquared++ = tmpDisplacementsSquaredPtr[i];
			//changing memcpy to memmove eliminates "Overlapping source and destination blocks"
			//errors reported by valgrind, but causes rougly a 30% increase in cpu time and does
			//not appear to affect memory usage
			memcpy(displacements, tmpDisplacementsPtr[i], DIMENSION*sizeof(double));
			displacements++;
			numberOfNeighbors++;
#ifdef PRUNESTATS
			okay ++;
		}
		else rejected++;
#else
		}
#endif //PRUNESTATS
	}
}

////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////

void NeighborLocator::Save(DMWriter *writer) const
{
	writer->PutDouble(cutoffDistance, "cutoffDistance");
}

void NeighborLocator::Load(DMReader *reader)
{
	SetCutoffDistance(reader->GetDouble("cutoffDistance"));
}


const std::string &NeighborLocator::GetType() const
{
    return className;
}


// ***********************************************************************


NoNeighborLocator::NoNeighborLocator(double cutoffDistance): NeighborLocator(cutoffDistance)
{
	this->cutoffDistance = cutoffDistance;
}

bool NoNeighborLocator::Neighbors(const double *center, int* neighbors, int& numNeighbors,
				double (*displacements)[DIMENSION],	double* displacementsSquared,
				int& spacesLeftInArrays, double cutoff)
{
	throw AttributeError("NoNeighborLocator","Neighbors");
	return false;
}

bool NoNeighborLocator::ReturnTrue(int &numNeighbors)
{
    if(cutoffDistance == 0.) numNeighbors = 0;
    else
		throw DMException("NoNeighborLocator","ReturnTrue",
						  "A function called Neighbors. Please set a neighborLocator");
	return true;
}

NeighborLocator *NoNeighborLocator::Copy() const
{
  NoNeighborLocator *newNeighborLocator = new NoNeighborLocator(cutoffDistance);
  
  return newNeighborLocator;
}


