
//     Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
//     Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

//     This library is free software; you can redistribute it and/or modify it
//     under the terms of the GNU Lesser General Public License as published by the
//     Free Software Foundation; either version 2.1 of the License, or (at your
//     option) any later version.

//     This library is distributed in the hope that it will be useful, but WITHOUT
//     ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//     FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
//     for more details.

//     You should have received a copy of the GNU Lesser General Public License
//     along with this library; if not, write to the Free Software Foundation,
//     Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
//     http://www.gnu.org/licenses/lgpl.html

//     Author Contact Information:
//     Jim Sethna      
//     sethna@lassp.cornell.edu  
//     117 Clark Hall
//     Cornell University
//     Ithaca, NY 14850


%{
#include "NeighborLocator.h"
#include "SimpleNeighborLocator.h"
#include "CellNeighborList.h"
#include "CellNeighborLocator.h"
#include "CompositeNeighborLocator.h"
#include "BondBreakerNeighborLocator.h"
%}
/* AtomIndex is assumed to be typedef int[2] */
%typemap(python, in) AtomIndex , int[2]   
{
	$1 = new AtomIndex; //titi
	if(PyList_Check($input))
	{
		int i;
		PyObject *o;
		for(i=0;i<2;++i)
		{
			o = PyList_GetItem($input,i);
			if(PyInt_Check(o)) $1[i] = PyInt_AsLong(o);
			else
			{
				PyErr_SetString(PyExc_TypeError,"list must contain int");
				return NULL;
			}
		}
	}
	else if (PyArray_Check($input))
	{
		PyArrayObject *o = (PyArrayObject *)($input);
		if((o->nd == 1) && (o->dimensions[0] == 2) && (o->descr->type_num == PyArray_INT) )
			for(int coord=0; coord<2; ++coord) *($1+coord) = *(int*)(o->data + coord*o->strides[0]);
		else
		{
			PyErr_SetString(PyExc_TypeError,"not an Array of correct size and/or type");
			return NULL;
		}
	}
	else
	{
		PyErr_SetString(PyExc_TypeError,"not a list nor an array");
		return NULL;
	}
}
%typemap(python,freearg) AtomIndex
{
	delete [] $1;
}

/* we first declare 2 "helper" functions that are called before and after usual call for Neighbors */
%{
bool AllocateForNeighborLocator(int length, int *&neighbors, double (*&disp)[DIMENSION], double *&dispSq)
{
	if(neighbors) delete [] neighbors;
	if(disp) delete [] disp;
	if(dispSq) delete [] dispSq;
	neighbors = new int[length];
	disp = new double [length][DIMENSION];
	dispSq = new double[length];
	return (neighbors && disp && dispSq);
}
PyObject *AgregateNeighborLocatorOutput(int *neighbors, int numNeighbors, double (*disp)[DIMENSION], double *dispSq)
{
	int dimensionOfNumPy[2];
	dimensionOfNumPy[0] = numNeighbors;
	dimensionOfNumPy[1] = DIMENSION;
	PyObject *pRes = PyList_New(0);
// append the number of neighbors
	PyObject *pNumNeighbors = PyInt_FromLong(numNeighbors);
	PyList_Append(pRes, pNumNeighbors);
	Py_DECREF(pNumNeighbors);
// append the array of the neighbors' index
	PyObject *pNeighbors = PyArray_FromDims(1,dimensionOfNumPy,'i');
	memcpy(((PyArrayObject *)pNeighbors)->data, neighbors, numNeighbors*sizeof(int));
	PyList_Append(pRes, pNeighbors);
	Py_DECREF(pNeighbors);
	delete [] neighbors;
// append the displacements array
	PyObject *pDisp = PyArray_FromDims(2,dimensionOfNumPy,'d');
	memcpy(((PyArrayObject *)pDisp)->data, disp, DIMENSION*numNeighbors*sizeof(double));
	PyList_Append(pRes, pDisp);
	Py_DECREF(pDisp);
	delete [] disp;
// append the displacements squared array
	PyObject *pDispSq = PyArray_FromDims(1,dimensionOfNumPy,'d');
	memcpy(((PyArrayObject *)pDispSq)->data, dispSq, numNeighbors*sizeof(double));
	PyList_Append(pRes, pDispSq);
	Py_DECREF(pDispSq);
	delete [] dispSq;
// return the PyList
	return pRes;
}
%}


/* BASE CLASS */
%nodefault NeighborLocator;
class NeighborLocator : public Serializable
{
public:
	virtual void UpdatePosition();
	virtual void SetBoundaryConditions(const BoundaryConditions *bc);
	virtual NeighborLocator *Copy() const;
	virtual void SetCutoffDistance(double cutoff);	
	virtual double GetCutoffDistance() const; 
	%extend
	{
		void SetPositionArray(const DMArray<double, DIMENSION> *positions){self->SetPositionArray(positions);}
		PyObject *NeighborsFromCenter(double center[DIMENSION], double cutoff=-1.)
		{
			int length = 200, numNeighbors;
			int *neighbors = 0;
			double (*disp)[DIMENSION] = 0;
			double *dispSq = 0;
			bool done = false;
			while(!done)
			{
				if (!AllocateForNeighborLocator(length, neighbors, disp, dispSq))
				{
					PyErr_SetString(PyExc_TypeError,"Unable to Allocate in NeighborLocator");
					return NULL;
				}
				int iTmp = length;
				done = self->Neighbors(center, neighbors, numNeighbors, disp, dispSq, iTmp, cutoff);
				if (!done) length *= 5;
			}
			return AgregateNeighborLocatorOutput(neighbors, numNeighbors, disp, dispSq);
		}
		PyObject *Neighbors(int atomNumber)
		{
			int length = 200, numNeighbors;
			int *neighbors = 0;
			double (*disp)[DIMENSION] = 0;
			double *dispSq = 0;
			bool done = false;
			while(!done)
			{
				if (!AllocateForNeighborLocator(length, neighbors, disp, dispSq))
				{
					PyErr_SetString(PyExc_TypeError,"Unable to Allocate in NeighborLocator");
					return NULL;
				}
				int iTmp = length;
				done = self->Neighbors(atomNumber, neighbors, numNeighbors, disp, dispSq, iTmp);
				if (!done) length *= 5;
			}
			return AgregateNeighborLocatorOutput(neighbors, numNeighbors, disp, dispSq);
		}
		PyObject *HalfNeighbors(int atomNumber)
		{
			int length = 200, numNeighbors;
			int *neighbors = 0;
			double (*disp)[DIMENSION] = 0;
			double *dispSq = 0;
			bool done = false;
			while(!done)
			{
				if (!AllocateForNeighborLocator(length, neighbors, disp, dispSq))
				{
					PyErr_SetString(PyExc_TypeError,"Unable to Allocate in NeighborLocator");
					return NULL;
				}
				int iTmp = length;
				done = self->HalfNeighbors(atomNumber, neighbors, numNeighbors, disp, dispSq, iTmp);
				if (!done) length *= 5;
			}
			return AgregateNeighborLocatorOutput(neighbors, numNeighbors, disp, dispSq);
		}
		PyObject *TypedNeighbors(int atomNumber, int type1, int type2)
		{
			int length = 200, numNeighbors;
			int *neighbors = 0;
			double (*disp)[DIMENSION] = 0;
			double *dispSq = 0;
			bool done = false;
			while(!done)
			{
				if (!AllocateForNeighborLocator(length, neighbors, disp, dispSq))
				{
					PyErr_SetString(PyExc_TypeError,"Unable to Allocate in NeighborLocator");
					return NULL;
				}
				int iTmp = length;
				
				done = self->Neighbors(atomNumber, type1, type2, neighbors, numNeighbors, disp, dispSq, iTmp);
				if (!done) length *= 5;
			}
			return AgregateNeighborLocatorOutput(neighbors, numNeighbors, disp, dispSq);
		}
		PyObject *NeighborsPosition(PyObject *positionObj, double cutoffDistance = -1)
		{
			if(!PyArray_Check(positionObj))
			{
			PyErr_SetString(PyExc_TypeError,"Must pass a NumPy array");
			return NULL;
			}
			PyArrayObject *array = (PyArrayObject *)(positionObj);
			if(array->nd != 1 || array->dimensions[0] != DIMENSION || array->descr->type_num != PyArray_DOUBLE)
			{
			PyErr_SetString(PyExc_TypeError,"Array must be 1 dimensional, of length DIMENSION, type double");
			return NULL;
			}
			double *center = (double *)array->data;
			int length = 200, numNeighbors;
			int *neighbors = 0;
			double (*disp)[DIMENSION] = 0;
			double *dispSq = 0;
			bool done = false;
			while(!done)
			{
				if (!AllocateForNeighborLocator(length, neighbors, disp, dispSq))
				{
					PyErr_SetString(PyExc_TypeError,"Unable to Allocate in NeighborLocator");
					return NULL;
				}
				int iTmp = length;
				done = self->Neighbors(center, neighbors, numNeighbors, disp, dispSq, iTmp, cutoffDistance);
				if (!done) length *= 5;
			}
			return AgregateNeighborLocatorOutput(neighbors, numNeighbors, disp, dispSq);
		}
	}
};

class NoNeighborLocator: public NeighborLocator
{
public:
	NoNeighborLocator(double cutoff = 1.);
};
%include "SimpleNeighborLocator.h"
%include "CellNeighborLocator.h"
%include "CellNeighborList.h"
%include "CompositeNeighborLocator.h"
%include "BondBreakerNeighborLocator.h"
