// LocalForceObserver.cpp: implementation of the LocalForceObserver class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
#include "DigitalMaterial.h"
#include "LocalForceObserver.h"
#include "Potential.h"
#include "ListOfAtoms.h"
#include "DMArray.h"
#include "DMLastHeader.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

LocalForceObserver::LocalForceObserver(Potential *potential, double forceCutoff, std::string fileName, int notifyLevel) : ListOfAtomsObserver(notifyLevel)
{
	this->potential = potential;
	this->forceCutoff = forceCutoff;
	fileObject = new ofstream(fileName.c_str());
	nAtoms = 0;
	forces = 0;
}

LocalForceObserver::~LocalForceObserver()
{
	delete fileObject;
	if(forces)
		delete forces;
}

void LocalForceObserver::Allocate(ListOfAtoms *atoms)
{
	if(forces)
		delete forces;
	forces = new DMArray<double, DIMENSION>(atoms->GetShape());
}

void LocalForceObserver::Update(ListOfAtoms *atoms)
{
	if (!forces || nAtoms != atoms->GetNumber())
		Allocate(atoms);
	
	potential->CalculateForces(atoms,forces);
	
	int coord, maxAtom[DIMENSION], atomIdx;
	double maxForce[DIMENSION];
	for( coord = 0; coord < DIMENSION; ++coord )
	{
		maxForce[coord] = 0.;
		maxAtom[coord] = 0;
	}
	int nLargeForces = 0;
	double fCOSq = forceCutoff * forceCutoff;

	DMArrayIterator<double, DIMENSION> forceIter(forces);
	for( forceIter.Init(),atomIdx = 0; !forceIter.IsDone(); forceIter.Next(),++atomIdx )
	{
		double *currentForce = forceIter.Current();
		double forceMagSq = 0.;
		for( coord = 0; coord < DIMENSION; ++coord )
		{
			if(fabs(currentForce[coord]) > maxForce[coord]) 
			{
				maxForce[coord] = currentForce[coord];
				maxAtom[coord] = atomIdx;
			}
			forceMagSq += currentForce[coord] * currentForce[coord];
			if(forceMagSq > fCOSq)
				++nLargeForces;
		}

	}
	const DMArray<double, DIMENSION> *positions = atoms->GetCartesianPositions();
	cout << "For each coordinate, atom with maximum that component of force, with position and full force:" << endl;
	int oldPrec = cout.precision();
	cout.precision(5);
	for( coord = 0; coord < DIMENSION; ++coord )
	{		
		int coord2;
		cout << maxAtom[coord] << " (";
		for ( coord2 = 0; coord2 < DIMENSION; ++coord2 )
			cout << (*positions)[maxAtom[coord]][coord2] << " ";
		cout << "), (";
		for ( coord2 = 0; coord2 < DIMENSION; ++coord2 )
			cout << (*forces)[maxAtom[coord]][coord2] << " ";
		cout << ")" << endl;
	}
	cout << "Number of large forces " << nLargeForces << endl << endl;
	cout.precision(oldPrec);
}

