// CuRbPotential.cpp: implementation of the CuRbPotential class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
#include "DigitalMaterial.h"
#include <iostream>
#include "CuRbPotential.h"
#include "EMTPotential.h"
#include "ListOfAtoms.h"
#include "DMArray.h"
#include "IndexError.h"
#include "NotEnoughMemory.h"
#include "NeighborLocator.h"
#include "SerialBuilder.h"
#include "BoundaryConditions.h"

#include "DMLastHeader.h"

using namespace std;

#define LEN_NBR_ARRAYS 1500


static SerialBuilder<CuRbPotential> *_CKcreator = 
    new SerialBuilder<CuRbPotential>("CuRbPotential");

const string CuRbPotential::className = string("CuRbPotential");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


/*!

The parameters are set in the constructor. Units are eV and Bohr


*/

CuRbPotential::CuRbPotential(Potential * EMTptr) : CuRb1_A1(1444.47),
												   CuRb1_B1(3.38),
												   CuRb1_A2(37495.36),
												   CuRb1_B2(6.30),
												   CuRb2_A (4073.53),
												   CuRb2_B(3.58961),
												   CuRb_alpha(0.05),
												   CuNa_A1(1335.735),
												   CuNa_B1(3.623051),
												   CuNa_A2(55159.81),
												   CuNa_B2(8.184134)
{
	double bohrRadius = 0.5291772;
	bool usingBohr = true;
	if(usingBohr)
	{
		CuRb1_B1 *= bohrRadius;
		CuRb1_B2 *= bohrRadius;
		CuRb2_B *= bohrRadius;
		CuNa_B1 *= bohrRadius;
		CuNa_B2 *= bohrRadius;		
	}

	// default is to use Chad's potential (1)
	CuRb_switch = 1;
	surfaceInteractions = true;

	this->EMTptr = EMTptr;

	double cutoff = EMTptr->GetCutoffDistance();
	cout << "Cu-Rb cutoff distance: " << cutoff << endl;

	SetCutoffDistance(cutoff);
	nPairs = 0;


	displacementsSq = 0;
	displacements = 0;
	neighbors = 0;
	AllocateForNeighborLocator(50); // a rough estimate of the expected number of neighbors
}

CuRbPotential::~CuRbPotential()
{
	if(neighbors) 
	{	
		delete [] neighbors;
		delete [] displacements;
		delete [] displacementsSq;
	}
}




void CuRbPotential::AllocateForNeighborLocator(int nPairs)
{
	// nPairs is playing the role of EMTPotential's arrayLim here
	this->nPairs = nPairs;
	
	if(neighbors) 
	{	
		delete [] neighbors;
		delete [] displacements;
		delete [] displacementsSq;
	}

	if(nPairs > 0)
	{
		neighbors       = new (std::nothrow) int[nPairs];
		displacements   = new (std::nothrow) double[nPairs][DIMENSION];
		displacementsSq = new (std::nothrow) double[nPairs];
		
		ASSERT(neighbors);
		ASSERT(displacements);
		ASSERT(displacementsSq);
		
		if(0 == neighbors)
			throw NotEnoughMemory("CuRbPotential",
						"AllocateForNeighborLocator",
						SprintString("Problem allocating %d doubles", nPairs));
		if(0 == displacements)
			throw NotEnoughMemory("CuRbPotential",
						"AllocateForNeighborLocator",
						SprintString("Problem allocating %d doubles", nPairs*DIMENSION));
		if(0 == displacementsSq)
			throw NotEnoughMemory("CuRbPotential",
						"AllocateForNeighborLocator",
						SprintString("Problem allocating %d integers", nPairs));
	}
	else
	{
		neighbors       = 0;
		displacements   = 0;
		displacementsSq = 0;
	}
}




void CuRbPotential::CalculateForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces)
{
	int coord;
	double Rcut = GetCutoffDistance();
	double R0 = Rcut * CuRb_alpha;
	forces->Zero();

	if(atoms->NumberOfBranches() != 2)
		throw DMException("CuRbPotential","CalculateForces","LOA has wrong number of branches");

	if(surfaceInteractions)
		EMTptr->CalculateForces(atoms->GetBranch(0), forces->GetBranch(0));
	
	if(CuRb_switch == 0)
		return;

	// now add the contributions from the Cu-Rb interaction
	if(atoms->GetBranch(1)->GetNumber() != 1)
		throw DMException("CuRbPotential","CalculateForces","There must be exactly 1 atom in branch 1");

	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	int nAtoms = atoms->GetNumber();
	int RbIdx = nAtoms - 1;

	int numberOfNeighbors;
	int spacesLeftInArrays = nPairs;

	bool arrayNotTooFull = neighborLocator->Neighbors(RbIdx,neighbors,
		numberOfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
	while(!arrayNotTooFull)
	{
		AllocateForNeighborLocator(2*nPairs);
		spacesLeftInArrays = nPairs; // nPairs has been doubled
		arrayNotTooFull = neighborLocator->Neighbors(RbIdx,neighbors,
			numberOfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
	}

	
	double *forceI, *forceJ;
	forceI = (*forces)[RbIdx];
	for(int jx = 0; jx < numberOfNeighbors; ++jx)
	{
		int j = neighbors[jx];
		forceJ = (*forces)[j];
		double distance = sqrt(displacementsSq[jx]);
		double distMinusRcut = distance - Rcut;
		double expFactorCut = exp(R0 / distMinusRcut);
		double forceR;
		if(CuRb_switch == 1)
		{
			double expFactor1 =  CuRb1_A1 * exp(-CuRb1_B1 * distance);
			double expFactor2 =  CuRb1_A2 * exp(-CuRb1_B2 * distance);
			
			forceR = (CuRb1_B1 * expFactor1  + CuRb1_B2 * expFactor2 + (expFactor1 + expFactor2) * R0 /(distMinusRcut*distMinusRcut)) * expFactorCut / distance;
		}
		else if(CuRb_switch == 2)
		{
			double expFactor = CuRb2_A * exp(-CuRb2_B * distance);

			forceR = (CuRb2_B + R0 /(distMinusRcut*distMinusRcut)) * expFactor * expFactorCut / distance;
		}
		else if(CuRb_switch == 3)
		{
			double expFactor1 =  CuNa_A1 * exp(-CuNa_B1 * distance);
			double expFactor2 =  CuNa_A2 * exp(-CuNa_B2 * distance);
			
			forceR = (CuNa_B1 * expFactor1  + CuNa_B2 * expFactor2 + (expFactor1 + expFactor2) * R0 /(distMinusRcut*distMinusRcut)) * expFactorCut / distance;
		}

		for (coord=0; coord<DIMENSION; coord++)
		{
			forceI[coord] += -displacements[jx][coord]*forceR;
			forceJ[coord] += +displacements[jx][coord]*forceR;
		}
		
	
	}
	

}


double CuRbPotential::CalculateEnergy(ListOfAtoms *atoms)
{
	double Rcut = GetCutoffDistance();
	double RcutInv2 = 1./(Rcut * Rcut);
	double R0 = Rcut * CuRb_alpha;

	if(atoms->NumberOfBranches() != 2)
		throw DMException("CuRbPotential","CalculateForces","LOA has wrong number of branches");

	double energy = 0.;
	if(surfaceInteractions)
		energy = EMTptr->CalculateEnergy(atoms->GetBranch(0));

	if(CuRb_switch == 0)
		return energy;

	// now add the contributions from the Cu-Rb interaction
	if(atoms->GetBranch(1)->GetNumber() != 1)
		throw DMException("CuRbPotential","CalculateForces","There must be exactly 1 atom in branch 1");

	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	int nAtoms = atoms->GetNumber();
	int RbIdx = nAtoms - 1;

	int numberOfNeighbors;
	int spacesLeftInArrays = nPairs;

	bool arrayNotTooFull = neighborLocator->Neighbors(RbIdx,neighbors,
		numberOfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
	while(!arrayNotTooFull)
	{
		AllocateForNeighborLocator(2*nPairs);
		spacesLeftInArrays = nPairs; // nPairs has been doubled
		arrayNotTooFull = neighborLocator->Neighbors(RbIdx,neighbors,
			numberOfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
	}

	
	for(int jx = 0; jx < numberOfNeighbors; ++jx)
	{
		int j = neighbors[jx];
		double distance = sqrt(displacementsSq[jx]);
		double distMinusRcut = distance - Rcut;
		double expFactorCut = exp(R0 / distMinusRcut);
		double pairEnergy;
		if(CuRb_switch == 1)
		{
			double expFactor1 =  CuRb1_A1 * exp(-CuRb1_B1 * distance);
			double expFactor2 =  CuRb1_A2 * exp(-CuRb1_B2 * distance);
			pairEnergy = (expFactor1 + expFactor2) * expFactorCut;
		}
		else if(CuRb_switch == 2)
		{
			double expFactor = CuRb2_A * exp(-CuRb2_B * distance);
			pairEnergy = expFactor * expFactorCut;
		}
		else if(CuRb_switch == 3)
		{
			double expFactor1 =  CuNa_A1 * exp(-CuNa_B1 * distance);
			double expFactor2 =  CuNa_A2 * exp(-CuNa_B2 * distance);
			pairEnergy = (expFactor1 + expFactor2) * expFactorCut;
		}

		energy += pairEnergy;
	}
	
	return energy;
}



double CuRbPotential::GetCuRbInverseInteraction(double energy) const
{
	if(CuRb_switch == 2)
		return SimpleInverse(energy, CuRb2_A, CuRb2_B);
	else
	{
		double A1 = CuRb1_A1;
		double A2 = CuRb1_A2;
		double B1 = CuRb1_B1;
		double B2 = CuRb1_B2;
		if(CuRb_switch == 3)
		{
			A1 = CuNa_A1;
			A2 = CuNa_A2;
			B1 = CuNa_B1;
			B2 = CuNa_B2;
		}

		double firstIt = SimpleInverse(energy, A1, B1);
		double firstCorr = GetCorrFactor(firstIt, A1, A2, B1, B2);
		double secondIt = SimpleInverse(energy/firstCorr, A1, B1);
		double secondCorr = GetCorrFactor(secondIt, A1, A2, B1, B2);
		return SimpleInverse(energy/secondCorr, A1, B1);
	}
}
