// CuRbPotential.h: interface for the CuRbPotential class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef CUKRPOTENTIAL_H
#define CUKRPOTENTIAL_H

#include "DMFirstHeader.h"

#include "DigitalMaterial.h"
#include "Potential.h"

//! Potential for Kr-Cu interactions, single decaying exponential

class CuRbPotential : public Potential
{
public:
	CuRbPotential(Potential *EMTptr = 0);
	~CuRbPotential();
	
	void SetEMTptr(Potential *EMTptr) { this->EMTptr = EMTptr; }
	
	void CalculateForces (ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);
	//double CalculateEnergyAndForces (ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);
	double CalculateEnergy(ListOfAtoms *atoms);

	double GetLengthScale() const {return EMTptr->GetLengthScale();}

	double GetCuRbInverseInteraction(double energy) const;

	void SetCuRbInteraction(int CuRb_switch) { 
		if(CuRb_switch < 0 || CuRb_switch > 6) 
			throw DMException("CuRbPotential","SetCuRbInteraction","Must be in range 0--6");
		this->CuRb_switch = CuRb_switch;
	}
	void SetSurfaceInteractions(int si) {if(si) surfaceInteractions = true; else surfaceInteractions = false;}

private:
	void AllocateForNeighborLocator(int nPairs);
	double SimpleInverse(double energy, double A, double B) const {return (-1./B) * log(energy/A);}
	double GetCorrFactor(double r, double A1, double A2, double B1, double B2) const {return (1. + (A2/A1)*exp(-(B2 - B1)*r));}
	//! a rough (over-)estimation of the number of neighbors
	int         nPairs;
	double      *displacementsSq;
	double      (*displacements)[DIMENSION];
	int         *neighbors;

	// Chad's potential
	double CuRb1_A1;
	double CuRb1_B1;
	double CuRb1_A2;
	double CuRb1_B2;
	
	// NB's potential
	double CuRb2_A;
	double CuRb2_B;

	// Chad's Na-Cu potential
	double CuNa_A1;
	double CuNa_B1;
	double CuNa_A2;
	double CuNa_B2;

	// chooses which form of Cu-Rb interaction we want
	// 0 means none; 1 means Chad's; 2 means NB's; 3 means Cha'ds Cu-Na one
	int CuRb_switch;
	// whether Cu-Cu interactions are turned on or not
	bool surfaceInteractions;


	double CuRb_alpha;
	static const std::string className;

	//! for Cu
	Potential *EMTptr;
};


#endif // !defined(CUKRPOTENTIAL_H
