// EMT.h  --  Implements the EMT potential for multiple elements.
// When done, this will make EMTPotential.{h,cpp} obsolete.
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
#ifndef _EMT_H
#define _EMT_H

#include "DigitalMaterial.h"
#include "DMArray.h"
#include "Potential.h"
#include "EMTParameterProvider.h"
#include "BoundaryConditions.h"
#include "TinyMatrix.h"
// #include "Tiny3DArray.h"

#if DIMENSION == 3  // Only works in 3D

class EMT : public Potential
{
public:
	EMT(ListOfAtoms *atoms, EMTParameterProvider *prov = 0);
	virtual ~EMT();
	virtual double CalculateEnergy(ListOfAtoms *atoms);
	virtual void CalculateForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);
	virtual double CalculateEnergyAndForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);
	virtual double GetCutoffDistance() {return cutoffdistance;}

	DMArray<double, DIMENSION> *AllocateForces(); // Helper function returning an array suitable for
	                                              // the forces
	const DMArray<double> *DebugSigma1(int n=0) {return sigma1[n];}
	const DMArray<double> *DebugSigma2(int n=0) {return sigma2[n];}
	const DMArray<double> *GetEc() {return Ec;}
	const DMArray<double> *GetEas() {return Eas;}
	const DMArray<double> *GetRadius() {return radius;}
	const DMArray<double> *GetdEds() {return dEds;}
	
	const std::string &GetType() { return className; }
protected:
	virtual void Allocate(bool first = false);
	virtual void DeAllocate();
	virtual void InitParameters();
	virtual void CheckListOfAtoms(ListOfAtoms *atoms);
	virtual void CalculateSigmas();
	virtual double CalculateEnergyAfterSigmas();
	virtual void CalculateForcesAfterEnergy(DMArray<double, DIMENSION> *forces);
private:  // Methods.
	
	// sigma_batch does the hard work in CalculateSigmas().  Not virtual, do not reimplement without 
	// reimplementing CalculateSigmas().
	void sigma_batch(const int n, const emt_parameters selfparam, const emt_parameters otherparam,
					 const int *self, const int *other, const double (*rnb)[3], const double *dist2,
					 DMArray<double> *sigma1self, DMArray<double> *sigma1other,
					 DMArray<double> *sigma2self, DMArray<double> *sigma2other);
	// force_batch does the hard work in CalculateForcesAfterEnergy().  Not virtual, do not
	// reimplement without reimplementing CalculateForcesAfterEnergy().
	void force_batch(int n, const emt_parameters selfparam, const emt_parameters otherparam,
					 const int *self, const int *other, const double (*rnb)[3],
					 const double *dist2, const double *dEdss, const double *dEdso,
					 DMArray<double, DIMENSION> *forceself, DMArray<double, DIMENSION> *forceother);

private:  // Data
	ListOfAtoms *atoms;             // The atoms we are working on
	EMTParameterProvider *provider; // The source of the EMT parameters
	bool ownProvider;               // May we delete the provider?
	std::vector<const emt_parameters *> parameters;
	const TinyDoubleMatrix *chi;
	int nelements;                  // The number of different elements in the simulation
	int nleaves;                    // The number of leaves in the list of atoms
	double cutoffdistance;          // Cutoff parameters (from EMTParameterProvider).
	double cutoffslope;
	int maxleaflen;                 // Lenght of the longest leaf.
	
	/* Temporary data for the atoms */
	// Each atom has two sigmas for each possible neighboring element.  So these are arrays
	// (nelements long).
	std::vector<DMArray<double>*> sigma1;
	std::vector<DMArray<double>*> sigma2;
	
	// Each atom has a single Ec, Eas, radius etc.  These are pointers to arrays that are allocated later.
	DMArray<double> *Ec;      
	DMArray<double> *Eas;
	DMArray<double> *radius;
	DMArray<double> *dEds;

	/* The buffers for batch processing */
	static const int BUFLEN;  // The batch--buffer size.  Avoid powers of 2
//	TinyMatrix<int[BUFLEN]> me;              // The index of the first atom
//	TinyMatrix<int[BUFLEN]> other;           // The index of the second atom
//	TinyMatrix<double[BUFLEN][3]> rnb;       // The relative position
//	TinyMatrix<double[BUFLEN]> dist2;        // The square of the distance
	static const std::string className;
}; 



#endif // DIMENSION == 3
#endif // ! _EMT_H
