// EMTParameterProvider.h  --  Abstract class for objects providing EMT parameters.

/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef _EMTPARAMETERPROVIDER_H
#define _EMTPARAMETERPROVIDER_H

#include "DMFirstHeader.h"
#include "TinyMatrix.h"
#include <string>

// The stucture used to for store EMT parameters.
struct emt_parameters
{
	double e0, seq, neq, V0, eta2, kappa, lambda, mass, invmass;
	double gamma1, gamma2;
	// double pairA, pairD;
	double lengthscale;  // Not an EMT parameter.  1/sqrt(2) of the lattice constant.
	int Z;
	std::string name;
	int index;   // An index into various arrays.  Counts the elements in the simulation.
};

// A geometric constant.
// static const double Beta = 1.80939979; // ((16*pi/3)^(1/3))/sqrt(2)
static const double Beta = 1.809; // Preserve same rounding as in ARTwork.

// An EMTParameterProvider provides the EMT parameters to the potential.  The GetParameters() method 
// returns the parameters for a given element.  If it is called multiple times with the same element 
// it must return a pointer to the same emt_parameters struct.  The Provider owns the parameters,
// they are deleted when the provider is deleted.  After getting _all_ the elements, CalcGammaEtc()
// should be called to calculate the cutoff radius, the gammas and other quantities that cannot be
// calculated before it is known which elements are in the simulation.  The gammas and the cutoff
// are determined by the Provider since changing the way they are determined corresponds to changing 
// the EMT potential.
class EMTParameterProvider 
{
public:
	// virtual ~EMTParameterProvider() = 0;
	virtual const emt_parameters *GetParameters(const std::string *element) = 0;
	virtual void CalcGammaEtc() = 0;
	virtual double GetCutoffDistance() = 0; // Can it be made element-dependent?
	virtual double GetCutoffSlope() = 0;
	virtual double GetLengthScale() = 0;    // The potential delegates this to the provider.
	virtual const TinyDoubleMatrix *GetChi() = 0;
	virtual int GetNumberOfElements() = 0;
};

#endif // ! _EMTPARAMETERPROVIDER_H
