// EMTPotential.cpp: implementation of the EMTPotential class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/


// Jakob Schiotz' FAST code, converted to the Digital Material Format
// We (Sethna, Cheong, and Bailey) have tried to leave the code in its 
// original format as much as we could

#include "DMFirstHeader.h"
#include <iostream>
#include <fstream>
#include <assert.h>

#include "DigitalMaterial.h"

#include "DMArrayIterator.h"
#include "EMTPotential.h"
#include "NeighborLocator.h"
#include "ListOfAtoms.h"
#include "LeafToLeafIterator.h"
#include "DMException.h"
#include "vectools.h"
#include "mass.h"

#include "SerialBuilder.h"

#if DIMENSION == 3

#include "DMLastHeader.h"
using namespace std;

static SerialBuilder<EMTPotential> *_EMTcreator = 
    new SerialBuilder<EMTPotential>("EMTPotential");

const string EMTPotential::className = string("EMTPotential");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

EMTPotential::EMTPotential(string element)
// XXX Constructor for single-element Cu simulation
{
    this->element = element;
    
    int i;	// Stupid Microsoft incompatibility
    
    emt = new emt_struct(element.c_str());
    cut.setcut(emt, 1);
    emt->CalculateGamma(cut);
    
    SetCutoffDistance(cut.cutdist);
    
    sigma1 = 0; 
    sigma2 = 0;
    radius = 0;
    Ec	   = 0;
    Eas    = 0;
    dE0ds  = 0;
    
    // Global arrays for local temporary data.  A large block of memory
    // is allocated, it is then divided into parts (dTmp[n]) each holding
    // nAtoms numbers. Since the data is contiguous it can be used for arrays
    // that are any multiple of the number of atoms, eg. if a temporary
    // 3-vector is needed for each atom.  The space can also be used for
    // single precision numbers (fTmp[n]) or integers (iTmp[n]).
    //
    // BEWARE: The double, float and integer arrays occupy the SAME SPACE in
    // memory.  Do not use them simultaneously.
    //
    
    for (i = 0; i < MAXTMP; i++)
    {
        dTmp[i] = NULL;
    }
    
    for (i = 0; i < 2*MAXTMP; i++)
    {
        fTmp[i] = NULL;
        iTmp[i] = NULL;
    }
}

EMTPotential::~EMTPotential()
{
    delete emt;
    
    if(sigma1) delete sigma1;
    if(sigma2) delete sigma2;
    if(radius) delete radius;
    if(Ec)     delete Ec;
    if(Eas)    delete Eas;
    if(dE0ds)  delete dE0ds;
    
    if (dTmp[0]) delete [] dTmp[0];
    
}

void EMTPotential::Allocate(const ListOfAtoms *atoms)
{
    this->atoms = atoms;
    this->nAtoms = atoms->GetNumber();
    
    if(sigma1) delete sigma1;
    if(sigma2) delete sigma2;
    if(radius) delete radius;
    if(Ec)     delete Ec;
    if(Eas)    delete Eas;
    if(dE0ds)  delete dE0ds;
    
    sigma1 = new DMArray<double>(atoms->GetCartesianPositions());
    sigma2 = new DMArray<double>(atoms->GetCartesianPositions());
    radius = new DMArray<double>(atoms->GetCartesianPositions());
    Ec     = new DMArray<double>(atoms->GetCartesianPositions());
    Eas    = new DMArray<double>(atoms->GetCartesianPositions());
    dE0ds  = new DMArray<double>(atoms->GetCartesianPositions());
    
    sigma1->GetSynchronizationKit()->Init();
    sigma2->GetSynchronizationKit()->Init();
    radius->GetSynchronizationKit()->Init();
    Ec    ->GetSynchronizationKit()->Init();
    Eas	  ->GetSynchronizationKit()->Init();
    dE0ds ->GetSynchronizationKit()->Init();
    dE0dsExtended = dE0ds ->GetSynchronizationKit()->GetExtended();
    
    int ndouble = 7;
    int nfloat = ndouble * 2;
    
    // Allocate temp data as a contiguous block 
    
    assert(ndouble <= MAXTMP);
    
    if (dTmp[0]) delete [] dTmp[0];

    dTmp[0] = new double[ndouble * nAtoms];
    //    TESTPTR(dTmp[0]);
    fTmp[0] = (float *) dTmp[0];
    iTmp[0] = (int *) dTmp[0];
    
    int i;	// Stupid Microsoft incompatibility
    
    for (i = 1; i < ndouble; i++)
        dTmp[i] = dTmp[i-1] + nAtoms;
    for (i = ndouble; i < MAXTMP; i++)
        dTmp[i] = NULL;
    for (i = 1; i < nfloat; i++)
    {
        fTmp[i] = fTmp[i-1] + nAtoms;
        iTmp[i] = iTmp[i-1] + nAtoms;
    }
    for (i = nfloat; i < 2*MAXTMP; i++)
    {
        fTmp[i] = NULL;
        iTmp[i] = NULL;
    }
}

void EMTPotential::cut_struct::setcut(emt_struct *e, int n) 
{
    // XXX These were initialized in Jakob's init.c; we don't know whether 
    // putting them into this constructor makes sense
    
    shells[0] = 3;
    shells[1] = 4;
    
    // This was in setcut() in Jakob's gamma.c
    
    double r = 0.0;
    for (int i=0; i < n; i++)
        if (e[i].seq > r)
            r = e[i].seq;
        
        cutdist = r * BETA * 0.5 * (sqrt((double) shells[0]) +
            sqrt((double) shells[0] + 1.0));
        cutlist = r * BETA * 0.5 * (sqrt((double) shells[1]) +
            sqrt((double) shells[1] + 1.0));
        
        r = cutdist * 2.0 * sqrt((double) shells[0] + 1.0) /
            (sqrt((double) shells[0]) + sqrt((double) shells[0] + 1.0));
        cutslope = log(9999.0)/(r - cutdist);
}

EMTPotential::emt_struct::emt_struct(const char *element) 
{
    // This constructor gets all the parameters from the file, but DOES NOT 
    // calculate gammas.
    // You need to construct the cut_struct, and then use calcgamma, to get 
    // things set up right
    
    // The Danish code for reading in EMT parameters seems very general and 
    // complicated
    // We're making a much less general reader which works for their current 
    // emt.def file
    // Our reader will break, for example, if the order for the parameters in 
    // the element file
    // changes.
    // XXX We'd like to preserve the nice generality, but not right now
    
    double E0, S0, n0;
    if(!strcmp(element,"Al")){
        E0=-3.280; S0=3.000; V0=1.493; Z=13; eta2=1.240; kappa=2.000; lambda=1.169; mass=26.98 ; n0=0.007  ; 
    } 
    else if(!strcmp(element,"Cu")){
        E0=-3.510; S0=2.67 ; V0=2.476; Z=29; eta2=1.652; kappa=2.74 ; lambda=1.906; mass=63.54 ; n0=0.0091 ;
    }
    else if(!strcmp(element,"Ag")){ 
        E0=-2.96 ; S0=3.01 ; V0=2.132; Z=47; eta2=1.652; kappa=2.790; lambda=1.892; mass=107.87; n0=0.00547;
    }
    else if(!strcmp(element,"Au")){ 
        E0=-3.80 ; S0=3.00 ; V0=2.321; Z=79; eta2=1.674; kappa=2.873; lambda=2.182; mass=196.97; n0=0.00703; 
    }
    else if(!strcmp(element,"Ni")){
        E0=-4.44 ; S0=2.60 ; V0=3.673; Z=28; eta2=1.669; kappa=2.757; lambda=1.948; mass=58.71 ; n0=0.0103 ;
    }
    else if(!strcmp(element,"Pd")){
        E0=-3.90 ; S0=2.87 ; V0=2.773; Z=46; eta2=1.818; kappa=3.107; lambda=2.155; mass=106.4 ; n0=0.00688;
    }
    else if(!strcmp(element,"Pt")){
        E0=-5.85 ; S0=2.90 ; V0=4.067; Z=78; eta2=1.812; kappa=3.145; lambda=2.192; mass=195.09; n0=0.00802;
    }
    else if(!strcmp(element,"H")){
        E0=-2.15 ; S0=1.9  ; V0=0.   ; Z=1 ; eta2=3.2  ; kappa=4.1  ; lambda=2.096; mass=1.    ; n0=0.007;
    }
    else throw DMException("EMTPotential::emt_struct","emt_struct",
        "This element isn't defined in EMT");
    
    e0 = E0;
    seq = S0;
    invmass = 1.0/mass;
    neq = n0;
}

void EMTPotential::emt_struct::CalculateGamma(cut_struct cut)
{
    // This is from calcgamma in Jakob's gamma.c
    // Call after cut_struct has been initialized
    
    // It calculates gammas for only one element at a time 
    // If you have an array of emt_structs; you need to loop over them
    
    int is;
    static int shellpop[5] = {12, 6, 24, 12, 24};
    double w, d;
    
    assert(cut.cutdist != 0.0);
    
    gamma1 = gamma2 = 0.0;
    for (is = 0; is < cut.shells[1] && is < 5; is++) 
    {
        d = sqrt((double) (is + 1)) * BETA * seq;
        w = 1. / (1. + exp(cut.cutslope * (d - cut.cutdist)));
        gamma1 += w * shellpop[is] * exp(-d * eta2);
        gamma2 += w * shellpop[is] * exp(-d * kappa / BETA);
    }
    gamma1 /= shellpop[0] * exp(-BETA * seq * eta2);
    gamma2 /= shellpop[0] * exp(-seq * kappa);
}

// Calculate the energies.  
double EMTPotential::CalculateEnergy(ListOfAtoms *atoms)
{
    if (!(atoms == this->atoms && atoms->GetShape()== sigma1->GetShape()))
        Allocate(atoms);
    
    //	CalculateSigmas() MUST have been called after the
    //	last time the positions were updated
    
    CalculateSigmas(atoms);
    return CalculateEnergyAfterSigmas(atoms);
}

double EMTPotential::CalculateEnergyAfterSigmas(ListOfAtoms *atoms)
{
    int i;
    
    // Calculate combinations of EMT parameters
    double emkappaseq = emt->kappa * emt->seq;
    double eminv12gamma1 = 1.0 / (12.0 * emt->gamma1);
    double emseq = emt->seq;
    double emkappa = emt->kappa;
    double emlambda = emt->lambda;
    double eme0lambdalambda = emt->e0 * emt->lambda * emt->lambda;
    double eme0lambdalambdaseq = emt->e0 * emt->lambda * emt->lambda * emt->seq;
    double em6V0kappa = 6.0 * emt->V0 * emt->kappa;
    double eminvbetaeta2 = 1.0 / (emt->eta2 * BETA);
    double emlambdaseq = emt->lambda * emt->seq;
    double eme0lambda = emt->e0 * emt->lambda;
    double em6V0 = 6.0 * emt->V0;
    double emneghalfV0overgamma2 = -0.5 * emt->V0 / emt->gamma2;
    
    double *tmp1, *tmp2, *ex1, *ex2;

    // Allocate temporary arrays 
    tmp1 = dTmp[0];
    tmp2 = dTmp[1];
    ex1 = dTmp[2];
    ex2 = dTmp[3];
    ASSERT(tmp1 && tmp2 && ex1 && ex2);
    
    double totEc  = 0.0;
    double totEas = 0.0;
    LeafToLeafIterator<DMArray<double> >sigma1It(sigma1);
    LeafToLeafIterator<DMArray<double> >sigma2It(sigma2);
    LeafToLeafIterator<DMArray<double> >radiusIt(radius);
    LeafToLeafIterator<DMArray<double> >EcIt    (Ec    );
    LeafToLeafIterator<DMArray<double> >EasIt   (Eas   );
    for(sigma1It.Init(), sigma2It.Init(), radiusIt.Init(), EcIt.Init(), EasIt.Init();
    !sigma1It.Done();
    sigma1It.Next(), sigma2It.Next(), radiusIt.Next(), EcIt.Next(), EasIt.Next())
    {
        const int size = sigma1It.Current()->Size();
        if (size == 0) continue;
        const double *pSig1 = &(*sigma1It.Current())[0];
        const double *pSig2 = &(*sigma2It.Current())[0];
        double *pRadi = &(*radiusIt.Current())[0];
        double *pEc   = &(*EcIt.Current())[0];
        double *pEas  = &(*EasIt.Current())[0];
		// radius = emseq - eminvbetaeta2 * log(sigma1[is] * eminv12gamma1);
        vec_mul_scal_add_scal(tmp1, pSig1, eminv12gamma1, 0.0, size);
        vlog(tmp2, tmp1, &size);
        vec_mul_scal_add_scal(pRadi, tmp2, -eminvbetaeta2, emseq, size);

        // ex1 = exp(-emkappa * radius + emkappaseq); 
        vec_mul_scal_add_scal(tmp1, pRadi, -emkappa, emkappaseq, size);
        vexp(ex1, tmp1, &size);  // tmp2 = ex1 

        // ex2 = exp(-emlambda * radius + emlambdaseq); 
        vec_mul_scal_add_scal(tmp1, pRadi, -emlambda, emlambdaseq, size);
        vexp(ex2, tmp1, &size);  // tmp1 = ex2 
#ifdef CALCNBAR
        // nbar = emneq * ex1; 
        vec_mul_scal_add_scal(nbar, ex1, emneq, 0.0, size);
#endif
        // Ec = emt->e0 * (1 + emt->lambda * (radius - seq)) * ex2 
        const double cnst = emt->e0 * (1.0 - emt->lambda * emt->seq);
        vec_mul_scal_add_scal_mul(pEc, pRadi, eme0lambda, cnst, ex2, size);
        // Eas = 6 * emt->V0 * ex1 - 0.5 * emt->V0 * sigma2 / gamma2 
        vec_dbl_mul_scal_add(pEas, ex1, em6V0, pSig2, emneghalfV0overgamma2,size);			      
        
        // Calculate total cohesive and atomic sphere correction energy for this leaf
        for (i = 0; i < size; i++)
        {
			if(pSig1[i] != 0.0)
			{
				totEc  += pEc[i] - emt->e0;
				totEas += pEas[i];
			}
		}
    }
	
    return (totEc + totEas);
}

double EMTPotential::CalculateEnergyAndForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces)
{
    if (!(atoms == this->atoms && atoms->GetShape() == sigma1->GetShape()))
        Allocate(atoms);
    
    CalculateSigmas(atoms);
    
    double totE = CalculateEnergyAfterSigmas(atoms);
    
    CalculateForcesAfterSigmas(atoms,forces);
    
    return totE;
}

void EMTPotential::CalculateSigmas(ListOfAtoms *atoms, const bool forcesOnly)
{
    int is, nb, is1;
    double rnb[ARRLIM][DIMENSION];	// The scaled coordinates for us (Jakob's rnbs) 								// are inside the atoms class: forces use 
    // Cartesian coordinates
    double rnb2[ARRLIM];
    
    // Calculate combinations of EMT parameters
    
    double cutslope = cut.cutslope;
    double cutslopecutdist = cut.cutslope*cut.cutdist;
    double emeta2 = emt->eta2;
    double emeta2betaseq = emt->eta2*BETA*emt->seq;
    double emkappaoverbeta = emt->kappa/BETA;
    double emkappaseq = emt->kappa*emt->seq;
    double emneg_eta2 = -emeta2;
    double emneg_kob = -emkappaoverbeta;
    double vdist[ARRLIM]; //, sqdist[ARRLIM];
    double exarg[ARRLIM], vex1[ARRLIM], vex2[ARRLIM], vex3[ARRLIM];
    double invwght[ARRLIM], wght[ARRLIM];
    double dsigma1[ARRLIM], dsigma2[ARRLIM];
    int nbatch;					// used for distributing results from stack
    int slotsLeft;				// used for accumulating stack with HalfNeighbors()
    
    
    //	if (nAtoms != atoms->GetNumber()) Allocate(this->nAtoms);
    
    // Set sigmas to zero
    sigma1->Zero();
    sigma2->Zero();
    
    // Calculate sigmas
    
    // Loop over all atoms. 'is' is incremented below
    
    NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
    
    ASSERT(neighborLocator);	// Needs neighborlocator attached
    
    int neighbors[ARRLIM];
    int numnbr[ARRLIM];
    
    nAtoms = atoms->GetNumber();

    for (is = 0; is < nAtoms;)
    {
        // Collect up to ARRLIM displacements and process them in a
        // vectorized manner
        
        is1 = is;				// Used to check syncronization
        slotsLeft = ARRLIM;		// Nothing in the vector yet
        
        bool accumulate = true;
        
        while (accumulate && (is < nAtoms) && (is-is1 < ARRLIM))
        {
            // WARNING We've changed the convention for rnb: it now gives position of j with respect to i
            accumulate = 	neighborLocator->
                HalfNeighbors(	is,
                neighbors+(ARRLIM-slotsLeft),
                numnbr[is-is1],
                rnb+(ARRLIM-slotsLeft),
                rnb2+(ARRLIM-slotsLeft),
                slotsLeft);
            if (accumulate) is++;
        }
        nbatch = ARRLIM - slotsLeft;	// This is how many slots we have used
        // Need this, otherwise the following
        // vectorized routines would not work
        
        
        // vdist = sqrt(sqdist), distance between atoms
        vsqrt(vdist,rnb2,&nbatch);
        
        // Calculate contributions to the sigmas
        // ex1 = exp(cutslope * dist - cutslopecutdist);
        vec_mul_scal_add_scal(exarg, vdist, cutslope, 
            -cutslopecutdist, nbatch);
        vexp(vex1, exarg, &nbatch);
        // ex2 = exp(-emeta2 * dist + emeta2betaseq);
        vec_mul_scal_add_scal(exarg, vdist, emneg_eta2, emeta2betaseq, nbatch);
        vexp(vex2, exarg, &nbatch);
        // wght = 1 / (ex1 + 1.0) 
        vec_mul_scal_add_scal(invwght, vex1, 1.0, 1.0, nbatch);
        vrec(wght, invwght, &nbatch);
        // dsigma1 = wght * ex2 
        vec_mul(dsigma1, wght, vex2, nbatch);
        if(!forcesOnly)  // We don't need sigma2 to get the forces.
        {
            // ex3 = exp(-emkappaoverbeta * dist + emkappaseq); 
            vec_mul_scal_add_scal(exarg, vdist, emneg_kob, emkappaseq, nbatch);
            vexp(vex3, exarg, &nbatch);
            // dsigma2 = wght * ex3 
            vec_mul(dsigma2, wght, vex3, nbatch);
        }
        // Distribute contributions to the sigmas 
        
        int numAtomsDone = is - is1;
        nbatch = 0;
        
        // Danish way of synchronization
        
        // while (is1 < nAtoms && nbatch + 
        //	atoms->GetNeighborLocator()->MaxNumberOfHalfNeighbors(is1) < ARRLIM)
        // {
        //	for (nb = 0; nb < numnbr[nstack]; nb++)
        //	{
        
        // Our way of synchronization
        
        for ( ; is1 < is; is1++)
        {
            for (nb = 0; nb < numnbr[numAtomsDone-is+is1]; nb++)
            {
                const int nbr = neighbors[nbatch];
                (*sigma1)[is1] += dsigma1[nbatch];
                if(nbr<nAtoms) (*sigma1)[nbr] += dsigma1[nbatch];
                if(!forcesOnly)  // We don't need sigma2 to get the forces.
                {
                    (*sigma2)[is1] += dsigma2[nbatch];
                    if(nbr<nAtoms) (*sigma2)[nbr] += dsigma2[nbatch]; // otherwise it's a ghost
                }
                nbatch++;
            }
        }
        assert(is == is1);
    }
    sigma1->GetSynchronizationKit()->Synchronize();
    sigma2->GetSynchronizationKit()->Synchronize();
    slotsLeft = ARRLIM;
}


void EMTPotential::CalculateForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces)
{
    if (!(atoms == this->atoms && atoms->GetShape() == sigma1->GetShape()))
        Allocate(atoms);
    // sigmas and dEd0ds
    CalculateSigmas(atoms,true);
    CalculateForcesAfterSigmas(atoms, forces);
}

void EMTPotential::CalculateForcesAfterSigmas(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces)
{
    int is, nb, is1;   
    // double rd[3];
    double rnb[ARRLIM][DIMENSION];
    double rnb2[ARRLIM];
    double cutslope = cut.cutslope;
    double cutslopecutdist = cut.cutslope * cut.cutdist;
    double emeta2 = emt->eta2;
    double emeta2betaseq = emt->eta2 * BETA * emt->seq;
    double eminv12gamma1 = 1.0 / (12.0 * emt->gamma1);
    double eminvbetaeta2 = 1.0 / (emt->eta2 * BETA);
    double emseq = emt->seq;
    double emkappa = emt->kappa;
    double emlambda = emt->lambda;
    double emlambdaseq = emt->lambda * emt->seq;
    double eme0lambdalambda = emt->e0 * emt->lambda * emt->lambda;
    double eme0lambdalambdaseq = emt->e0 * emt->lambda * emt->lambda * emt->seq;
    double emkappaoverbeta = emt->kappa / BETA;
    double em6V0kappa = 6.0 * emt->V0 * emt->kappa;
    double emkappaseq = emt->kappa * emt->seq;
    double emV0overgamma2 = emt->V0 / emt->gamma2;
    double df, dfx, dfy, dfz;
    double emneg_eta2 = -emeta2;
    double emneg_kob = -emkappaoverbeta;
    double vdist[ARRLIM], invr[ARRLIM]; //, sqdist[ARRLIM]
    double exarg[ARRLIM], vex1[ARRLIM], vex2[ARRLIM], vex3[ARRLIM];
    double invwght[ARRLIM], wght[ARRLIM], dwdr[ARRLIM];
    double dsg1dr[ARRLIM], dsg2dr[ARRLIM];
    int nbatch;				// Used for distributing results on the stack
    int slotsLeft;			// Used to accumulate the stack using HalfNeighbors()
    double *tmp1, *tmp2, *ex1, *ex2;
    
    // Set forces to zero
    forces->Zero();
    
    // Allocate temporary arrays 
    tmp1 = dTmp[0];
    tmp2 = dTmp[1];
    ex1 = dTmp[2];
    ex2 = dTmp[3];
    ASSERT(tmp1 && tmp2 && ex1 && ex2);
    // Calculate gradients 
    // Prevent underflows 
    DMArrayIterator<double> sigIter(sigma1);
    for (sigIter.Init(); !sigIter.IsDone(); sigIter.Next()) //is = 0; is < nAtoms; is++)
    {
        if (sigIter.Current() < 1.0e-9)
            sigIter.Current() = 1.0e-9;
    }
    
    LeafToLeafIterator<DMArray<double> > sigma1It(sigma1);
    LeafToLeafIterator<DMArray<double> > radiusIt(radius);
    LeafToLeafIterator<DMArray<double> > dE0dsIt(dE0ds);
    for(sigma1It.Init(), radiusIt.Init(), dE0dsIt.Init();
    !sigma1It.Done();
    sigma1It.Next(), radiusIt.Next(), dE0dsIt.Next())
    {
        const int size = sigma1It.Current()->Size();
        if(size == 0) continue;
        const double *pSig1 = &(*sigma1It.Current())[0];
        double *pRadi = &(*radiusIt.Current())[0];
        double *pdE0ds= &(*dE0dsIt.Current())[0];
        
        // radius = emseq - eminvbetaeta2 * log(sigma1[is] * eminv12gamma1); 
        vec_mul_scal_add_scal(tmp1, pSig1, eminv12gamma1, 0.0, size);
        vlog(tmp2, tmp1, &size);
        vec_mul_scal_add_scal(pRadi, tmp2, -eminvbetaeta2, emseq, size);
        // ex1 = exp(-emkappa * radius + emkappaseq); 
        // ex2 = exp(-emlambda * radius + emlambdaseq); 
        vec_mul_scal_add_scal(tmp1, pRadi, -emkappa, emkappaseq, size);
        vexp(ex1, tmp1, &size);  // tmp2 = ex1 
        vec_mul_scal_add_scal(tmp1, pRadi, -emlambda, emlambdaseq, size);
        vexp(ex2, tmp1, &size);  // tmp3 = ex2 
        
        // dEds = (- eme0lambdalambda * radius + eme0lambdalambdaseq) * ex2
        //  - em6V0kappa * ex1; 
        vec_mul_scal_add_scal(tmp1, pRadi, - eme0lambdalambda,
            eme0lambdalambdaseq, size);
        vec_eq_self_mul_addscal_mul(tmp1, ex2, -em6V0kappa, ex1, size);
        // tmp1 = dEds
        // dEd0ds[is] =  - eminvbetaeta2 / sigma1[is] * dEds; 
        vrec(tmp2, pSig1, &size);
        vec_mul_mul_scal(pdE0ds, tmp1, tmp2, -eminvbetaeta2, size);
        
    }
    
    dE0ds->GetSynchronizationKit()->Synchronize();
    
    // Calculate forces 
    
    NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
    
    int neighbors[ARRLIM];
    int numnbr[ARRLIM];
    
    for (is = 0; is < nAtoms; )
    {
        is1 = is;
        slotsLeft = ARRLIM;
        
        bool accumulate = true;
        
        while (accumulate && (is < nAtoms) && (is-is1 < ARRLIM))
        {
            // WARNING We've changed the convention for rnb: it now gives position of j with respect to i
            accumulate = 	neighborLocator->
                HalfNeighbors(	is,
                neighbors+(ARRLIM-slotsLeft),
                numnbr[is-is1],
                rnb+(ARRLIM-slotsLeft),
                rnb2+(ARRLIM-slotsLeft),
                slotsLeft);
            if (accumulate) is++;
        }
        nbatch = ARRLIM - slotsLeft;	// This is how many slots we have used
        // This variable is needed for the
        // following vectorized routines to
        // work
        
        vsqrt(vdist,rnb2,&nbatch);
        
        // Calculate contributions to the sigmas
        // ex1 = exp(cutslope * dist - cutslopecutdist);
        vec_mul_scal_add_scal(exarg, vdist, cutslope, 
            -cutslopecutdist, nbatch);
        vexp(vex1, exarg, &nbatch);
        // ex2 = exp(-emeta2 * dist + emeta2betaseq);
        vec_mul_scal_add_scal(exarg, vdist, emneg_eta2, emeta2betaseq, nbatch);
        vexp(vex2, exarg, &nbatch);
        // ex3 = exp(-emkappaoverbeta * dist + emkappaseq);
        vec_mul_scal_add_scal(exarg, vdist, emneg_kob, emkappaseq, nbatch);
        vexp(vex3, exarg, &nbatch);
        
        // wght = 1.0 / (1.0 + ex1);
        // dwdr = - cutslope * wght * (1 - wght);
        vec_mul_scal_add_scal(invwght, vex1, 1.0, 1.0, nbatch);
        vrec(wght, invwght, &nbatch);
        vec_diffweight(dwdr, wght, -cutslope, nbatch);
        
        // dsg1dr = (dwdr - wght * emeta2) * ex2;
        vec_mul_scal_add_mul(dsg1dr, wght, -emeta2, dwdr, vex2, nbatch);
        // dsg2dr = (dwdr - wght * emkappaoverbeta) * ex3;
        vec_mul_scal_add_mul(dsg2dr, wght, -emkappaoverbeta, 
            dwdr, vex3, nbatch);
        
        // invr = 1.0 / vdist
        vrec(invr, vdist, &nbatch);
        
        // Distribute force contributions
        
        int numAtomsDone = is - is1;
        nbatch = 0;
        
        // while (is1 < nAtoms && nbatch + 
        //	atoms->GetNeighborLocator()->NumberOfHalfNeighbors(is1) < ARRLIM)
        // {
        //	for (nb = 0; nb < numnbr[nstack]; nb++)
        //	{
        for ( ; is1 < is; is1++)
        {
            double *forces1 = (*forces)[is1];
            for (nb = 0; nb < numnbr[numAtomsDone-is+is1]; nb++)
            {
                int nbr = neighbors[nbatch];
                df = ((*dE0dsExtended)[is1] + (*dE0dsExtended)[nbr]) * dsg1dr[nbatch] -
                    emV0overgamma2 * dsg2dr[nbatch];
                df *= invr[nbatch];
                dfx = df * rnb[nbatch][0];
                dfy = df * rnb[nbatch][1];
                dfz = df * rnb[nbatch][2];
                
                // WARNING We've changed the convention for rnb: it now gives position of j with respect to i
                
                //(*forces)[is1][0] += dfx;
                //(*forces)[neighbors[nbatch]][0] -= dfx;
                //(*forces)[is1][1] += dfy;
                //(*forces)[neighbors[nbatch]][1] -= dfy;
                //(*forces)[is1][2] += dfz;
                //(*forces)[neighbors[nbatch]][2] -= dfz;
                forces1[0] += dfx;
                forces1[1] += dfy;
                forces1[2] += dfz;
                if(nbr<nAtoms)
                {
                    double *forces2 = (*forces)[nbr];
                    forces2[0] -= dfx;
                    forces2[1] -= dfy;
                    forces2[2] -= dfz;
                }
                nbatch++;
            }
        }
        assert(is == is1);
    }
    
    // XXX No scaled coordinates (yet, perhaps ever?)
    
    // Convert forces to scaled forces
    // Should be done by a call, but this is such a small part of the time
    // for (is = 0; is < nAtoms; is++)
    // {
    //	rd[0] = fs[is][0];
    //	rd[1] = fs[is][1];
    //	rd[2] = fs[is][2];
    //	fs[is][0] = invbasis[0][0] * rd[0] + invbasis[0][1] * rd[1]
    //	+ invbasis[0][2] * rd[2];
    //	fs[is][1] = invbasis[1][0] * rd[0] + invbasis[1][1] * rd[1]
    //	+ invbasis[1][2] * rd[2];
    //	fs[is][2] = invbasis[2][0] * rd[0] + invbasis[2][1] * rd[1]
    //	+ invbasis[2][2] * rd[2];
    // }
}

void EMTPotential::CalculateStressTensor(ListOfAtoms *atoms, int i, double stress[6])
{
#ifdef DMPARALLEL 
    throw DMException("EMTPotential", "CalculateStressTensor",  
        "this function cannot be used in parallel, see comment in source file");
#endif
   /* the problem with this function is that it needs the radius of neighboring atoms,
        which can be ghosts. At this time, with the current implementation, we cannot get
        them. To get them, we could either:
        A. Synchronize the radius array, but this cannot be done in this function, since 
        not all the processors necessarily call this function.
        B. Since we know the ghosts' sigma1, we can recalculate the radius array on the ghosts.
    */
    int ii,nb, alpha, beta;
    double rnb[ARRLIM][DIMENSION];
    double rnb2[ARRLIM];
    double emnegeta2 = - emt->eta2;
    double emnegeta2beta = emnegeta2 * BETA;
    double emeta2betaseq = - emnegeta2beta * emt->seq;
    double emneglambda = -emt->lambda;
    double emlambdaseq = emt->lambda * emt->seq;
    double emnege0 = -emt->e0;
    double emnegkappa = -emt->kappa;
    double emkappaseq = emt->kappa * emt->seq;
    double emneg6V0kappa = - 6.0 * emt->V0 * emt->kappa;
    double cutslope = cut.cutslope;
    double cutslopecutdist = cut.cutslope * cut.cutdist;
    double emnegkob = -emt->kappa / BETA;
    double emnegslopepluseta2 = - (cutslope + emt->eta2);
    double emnegslopepluskob = - (cutslope + emt->kappa / BETA);
    double emnegV0overtwogamma2 = - emt->V0 / (2.0*emt->gamma2);
    
    double vdist[ARRLIM], tmp[ARRLIM];
    double vex1[ARRLIM], vex2[ARRLIM], vex3[ARRLIM];
    double wght[ARRLIM];
    double dsigma1ddist[ARRLIM], dsigma2ddist[ARRLIM];
    double dEddist[ARRLIM], oprod[ARRLIM], stresscontrib[6][ARRLIM];
    
    // Initialize the local stress to 0. 
    memset(stress, 0, 6*sizeof(double));
    
    // Assign temporary pointers
    double *thisRadius = dTmp[0];
    double *thisdsdsigma1 = dTmp[1];
    double *thisdEcdsigma1 = dTmp[2];
    double *thisdEasdsigma1 = dTmp[3];
    double *tmp1 = dTmp[4];
    double *thisEc = dTmp[5];
    double *thisSigma1 = dTmp[6];
    
    // Find all the neighbors of atoms i
    NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
    int neighbors[ARRLIM];
    int numnbr;
    int slotsLeft = ARRLIM-1;			// Used to accumulate the stack using Neighbors() 
    // (Here we reserve that last place for atom i)
    
    bool ok = neighborLocator->Neighbors(i, neighbors, numnbr, rnb, rnb2, slotsLeft);
    if(!ok) throw DMException("EMTPotential", "CalculateStressTensor", "Insufficient Memory Allocation");
    
    neighbors[numnbr++] = i ; // because we need to treat the values of Ec, dEcdsigma1, etc... also for atom i 
    
    // We need the following quantities on the neighboring atoms (and i itself)
    
    
    for(int at=0; at<numnbr; ++at)
    {
        const int nb = neighbors[at];
        thisRadius[at] = (*radius)[nb];
        thisSigma1[at] = (*sigma1)[nb];
        thisEc[at] = (*Ec)[nb];
    }
    // dsdsigma1 = 1.0 / (- beta eta2 sigma1)
    vec_mul_scal(tmp1, thisSigma1, emnegeta2beta, numnbr);
    vrec(thisdsdsigma1, tmp1, &numnbr);
    
    // tmp1 = exp(-emlambda * radius + emlambdaseq)
    // dEcdsigma1 = (eme0 * tmp1 - Ec) * lambda
    // Use dEasdsigma1 as a temp variable
    vec_mul_scal_add_scal(thisdEasdsigma1, thisRadius, emneglambda, emlambdaseq, numnbr);
    vexp(tmp1, thisdEasdsigma1, &numnbr);
    
    // dEcdsigma1 = (-emt->e0 * tmp1 + Ec) * -emt->lambda
    vec_mul_scal_add_mul_scal_mul(thisdEcdsigma1, tmp1, emnege0, thisEc, emneglambda, thisdsdsigma1, numnbr);
    
    // tmp1 = exp(-kappa * radius + kappa * seq)
    // dEasdsigma1 = (-6V0 kappa * tmp1 * dsdsigma1
    vec_mul_scal_add_scal(thisdEasdsigma1, thisRadius, emnegkappa, emkappaseq, numnbr);
    vexp(tmp1, thisdEasdsigma1, &numnbr);
    vec_mul_mul_scal(thisdEasdsigma1, tmp1, thisdsdsigma1, emneg6V0kappa, numnbr);
    
    numnbr--; // from here we can consider the real neighbors
    
    for(nb=0; nb<numnbr; ++nb)
    {
        thisdEcdsigma1[nb] += thisdEcdsigma1[numnbr];
        thisdEasdsigma1[nb] += thisdEasdsigma1[numnbr];
    }
    
    
    // interatomic distances
    vsqrt(vdist,rnb2,&numnbr);
    
    // vex1 = exp(cutslope * dist - cutslopecutdist);
    vec_mul_scal_add_scal(tmp, vdist, cutslope, -cutslopecutdist, numnbr);
    vexp(vex1, tmp, &numnbr);
    // vex2 = exp(-emeta2 * dist + emeta2betaseq);
    vec_mul_scal_add_scal(tmp, vdist, emnegeta2, emeta2betaseq, numnbr);
    vexp(vex2, tmp, &numnbr);
    // vex3 = exp(-emkappaoverbeta * dist + emkappaseq);
    vec_mul_scal_add_scal(tmp, vdist, emnegkob, emkappaseq, numnbr);
    vexp(vex3, tmp, &numnbr);
    
    // wght = 1.0 / (vex1 + 1.0);
    vec_mul_scal_add_scal(tmp, vex1, 1.0, 1.0, numnbr);
    vrec(wght, tmp, &numnbr);
    
    // dsigma1ddist = (cutslope * wght - (cutslope + eta2)) * wght * vex2
    vec_P_av_times_b_plus_c_p_times_av_times_dv(dsigma1ddist, wght, cutslope, emnegslopepluseta2, vex2, numnbr);
    
    // dsigma2ddist = (cutslope * wght - (cutslope + kappa/beta)) * wght * vex3
    vec_P_av_times_b_plus_c_p_times_av_times_dv(dsigma2ddist, wght, cutslope, emnegslopepluskob, vex3, numnbr);
    
    // dEddist = dEcdsigma1 * dsigma1ddist + dEasdsigma1 * dsigma1ddist - (V0 / (2 gamma2)) * dsigma2ddist
    vec_av_times_bv_plus_cv_times_bv_plus_d_times_ev(dEddist, thisdEcdsigma1, dsigma1ddist, thisdEasdsigma1, 
        2.*emnegV0overtwogamma2, dsigma2ddist, numnbr);
    
    // tmp = 0.5 * dEddist / dist
    vrec(tmp, vdist, &numnbr);
    vec_self_mul_mul_scal(tmp, dEddist, 0.5, numnbr);
    
    // stresscontrib[alpha, beta] = 0.5 * dEddist * rnb[][alpha] * rnb[][beta] / dist
    for(alpha = 0; alpha < 3; ++alpha)
        for(beta = alpha; beta < 3; ++beta)
        {
            ii = stresscomp[alpha][beta];
            // oprod[] = rnb[][alpha] * rnb[][beta]
            vec_outherprod(oprod, rnb, alpha, beta, numnbr);
            // stresscontrib[ii] = tmp*oprod
            vec_mul(stresscontrib[ii], tmp, oprod, numnbr);
        }
        
        // Sum stress contributions   
        for (nb = 0; nb < numnbr; nb++)
        {
            for(ii=0; ii<6; ++ii)
            {
                stress[ii] += stresscontrib[ii][nb];
            }
        }
        
        // Finally add the dynamic part and scale with the atomic volume
        double invvol = 3./(4. * PI * emt->seq * emt->seq * emt->seq);
        double mass = emt->mass;
        const double *veloCurrent = (*(atoms->GetCartesianVelocities()))[i] ;
        for(alpha=0; alpha<3; ++alpha)
            for(beta=alpha; beta<3; ++beta)
            {
                ii = stresscomp[alpha][beta];
                stress[ii] -= veloCurrent[alpha]*veloCurrent[beta]*mass;
                stress[ii] *= invvol;
            }
}

void EMTPotential::CalculateStressTensor(ListOfAtoms *atoms, DMArray<double, (DIMENSION+1)*DIMENSION/2> *stress)
{
    int is, ii, is1, nb, alpha, beta;
    double rnb[ARRLIM][DIMENSION];
    double rnb2[ARRLIM];
    double emnegeta2 = - emt->eta2;
    double emnegeta2beta = emnegeta2 * BETA;
    double emeta2betaseq = - emnegeta2beta * emt->seq;
    double emneglambda = -emt->lambda;
    double emlambdaseq = emt->lambda * emt->seq;
    double emnege0 = -emt->e0;
    double emnegkappa = -emt->kappa;
    double emkappaseq = emt->kappa * emt->seq;
    double emneg6V0kappa = - 6.0 * emt->V0 * emt->kappa;
    double cutslope = cut.cutslope;
    double cutslopecutdist = cut.cutslope * cut.cutdist;
    double emnegkob = -emt->kappa / BETA;
    double emnegslopepluseta2 = - (cutslope + emt->eta2);
    double emnegslopepluskob = - (cutslope + emt->kappa / BETA);
    double emnegV0overtwogamma2 = - emt->V0 / (2.0*emt->gamma2);
    
    double vdist[ARRLIM], tmp[ARRLIM];
    double vex1[ARRLIM], vex2[ARRLIM], vex3[ARRLIM];
    double wght[ARRLIM];
    double v_dEcdsigma1[ARRLIM], v_dEasdsigma1[ARRLIM], dsigma1ddist[ARRLIM], dsigma2ddist[ARRLIM];
    double dEddist[ARRLIM], oprod[ARRLIM], stresscontrib[6][ARRLIM];
    int nbatch;				// Used for distributing results on the stack
    int slotsLeft;			// Used to accumulate the stack using HalfNeighbors()
    
    stress->Zero();
    radius->GetSynchronizationKit()->Synchronize();
    Ec    ->GetSynchronizationKit()->Synchronize();
    
    double *dsdsigma1   = dTmp[0];
    double *dEcdsigma1  = dTmp[1];
    double *dEasdsigma1 = dTmp[2];
    double *tmp1        = dTmp[3];
    double *tmp2        = dTmp[4];
    
    int sum = 0;
    LeafToLeafIterator<DMArray<double> > sigma1It(sigma1);
    LeafToLeafIterator<DMArray<double> > radiusIt(radius);
    LeafToLeafIterator<DMArray<double> > EcIt    (Ec    );
    for(sigma1It.Init(), radiusIt.Init(), EcIt.Init();
    !sigma1It.Done();
    sigma1It.Next(), radiusIt.Next(), EcIt.Next())
    {
        const int size = sigma1It.Current()->Size();
        if(size == 0) continue;
        const double *pSig1 = &(*sigma1It.Current())[0];
        const double *pRadi = &(*radiusIt.Current())[0];
        const double *pEc   = &(*EcIt.Current())[0];
        
        // dsdsigma1 = 1.0 / (- beta eta2 sigma1)
        vec_mul_scal(tmp1, pSig1, emnegeta2beta, size);
        vrec(dsdsigma1, tmp1, &size);
        
        // tmp1 = exp(-emlambda * radius + emlambdaseq)
        vec_mul_scal_add_scal(tmp2, pRadi, emneglambda, emlambdaseq, size);
        vexp(tmp1, tmp2, &size);
        
        // dEcdsigma1 = (-emt->e0 * tmp1 + Ec) * -emt->lambda
        vec_mul_scal_add_mul_scal_mul(dEcdsigma1+sum, tmp1, 
            emnege0, pEc, emneglambda, dsdsigma1, size);
        
        // tmp1 = exp(-kappa * radius + kappa * seq)
        // dEasdsigma1 = (-6V0 kappa * tmp1 * dsdsigma1
        vec_mul_scal_add_scal(dEasdsigma1+sum, pRadi, emnegkappa, emkappaseq, size);
        vexp(tmp1, dEasdsigma1+sum, &size);
        vec_mul_mul_scal(dEasdsigma1+sum, tmp1, dsdsigma1, emneg6V0kappa, size);
        
        sum += size;
    }
    ASSERT(sum == nAtoms);
    
    // Now calculate derivative with respect to the interactomic distances
    NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
    int neighbors[ARRLIM];
    int numnbr[ARRLIM];
    for (is = 0; is < nAtoms; )
    {
        is1 = is;
        slotsLeft = ARRLIM;
        nbatch = 0;
        bool accumulate = true;
        
        while (accumulate && (is < nAtoms) && (is-is1 < ARRLIM))
        {
            accumulate = 	neighborLocator->Neighbors(	is,
                neighbors+(ARRLIM-slotsLeft),
                numnbr[is-is1],
                rnb+(ARRLIM-slotsLeft),
                rnb2+(ARRLIM-slotsLeft),
                slotsLeft);
            for(int i=0; i<numnbr[is-is1]; ++i)
            {
                v_dEcdsigma1[nbatch] = dEcdsigma1[is];
                v_dEasdsigma1[nbatch++] = dEasdsigma1[is];
            }
            if (accumulate) is++;
        }
        ASSERT(nbatch == ARRLIM - slotsLeft);	// This is how many slots we have used
        // This variable is needed for the
        // following vectorized routines to
        // work
        
        // interatomic distances
        vsqrt(vdist,rnb2,&nbatch);
        
        // Calculate contributions to the sigmas
        // vex1 = exp(cutslope * dist - cutslopecutdist);
        vec_mul_scal_add_scal(tmp, vdist, cutslope, -cutslopecutdist, nbatch);
        vexp(vex1, tmp, &nbatch);
        // vex2 = exp(-emeta2 * dist + emeta2betaseq);
        vec_mul_scal_add_scal(tmp, vdist, emnegeta2, emeta2betaseq, nbatch);
        vexp(vex2, tmp, &nbatch);
        // vex3 = exp(-emkappaoverbeta * dist + emkappaseq);
        vec_mul_scal_add_scal(tmp, vdist, emnegkob, emkappaseq, nbatch);
        vexp(vex3, tmp, &nbatch);
        
        // wght = 1.0 / (vex1 + 1.0);
        vec_mul_scal_add_scal(tmp, vex1, 1.0, 1.0, nbatch);
        vrec(wght, tmp, &nbatch);
        
        // dsigma1ddist = (cutslope * wght - (cutslope + eta2)) * wght * vex2
        vec_P_av_times_b_plus_c_p_times_av_times_dv(dsigma1ddist, wght, cutslope, emnegslopepluseta2, vex2, nbatch);
        
        // dsigma2ddist = (cutslope * wght - (cutslope + kappa/beta)) * wght * vex3
        vec_P_av_times_b_plus_c_p_times_av_times_dv(dsigma2ddist, wght, cutslope, emnegslopepluskob, vex3, nbatch);
        
        // dEddist = v_dEcdsigma1 * dsigma1ddist + v_dEasdsigma1 * dsigma1ddist - (V0 / (2 gamma2)) * dsigma2ddist
        vec_av_times_bv_plus_cv_times_bv_plus_d_times_ev(dEddist, v_dEcdsigma1, dsigma1ddist, v_dEasdsigma1, 
            emnegV0overtwogamma2, dsigma2ddist, nbatch);
        
        // tmp = 0.5 * dEddist / dist
        vrec(tmp, vdist, &nbatch);
        vec_self_mul_mul_scal(tmp, dEddist, 0.5, nbatch);
        
        // stresscontrib[alpha, beta] = 0.5 * dEddist * rnb[][alpha] * rnb[][beta] / dist
        for(alpha = 0; alpha < 3; ++alpha)
            for(beta = alpha; beta < 3; ++beta)
            {
                ii = stresscomp[alpha][beta];
                // oprod[] = rnb[][alpha] * rnb[][beta]
                vec_outherprod(oprod, rnb, alpha, beta, nbatch);
                // stresscontrib[ii] = tmp*oprod
                vec_mul(stresscontrib[ii], tmp, oprod, nbatch);
            }
            
            // Distribute stress contributions
            
            int numAtomsDone = is - is1;
            nbatch = 0;
            
            double dummy[6];
            for ( ; is1 < is; is1++)
            {
                double *stress1 = (*stress)[is1];
                for (nb = 0; nb < numnbr[numAtomsDone-is+is1]; nb++)
                {
                    double *stress2;
                    if(neighbors[nbatch] < nAtoms) stress2 = (*stress)[neighbors[nbatch]];
                    else stress2 = dummy;
                    for(ii=0; ii<6; ++ii)
                    {
                        stress1[ii] += stresscontrib[ii][nbatch];
                        stress2[ii] += stresscontrib[ii][nbatch];
                    }
                    nbatch++;
                }
            }
            assert(is == is1);
    }
    
    // Finally add the dynamic part and scale with the atomic volume
    double invvol = 3./(4. * PI * emt->seq * emt->seq * emt->seq);
    double mass = emt->mass;
    DMArrayIterator<double, (DIMENSION+1)*DIMENSION/2> stressIter(stress);
    DMArrayIterator<double, DIMENSION> veloIter(atoms->GetCartesianVelocities());
    double *stressCurrent, *veloCurrent;
    for(stressIter.Init(), veloIter.Init(); !stressIter.IsDone(); stressIter.Next(), veloIter.Next())
    {
        veloCurrent = veloIter.Current();
        stressCurrent = stressIter.Current();
        for(alpha=0; alpha<3; ++alpha)
            for(beta=alpha; beta<3; ++beta)
            {
                ii = stresscomp[alpha][beta];
                stressCurrent[ii] -= veloCurrent[alpha]*veloCurrent[beta]*mass;
                stressCurrent[ii] *= invvol;  // because now we cannot use the *= operator on the DMArray
            }
    }
    //	(*stress) *= invvol;
    
}


double EMTPotential::CalculateAtomicEnergy(ListOfAtoms *atoms, int i)
{
    if (!(atoms == this->atoms && atoms->GetShape() == sigma1->GetShape()))
        Allocate(atoms);
    
    ASSERT((i < atoms->GetNumber()) && (i>=0));
    double sigma1i = 0;
    double sigma2i = 0;
    
    double *vdist = dTmp[0], *sqdist = dTmp[1];
    double *exarg= dTmp[2], *vex1 = dTmp[3], *vex2 = dTmp[4], *vex3 = dTmp[5];
    double *invwght = dTmp[6], *wght = new double[ARRLIM];
    double *dsigma1 = new double[ARRLIM], *dsigma2 = new double[ARRLIM];
    double (*rnb)[DIMENSION] = new double[ARRLIM][DIMENSION];  //XXXDM
    int nNeighbors;
    int *neighbors = new int[ARRLIM];
    
    NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
    int slotsLeft = ARRLIM;
    bool ok = neighborLocator->Neighbors(i,neighbors,nNeighbors,rnb, sqdist,slotsLeft);
    
    if (!ok) throw DMException("EMTPotential", "CalculateSigma", "Insufficient allocated memory");
    
    // vdist = sqrt(sqdist), distance between atoms
    vsqrt(vdist,sqdist,&nNeighbors);
    // ex1 = exp(cutslope * dist - cutslopeCutdist)
    vec_mul_scal_add_scal(exarg, vdist, cut.cutslope,-cut.cutslope*cut.cutdist, nNeighbors);
    vexp(vex1, exarg, &nNeighbors);
    // ex2 = exp(-emeta2 * dist + emeta2BetaSeq)
    vec_mul_scal_add_scal(exarg, vdist, -emt->eta2, emt->eta2 * BETA * emt->seq, nNeighbors);
    vexp(vex2, exarg, &nNeighbors);
    // ex3 = exp(-emKappaOverBeta * dist + emKappaSeq)
    vec_mul_scal_add_scal(exarg, vdist, -emt->kappa/BETA, emt->kappa * emt->seq, nNeighbors);
    vexp(vex3, exarg, &nNeighbors);
    // wght = 1 / (ex1 + 1.0)
    vec_mul_scal_add_scal(invwght, vex1, 1.0, 1.0, nNeighbors);
    vrec(wght, invwght, &nNeighbors);
    // dsigma1 = wght * ex2
    vec_mul(dsigma1, wght, vex2, nNeighbors);
    // dsigma2 = wght * ex3
    vec_mul(dsigma2, wght, vex3, nNeighbors);
    
    for(int at=0;at<nNeighbors;at++) 
    {
        sigma1i += dsigma1[at];
        sigma2i += dsigma2[at];
    }
    if(sigma1i<1.e-9) sigma1i = 1.e-9;
    if(sigma2i<1.e-9) sigma2i = 1.e-9;
    
    delete[] wght; delete[] dsigma1; delete[] dsigma2; delete[] rnb;
    delete[] neighbors;
    
    double tmp = sigma1i/(12.*emt->gamma1);
    double radiusMinusSeq = - log(tmp)/(BETA*emt->eta2);
    double ex1 = exp(-emt->kappa * radiusMinusSeq);
    tmp = emt->lambda * radiusMinusSeq;
    double Eci = emt->e0 * ((tmp + 1.0) * exp(-tmp) - 1.0);
    double Easi = emt->V0 * (6.0*ex1 - 0.5*sigma2i/emt->gamma2);
    
    return Eci + Easi;
}


// Roughly returns the diameter of the atom
// For fcc, should return 1/sqrt(2) of the lattice constant, for example
// Currently used to set up the lattice structure: a good initial number is important
double EMTPotential::GetLengthScale() const
{
    if(element == "Cu") return  6.789382809/sqrt(2.0);
    if(element == "Au") return  7.66504117182/sqrt(2.0);
    if(element == "Ag") return  7.6790043/sqrt(2.0);
    if(element == "Al") return  7.54871784/sqrt(2.0);
    if(element == "Pt") return  7.41119853/sqrt(2.0);
    return 6./sqrt(2.0);
}

// For each potential, we need some kind of energy scale to set time constants
// for movers and such
// It doesn't have to be precisely defined! A crude measure is OK.
// This should return, for a pair potential V(x) with LengthScale ell, V''(ell)
// For now, we're getting this from twice the binding energy per atom divided by the length scale squared
// (as if the potential were parabolic, cut off at LengthScale

double EMTPotential::GetEnergyCurvatureScale() const
{
    double bindingEnergy = 3.7;	// XXX ???
    if(element == "Cu") bindingEnergy = 3.5115924695;
    if(element == "Au") bindingEnergy = 3.8001349225;
    if(element == "Ag") bindingEnergy = 2.96036651332;
    if(element == "Al") bindingEnergy = 3.2848632506;
    if(element == "Pt") bindingEnergy = 5.85013367;
    return 2.0 * bindingEnergy / (GetLengthScale()*GetLengthScale());
}


////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////

void EMTPotential::Save(DMWriter *writer) const
{
	Potential::Save(writer);
	writer->PutString(element, "elem");

    // to check that the parameters have not changes when we reload.
    writer->PutDouble(emt->e0, "e0");
	writer->PutDouble(emt->seq, "seq");
	writer->PutDouble(emt->neq, "neq");
	writer->PutDouble(emt->V0, "V0");
	writer->PutDouble(emt->eta2, "eta2");
	writer->PutDouble(emt->kappa, "kappa");
	writer->PutDouble(emt->lambda, "lambda");
	writer->PutDouble(emt->mass, "mass");
	writer->PutDouble(cut.cutslope, "cutslope");
	writer->PutDouble(cut.cutdist, "cutdist");
}

void EMTPotential::Load(DMReader *reader)
{
	Potential::Load(reader);
	// We don't change these parameters. We're only saving them in case 
	// somebody does in the future, and in the future after that wants to 
	// re examine old data in which case. The function CheckPotentialParameter
	// throws an error if the double retrieved via the reader using the string
	// passed is not equal to the double passed
    
    this->element = reader->GetString("elem");

    delete emt;
    emt = new emt_struct(element.c_str());
    cut.setcut(emt, 1);
    emt->CalculateGamma(cut);
    
    SetCutoffDistance(cut.cutdist);

	CheckPotentialParameter(reader, emt->e0, "e0");
	CheckPotentialParameter(reader, emt->seq, "seq");
	CheckPotentialParameter(reader, emt->neq, "neq");
	CheckPotentialParameter(reader, emt->V0, "V0");
	CheckPotentialParameter(reader, emt->eta2, "eta2");
	CheckPotentialParameter(reader, emt->kappa, "kappa");
	CheckPotentialParameter(reader, emt->lambda, "lambda");
	CheckPotentialParameter(reader, emt->mass, "mass");
	CheckPotentialParameter(reader, cut.cutslope, "cutslope");
	CheckPotentialParameter(reader, cut.cutdist, "cutdist");
}



#endif // DIMENISONIS3
