// Granular Potential
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
#include "DigitalMaterial.h"
#include <iostream>
#include "GranularPotential.h"

#include "ListOfAtoms.h"
#include "DMArray.h"
#include "IndexError.h"
#include "NotEnoughMemory.h"
#include "NeighborLocator.h"
#include "SerialBuilder.h"
#include "BoundaryConditions.h"

#include "DMLastHeader.h"

using namespace std;

#define LEN_NBR_ARRAYS 1500


static SerialBuilder<GranularPotential> *_Gcreator = 
    new SerialBuilder<GranularPotential>("GranularPotential");

const string GranularPotential::className = string("GranularPotential");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


GranularPotential::GranularPotential() 
{
	particleDiameter = 1.0;
	elasticConstant = 10000.0;
	dampingParameter = 0;
}

GranularPotential::GranularPotential(double pD, double eC, double dP) 
{
	particleDiameter = pD;
	elasticConstant = eC;
	dampingParameter = dP;
}


void GranularPotential::CalculateForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces)
{
	int i,coord,jx;
	int j;
	int numberHalfNeighbors;
	forces->Zero();
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	double *forceI, *forceJ;
	int nAtoms = atoms->GetNumber();
	double dummyForce[DIMENSION]; // where we send the force contribution to a ghost atom


	int atomIdx1, atomIdx2;
	int fastNeighbors[LEN_NBR_ARRAYS];
	int numbersOfNeighbors[LEN_NBR_ARRAYS];
	double fastDisplacements[LEN_NBR_ARRAYS][DIMENSION];
	double fastDisplacementsSq[LEN_NBR_ARRAYS];
	double forceR[LEN_NBR_ARRAYS];
	int batchSize;
	int batchIdx;					// used for distributing results from stack
    int slotsLeft;				// used for accumulating stack with HalfNeighbors()


	for (atomIdx1 = 0; atomIdx1 < nAtoms;)
    {
        // Collect up to LEN_NBR_ARRAYS displacements and process them in a
        // vectorized manner
		
		atomIdx2 = atomIdx1;			
		slotsLeft = LEN_NBR_ARRAYS;		// Nothing in the vector yet
		
		bool accumulate = true;
		
		while (accumulate && (atomIdx1 < nAtoms) && (atomIdx1-atomIdx2 < LEN_NBR_ARRAYS))
		{
			accumulate = 	neighborLocator->
				HalfNeighbors(	atomIdx1,
							fastNeighbors+(LEN_NBR_ARRAYS-slotsLeft),
							numbersOfNeighbors[atomIdx1-atomIdx2],
							fastDisplacements+(LEN_NBR_ARRAYS-slotsLeft),
							fastDisplacementsSq+(LEN_NBR_ARRAYS-slotsLeft),
							slotsLeft);
			if (accumulate) atomIdx1++;
		}
		batchSize = LEN_NBR_ARRAYS - slotsLeft;	// This is how many slots we have used
		ASSERT(batchSize >= 0);

// 		for( batchIdx = 0; batchIdx < batchSize; ++batchIdx )
// 		{
// 			int j = fastNeighbors[batchIdx];
// 			normOfRelativeDisplacement = sqrt(fastDisplacementsSq[batchIdx]);
// 			for(coord = 0; coord < DIMENSION; ++coord){
// 				normalVector[coord] = fastDisplacements[batchIdx][i] / normOfRelativeDisplacement;
// 				relativeVelocity[coord] = atoms->GetCartesianVelocity(
// 					// calculate relative velocity
				
// 			}
// 			// calculate force
// 			forceR[batchIdx] = elasticConstant*(particleDiameter - normOfRelativeDisplacement) - ;
			
// 		}


		int numAtomsDone = atomIdx1 - atomIdx2;
        batchIdx = 0;
 
		double tempForce[DIMENSION];
		double *forceI, *forceJ;
		for ( ; atomIdx2 < atomIdx1; atomIdx2++)
        {

			int i = atomIdx2;

			forceI = (*forces)[i];
			numberHalfNeighbors = numbersOfNeighbors[numAtomsDone-atomIdx1+atomIdx2];
			for (int jx = 0; jx < numberHalfNeighbors; ++jx)
			{
				int j = fastNeighbors[batchIdx];
				double normalRelativeVelocity = 0;

				double normOfRelativeDisplacement = sqrt(fastDisplacementsSq[batchIdx]);
				for(coord = 0; coord < DIMENSION; ++coord){
					double normalVector = fastDisplacements[batchIdx][coord] / normOfRelativeDisplacement;
					double relativeVelocity = atoms->GetCartesianVelocity(i,coord) - atoms->GetCartesianVelocity(j,coord);
					normalRelativeVelocity += normalVector*relativeVelocity;
				}
					
				// calculate force
				forceR[batchIdx] = elasticConstant*(particleDiameter - normOfRelativeDisplacement) 
					+ dampingParameter*normalRelativeVelocity;

				if( j < nAtoms)
					forceJ = (*forces)[j];
				else
					forceJ = dummyForce;

				for (coord=0; coord<DIMENSION; coord++)
				{
					forceI[coord] += -fastDisplacements[batchIdx][coord]*forceR[batchIdx];
					forceJ[coord] += +fastDisplacements[batchIdx][coord]*forceR[batchIdx];
				}
				batchIdx++;
			}
			
        }
        ASSERT(atomIdx1 == atomIdx2);
	}
}

double GranularPotential::CalculateEnergy(ListOfAtoms *atoms)
{
	double energy = 0;
// 	for (int atom=0; atom < atoms->GetNumber(); atom++)
// 	{
// 		energy += g * atoms->GetCartesianPosition(atom, dim);
// 	}
	return energy;
}

