// HolianCutLennardJonesPotential.cpp: implementation of the HolianCutLennardJonesPotential class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
#include <assert.h>
#include "DigitalMaterial.h"

#include "HolianCutLennardJonesPotential.h"
#include "ListOfAtoms.h"
#include "NeighborLocator.h"
#include "DMArray.h"
#include "SprintString.h"
#include "IndexError.h"
#include "NotEnoughMemory.h"
#include "SerialBuilder.h"
#include "DMLastHeader.h"

static SerialBuilder<HolianCutLennardJonesPotential> *_HolianCLJcreator = 
    new SerialBuilder<HolianCutLennardJonesPotential>("HolianCutLennardJonesPotential");

const std::string HolianCutLennardJonesPotential::className = std::string("HolianCutLennardJonesPotential");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

HolianCutLennardJonesPotential::HolianCutLennardJonesPotential(): CutLennardJonesPotential(),
																  HCLJ_rSpl(1.2444550602598078860),
																  HCLJ_a2(0.5424493976339965),
																  HCLJ_a3(0.09350526707070671),
																  HCLJ_cutoffDistance(1.711238225055858)
{
	r2Spl = HCLJ_rSpl*HCLJ_rSpl;
    SetCutoffDistance(HCLJ_cutoffDistance);
}


HolianCutLennardJonesPotential::~HolianCutLennardJonesPotential()
{

}


double HolianCutLennardJonesPotential::CalculatePairForceEnergy(double displacement[DIMENSION], double displacementSq, double *force)
{
	double pairEnergy = 0.;
	ASSERT(displacementSq < GetCutoffDistance2());
	double forceR = 0.;

	if (displacementSq < r2Spl) 
	{
		double r2Inv = 1. / displacementSq;
		double r6Inv = r2Inv * r2Inv * r2Inv;
		double r8Inv = r6Inv * r2Inv;
		
		double r12Inv = r6Inv*r6Inv;
		pairEnergy = 4*(r12Inv - r6Inv);
		
		double r14Inv = r12Inv*r2Inv;
		forceR = 24.0*(2.0*r14Inv - r8Inv);// note factor of 4 replaced
	}
	else 
	{
		double temp = GetCutoffDistance2() - displacementSq;		
		forceR = -4 * HCLJ_a2 * temp + 6 * HCLJ_a3 * temp * temp;	
		pairEnergy = - HCLJ_a2 * temp * temp + HCLJ_a3 * temp * temp * temp;
	}

	for ( int coord = 0; coord < DIMENSION; ++coord )
		force[coord] += -displacement[coord] * forceR;

	return pairEnergy;
}



void HolianCutLennardJonesPotential::CalculateForces(ListOfAtoms *atoms, 
												  DMArray<double, DIMENSION> *forces)
{
	int i, coord, jx, j;
	int numberHalfNeighbors;
	forces->Zero();
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	double *forceI, *forceJ;
	int nAtoms = atoms->GetNumber();
	double dummyForce[DIMENSION]; // where we send the force contribution to a ghost atom
	for( i = 0; i < nAtoms; ++i )
    {	
		int spacesLeftInArrays = nPairs;
		bool arrayNotTooFull = neighborLocator->HalfNeighbors(i,neighbors,
			numberHalfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		while(!arrayNotTooFull)
		{
			AllocateForNeighborLocator(2*nPairs);
			spacesLeftInArrays = nPairs; // nPairs has been doubled
			arrayNotTooFull = neighborLocator->HalfNeighbors(i,neighbors,
				numberHalfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		}
		forceI = (*forces)[i];
		double forceR;
		for(jx=0;jx<numberHalfNeighbors;jx++)
		{
			j=neighbors[jx];
			if(j<nAtoms) forceJ = (*forces)[j];
			else forceJ = dummyForce;
			if (displacementsSq[jx] < r2Spl) 
			{
				double r2Inv = 1.0/ displacementsSq[jx];
				double r4Inv = r2Inv*r2Inv;
				double r8Inv = r4Inv*r4Inv;
				double r14Inv = r8Inv*r4Inv*r2Inv;
				forceR = 24.0*(2.0*r14Inv - r8Inv);
			}
			else 
			{
				double temp = GetCutoffDistance2() - displacementsSq[jx];
				forceR = -4 * HCLJ_a2 * temp + 6 * HCLJ_a3 * temp * temp;	
			}
			// these are the only two cases since we've already checked that
			// we're inside the outer cutoff
			// Note that the displacements given by NeighborsHalf are of
			// particles j wrt particles i
			for (coord=0; coord<DIMENSION; coord++)
			{
				forceI[coord] += -displacements[jx][coord]*forceR;
				forceJ[coord] += +displacements[jx][coord]*forceR;
			}
			
		}	// Finished with atom j, given by index jx
	} // end of i loop
}



double HolianCutLennardJonesPotential::CalculateEnergy(ListOfAtoms *atoms)
{
	double potential = 0.0;
	int numberHalfNeighbors;
	int i,jx;
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	int nAtoms = atoms->GetNumber();
	for(i=0;i<nAtoms;i++)
    {
		int spacesLeftInArrays = nPairs;
		bool arrayNotTooFull = neighborLocator->HalfNeighbors(i,neighbors,
			numberHalfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		while(!arrayNotTooFull)
		{
			AllocateForNeighborLocator(2*nPairs);
			spacesLeftInArrays = nPairs; // nPairs has been doubled
			arrayNotTooFull = neighborLocator->HalfNeighbors(i,neighbors,
				numberHalfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		}
		for(jx=0;jx<numberHalfNeighbors;jx++)
		{
			double pairEnergy;
			
			if(displacementsSq[jx] < r2Spl)
			{
				double r2inv = 1.0/displacementsSq[jx];
				double r4inv = r2inv*r2inv;
				double r6inv = r4inv*r2inv;
				double r12inv = r6inv*r6inv;
				pairEnergy = 4*(r12inv - r6inv);
			}
			else
			{
				double temp = GetCutoffDistance2() - displacementsSq[jx];
				pairEnergy = - HCLJ_a2*temp*temp + HCLJ_a3*temp*temp*temp;
			}
			if(neighbors[jx] >= nAtoms) // means this pair is between a real and a ghost atom
				pairEnergy /= 2;
			potential += pairEnergy;
		}	// Finished with atom j, given by index jx
    }
	return potential;
}


double HolianCutLennardJonesPotential::CalculateEnergyAndForces(
										ListOfAtoms *atoms,
										DMArray<double, DIMENSION> *forces)
{
	double potential = 0.0;
	int i,j,jx,coord;
	int numberHalfNeighbors;

	// zero the forces:
	forces->Zero();

	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);

	int nAtoms = atoms->GetNumber();
	for(i=0;i<nAtoms;i++)
	{
		int spacesLeftInArrays = nPairs;
		bool arrayNotTooFull = neighborLocator->HalfNeighbors(i,neighbors,
			numberHalfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		while(!arrayNotTooFull)
		{
			AllocateForNeighborLocator(2*nPairs);
			spacesLeftInArrays = nPairs; // nPairs has been doubled
			arrayNotTooFull = neighborLocator->HalfNeighbors(i,neighbors,
				numberHalfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		}    
		for(jx=0;jx<numberHalfNeighbors;jx++)
		{
			j=neighbors[jx];
			double forceR;

			if (displacementsSq[jx] < r2Spl) 
			{
				double r2inv = 1.0/ displacementsSq[jx];
				double r6inv = r2inv*r2inv*r2inv;
				double r8inv = r2inv*r6inv;
				double r12inv = r6inv*r6inv;
				double r14inv = r6inv*r6inv*r2inv;		
				forceR = 24.0*(2.0*r14inv - r8inv); // note factor of 4 replaced
				potential += 4*(r12inv - r6inv);
			}
			else
			{
				double temp = GetCutoffDistance2() - displacementsSq[jx];
				potential += - HCLJ_a2*temp*temp + HCLJ_a3*temp*temp*temp;
				forceR = -4*HCLJ_a2*temp + 6*HCLJ_a3*temp*temp;	
			}
			// these are the only two cases since we've already checked that we're inside the outer cutoff
			
			// Note that the displacements given by NeighborsHalf are of particles j wrt particles i
			// (position i is center of universe, so displacements are rj - ri)
			for (coord=0; coord<DIMENSION; coord++)
			{
				(*forces)[i][coord] += -displacements[jx][coord]*forceR;
				(*forces)[j][coord] += +displacements[jx][coord]*forceR;
			}
			
			
		}	// Finished with atom j, given by index jx
    } // end of i loop
	return potential;
}



void HolianCutLennardJonesPotential::StressTensorKernel(ListOfAtoms *atoms, int i, double stress[(DIMENSION+1)*DIMENSION/2])
{
	int j,jx;
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	int spacesLeftInArrays = nPairs, numberOfNeighbors,nAtoms = atoms->GetNumber();
	bool arrayNotTooFull = neighborLocator->Neighbors(i,neighbors, numberOfNeighbors, displacements,displacementsSq,spacesLeftInArrays);

	while(!arrayNotTooFull)
	{
		AllocateForNeighborLocator(2*nPairs);
		spacesLeftInArrays = nPairs; // nPairs has been doubled
		arrayNotTooFull = neighborLocator->Neighbors(i,neighbors, numberOfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
	}
	ASSERT(arrayNotTooFull);

	for(jx=0;jx<numberOfNeighbors;jx++)
	{
		j=neighbors[jx];
		double forceR;
		if (displacementsSq[jx] < r2Spl) 
		{
			double r2Inv = 1.0/ displacementsSq[jx];
			double r4Inv = r2Inv*r2Inv;
			double r8Inv = r4Inv*r4Inv;
			double r14Inv = r8Inv*r4Inv*r2Inv;
			forceR = 24.0*(2.0*r14Inv - r8Inv);
		}
		else
		{
			double temp = GetCutoffDistance2() - displacementsSq[jx];
			forceR = -4 * HCLJ_a2*temp + 6 * HCLJ_a3 * temp * temp;	
		}	
		
		for( int alpha = 0; alpha < DIMENSION; ++alpha )
			for( int beta = alpha; beta < DIMENSION; ++beta )
			{
				const int ii = stresscomp[alpha][beta];
				stress[ii] += - 0.5 * forceR*displacements[jx][alpha]*displacements[jx][beta];
			}
	}

}



////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////

void HolianCutLennardJonesPotential::Save(DMWriter *writer) const
{
	Potential::Save(writer);

	writer->PutDouble(HCLJ_rSpl, "HCLJ_rSpl");
	writer->PutDouble(HCLJ_a2, "HCLJ_a2");
	writer->PutDouble(HCLJ_a3, "HCLJ_a3");
	writer->PutDouble(HCLJ_cutoffDistance, "HCLJ_cutoffDistance");
}

void HolianCutLennardJonesPotential::Load(DMReader *reader)
{
	Potential::Load(reader);

	// We don't change these parameters. We're only saving them in case 
	// somebody does in the future, and in the future after that wants to 
	// re examine old data in which case 

	CheckPotentialParameter(reader, HCLJ_rSpl, "HCLJ_rSpl");
	CheckPotentialParameter(reader, HCLJ_a2, "HCLJ_a2");
	CheckPotentialParameter(reader, HCLJ_a3, "HCLJ_a3");
	// the following variable should really be taken care by the base class
	// function, which deals with the square of the cutoff
	CheckPotentialParameter(reader, HCLJ_cutoffDistance, "HCLJ_cutoffDistance");
}
