// Johnson1964IronPotential.cpp: implementation of the Johnson1964IronPotential class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
#include "DigitalMaterial.h"

#include "Johnson1964IronPotential.h"
#include "ListOfAtoms.h"
#include "DMArray.h"
#include "IndexError.h"
#include "NotEnoughMemory.h"
#include "NeighborLocator.h"
#include "SerialBuilder.h"


#include "DMLastHeader.h"

static SerialBuilder<Johnson1964IronPotential> *_J1964creator = 
    new SerialBuilder<Johnson1964IronPotential>("Johnson1964IronPotential");

const std::string Johnson1964IronPotential::className = std::string("Johnson1964IronPotential");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


Johnson1964IronPotential::Johnson1964IronPotential(): J1964_param1(2.195976),
													  J1964_param2(3.097910),
													  J1964_param3(2.704060),
													  J1964_param4(7.436448),
													  J1964_param5(0.639230),
													  J1964_param6(3.115829),
													  J1964_param7(0.477871),
													  J1964_param8(1.581570),
													  J1964_param9(1.115035),
													  J1964_param10(3.066403),
													  J1964_param11(0.466892),
													  J1964_param12(1.547967),
													  cutoffDistance(3.44),
													  innerCutoff1Sq(2.4*2.4),
													  innerCutoff2Sq(3.0*3.0)
{
	SetCutoffDistance(cutoffDistance);

	
	nPairs = 0;

	neighbors		= NULL;
	displacements	= NULL;
	displacementsSq = NULL;

	AllocateForNeighborLocator(150); // a rough estimate of the expected number of neighbors
}

Johnson1964IronPotential::~Johnson1964IronPotential()
{
	if(neighbors)
	{
		delete [] neighbors;
		delete [] displacements;
		delete [] displacementsSq;
	}
}

void Johnson1964IronPotential::AllocateForNeighborLocator(int nPairs)
{
// nPairs is playing the role of EMTPotential's arrayLim here
	this->nPairs = nPairs;
	
	if(neighbors) 
	{	
		delete [] neighbors;
		delete [] displacements;
		delete [] displacementsSq;
	}

	if(nPairs > 0)
	{
		neighbors       = new (std::nothrow) int[nPairs];
		displacements   = new (std::nothrow) double[nPairs][DIMENSION];
		displacementsSq = new (std::nothrow) double[nPairs];
		
		ASSERT(neighbors);
		ASSERT(displacements);
		ASSERT(displacementsSq);
		
		if(0 == neighbors)
			throw NotEnoughMemory("Johnson1964IronPotential",
					"AllocateForNeighborLocator",
					SprintString("Problem allocating %d doubles", nPairs));
		if(0 == displacements)
			throw NotEnoughMemory("Johnson1964IronPotential",
					"AllocateForNeighborLocator",
					SprintString("Problem allocating %d doubles", nPairs*DIMENSION));
		if(0 == displacementsSq)
			throw NotEnoughMemory("Johnson1964IronPotential",
					"AllocateForNeighborLocator",
					SprintString("Problem allocating %d integers", nPairs));
	}
	else
	{
		neighbors       = NULL;
		displacements   = NULL;
		displacementsSq = NULL;
	}
}

void Johnson1964IronPotential::CalculateForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces)
{
	int i,coord,jx;
	int j;
	int numberHalfNeighbors;
	forces->Zero();
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	double *forceI, *forceJ;
	int nAtoms = atoms->GetNumber();
	double dummyForce[DIMENSION]; // where we send the force contribution to a ghost atom
	for(i=0;i<nAtoms;i++)
    {	
		int spacesLeftInArrays = nPairs;
		bool arrayNotTooFull = neighborLocator->HalfNeighbors(i,neighbors,
			numberHalfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		while(!arrayNotTooFull)
		{
			AllocateForNeighborLocator(2*nPairs);
			spacesLeftInArrays = nPairs; // nPairs has been doubled
			arrayNotTooFull = neighborLocator->HalfNeighbors(i,neighbors,
				numberHalfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		}
		
		forceI = (*forces)[i];
		for(jx=0;jx<numberHalfNeighbors;jx++)
		{
			j=neighbors[jx];
			if(j<nAtoms) forceJ = (*forces)[j];
			else forceJ = dummyForce;
			double distance = sqrt(displacementsSq[jx]);
			double forceR;


			if(displacementsSq[jx] < innerCutoff1Sq)
			{
				forceR = (3.*J1964_param1 * (distance - J1964_param2) * 
						  (distance - J1964_param2) - J1964_param3 )/distance;
			}
			else if(displacementsSq[jx] < innerCutoff2Sq)
			{
				forceR = (3.*J1964_param5 * (distance - J1964_param6) * 
						  (distance - J1964_param6) - J1964_param7 )/distance;
			}
			else
			{
				forceR = (3.*J1964_param9 * (distance - J1964_param10) * 
						  (distance - J1964_param10) - J1964_param11 )/distance;
			}


			// Note that the displacements given by NeighborsHalf are of particles j wrt particles i
			for (coord=0; coord<DIMENSION; coord++)
			{
				forceI[coord] += -displacements[jx][coord]*forceR;
				forceJ[coord] += +displacements[jx][coord]*forceR;
			}
			
		}	// Finished with atom j, given by index jx
	} // end of i loop
}

void Johnson1964IronPotential::CalculateForcesBetweenTypes(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces, int type1, int type2) 
// the force array isn't initialized in this function
{
	int i,j,coord,jx;
	int numNeighbors;
	
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	int spacesLeftInArrays;
	double *forceI, *forceJ;
	int nAtoms1 = atoms->GetLeaf(type1)->GetNumber();
	for(i=0;i<nAtoms1;i++)
    {	
		spacesLeftInArrays = nPairs;
		bool arrayNotTooFull = neighborLocator->Neighbors(i, type1, type2,
			neighbors,numNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		while(!arrayNotTooFull)
		{
			AllocateForNeighborLocator(2*nPairs);
			spacesLeftInArrays = nPairs; // nPairs has been doubled
			arrayNotTooFull = neighborLocator->Neighbors(i, type1, type2,
				neighbors,numNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		}

		forceI = (*(forces->GetLeaf(type1)))[i];
		for(jx=0;jx<numNeighbors;jx++)
		{
			j=neighbors[jx];
			forceJ = (*forces)[j];
			
			double distance = sqrt(displacementsSq[jx]);
			double forceR;
			
			if(displacementsSq[jx] < innerCutoff1Sq)
			{
				forceR = (3.*J1964_param1 * (distance - J1964_param2) * 
						  (distance - J1964_param2) - J1964_param3 )/distance;
			}
			else if(displacementsSq[jx] < innerCutoff2Sq)
			{
				forceR = (3.*J1964_param5 * (distance - J1964_param6) * 
						  (distance - J1964_param6) - J1964_param7 )/distance;
			}
			else
			{
				forceR = (3.*J1964_param9 * (distance - J1964_param10) * 
						  (distance - J1964_param10) - J1964_param11 )/distance;
			}

			
			// Note that the displacements given by NeighborsHalf are of particles j wrt particles i
			for (coord=0; coord<DIMENSION; coord++)
			{
				forceI[coord] += -displacements[jx][coord]*forceR;
				forceJ[coord] += +displacements[jx][coord]*forceR;
			}
		}	// Finished with atom j, given by index jx
	} // end of i loop
}


double Johnson1964IronPotential::CalculateEnergy(ListOfAtoms *atoms)
{		
	double potential = 0.0;
	int numberHalfNeighbors;
	int i,jx;
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	int nAtoms = atoms->GetNumber();
	for(i=0;i<nAtoms;i++)
    {
		int spacesLeftInArrays = nPairs;
		bool arrayNotTooFull = neighborLocator->HalfNeighbors(i,neighbors,
			numberHalfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		while(!arrayNotTooFull)
		{
			AllocateForNeighborLocator(2*nPairs);
			spacesLeftInArrays = nPairs; // nPairs has been doubled
			arrayNotTooFull = neighborLocator->HalfNeighbors(i,neighbors,
				numberHalfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		}
		for(jx=0;jx<numberHalfNeighbors;jx++)
		{
			double distance = sqrt(displacementsSq[jx]);
			double pairEnergy;
			
			if(displacementsSq[jx] < innerCutoff1Sq)
			{
				pairEnergy = -J1964_param1 * (distance - J1964_param2) * 
					(distance - J1964_param2) * (distance - J1964_param2) + 
					J1964_param3*distance - J1964_param4;
			}
			else if(displacementsSq[jx] < innerCutoff2Sq)
			{
			  pairEnergy = -J1964_param5 * (distance - J1964_param6) * 
				  (distance - J1964_param6) * (distance - J1964_param6) +
				  J1964_param7*distance - J1964_param8;
			}
			else
			{
				pairEnergy = -J1964_param9 * (distance - J1964_param10) *
					(distance - J1964_param10) * (distance - J1964_param10) +
					J1964_param11*distance - J1964_param12;
			}

			if(neighbors[jx] >= nAtoms) // means this pair is between a real and a ghost atom
				pairEnergy /= 2;
			potential += pairEnergy;
		}	// Finished with atom j, given by index jx
    }
	return potential;
}

double Johnson1964IronPotential::CalculateEnergyAndForces(ListOfAtoms *atoms,DMArray<double, DIMENSION> *forces)
{
	double potential = 0.0;
	int i,j,jx,coord;
	int numberHalfNeighbors;
	// zero the forces:
	forces->Zero();
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	int nAtoms = atoms->GetNumber();
	for(i=0;i<nAtoms;i++)
	{
		int spacesLeftInArrays = nPairs;
		bool arrayNotTooFull = neighborLocator->HalfNeighbors(i,neighbors,
			numberHalfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		while(!arrayNotTooFull)
		{
			AllocateForNeighborLocator(2*nPairs);
			spacesLeftInArrays = nPairs; // nPairs has been doubled
			arrayNotTooFull = neighborLocator->HalfNeighbors(i,neighbors,
				numberHalfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		}		
		for(jx=0;jx<numberHalfNeighbors;jx++)
		{
			j=neighbors[jx];
			double forceR,pairEnergy;
			double distance = sqrt(displacementsSq[jx]);
	   
			if(displacementsSq[jx] < innerCutoff1Sq)
			{
				forceR = (3.*J1964_param1 * (distance - J1964_param2) * 
						  (distance - J1964_param2) - J1964_param3 )/distance;
		
				pairEnergy = -J1964_param1 * (distance - J1964_param2) * 
					(distance - J1964_param2) * (distance - J1964_param2) + 
					J1964_param3*distance - J1964_param4;		
			}
			else if(displacementsSq[jx] < innerCutoff2Sq)
			{
				forceR = (3.*J1964_param5 * (distance - J1964_param6) * 
						  (distance - J1964_param6) - J1964_param7 )/distance;

				pairEnergy = -J1964_param5 * (distance - J1964_param6) * 
				  (distance - J1964_param6) * (distance - J1964_param6) +
				  J1964_param7*distance - J1964_param8;
			}
			else
			{
				forceR = (3.*J1964_param9 * (distance - J1964_param10) * 
						  (distance - J1964_param10) - J1964_param11 )/distance;
				pairEnergy = -J1964_param9 * (distance - J1964_param10) *
					(distance - J1964_param10) * (distance - J1964_param10) +
					J1964_param11*distance - J1964_param12;
			}
			
			
			// Note that the displacements given by NeighborsHalf are of particles j wrt particles i
			// (position i is center of universe, so displacements are rj - ri)
			for (coord=0; coord<DIMENSION; coord++)
			{
				(*forces)[i][coord] += -displacements[jx][coord]*forceR;
				(*forces)[j][coord] += +displacements[jx][coord]*forceR;
			}
			
		}	// Finished with atom j, given by index jx
	} // end of i loop
	return potential;
}

double Johnson1964IronPotential::CalculateAtomicEnergy(ListOfAtoms *atoms, int i)
{  // here the energy of an atom is half the energy in each of its bonds.
    if(i >= atoms->GetNumber() || i<0)
		throw IndexError("Johnson1964IronPotential","CalculateAtomicEnergy",
						 "No such atom.");
	double potential = 0.0;
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	int numberOfNeighbors;
	int spacesLeftInArrays = nPairs;
	bool arrayNotTooFull = neighborLocator->Neighbors(i,neighbors,numberOfNeighbors, 
					       displacements,displacementsSq,spacesLeftInArrays);
	while(!arrayNotTooFull)
	{
		AllocateForNeighborLocator(2*nPairs);
		spacesLeftInArrays = nPairs; // nPairs has been doubled
		arrayNotTooFull = neighborLocator->Neighbors(i,neighbors,
			numberOfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
	}
	
	for(int jx=0;jx<numberOfNeighbors;jx++)
	{
		double distance = sqrt(displacementsSq[jx]);

		if(displacementsSq[jx] < innerCutoff1Sq)
		{
			potential += -J1964_param1 * (distance - J1964_param2) * 
				(distance - J1964_param2) * (distance - J1964_param2) + 
				J1964_param3*distance - J1964_param4;
		}
		else if(displacementsSq[jx] < innerCutoff2Sq)
		{
			potential += -J1964_param5 * (distance - J1964_param6) * 
				(distance - J1964_param6) * (distance - J1964_param6) +
				J1964_param7*distance - J1964_param8;
		}
		else
		{
			potential += -J1964_param9 * (distance - J1964_param10) *
				(distance - J1964_param10) * (distance - J1964_param10) +
				J1964_param11*distance - J1964_param12;
		}
		
	}
	return potential/2.;   // since we have split energy of bonds among the atoms.
}

////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////

void Johnson1964IronPotential::Save(DMWriter *writer) const
{
	Potential::Save(writer);
	
	writer->PutDouble(innerCutoff1Sq,"innerCutoff1Sq");
	writer->PutDouble(innerCutoff2Sq,"innerCutoff2Sq");
	writer->PutDouble(J1964_param1, "J1964_para1");
	writer->PutDouble(J1964_param2, "J1964_para2");
	writer->PutDouble(J1964_param3, "J1964_para3");
	writer->PutDouble(J1964_param4, "J1964_para4");
	writer->PutDouble(J1964_param5, "J1964_para5");
	writer->PutDouble(J1964_param6, "J1964_para6");
	writer->PutDouble(J1964_param7, "J1964_para7");
	writer->PutDouble(J1964_param8, "J1964_para8");
	writer->PutDouble(J1964_param9, "J1964_para9");
	writer->PutDouble(J1964_param10, "J1964_para10");
	writer->PutDouble(J1964_param11, "J1964_para11");
	writer->PutDouble(J1964_param12, "J1964_para12");

}

void Johnson1964IronPotential::Load(DMReader *reader)
{
	Potential::Load(reader);

	CheckPotentialParameter(reader, J1964_param1, "J1964_para1");
	CheckPotentialParameter(reader, J1964_param2, "J1964_para2");
	CheckPotentialParameter(reader, J1964_param3, "J1964_para3");
	CheckPotentialParameter(reader, J1964_param4, "J1964_para4");
	CheckPotentialParameter(reader, J1964_param5, "J1964_para5");
	CheckPotentialParameter(reader, J1964_param6, "J1964_para6");
	CheckPotentialParameter(reader, J1964_param7, "J1964_para7");
	CheckPotentialParameter(reader, J1964_param8, "J1964_para8");
	CheckPotentialParameter(reader, J1964_param9, "J1964_para9");
	CheckPotentialParameter(reader, J1964_param10, "J1964_para10");
	CheckPotentialParameter(reader, J1964_param11, "J1964_para11");
	CheckPotentialParameter(reader, J1964_param12, "J1964_para12");
}

