// MEAMPotential.h: interface for the MEAMPotential class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef MEAMPOTENTIAL_H
#define MEAMPOTENTIAL_H

#include "DMFirstHeader.h"

#include "DigitalMaterial.h"
#include "Potential.h"
#include <string>
#include <map>

//! Modified Embedded Atom Potential, after Baskes, Phys. Rev. B 46, 2727 (1992)
//  and Lee and Baskes, Phys. Rev. B 62, 8564 (2000)


/*

  My implementation of Mike Baskes's MEAM potential. It has two major additions
  compared to the EAM: angular terms in the density (rather than just using
  spherically averaged aotmic densities) which correspond to s, p, d, f
  electron densities; and a 3-body screening, such that two atoms even if 
  within the standard 2-body cutoff distance, will not see each other if a
  third atom is between them. The amount by which the third atom screens the 
  other two depends on the geometry.

  Incompletenesses:

  (1) Can only right now, do one element at a time. I will generalize this
  if and when necessary. Main things this will involve: arrays for all 
  parameters, thinking baout how to loop over types (leaves of LOA). In certain
  places you need to do averages e.g. averages over the t-parameters of the
  neighboring atoms contributing to the density at a
  given embedding location
  
  (2) Not optimized for speed. Several parts could be probably sped up
  by making heavier use of arrays (like EMT does)
 
  \author nbailey
  \ingroup moleculardynamics
  \date May 2002

  
 */

#if DIMENSION == 3


struct MEAM_ParameterList
{
	MEAM_ParameterList( double E0, double R0, double alpha, double A, double beta0, double beta1, double beta2, double beta3, double t0, double t1, double t2, double t3, int Zi, std::string referenceLattice, double rCut, double deltaRcut, double scrnCmin, double scrnCmax):  E0(E0), R0(R0), alpha(alpha), A(A), Zi(Zi), referenceLattice(referenceLattice), rCut(rCut), deltaRcut(deltaRcut), scrnCmin(scrnCmin), scrnCmax(scrnCmax)
		{
			beta[0] = beta0; beta[1] = beta1; beta[2] = beta2; beta[3] = beta3;
			t[0] = t0; t[1] = t1; t[2] = t2; t[3] = t3;
		}
	double E0;
	double R0;
	double alpha;
	double A;
	double beta[4];
	double t[4];
	int Zi;
	std::string referenceLattice;

	double rCut;
	double deltaRcut;
	double scrnCmin;
	double scrnCmax;
};


class MEAMPotential : public Potential
{
public:
	MEAMPotential(MEAM_ParameterList *paramList = 0);
	virtual ~MEAMPotential();
	void CalculateForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);

	// void CalculateForcesBetweenTypes(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces, int type1, int type2) ; // the force array isn't initialized in this function



	double CalculateEnergy(ListOfAtoms *atoms);
	// double CalculateEnergyAndForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);
	double CalculateAtomicEnergy(ListOfAtoms *atoms, int i);
	double GetInnerCutoffDistance() const {return rCut;}

	// void CalculateStressTensor(ListOfAtoms *atoms, int i, double stress[(DIMENSION+1)*DIMENSION/2]);
	// void CalculateStressTensor(ListOfAtoms *atoms, DMArray<double, DIMENSION*(DIMENSION+1)/2> *stress);
	

	// Roughly returns the diameter of the atom
	// For fcc, should return 1/sqrt(2) of the lattice constant, for example
	// Currently used to set up the lattice structure: a good initial number is important
	double GetLengthScale() const
	{
		return 1. / MEAM_invR0; 
	}

	// For each potential, we need some kind of energy scale to set time constants
	// for movers and such
	// It doesn't have to be precisely defined! A crude measure is OK.
	// This should return, for a pair potential V(x) with LengthScale ell, V''(ell)
	double GetEnergyCurvatureScale() const
	{
		return MEAM_E0 * MEAM_invR0 * MEAM_invR0 ;
	}

	
	void Save(DMWriter *writer) const;
	void Load(DMReader *reader);
	//! Get the string name of the class for saving or loading.
	const std::string &GetType() const
	{
		return className;
	}

 protected:
#ifndef SWIG
	//! Do the main work for calculating atomic stress. This is called by the two versions of CalculateStressTensor
	void StressTensorKernel(ListOfAtoms *atoms, int i, double stress[(DIMENSION+1)*DIMENSION/2]);
	//! Allocate the \a neighbors, \a displacements, and \a displacementsSq arrays
	void AllocateForNeighborLocator(int nPairs);
	
	void FreeArrays();
	//! a rough (over-)estimation of the number of neighbors
	int         nPairs;
	//! Squared distances to neighbors of i; filled in by NeighborLocator
	double      *displacementsSq;
	double      (*displacements)[DIMENSION];
	int         *neighbors;

	//! For a given i, screening function for neighbors jx of atom i
	double *screeningFactors;
	//! Derivative of screening funciton wrt position i and all the jx's
	double (*screeningDerivs)[DIMENSION];
	//! Distances to neighbors of i
	double *pairDistance;
	//! Atomic electron densities form neighbors of i, different angular terms
	double (*atomicElecDensity)[4];
	//! array of derivatives of the function used for interpolating screening
	double *kx_scrnDer_over_Sijk;
	//! array of bool to store whether a neighbor contributed to screening
	bool *kx_scrn_contrib;
	//! locations in forces array corresponding to the current atom's neighbors
	double **forcePtr;
	

	// parameters associated with screening
	double rCut;
	double deltaRcut;
	double scrnCmin;
	double scrnCmax;
	// parameters of the MEAM potential itself
	double MEAM_E0;
	double MEAM_invR0;
	double MEAM_alpha;
	double MEAM_A;
	double MEAM_beta[4];
	double MEAM_t[4];
	int MEAM_Zi;
	double zBar;
	std::string referenceLattice;

	

	// screening functions
	double TwoBodyScreen(double rij, double *twoBodyScrnDeriv) const 
		{
			double x = (rCut - rij) / deltaRcut;
			if(x >= 1)
			{
				*twoBodyScrnDeriv = 0.;
				return 1.;
			}
			else if(x <= 0.0)
			{
				*twoBodyScrnDeriv = 0.;
				return 0.;
			}
			else
			{
				double y = (1 - x) * (1 - x) * (1 - x) * (1 - x);
				*twoBodyScrnDeriv = -8 * (1 - y) * (1 - x) * (1 - x) * (1 - x) / deltaRcut;
				return (1 - y) * (1 - y);
			}
		}
	double ThreeBodyScreen(double C, double *threeBodyScrnDeriv) const
		{
			double x = (C - scrnCmin) / (scrnCmax - scrnCmin);
			if(x >= 1)
			{
				*threeBodyScrnDeriv = 0.;
				return 1.;
			}
			else if(x <= 0.0) 
			{
				*threeBodyScrnDeriv = 0.;
				return 0.;
			}
			else
			{
				double y = (1 - x) * (1 - x) * (1 - x) * (1 - x);
				*threeBodyScrnDeriv = 8 * (1 - y) * (1 - x) * (1 - x) * (1 - x) / (scrnCmax - scrnCmin);
				return (1 - y) * (1 - y);
			}
		}
	// derivatives of ... ?


	double AddPartialDensities(double zeroPart, double angularPartSq, double *derivWRTZeroPart, double *derivWRTAngularPartSq)
		{
			// following has the problems when atoms are very close
			// leading tmp to become negative
			/*double tmp = basePart * basePart + angularPartSq;
			  return sqrt(tmp); */

			// with multiple elements, will need to divide
			// by an averaged zeroPart
			double expt = 0.5 * angularPartSq / (zeroPart * zeroPart + TINY);
			double exp_expt = exp(expt);
			
			// do derivatives as well (this function will be more frequently called by CalculateForces; calculating just the energy is rare)
			*derivWRTZeroPart = exp_expt * (1.0 - 2. * expt);
			*derivWRTAngularPartSq = 0.5 * exp_expt/(zeroPart + TINY);

			return zeroPart * exp_expt;
		}
	
	

private:
	void ScreeningAndAtomicDensities(int numberOfNeighbors);
	void CalculatePartialDensities123(int numberOfNeighbors, double partialDensitySq[4], double density1factor[DIMENSION], double density2term1factor[DIMENSION][DIMENSION], double *density2term2, double density3factor[DIMENSION][DIMENSION][DIMENSION]);
	double CalculatePairInteraction(int jx, double *forceFactorPair);


	void CalculateDensityScale()
		{
			double s[3] = {0.,0.,0.};
			if(referenceLattice ==  "diamond")
				s[2] = 32./9.;
			else if (referenceLattice ==  "hcp")
				s[2] = 1./3.;
			else if(referenceLattice == "dimer")
				{  
					s[0] = 1.;
					s[1] = 2./3.;
					s[2] = 1.;
				}
			double dummy1, dummy2;
			this->zBar = AddPartialDensities( MEAM_Zi, s[0] * MEAM_t[1] + s[1] * MEAM_t[2] + s[2] * MEAM_t[3], &dummy1, &dummy2 );
		}
	static const std::string className;	
	const double TINY;
#endif // SWIG 
};

#endif // DIMENSION == 3

#endif // !defined(MEAMPOTENTIAL_H
