// MultiCutLennardJonesPotential.cpp: implementation of the MultiCutLennardJonesPotential class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
#include "DMFirstHeader.h"
#include "DigitalMaterial.h"

#include <string>
#include "MultiCutLennardJonesPotential.h"
#include "ListOfAtoms.h"
#include "DMArray.h"
#include "IndexError.h"
#include "NotEnoughMemory.h"
#include "NeighborLocator.h"
#include "LeafToLeafIterator.h"
#include "SerialBuilder.h"


#include "DMLastHeader.h"

using namespace std;

static SerialBuilder<MultiCutLennardJonesPotential> *_MultiCLJcreator = 
    new SerialBuilder<MultiCutLennardJonesPotential>("MultiCutLennardJonesPotential");

const string MultiCutLennardJonesPotential::className = string("MultiCutLennardJonesPotential");

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


MultiCutLennardJonesPotential::MultiCutLennardJonesPotential() : cut1(2.41308778858241),
													   cut2(2.7),
													   CLJ_alpha(0.0031445385865938453275),
													   CLJ_beta(-0.04678508324116488467),
													   CLJ_gamma(0.01283541378358433),
													   CLJ_delta(-0.000880343880904275)
{
	// cut1 marks the boundary between the standard LJ part and the polynomial part
	// cut2 is the actual cutoff, greater than which separation the potential is zero
	r2cut1 = cut1*cut1; // unscaled by sigma--instead we divide
	// the displacementSq by sigma squared and compare to the unscaled cutoff
	r2cut2 = cut2 * cut2; // unscaled
	SetCutoffDistance(cut2);
	nAtoms = 0;

	displacementsSq = NULL;
	displacements = NULL;
	neighbors = NULL;


	// default: like CutLennardJonesPotential;
	epsilons = NULL;
	sigmas = NULL;
	AllocateParameters(1);
	epsilons[0][0] = sigmas[0][0] = 1.0;
}

MultiCutLennardJonesPotential::~MultiCutLennardJonesPotential()
{
if(neighbors)
	{
		delete [] neighbors;
		delete [] displacements;
		delete [] displacementsSq;
	}
}

void MultiCutLennardJonesPotential::Allocate(int nAtoms)
{
// nAtoms is playing the role of EMTPotential's arrayLim here
	this->nAtoms = nAtoms;	
	
	if(neighbors) 
	{	
		delete [] neighbors;
		delete [] displacements;
		delete [] displacementsSq;
	}

	if(nAtoms > 0)
	{
		neighbors       = new (std::nothrow) int[nAtoms];
		displacements   = new (std::nothrow) double[nAtoms][DIMENSION];
		displacementsSq = new (std::nothrow) double[nAtoms];
		
		ASSERT(neighbors);
		ASSERT(displacements);
		ASSERT(displacementsSq);
		
		if(0 == neighbors)
			throw NotEnoughMemory("MultiCutLennardJonesPotential",
								  "Allocate",
								  "Allocating nAtoms doubles");
		if(0 == displacements)
			throw NotEnoughMemory("MultiCutLennardJonesPotential",
								  "Allocate",
								  "Allocating nAtoms*DIMENSION doubles");
		if(0 == displacementsSq)
			throw NotEnoughMemory("MultiCutLennardJonesPotential",
								  "Allocate",
								  "Allocating nAtoms doubles");
	}
	else
	{
		neighbors       = NULL;
		displacements   = NULL;
		displacementsSq = NULL;
	}
}


void MultiCutLennardJonesPotential::AllocateParameters(int numTypes)
{
	if(epsilons)
	{
		for(int a=0;a<numberOfTypes;++a) // old numberOfTypes !
		{
			delete epsilons[a];
			delete sigmas[a];
		}
		delete epsilons;
		delete sigmas;
	}
	numberOfTypes = numTypes;
	
	epsilons  = new double*[numberOfTypes];
	sigmas    = new double*[numberOfTypes];
	
	for(int a=0;a<numberOfTypes;++a) // old numberOfTypes !
	{
		epsilons[a]  = new double[numberOfTypes];
	    sigmas[a]    = new double[numberOfTypes];
	}

}

void MultiCutLennardJonesPotential::SetSigmas(double **sigma)
{
	double maxSigma = 0.;

	for(int a=0;a<numberOfTypes;++a)
		for(int b=0;b<=a;++b)
		{
			if(sigma[a][b] < 0.)
				throw DMException("MultiCutLennardJonesPotential","SetSigmas","Negative value for sigma");
			
			this->sigmas[a][b] = this->sigmas[b][a]
				= 0.5*(sigma[a][b] + sigma[b][a]);
			if(sigma[a][b] > maxSigma)
				maxSigma = sigma[a][b];	
		}	
	
	SetCutoffDistance(cut2*maxSigma); // for now, using a single cutoff for the NeighborLocator. Note right now there's no way to ensure the NeighborLocator is updated if the sigmas get changed after the NeighborLocator has been created. It's up to the user.
	
}


void MultiCutLennardJonesPotential::SetEpsilons(double **epsilon)
{
	for(int a=0;a<numberOfTypes;++a)
		for(int b=0;b<=a;++b)
			this->epsilons[a][b] = this->epsilons[b][a]
				= 0.5*(epsilon[a][b]+epsilon[b][a]);
	// note that it is symmetrized
}

void MultiCutLennardJonesPotential::CalculateForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces)
{
	int i,coord,jx;
	int j;
	int numberNeighbors;
	bool arrayNotTooFull;
	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber());
	forces->Zero();
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	int spacesLeftInArrays;
	double *forceI, *forceJ;
	double dummyForce[DIMENSION]; // where we send the force contribution to a ghost atom

	// loop over leaves
	int numberOfLeaves = atoms->NumberOfLeaves();
	int typeIdx1, typeIdx2;
	std::map<std::string, int>::iterator foundElement;
	for(int leaf1 = 0; leaf1 < numberOfLeaves; ++leaf1 )
	{
		// use the typename for this leaf to get which element it is
		if((foundElement = elementNames.find(atoms->GetLeaf(leaf1)->GetTypeName())) != elementNames.end())
			typeIdx1 = foundElement->second;
				else typeIdx1 = 0;
		
		for( int leaf2 = 0; leaf2 <= leaf1; ++leaf2 )
		{
			// use the typename for this leaf to get which element it is
			if((foundElement = elementNames.find(atoms->GetLeaf(leaf2)->GetTypeName())) != elementNames.end())
				typeIdx2 = foundElement->second;
			else typeIdx2 = 0;
			int nAtoms1 = atoms->GetLeaf(leaf1)->GetNumber();
			for(i=0;i<nAtoms1;++i)
			{	
				spacesLeftInArrays = nAtoms;
				arrayNotTooFull	= neighborLocator->Neighbors(i, leaf1, leaf2,neighbors,numberNeighbors, 
															 displacements,displacementsSq,spacesLeftInArrays);
				ASSERT(arrayNotTooFull);
	

				forceI = (*forces->GetLeaf(leaf1))[i];
			
				for(jx=0;jx<numberNeighbors;jx++)
				{
					j=neighbors[jx];
					if(j<nAtoms) forceJ = (*forces->GetLeaf(leaf2))[j];
					else forceJ = dummyForce;
					double forceR = 0.;		
				
					// for different leaves double counting is not an issue
					// so all j are valid. For the same leaf only j > i are
					// used. The following expression takes care of both cases
					if(leaf1 != leaf2 || j>i)
					{
						// scale the displacements by sigma squared
						displacementsSq[jx] /= (sigmas[typeIdx1][typeIdx2]*sigmas[typeIdx1][typeIdx2]);
						if (displacementsSq[jx] < r2cut1) 
						{
							double r2Inv = 1.0 / displacementsSq[jx];
							double r4Inv = r2Inv*r2Inv;
							double r8Inv = r4Inv*r4Inv;
							double r14Inv = r8Inv*r4Inv*r2Inv;
							forceR = 24.0*(2.0*r14Inv - r8Inv);// note factor of 4 replaced
						}
						else  if (displacementsSq[jx] < r2cut2)
							// needed because we are using a single cutoff for the NeighborLocator
							forceR = 4*(- CLJ_gamma*2. + - 4.*CLJ_delta*displacementsSq[jx]);
					} // if not within cutoff for that pair, or if leaf1= leaf2 and j<i, then forceR will remain zero.
					
					forceR *= epsilons[typeIdx1][typeIdx2]/(sigmas[typeIdx1][typeIdx2] * sigmas[typeIdx1][typeIdx2]);
					// Note that the displacements given by NeighborsHalf are of particles j wrt particles i
					for (coord=0; coord<DIMENSION; coord++)
					{
						forceI[coord] += -displacements[jx][coord]*forceR;
						forceJ[coord] += +displacements[jx][coord]*forceR;
					}
					
				}	// Finished with atom j, given by index jx
				
			} // finished with atom i
		} // end of loops over leaf2
	} // end loop over leaf1
}

void MultiCutLennardJonesPotential::CalculateForcesBetweenTypes(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces, int leaf1, int leaf2) 
// the force array isn't initialized in this function
{
	int i,j,coord,jx;
	int numNeighbors;
	bool arrayNotTooFull;
	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber());

	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	int spacesLeftInArrays;
	double *forceI, *forceJ;
	int nAtoms1 = atoms->GetLeaf(leaf1)->GetNumber();

	int typeIdx1, typeIdx2;
	std::map<std::string, int>::iterator foundElement;
	// use the typename for the two leaves to get which elements
	if((foundElement = elementNames.find(atoms->GetLeaf(leaf1)->GetTypeName())) != elementNames.end())
		typeIdx1 = foundElement->second;
	else typeIdx1 = 0;
	if((foundElement = elementNames.find(atoms->GetLeaf(leaf2)->GetTypeName())) != elementNames.end())
		typeIdx2 = foundElement->second;
	else typeIdx2 = 0;

	for(i=0;i<nAtoms1;i++)
    {	
		spacesLeftInArrays = nAtoms;
		arrayNotTooFull	= neighborLocator->Neighbors(i, leaf1, leaf2,neighbors,numNeighbors, 
											displacements,displacementsSq,spacesLeftInArrays);
		ASSERT(arrayNotTooFull);

		forceI = (*(forces->GetLeaf(leaf1)))[i];
		for(jx=0;jx<numNeighbors;jx++)
		{
			j=neighbors[jx];
			forceJ = (*forces)[j];
			double forceR = 0.;		
			
			if(leaf1 != leaf2 || j>i)
			{
				// scale the displacements by sigma squared
				displacementsSq[jx] /= (sigmas[typeIdx1][typeIdx2]*sigmas[typeIdx1][typeIdx2]);
				
				if (displacementsSq[jx] < r2cut1) 
				{
					double r2Inv = 1.0/ displacementsSq[jx];
					double r4Inv = r2Inv*r2Inv;
					double r8Inv = r4Inv*r4Inv;
					double r14Inv = r8Inv*r4Inv*r2Inv;
					forceR = 24.0*(2.0*r14Inv - r8Inv);// note factor of 4 replaced
				}
				else if (displacementsSq[jx] < r2cut2)
					forceR = 4*(- CLJ_gamma*2. + - 4.*CLJ_delta*displacementsSq[jx]);
			}

			forceR *= epsilons[typeIdx1][typeIdx2]/(sigmas[typeIdx1][typeIdx2] *
											  sigmas[typeIdx1][typeIdx2]);
			
			// Note that the displacements given by NeighborsHalf are of particles j wrt particles i
			for (coord=0; coord<DIMENSION; coord++)
			{
				forceI[coord] += -displacements[jx][coord]*forceR;
				forceJ[coord] += +displacements[jx][coord]*forceR;
			}
		}	// Finished with atom j, given by index jx
	} // end of i loop
}


double MultiCutLennardJonesPotential::CalculateEnergy(ListOfAtoms *atoms)
{
	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber());
	double potential = 0.0;
	int numberNeighbors;
	bool arrayNotTooFull;
	int i,j,jx;
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	int spacesLeftInArrays;


	// loop over leaves
	int numberOfLeaves = atoms->NumberOfLeaves();
	int typeIdx1, typeIdx2;
	std::map<std::string, int>::iterator foundElement;
	for(int leaf1=0;leaf1<numberOfLeaves;++leaf1)
	{
		// use the typename for the two leaves to get which elements
		if((foundElement = elementNames.find(atoms->GetLeaf(leaf1)->GetTypeName())) != elementNames.end())
			typeIdx1 = foundElement->second;
		else typeIdx1 = 0;
		
		for(int leaf2=0;leaf2<=leaf1;++leaf2)
		{
			if((foundElement = elementNames.find(atoms->GetLeaf(leaf2)->GetTypeName())) != elementNames.end())
				typeIdx2 = foundElement->second;
			else typeIdx2 = 0;
			int nAtoms1 = atoms->GetLeaf(leaf1)->GetNumber();
			for(i=0;i<nAtoms1;i++)
			{	
				spacesLeftInArrays = nAtoms;
				arrayNotTooFull	= neighborLocator->Neighbors(i, leaf1, leaf2,neighbors,numberNeighbors, 
															 displacements,displacementsSq,spacesLeftInArrays);
				ASSERT(arrayNotTooFull);
				
				
				for(jx=0;jx<numberNeighbors;jx++)
				{
					double pairEnergy = 0.;
					j=neighbors[jx];
					
					if(leaf1 != leaf2 || j>i)
					{
						// scale the displacements by sigma squared
						displacementsSq[jx] /= (sigmas[typeIdx1][typeIdx2]*sigmas[typeIdx1][typeIdx2]);
						if(displacementsSq[jx] < r2cut1)
						{
							double r2inv = 1.0/displacementsSq[jx];
							double r4inv = r2inv*r2inv;
							double r6inv = r4inv*r2inv;
							double r12inv = r6inv*r6inv;
							pairEnergy = 4*(r12inv - r6inv + CLJ_alpha);
						}
						else  if (displacementsSq[jx] < r2cut2)
							// needed because we are using a single cutoff for the NeighborLocator
						{
							pairEnergy = 4*(CLJ_beta + CLJ_gamma*displacementsSq[jx] + CLJ_delta*displacementsSq[jx]*displacementsSq[jx]);
						}
						
					} // if not within cutoff for that pair, or if leaf1= leaf2 and j<i, then forceR will remain zero.
					
					if(neighbors[jx] >= nAtoms) // means this pair is between a real and a ghost atom
						pairEnergy /= 2;

					pairEnergy *= epsilons[typeIdx1][typeIdx2];
					potential += pairEnergy;
				}	// Finished with atom j, given by index jx
			} // end of i loop
			
		} // end of leaf1
	} // end of leaf2 loop

	return potential;
}

double MultiCutLennardJonesPotential::CalculateEnergyAndForces(ListOfAtoms *atoms,DMArray<double, DIMENSION> *forces)
{
	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber());
	double potential = 0.0;
	int i,j,jx,coord;
	int numberHalfNeighbors;
	bool arrayNotTooFull;

	// zero the forces:
	forces->Zero();

	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);

	int spacesLeftInArrays;

	// loop over leaves
	int numberOfLeaves = atoms->NumberOfLeaves();
	int typeIdx1, typeIdx2;
	std::map<std::string, int>::iterator foundElement;
	for(int leaf1=0;leaf1<numberOfLeaves;++leaf1)
	{
		// use the typename for the two leaves to get which elements
		if((foundElement = elementNames.find(atoms->GetLeaf(leaf1)->GetTypeName())) != elementNames.end())
			typeIdx1 = foundElement->second;
		else typeIdx1 = 0;

		int nAtoms1 = atoms->GetLeaf(leaf1)->GetNumber();
		for(int leaf2=0;leaf2<=leaf1;++leaf2)
		{
			if((foundElement = elementNames.find(atoms->GetLeaf(leaf2)->GetTypeName())) != elementNames.end())
				typeIdx2 = foundElement->second;
			else typeIdx2 = 0;

			for(i=0;i<nAtoms1;++i)
			{
				spacesLeftInArrays = nAtoms;
				arrayNotTooFull	= neighborLocator->HalfNeighbors(i,neighbors,numberHalfNeighbors, 
																 displacements,displacementsSq,spacesLeftInArrays);
				ASSERT(arrayNotTooFull);
				
				
				for(jx=0;jx<numberHalfNeighbors;jx++)
				{
					j=neighbors[jx];
					double forceR = 0.,pairEnergy = 0.;		
					
					if(leaf1 != leaf2 || j>i)
					{
						// scale the displacements by sigma squared
						displacementsSq[jx] /= (sigmas[typeIdx1][typeIdx2]*sigmas[typeIdx1][typeIdx2]);
						
						if (displacementsSq[jx] < r2cut1) 
						{
							double r2inv = 1.0/ displacementsSq[jx];
							double r6inv = r2inv*r2inv*r2inv;
							double r8inv = r2inv*r6inv;
							double r12inv = r6inv*r6inv;
							double r14inv = r6inv*r6inv*r2inv;		
							forceR = 24.0*(2.0*r14inv - r8inv); // note factor of 4 replaced
							pairEnergy = 4*(r12inv - r6inv + CLJ_alpha);
						}
						else if (displacementsSq[jx] < r2cut2)
						{
							forceR = 4*(- CLJ_gamma*2. + - 4.*CLJ_delta*displacementsSq[jx]);
							pairEnergy = 4*(CLJ_beta + CLJ_gamma*displacementsSq[jx]
											+ CLJ_delta*displacementsSq[jx]*displacementsSq[jx]);
						}
					}
					// Note that the displacements given by NeighborsHalf are of particles j wrt particles i
					// (position i is center of universe, so displacements are rj - ri)
					forceR *= epsilons[typeIdx1][typeIdx2]/(sigmas[typeIdx1][typeIdx2] *
													  sigmas[typeIdx1][typeIdx2]);
					
					
					if(neighbors[jx] >= nAtoms) //pair between real and ghost
						pairEnergy /= 2;
					
					pairEnergy *= epsilons[typeIdx1][typeIdx2];
					potential += pairEnergy;
					
					for (coord=0; coord<DIMENSION; coord++)
					{
						(*forces)[i][coord] += -displacements[jx][coord]*forceR;
						(*forces)[j][coord] += +displacements[jx][coord]*forceR;
					}

					
					
				}	// Finished with atom j, given by index jx
			} // end of i loop
			
		} // end of loop over leaves (leaf1)
	} // end of loop over leaves (leaf2)
	
	return potential;
}

double MultiCutLennardJonesPotential::CalculateAtomicEnergy(ListOfAtoms *atoms, int i)
{  // here the energy of an atom is half the energy in each of its bonds.
    if(i >= atoms->GetNumber() || i<0)
		throw IndexError("MultiCutLennardJonesPotential","CalculateAtomicEnergy",
						 "No such atom.");
	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber());
	double potential = 0.0
		;
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	int spacesLeftInArrays, numberOfNeighbors;
	spacesLeftInArrays = nAtoms;

	bool arrayNotTooFull = neighborLocator->Neighbors(i,neighbors,numberOfNeighbors, 
					       displacements,displacementsSq,spacesLeftInArrays);
	ASSERT(arrayNotTooFull);

	
	// find out what leaf i belongs to
	int leaf1=0,total=0;
	LeafToLeafIterator<ListOfAtoms> iter(atoms);
	while((total+=iter.Current()->GetNumber())<=i)
	{
		++leaf1;
		iter.Next();
	}


	int typeIdx1, typeIdx2;
	std::map<std::string, int>::iterator foundElement;
	// use the typename for the leaves to get which elements
	if((foundElement = elementNames.find(atoms->GetLeaf(leaf1)->GetTypeName())) != elementNames.end())
		typeIdx1 = foundElement->second;
	else typeIdx1 = 0;


	int numberOfLeaves = atoms->NumberOfLeaves();
	for(int leaf2=0;leaf2<numberOfLeaves;++leaf2)
		for(int jx=0;jx<numberOfNeighbors;jx++)
		{
			if((foundElement = elementNames.find(atoms->GetLeaf(leaf2)->GetTypeName())) != elementNames.end())
				typeIdx2 = foundElement->second;
			else typeIdx2 = 0;

			int j=neighbors[jx];
			double pairEnergy = 0.;		
			
			if(leaf1 != leaf2 || j>i)
			{
				// scale the displacements by sigma squared
				displacementsSq[jx] /= (sigmas[typeIdx1][typeIdx2]*sigmas[typeIdx1][typeIdx2]);
				
				if(displacementsSq[jx] < r2cut1)
				{
					double r2inv = 1.0/displacementsSq[jx];
					double r4inv = r2inv*r2inv;
					double r6inv = r4inv*r2inv;
					double r12inv = r6inv*r6inv;
					potential += 4.*(r12inv - r6inv + CLJ_alpha);
				}
				else  if(displacementsSq[jx] < r2cut2)
				{
					pairEnergy= 4.*(CLJ_beta + CLJ_gamma*displacementsSq[jx] + CLJ_delta*displacementsSq[jx]*displacementsSq[jx]);
				} 
				
			} // closes if(leaf1 != leaf2 || j>i)
			pairEnergy *= epsilons[typeIdx1][typeIdx2];
			potential += pairEnergy;
			
			
		} // end loop over jx and over leaf2
	return potential/2.;   // since we have split energy of bonds among the atoms.
}


////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////

void MultiCutLennardJonesPotential::Save(DMWriter *writer) const
{
	Potential::Save(writer);

	writer->PutDouble(cut1, "cut1");
	writer->PutDouble(cut2, "cut2");
	writer->PutDouble(r2cut1, "r2cut1");
	writer->PutDouble(CLJ_alpha, "CLJ_alpha");
	writer->PutDouble(CLJ_beta, "CLJ_beta");
	writer->PutDouble(CLJ_gamma, "CLJ_gamma");
	writer->PutDouble(CLJ_delta, "CLJ_delta");
}

void MultiCutLennardJonesPotential::Load(DMReader *reader)
{
	Potential::Load(reader);

	// as in CutLennardJonesPotential
	CheckPotentialParameter(reader, cut1, "cut1");
	CheckPotentialParameter(reader, cut2, "cut2");
	CheckPotentialParameter(reader, r2cut1, "r2cut1");
	CheckPotentialParameter(reader, CLJ_alpha, "CLJ_alpha");
	CheckPotentialParameter(reader, CLJ_beta, "CLJ_beta");
	CheckPotentialParameter(reader, CLJ_gamma, "CLJ_gamma");
	CheckPotentialParameter(reader, CLJ_delta, "CLJ_delta");

}
