
//     Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
//     Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

//     This library is free software; you can redistribute it and/or modify it
//     under the terms of the GNU Lesser General Public License as published by the
//     Free Software Foundation; either version 2.1 of the License, or (at your
//     option) any later version.

//     This library is distributed in the hope that it will be useful, but WITHOUT
//     ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//     FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
//     for more details.

//     You should have received a copy of the GNU Lesser General Public License
//     along with this library; if not, write to the Free Software Foundation,
//     Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
//     http://www.gnu.org/licenses/lgpl.html

//     Author Contact Information:
//     Jim Sethna      
//     sethna@lassp.cornell.edu  
//     117 Clark Hall
//     Cornell University
//     Ithaca, NY 14850


%{
#include "Potential.h"
#include "CuRbPotential.h"
#include "CutLennardJonesPotential.h"
#include "HolianCutLennardJonesPotential.h"
#include "Johnson1964IronPotential.h"
#include "EMTPotential.h"
#include "EMTDefaultParameterProvider.h"
#include "EMTAtomicUnitParameterProvider.h"
#include "EMT.h"
#include "MEAMPotential.h"
#include "MultiCutLennardJonesPotential.h"
#include "SiliconEDIPPotential.h"
#include "StillingerWeberPotential.h"
#include "GravityPotential.h"
#include "CompositePotential.h"
#include "GranularPotential.h"
#include <string>
%}



%include "Potential.h"
%extend Potential
{
public:
	ArrayDoubleVector* Forces(ListOfAtoms *atoms)
	{
		ArrayDoubleVector *forces = new DMArray<double,DIMENSION>(atoms->GetCartesianPositions());
		self->CalculateForces(atoms, forces);
		atoms->AdjustForces(forces);
		return forces;
	}
	PyObject *CalculateStress(ListOfAtoms *atoms, int i)
	{
		int dims = DIMENSION*(DIMENSION+1)/2;
		double *stress = new double[dims];
		self->CalculateStressTensor(atoms, i, stress);
		PyObject *o = PyArray_FromDimsAndData(1, &dims, 'd', (char *)stress);
		return o;
	}
	void CheckForces(ListOfAtoms *atoms, double epsilon, bool verbose = false)
	{   // a function that checks whether the force is the gradient of the energy
  		DMArray<double, DIMENSION> *forces = new DMArray<double, DIMENSION>(atoms->GetShape());
  		self->CalculateForces(atoms, forces);
		double max = 0;
		int maxAT = -1;
		int maxCoord = -1;
  		for(int at=0; at < atoms->GetNumber(); ++at) 
			for(int coord=0; coord<DIMENSION;++coord)
    		{
     			double e1, e2;
				atoms->IncrementCartesianPosition(at, coord, epsilon);
				e1 = self->CalculateEnergy(atoms);
				atoms->IncrementCartesianPosition(at, coord, -2.*epsilon);
				e2 = self->CalculateEnergy(atoms);
				atoms->IncrementCartesianPosition(at, coord, epsilon);
				double f1 = 0.5*(e2-e1)/epsilon, f2 = (*forces)[at][coord], diff = f1-f2;
				if(verbose) std::cout << at << "," << coord << "\t" << f1 << "\t" << f2 << "\t" << diff << std::endl;
				if(fabs(diff) > max)
				{
					max = fabs(diff);
					maxAT = at;
					maxCoord = coord;
				}
			}
		delete forces;
		std::cout << "Max diff = " << max << " for atom " << maxAT << " along coord " << maxCoord <<  std::endl;
	}
}



class CuRbPotential : public Potential
{	
public:
	CuRbPotential(Potential *EMTptr);
	void SetCuRbInteraction(int CuRb_switch);
	double GetCuRbInverseInteraction(double energy);	
	void SetSurfaceInteractions(int si);
};

class CutLennardJonesPotential: public Potential
{
public:
	CutLennardJonesPotential();
	double GetInnerCutoffDistance() const;
};

class HolianCutLennardJonesPotential: public Potential
{
public:
	HolianCutLennardJonesPotential();
};


class GravityPotential: public Potential
{
public:
	GravityPotential();
	void SetDirection(double direction[DIMENSION]);
	void SetG(double g);
};

class GranularPotential: public Potential
{
public:
	GranularPotential();
	GranularPotential(double pD, double eC, double dP);
};

class CompositePotential: public Potential
{
public:
	CompositePotential();
	CompositePotential(Potential *potential1, Potential *potential2);
};


%include MCLJtypemap.i
%include "MultiCutLennardJonesPotential.h"

#if DIMENSION == 3

class Johnson1964IronPotential: public Potential 
{
public:
	Johnson1964IronPotential();
};

class EMTPotential: public Potential
{
public:
    EMTPotential(char *element = "Cu");

};

%nodefault EMTParameterProvider;
class EMTParameterProvider{};

class EMTDefaultParameterProvider : public EMTParameterProvider
{
public:
	EMTDefaultParameterProvider();
};

class EMTAtomicUnitParameterProvider : public EMTDefaultParameterProvider
{
public:
	EMTAtomicUnitParameterProvider();
};

class EMT : public Potential
{
public:
	EMT(ListOfAtoms *atoms, EMTParameterProvider *prov = 0);
	virtual ~EMT();
	ArrayDoubleVector *AllocateForces();
	const ArrayDouble *DebugSigma1(int n=0);
	const ArrayDouble *DebugSigma2(int n=0);	
	const ArrayDouble *GetEc();
	const ArrayDouble *GetEas();
	const ArrayDouble *GetRadius();
	const ArrayDouble *GetdEds();
	const double GetCutoffDistance();
};

%inline %{
EMT *newEMT_Atomic(ListOfAtoms *atoms)
{
	EMTAtomicUnitParameterProvider *param = new EMTAtomicUnitParameterProvider();
	return new EMT(atoms, param);
}
%}



%include "MEAMPotential.h"



class StillingerWeberPotential: public Potential
{
public:
	StillingerWeberPotential(double BRITTLE = 1., std::string unitString = "natural");
	void CalculateHessianMatrix(ListOfAtoms *atoms, double *matrix);

};

class SiliconEDIPPotential: public Potential
{
public:
	SiliconEDIPPotential();
};


#endif // DIMENSION == 3

