// ExternalStress.cpp: implementation of the ExternalStress class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"
#include <vector>
#include <set>

#include "DigitalMaterial.h"
#include "ExternalStress.h"
#include "ListOfAtoms.h"
#include "DMException.h"
#include "lapack.h"


#include "DMLastHeader.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


ExternalStress::ExternalStress(double *C,double *origin, ListOfAtoms *undeformed)
{
	for( int a = 0; a < 3; ++a )
		this->C[a] = C[a]; // Elastic constants
	
	for( int coord1 = 0; coord1 < DIMENSION; ++coord1 )
    {
		for( int coord2 = 0; coord2 < DIMENSION; ++coord2 )
		{
			stress[coord1][coord2] = strain[coord1][coord2] = 0.;
		}
		this->origin[coord1] = origin[coord1];
    }
	this->undeformed = undeformed;
	
	
	
	
}


ExternalStress::~ExternalStress()
{

}

void ExternalStress::SetStress(double stress[DIMENSION][DIMENSION])
{
    int a,b,c,coord1,coord2,stringLength=1;
    double stressVector[N_CPTS];
    double elasticMatrix[N_CPTS * N_CPTS];
    char uplo='U';
    int matrixSize = N_CPTS,nrhs = 1,ipiv[N_CPTS],info, lwork=N_CPTS*10;
    double work[N_CPTS*10];
	
    for( coord1 = 0; coord1 < DIMENSION; ++coord1)
    {
		for(coord2=0;coord2<DIMENSION;++coord2)
		{
			this->stress[coord1][coord2] = stress[coord1][coord2];
		}
		stressVector[coord1] = stress[coord1][coord1]; 
    }    
    for( a = DIMENSION; a < N_CPTS; ++a)
    {
		// translate the off diagonal stress components to the one-index form. Using definitions given in Hirth and Lothe
		b = a + 1; while(b>=DIMENSION) b-= DIMENSION; // have to subtract twice when a=5
		c = b + 1; if(c>=DIMENSION) c-= DIMENSION;
		stressVector[a] = stress[b][c];
    }    
    

    // use elasticity to turn stress matrix into strain matrix
    
    // create the elastic matrix---must do each time stress is changed since it will be overwritten when
    // we solve for the strains
    for( a = 0; a < N_CPTS; ++a )
		for( b = 0; b < N_CPTS; ++b )
		{
			if( a < DIMENSION && b < DIMENSION)
				elasticMatrix[a*N_CPTS + b] = C[1]; // ie C_12 (will do diagonal elements below)
			else
				elasticMatrix[a*N_CPTS + b] = 0.;
		}
    
    for(a=0;a<DIMENSION;++a)
		elasticMatrix[a*(N_CPTS+1)] = C[0]; // ie C_11
    for(a=DIMENSION;a<N_CPTS;++a)
		elasticMatrix[a*(N_CPTS+1)] = C[2]; // ie C_44
    
    // solve the linear system to get velocity increments
	

	dsysv_(&uplo, &matrixSize, &nrhs,elasticMatrix, &matrixSize,ipiv,stressVector, &matrixSize, work, &lwork,&info);

    // stressVector nows contains the ENGINEERING strains (extra factor of two on the off-diagonal elements)
    for(a=0;a<DIMENSION;++a)
		strain[a][a] = stressVector[a];
    for(a=DIMENSION;a<N_CPTS;++a)
    {
		b = a + 1; if(b>=DIMENSION) b-= DIMENSION;
		c = b + 1; if(c>=DIMENSION) c-= DIMENSION;
		strain[b][c] = strain[c][b] = 0.5*stressVector[a];
    }

}

void ExternalStress::SetStrain(double strain[DIMENSION][DIMENSION])
{
	for( int coord1 = 0; coord1 < DIMENSION; ++coord1 )
		for( int coord2 = 0; coord2 < DIMENSION; ++coord2 )
		{
			this->strain[coord1][coord2] = strain[coord1][coord2];
		}
}


void ExternalStress::ElasticDisplacement(const double position[DIMENSION], double displacement[DIMENSION])
{

	for(int coord1=0;coord1<DIMENSION;++coord1)
		for(int coord2=0;coord2<DIMENSION;++coord2)
			displacement[coord1] += strain[coord1][coord2]*(position[coord2]-origin[coord2]);

}

void ExternalStress::GradientWRTPosition(const double position[DIMENSION],  double gradientWRTPosition[DIMENSION][DIMENSION])
{
	for(int coord1=0;coord1<DIMENSION;++coord1)
		for(int coord2=0;coord2<DIMENSION;++coord2)
			gradientWRTPosition[coord1][coord2] += strain[coord1][coord2];

}
