// SphereAtomsLocator.cpp: implementation of the SphereAtomsLocator class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DigitalMaterial.h"
#include "SphereAtomsLocator.h"
#include "ListOfAtoms.h"
#include "NeighborLocator.h"

#include "DMLastHeader.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

SphereAtomsLocator::SphereAtomsLocator(double center[DIMENSION], double radius)
{
	for( int coord = 0; coord < DIMENSION; ++coord )
		this->center[coord] = center[coord];
	this->radius = radius;

	neighbors       = 0;
	displacements   = 0;
	displacementsSq = 0;
}

SphereAtomsLocator::~SphereAtomsLocator()
{

}

void SphereAtomsLocator::AllocateForNeighborLocator(int nSpaces)
{
// nSpaces is playing the role of EMTPotential's arrayLim here
	this->nSpaces = nSpaces;
	
	if(neighbors) 
	{	
		delete [] neighbors;
		delete [] displacements;
		delete [] displacementsSq;
	}

	if(nSpaces > 0)
	{
		neighbors       = new (std::nothrow) int[nSpaces];
		displacements   = new (std::nothrow) double[nSpaces][DIMENSION];
		displacementsSq = new (std::nothrow) double[nSpaces];
		
		ASSERT(neighbors);
		ASSERT(displacements);
		ASSERT(displacementsSq);
	}
	else
	{
		neighbors       = 0;
		displacements   = 0;
		displacementsSq = 0;
	}
}



const std::vector<int> &SphereAtomsLocator::GetAtomNumbers(ListOfAtoms *atoms)
{
	listOfAtomNumbers.clear();
	
	int  numNeighbors, spacesLeftInArrays;
    bool arrayNotTooFull;
	int nAtoms = atoms->GetNumber();
	AllocateForNeighborLocator(nAtoms);	
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();


	spacesLeftInArrays = nSpaces;     
	arrayNotTooFull = neighborLocator->Neighbors(center, neighbors, numNeighbors, displacements, displacementsSq, spacesLeftInArrays, radius);
	
	while(!arrayNotTooFull)
	{
		AllocateForNeighborLocator(2*nSpaces);
		spacesLeftInArrays = nSpaces; // nSpaces has been doubled
		arrayNotTooFull = neighborLocator->Neighbors(center,neighbors, numNeighbors, displacements, displacementsSq, spacesLeftInArrays, radius);
	}
	
	for(int i = 0; i < numNeighbors; ++i)
	{
		listOfAtomNumbers.push_back(neighbors[i]);
	}

	return listOfAtomNumbers;
}

bool SphereAtomsLocator::Inside(const double *position)
{
	double rSq = 0.;
	for( int coord = 0; coord < DIMENSION; ++coord )
	{
		double diff = position[coord] - center[coord];
		rSq += diff * diff;
	}

	return (rSq < radius * radius);
}
