// PlotAtomsObserver.cpp: implementation of the PlotAtomsObserver class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"
#include <string>
#include "DigitalMaterial.h"

#include "PlotAtomsObserver.h"
#include "ListOfAtoms.h"
#include "SortAtoms.h"
#include "DMArray.h"
#include "LeafToLeafIterator.h"


#include "DMLastHeader.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

/*!	Opens a pipe to the PlotAtoms program and sets initial parameters.

	\param xMin left edge of box to display
	\param xMax right edge of box to display
	\param yMin bottom of box to display
	\param yMax top of box to display
	\param radius Radius of each atom.
	\param defaultColor is the color in which atoms are plotted if no other color information is given: 0 for white, otherwise black
	\param colorMethod class which provides a coloring function
	*/
PlotAtomsObserver::PlotAtomsObserver(double xMin, double xMax, double yMin, double yMax, int radius, int defaultColor, ColorMethod *colorMethod) 
{
	char command[1024];
	nColors = 20; // This seems to be the maximum allowed for PlotAtoms

	// there's got to be a better way of passing in a default color but I can't think of an easy one right now.

	if(defaultColor == 0) // default color is white
		sprintf(command, "PlotAtoms  -xmin %lf -xmax %lf -ymin %lf -ymax %lf -buffer -natoms %d -radius %d -color rgb:ff/ff/ff -c1 rgb:ff/ff/e0 -c2 rgb:ff/ff/c0 -c3 rgb:ff/ff/a0 -c4 rgb:ff/ff/80 -c5 rgb:ff/ff/60 -c6 rgb:ff/ff/40 -c7 rgb:ff/ff/20 -c8 rgb:ff/e0/ff -c9 rgb:ff/c0/ff -c10 rgb:ff/a0/ff -c11 rgb:ff/80/ff -c12 rgb:ff/60/ff -c13 rgb:ff/40/ff -c14 rgb:ff/20/ff -c15 rgb:e0/ff/ff -c16 rgb:c0/ff/ff -c17 rgb:a0/ff/ff -c18 rgb:80/ff/ff -c19 rgb:60/ff/ff -c20 rgb:40/ff/ff", xMin, xMax, yMin, yMax, nColors,radius);
	else // default is black
		sprintf(command, "PlotAtoms  -xmin %lf -xmax %lf -ymin %lf -ymax %lf -buffer -natoms %d -radius %d -color rgb:00/00/00 -c1 rgb:ff/ff/e0 -c2 rgb:ff/ff/c0 -c3 rgb:ff/ff/a0 -c4 rgb:ff/ff/80 -c5 rgb:ff/ff/60 -c6 rgb:ff/ff/40 -c7 rgb:ff/ff/20 -c8 rgb:ff/e0/ff -c9 rgb:ff/c0/ff -c10 rgb:ff/a0/ff -c11 rgb:ff/80/ff -c12 rgb:ff/60/ff -c13 rgb:ff/40/ff -c14 rgb:ff/20/ff -c15 rgb:e0/ff/ff -c16 rgb:c0/ff/ff -c17 rgb:a0/ff/ff -c18 rgb:80/ff/ff -c19 rgb:60/ff/ff -c20 rgb:40/ff/ff", xMin, xMax, yMin, yMax, nColors,radius);
	
	pipe_file = popen(command, "w");
	
	this->colorMethod = colorMethod;
	this->colorScale = 1.;
	this->colorScaleSetByUser = false;
	this->lengthScale = 1.;

	for( int a = 0; a < DIMENSION; ++a )
	{
		for( int b = 0; b < DIMENSION; ++b )
			this->rotationMatrix[a][b] = 0.;
		this->rotationMatrix[a][a] = 1.0;
	}	
	UpdateRotateScale();
}

/*!	Closes the pipe to the PlotAtoms program.
	*/
PlotAtomsObserver::~PlotAtomsObserver()
{
	// Close PlotAtoms Window (in version in jmt28 directory)...
	fprintf( pipe_file, "\nq\n" );
	pclose(pipe_file);

}


/*!
  
  Should we enforce the deterinant being one?

 */

void PlotAtomsObserver::SetRotationMatrix(double rotationMatrix[DIMENSION][DIMENSION])
{

	for( int a = 0; a < DIMENSION; ++a )
	{
		for( int b = 0; b < DIMENSION; ++b )
			this->rotationMatrix[a][b] = rotationMatrix[a][b];
	}	
	UpdateRotateScale(); 
}



void PlotAtomsObserver::UpdateRotateScale()
{
	for( int a = 0; a < DIMENSION; ++a )
	{
		for( int b = 0; b < DIMENSION; ++b )
			rotateScaleMatrix[a][b] = rotationMatrix[a][b] / lengthScale;
	}	

}

/*!	Sends ListOfAtoms positions to the PlotAtoms program.

	\param atoms Must be a pointer to a ListOfAtoms.
	*/
void PlotAtomsObserver::Update(ListOfAtoms *atoms)
{
	int numberOfLeaves = atoms->NumberOfLeaves(),color,i;
	double aveColorValue = 0.,maxColorValue = 0.;
	if(colorMethod)
		colorMethod->Update(atoms);
	
	// apply rotation and scaling
	const DMArray<double, DIMENSION> *positions = atoms->GetCartesianPositions();
	DMArray<double, DIMENSION> *rotatedScaledPositions = positions->ApplyMatrix(rotateScaleMatrix);
	
	
	if (DIMENSION < 3)
	{
		LeafToLeafIterator<ListOfAtoms> iter(atoms);
		for(int leaf=0;!iter.Done(); iter.Next(),++leaf)
		{
			std::map<int, int >::const_iterator colorToUse = colorMap.find(leaf);
			if(colorToUse == colorMap.end())
				color = 0;
			else color = colorToUse->second;
			
			DMArrayIterator<double, DIMENSION> posIter(iter.Current()->GetCartesianPositions());
			for(posIter.Init(),i=0; !posIter.IsDone(); posIter.Next(),++i)
			{
				if(colorMethod)
				{
					double value = fabs((*colorMethod)(atoms,i));
					aveColorValue += value;
					if(value > maxColorValue)
						maxColorValue = value;
					color = (int) (nColors * value/colorScale);
				}
				if( color >= 0)
				{
					double x = posIter.Current()[0];
					double y = posIter.Current()[1];
					fprintf( pipe_file, "%g %g %d\n",x,y,color);
				}		
			}	
		}
		fprintf( pipe_file, "\n" );
	}
	else	// DIMENSION >=3
	{
		static SortAtoms sort3D;
	

		const int *sortedAtomNumbers = sort3D.Sort(rotatedScaledPositions);
		for (long i=0; i<atoms->GetNumber(); i++)  // this isn't very efficient
		{
			int atom = sortedAtomNumbers[i];
			if(colorMethod)
			{
				double value = fabs((*colorMethod)(atoms,atom));
				aveColorValue += value;
				if(value > maxColorValue)
					maxColorValue = value;
				color = (int) (nColors * value/colorScale);
			}
			else
			{
				// find out what leaf sortedAtomNumbers[i] belongs to
				int leaf = 0,total = 0;
				LeafToLeafIterator<ListOfAtoms> iter(atoms);
				while((total+=iter.Current()->GetNumber())<=atom)
				{
					++leaf;
					iter.Next();
				}
				std::map<int, int >::const_iterator colorToUse = colorMap.find(leaf);
				if(colorToUse == colorMap.end())
					color = 0;
				else color = colorToUse->second;
			}
			if( color >= 0 )
			{
				double x = (*rotatedScaledPositions)[atom][0];
				double y = (*rotatedScaledPositions)[atom][1];
				fprintf( pipe_file, "%g %g %d\n",x,y,color);
			}
		}	
		fprintf( pipe_file, "\n" );


		delete rotatedScaledPositions;
	} // end of DIMENSION >= 3 else clause

	//colorScale = aveColorValue/atoms->GetNumber();
	if( !colorScaleSetByUser)
		colorScale = maxColorValue;
	fflush( pipe_file );
}


void PlotAtomsObserver::Plot(ListOfAtoms *atoms)
{
    Update(atoms);
}
