import string, os, Numeric, math, whrandom, tempfile
from DigitalMaterial.PX.Geometry.Common.Points import PointSet
from DigitalMaterial.PX.Geometry.Common.Vectors import Vector
from DigitalMaterial.PX.Tools.Utils.SystemCall import SystemCall
from DigitalMaterial.PX.Tools.Utils.FileIO import *
from DigitalMaterial.PX.Geometry.Common import nonuniformPoints


class PointSetSource:

    def GetOutput(self):
        pass

class PointSetInstance(PointSetSource):

    def __init__(self, pointset):
        self.pointset = pointset

    def GetOutput(self):
        return self.pointset


class PointSetReader(PointSetSource):

    def __init__(self, filename=None):
        self.filename = filename

    def SetFileName(self, filename):
        self.filename = filename

    def GetOutput(self):

        if self.filename is None:
            raise ValueError, "filename for PointSetReader not set"
        else:
            input = open(self.filename, 'r')
            line = string.split(input.readline())
            dim, source = string.atoi(line[0]), line[1:]
            npts = parse_atoi(input)
            pointset = PointSet([])
            for n in range(npts):
                coords = parse_map_atof(input)
                pointset.append(Vector(coords))
            input.close()
            return pointset


class RboxGenerator(PointSetSource):

    def __init__(self):
        self.seedstring = ''
        self.NumberOfPoints = 1
        self.options = None

    def SetOptions(self, options):
        self.options = options

    def SetNumberOfPoints(self, number):
        self.NumberOfPoints = number

    def GetNumberOfPoints(self):
        return self.NumberOfPoints

    def SetSeed(self, seed=None):
        self.seed = seed
        if seed is None:
            self.seedstring = 't'
        else:
            self.seedstring = 't' + `seed`

    def GetSeed(self):
        return self.seed

    def _buildSystemCall(self):
        if self.options is None:
            self.options = 'D3 ' + `self.NumberOfPoints` + ' ' + \
                           self.seedstring
        self._syscall = SystemCall('rbox', self.options)

    def GetOutputFileName(self):
        return self._syscall.outfile

    def GetOutput(self):
        self._buildSystemCall()
        self._syscall.SetTemporaryOutputFile()
        self._syscall.run()
        self._reader = PointSetReader(self._syscall.outfile)
        self.pointset = self._reader.GetOutput()
        os.remove(self._syscall.outfile)
        return self.pointset


class UniformPointSetGenerator(PointSetSource):

    def __init__(self):
        self.NumberOfPoints = 1
        self.component = None 

    def SetComponentGeometry(self, component):
        self.component = component
    
    def SetNumberOfPoints(self, num):
        self.NumberOfPoints = num
                 
    def GeneratePointSet(self):
        import RandomArray
        from Scientific.Geometry import Vector
        self.boundingBox()
        self.pointset = PointSet([])
        numG = 0
        while numG < self.NumberOfPoints:
            p = RandomArray.uniform(self.bmin, self.bmax, 3)
            pt = Vector(p[0],p[1],p[2])
            if (self.component.isInside(pt)):
                self.pointset.append(pt)
                numG = numG + 1
            
    def boundingBox(self):
        bx = []
        by = []
        bz = []
        for p in self.component.pointSet:
            bx.append(p[0])
            by.append(p[1])
            bz.append(p[2])
        
        self.bmax = Numeric.array([max(bx), max(by), max(bz)])
        self.bmin = Numeric.array([min(bx), min(by), min(bz)]) 

    def GetOutput(self):
        self.GeneratePointSet()
        return self.pointset
                    
class NonuniformPointSetGenerator(PointSetSource):

    def __init__(self):
        self.inputFile = None
        self.outputFile = None
        self.seed = None
        self.poissrv = None

    def SetInputFile(self, file):
        self.inputFile = file

    def SetOutputFile(self, file):
        self.outputFile = file
        
    def SetSeed(self, seed):
        self.seed = seed%256
        whrandom.seed(self.seed, self.seed, self.seed)

    def _initialize(self):

        self.outputFile = tempfile.mktemp()
        if self.seed is None:
            self.seed = whrandom.randint(10000, 500000)
        self._readInputFile()
        self.sp = self.pp[self.numsubdiv][3]

        for i in xrange(self.numsubdiv):
            self.pp[i][3] = self.pp[i][3]/self.sp

    def _generatePoisson_rv(self):
        self.poissrv = poissrnd(self.sp)

    def _generatePoints(self):

        if self.poissrv is None:
            self._generatePoisson_rv()
        self.pts = Numeric.zeros((self.poissrv, 3), Numeric.Float)

        for i in xrange(self.poissrv):
            index = 0
            r = whrandom.random()
            while (self.pp[index][3] < r):
                index = index + 1

            ii, jj, kk = self.pp[index][0:3]
            self.pts[i][0] = self.dx*whrandom.random() + ii*self.dx
            self.pts[i][1] = self.dy*whrandom.random() + jj*self.dy
            self.pts[i][2] = self.dz*whrandom.random() + kk*self.dz

    def _generate(self):
        if self.inputFile is None:
            raise ValueError, "nonuniformPoints input file must be set"
        self._initialize()
        self._generatePoints()
        self._write(self.outputFile)

    def GetOutput(self):
        self._generate()
        self._reader = PointSetReader(self.outputFile)
        self.pointset = self._reader.GetOutput()
        os.remove(self.outputFile)
        return self.pointset

    def _readInputFile(self):

        if self.inputFile is None:
            raise ValueError, "input file not specified"

        f = open(self.inputFile, 'r')
        self.xmax, self.ymax, self.zmax = \
                   map(string.atof, string.split(f.readline()))
        self.nx, self.ny, self.nz = \
                 map(string.atoi, string.split(f.readline()))
        self.dx = self.xmax/self.nx
        self.dy = self.ymax/self.ny
        self.dz = self.zmax/self.nz

        self.numsubdiv = self.nx * self.ny * self.nz
        self.p  = Numeric.zeros((self.numsubdiv,)*3, Numeric.Float)
        self.pp = Numeric.zeros((self.numsubdiv+1, 4))

        for i in xrange(self.numsubdiv):
            xtmp, ytmp, ztmp, savg = \
                       map(string.atof, string.split(f.readline()))
            ii = int(math.floor(xtmp/self.dx))
            jj = int(math.floor(ytmp/self.dy))
            kk = int(math.floor(ztmp/self.dz))

            self.p[ii][jj][kk] = (self.dx*self.dy*self.dz)/savg
            self.pp[i+1][0] = ii
            self.pp[i+1][1] = jj
            self.pp[i+1][2] = kk
            self.pp[i+1][3] = self.pp[i][3] + self.p[ii][jj][kk]

        f.close()


    def _write(self, outputFile=None):
        
        f = open(outputFile, 'w')
        f.write('3 nonuniformPoints\n')
        f.write('%d\n' % (self.poissrv))
        for i in xrange(self.poissrv):
            f.write('%20.16f %20.16f %20.16f\n' % \
                    (self.pts[i][0]-0.5*self.xmax, \
                     self.pts[i][1]-0.5*self.ymax, \
                     self.pts[i][2]-0.5*self.zmax))
        f.close()


def poissrnd(lam):

    t = 0
    N = 0
    laminv = 1./lam

    while t < 1:
        r = whrandom.random()
        r2 = -laminv*math.log(r)
        t = t + r2
        N = N+1

    return N



if __name__ == '__main__':
    print "RboxGenerator()"
    g2 = RboxGenerator()
    g2.SetNumberOfPoints(39)
    p2 = g2.GetOutput()
    print p2

    print "UniformPointSetGenerator()"
    g3 = UniformPointSetGenerator()
    g3.SetNumberOfPoints(32)
    from DigitalMaterial.PX.Geometry.Models import qhullModels
    g3.SetComponentGeometry(qhullModels.UnitCube(3))
    p3 = g3.GetOutput()
    print p3
    
    #g1 = NonuniformPointSetGenerator()
    #g1.SetInputFile('ptin.dat')
    #g1.SetSeed(1234567)
    #p1 = g1.GetOutput()
    #print p1

    

