# This module defines geometrical vectors in d dimensions with the standard
# operations on them. The elements are stored in an array.
#
# Written by: Konrad Hinsen <hinsen@cnrs-orleans.fr>
# Last revision: 1999-7-23
#
# Modified by: Chris Myers <myers@tc.cornell.edu> 
# to support arbitrary dimension vectors.
# Last revision: 2000-3-03
#

_undocumented = 1

import Numeric, types
from Scientific.Geometry import TensorModule
from Scientific.Geometry import VectorModule

class Vector(VectorModule.Vector):

    """Vector in R^d

    Constructor:

    - Vector([coordinates])   (N-vector from list of N coordinates)

    Vectors support the usual arithmetic operations
    ('v1', 'v2': vectors, 's': scalar): 

    -  'v1+v2'           (addition)
    -  'v1-v2'           (subtraction)
    -  'v1*v2'           (scalar product)
    -  's*v1', 'v1*s'    (multiplication with a scalar)
    -  'v1/s'            (division by a scalar)

    The coordinates can be extracted by indexing.

    Vectors are *immutable*, i.e. their elements cannot be changed.

    Vector elements can be any objects on which the standard
    arithmetic operations plus the functions sqrt and arccos are defined.
    """

    is_vector = 1

    def __init__(self, coordlist, dim=None):
	try:
	    tmp = coordlist[0]
	except:
	    raise TypeError, "input coordinates do not form a sequence"
	if dim is not None and len(coordlist) == 1:
	    self.array = coordlist[0]*dim
	else:
	    self.array = coordlist
	self.array = Numeric.array(self.array)
	self.dim = len(self.array)

    def __len__(self):
	return len(self.array)

    def __repr__(self):
	return 'Vector(%s)' % (`self.array.tolist()`)

    def __str__(self):
	return 'Vector(%s)' % (`self.array.tolist()`)

    def __add__(self, other):
	return Vector(self.array+other.array)
    __radd__ = __add__

    def __neg__(self):
	return Vector(-self.array)

    def __sub__(self, other):
	return Vector(self.array-other.array)

    def __rsub__(self, other):
	return Vector(other.array-self.array)

    def __mul__(self, other):
	if isVector(other):
	    return Numeric.add.reduce(self.array*other.array)
	elif TensorModule.isTensor(other):
	    product = TensorModule.Tensor(self.array).dot(other)
	    if product.rank == 1:
		return Vector(product.array)
	    else:
		return product
        elif hasattr(other, "_product_with_vector"):
            return other._product_with_vector(self)
	else:
	    return Vector(Numeric.multiply(self.array, other))

    def __rmul__(self, other):
	if TensorModule.isTensor(other):
	    product = other.dot(TensorModule.Tensor(self.array))
	    if product.rank == 1:
		return Vector(product.array)
	    else:
		return product
	else:
	    return Vector(Numeric.multiply(self.array, other))

    def __div__(self, other):
	if isVector(other):
	    raise TypeError, "Can't divide by a vector"
	else:
	    return Vector(Numeric.divide(self.array,1.*other))
	    
    def z(self):
        "Returns the z coordinate."
	if self.dim >= 3: 
	    return self.array[2]
	else:
	    raise AttributeError, "Vector is 2D; no z coordinate defined."

    def normalized(self):
        "Returns a normalized copy."
	len = Numeric.sqrt(Numeric.add.reduce(self.array*self.array))
	if len == 0:
	    raise ZeroDivisionError, "Can't normalize a zero-length vector"
	return Vector(Numeric.divide(self.array, len))

    def cross(self, other):
        "Returns the cross product with vector |other|."
	if self.dim != 3:
	    raise AttributeError, "Vector is 2D; cross product not defined."
	else:
	    return (VectorModule.Vector.cross(self, other))

    def length(self):
        import math
        return math.sqrt(Numeric.add.reduce(self.array*self.array))

# Type check

def isVector(x):
    "Return 1 if |x| is a vector."
    return hasattr(x,'is_vector')


# Some useful constants

ex3 = Vector([1,0,0])
ey3 = Vector([0,1,0])
ez3 = Vector([0,0,1])
null3 = Vector([0.,0.,0.])
ex2 = Vector([1,0])
ey2 = Vector([0,1])
ez2 = Vector([0,0])
null2 = Vector([0.,0.])
