from DigitalMaterial.PX.Geometry.Common import qhull, Points
from DigitalMaterial.PX.Geometry.Common.Vectors import *

eps = 1.0e-12

Insider={1:'inside', 0:'outside'}

class ConvexPolyhedron(qhull.qhull_ConvexHull):

    def __init__(self, dim, geometryOptions=None, pointSet=None):

        self.dim = dim
        self.geometryOptions = geometryOptions
        self.pointSet = pointSet
        if self.geometryOptions is not None:
            self.pointSetGenerator = \
                qhull.PointSetSource.RboxGenerator()
            self.pointSetGenerator.SetOptions(geometryOptions)  
        if self.pointSet is None:
            self.pointSet = self.pointSetGenerator.GetOutput()
        qhull.qhull_ConvexHull.__init__(self, self.pointSet)
    
        
    def isInside(self, point):
        """isInside is a simple inside function for a convex region. We 
        assume the Facet is oriented so the face normal determined by the
        right-hand rule points toward inside. We check the volume sign of 
        a tetrahedron determined by first three points of the facet and the 
        given point. The point is inside iff all the signs are negative. 
        Ref: pp. 129 O'Rourke 1998"""  
        
        from LinearAlgebra import determinant
        from Numeric import array, ones, Float
        inside = 1
        for f in self.Facets:
            mat = ones((4,4), Float)
            for i in range(len(mat)-1):
                mat[i, 0:3] = self.pointSet[f.vertices[i]].array
            mat[3,0:3] = point.array
            if (determinant(mat) >= 0.0 + eps):
                inside = 0
                break
        
        return inside        

    def NumberOfFacets(self):
        return len(self.Facets)

    def facet(self, index):
        return self.Facets[index]
        

class UnitCube (ConvexPolyhedron):

    def __init__(self, dim):
        self.dim = dim
        self.pointSetGenerator = qhull.PointSetSource.RboxGenerator()
        self.pointSetGenerator.SetOptions('D'+`dim` + ' c')
        self.pointSet = self.pointSetGenerator.GetOutput()
        qhull.qhull_ConvexHull.__init__(self, self.pointSet)
    

    def isInside(self, point):
        inside = 1
        for d in range(self.dim):
            if point[d] >= -0.5-eps and point[d] <= 0.5+eps:
                inside = inside and 1
            else:
                inside = 0
        return inside

class UnitDiamond (ConvexPolyhedron):

    def __init__(self, dim):
        self.dim = dim
        self.pointSetGenerator = qhull.PointSetSource.RboxGenerator()
        self.pointSetGenerator.SetOptions('D'+`dim` + ' d')
        self.pointSet = self.pointSetGenerator.GetOutput()
        qhull.qhull_ConvexHull.__init__(self, self.pointSet)    

class HexagonalCylinder(ConvexPolyhedron):

    def __init__(self, dim):
        from math import sqrt
        import Polygon
        
        self.dim = dim

        self.hex = Polygon.RegularPolygon(6, outerRadius=0.5)
        pts = []
        for p in self.hex.points:
            coord = p.array.tolist()
            coord.append(0.5)
            pts.append(Vector(coord))
            coord = p.array.tolist()
            coord.append(-0.5)
            pts.append(Vector(coord))
            
        self.pointSet = Points.PointSet(data=pts)

        qhull.qhull_ConvexHull.__init__(self, self.pointSet)
    

#    def isInside(self, point):
#        inside = 1
#
#        if point[2] >= -0.5-eps and point[2] <= 0.5+eps:
#            inside = inside and 1
#        else:
#            inside = 0
#        
#        p = Vector(point.array[0:2])
#        if self.hex.inside(p):
#            inside = inside and 1
#        else:
#            inside = 0
#
#        return inside



if __name__ == '__main__':

    from DigitalMaterial.PX.Geometry.Common.Vectors import Vector
    # cube
    print "UnitCube"
    c = UnitCube(3)

    p1 = Vector([0.,0.,0.])
    insideStatus = c.isInside(p1)
    print "Point %s is %s" % (p1, Insider[insideStatus])

    p2 = Vector([2.,0.,0.])
    insideStatus = c.isInside(p2)
    print "Point %s is %s" % (p2, Insider[insideStatus])

    p3 = Vector([0.5,0.5,0.5])
    insideStatus = c.isInside(p3)
    print "Point %s is %s" % (p3, Insider[insideStatus])

    p4 = Vector([0.5,0.5,0.500000001])
    insideStatus = c.isInside(p4)
    print "Point %s is %s" % (p4, Insider[insideStatus])
    # diamond
    print "UnitDiamond"
    c = UnitDiamond(3)

    p1 = Vector([0.,0.,0.])
    insideStatus = c.isInside(p1)
    print "Point %s is %s" % (p1, Insider[insideStatus])

    p2 = Vector([2.,0.,0.])
    insideStatus = c.isInside(p2)
    print "Point %s is %s" % (p2, Insider[insideStatus])

    p3 = Vector([0.5, 0.5, 0.5])
    insideStatus = c.isInside(p3)
    print "Point %s is %s" % (p3, Insider[insideStatus])

    p4 = Vector([0.2,0.2,0.0])
    insideStatus = c.isInside(p4)
    print "Point %s is %s" % (p4, Insider[insideStatus])
    # 10 random points in the unit cube centered at the origin
    print "Random Point Set"
    pointSetGenerator = qhull.PointSetSource.RboxGenerator()
    pointSetGenerator.SetOptions('100')
    c = ConvexPolyhedron(3, pointSet = pointSetGenerator.GetOutput())
    p1 = Vector([0.,0.,0.])
    insideStatus = c.isInside(p1)
    print "Point %s is %s" % (p1, Insider[insideStatus])

    p2 = Vector([2.,0.,0.])
    insideStatus = c.isInside(p2)
    print "Point %s is %s" % (p2, Insider[insideStatus])

    p3 = Vector([0.5,0.5,0.5])
    insideStatus = c.isInside(p3)
    print "Point %s is %s" % (p3, Insider[insideStatus])

    p4 = Vector([0.5,0.5,0.500000001])
    insideStatus = c.isInside(p4)
    print "Point %s is %s" % (p4, Insider[insideStatus])
