from DigitalMaterial.PX.Geometry.Tesselation.QhullVoronoi import *

def triangularArea(points):
    # from CRC Standard Math Tables, 24th edition, p. 7        
    import math
    side_a = points[0]-points[1]
    side_b = points[1]-points[2]
    side_c = points[2]-points[0]
    a = side_a.length()
    b = side_b.length()
    c = side_c.length()
    s = 0.5*(a+b+c)
    area = math.sqrt(s*(s-a)*(s-b)*(s-c))
    return area


class PolyhedralInterface:

    def __init__(self, tesselation=None, ID=None, vertices=[],
                 hyperplane=None):

        self.tesselation = tesselation
        self.ID = ID
        self.vertices = vertices
        self.hyperplane = hyperplane

    def NumberOfVertices(self):
        return len(self.vertices)

    def __repr__(self):
        s = 'PolyhedralInterface ( %s ) ' % (self.ID,)
        s = s + '{ vertices = %s }' % (self.vertices)
        s = s + '\n'
        return s

    def asConvexPolyhedron(self):
        from DigitalMaterial.PX.Geometry.Common import Points
        from DigitalMaterial.PX.Geometry.Models import qhullModels

        pts = Points.PointSet([])
        for v in self.vertices:
            pts.append(self.tesselation.Vertex(v))

        self.convexHull = qhullModels.ConvexPolyhedron(dim=3, pointSet=pts)

    def isInside(self, point):
        if not hasattr(self, 'convexHull'):
            self.asConvexPolyhedron()
        return self.convexHull.isInside(point)

    def area(self):
        interiorPoint = self._pickPointInCell()
        totalArea = 0.
        numvertices = self.NumberOfVertices()
        for nv in range(numvertices-1):
            v0 = self.vertices[nv]
            v1 = self.vertices[nv+1]
            p0 = self.tesselation.Vertex(v0)
            p1 = self.tesselation.Vertex(v1)
            p2 = interiorPoint
            subArea = triangularArea([p0,p1,p2])
            totalArea = totalArea+subArea
        v0 = self.vertices[numvertices-1]
        v1 = self.vertices[0]
        p0 = self.tesselation.Vertex(v0)
        p1 = self.tesselation.Vertex(v1)
        p2 = interiorPoint
        subArea = triangularArea([p0,p1,p2])
        totalArea = totalArea+subArea
        return totalArea
        
    def _pickPointInCell(self):
        halfway = self.NumberOfVertices()/2
        chord = self.tesselation.Vertex(halfway) - self.tesselation.Vertex(0)
        point = self.tesselation.Vertex(0) + 0.5*chord
        return point


class PolyhedralRegion:

    def __init__(self, tesselation=None, ID=None, interfaces=[]):
        self.tesselation = tesselation
        self.ID = ID
        self.interfaces = interfaces
        self.boundingBox()

    def NumberOfInterfaces(self):
        return len(self.interfaces)

    def __repr__(self):
        s = 'PolyhedralRegion ( %s )' % (self.ID,)
        s = s + '{ interfaces = %s }' % (self.interfaces)
        s = s + '\n'
        return s

    def boundingBox(self):
        bpts = []
        for i in self.interfaces:
            for v in self.tesselation.Interface(i).vertices:
                    bpts.append(self.tesselation.Vertex(v))
        bx = []
        by = []
        bz = []
        for p in bpts:
            bx.append(p[0])
            by.append(p[1])
            bz.append(p[2])
        
        self.bmax = [max(bx), max(by), max(bz)]
        self.bmin = [min(bx), min(by), min(bz)] 
        
    def isInsideBBox(self, point):
        if (point[0] <= self.bmax[0] and point[1] <= self.bmax[1] \
           and point[2] <= self.bmax[2] and point[0] >= self.bmin[0] \
           and point[1] >= self.bmin[1] and point[2] >= self.bmin[2]):
            return 1
        else: return 0    
    
    def asConvexPolyhedron(self):
        from DigitalMaterial.PX.Geometry.Common import Points
        from DigitalMaterial.PX.Geometry.Models import qhullModels

        vertexList = []
        for i in self.interfaces:
            for v in self.tesselation.Interface(i).vertices:
                if vertexList.count(v) == 0:
                    vertexList.append(v)

        pts = Points.PointSet([])
        for v in vertexList:
            pts.append(self.tesselation.Vertex(v))

        self.convexHull = qhullModels.ConvexPolyhedron(dim=3,pointSet=pts)

    def isInside(self, point):
        if not hasattr(self, 'convexHull'):
            self.asConvexPolyhedron()
        if (self.isInsideBBox(point)):
            return self.convexHull.isInside(point)
        else: return 0    


class Tesselation:

    pass

class PolyhedralTesselation(Tesselation):

    def __init__(self):
        pass
    
    def __repr__(self):
        s = 'PolyhedralTesselation ( ) ' 
        s = s + '{ %d regions, %d interfaces, %d vertices }' % \
            (self.NumberOfRegions(), self.NumberOfInterfaces(),
             self.NumberOfVertices())
        s = s + '\n'
        return s

    def dimension(self):
        return len(self._vertices.values()[0])

    def isInside(self, point):
        return self.componentGeometry.isInside(point)

    def Vertices(self):
        return self._vertices.values()

    def Vertex(self, ID):
        return self._vertices[ID]

    def Interfaces(self):
        return self._interfaces.values()

    def Interface(self, ID):
        return self._interfaces[ID]

    def Regions(self):
        return self._regions.values()

    def Region(self, ID):
        return self._regions[ID]

    def NumberOfVertices(self):
        return len(self._vertices)

    def NumberOfInterfaces(self):
        return len(self._interfaces)

    def NumberOfRegions(self):
        return len(self._regions)

    def ComponentGeometry(self):
        return self._componentGeometry
        
    def VertexIncidences(self):
        return self.incidenceInfo.vertices

    def VertexIncidence(self, ID):
        return self.incidenceInfo.vertices[ID]

    def InterfaceIncidences(self):
        return self.incidenceInfo.interfaces

    def InterfaceIncidence(self, ID):
        return self.incidenceInfo.interfaces[ID]


class TesselationSource:

    def GetTesselation(self):
        pass

class VoronoiTesselationBuilder(TesselationSource):

    def __init__(self, pointSetSource=None, componentGeometry=None):

        self.SetDefaultComponentGeometry()
        self.builder = None
        self.tesselation = None
        
    def SetPointSource(self, pointSource):
        self.pointSource = pointSource
    
    def SetComponentGeometry(self, componentGeometry):
        self.componentGeometry = componentGeometry

    def SetDefaultComponentGeometry(self):
        from DigitalMaterial.PX.Geometry.Models import qhullModels
        self.componentGeometry = qhullModels.UnitCube(3)

    def _buildBuilder(self):
        self.builder = QhullVoronoiBuilder(self.pointSource,
                                           self.componentGeometry)

    def _buildTesselation(self):
        self.voronoi = self.builder.GetVoronoiDiagram()
        self.tesselation = VoronoiDiagramToPolyhedralTesselation(self.voronoi)
        self.tesselation.componentGeometry = self.componentGeometry
        self.tesselation.incidenceInfo = IncidenceInfo(self.tesselation, 
                                                       self.componentGeometry)

    def GetTesselation(self):
        if self.builder is None: self._buildBuilder()
        if self.tesselation is None: self._buildTesselation()
        return self.tesselation


class IncidenceInfo:

    def __init__(self, tesselation, componentGeometry):

        self.tesselation = tesselation
        self.componentGeometry = componentGeometry
        self.vertices = self._findIncidentVertices()
        self.interfaces = self._findIncidentInterfaces(self.vertices)


    def _findIncidentVertices(self, tol=1.0e-10):
        iV = {}
        for ID, v in self.tesselation._vertices.items():
            for nh, hyp in self.componentGeometry.normalSet.items():
                pp = qhull.pointProject(v, hyp)
                diff = v-pp
                dist = diff.length()
                if dist < tol:
                    if not iV.has_key(ID): 
                        iV[ID] = []
                    iV[ID].append(nh)
                    continue
        return iV

    def _findIncidentInterfaces(self, iV):
        iI = {}
        for iID, i in self.tesselation._interfaces.items():
            numIV = self.componentGeometry.NumberOfFacets() * [0]
            for v in i.vertices:
                if iV.has_key(v):
                    facets = iV[v]
                    for f in facets:
                        numIV[f] = numIV[f] + 1

            g3 = filter(lambda x:x>2, numIV)
            if len(g3) > 0: 
                g3 = g3[0]
                iI[iID] = numIV.index(g3)

        return iI


def GetInterfacesOnComponentSurface(tesselation, surfaceIndex):

    interfaces = []
    for ID, surf in tesselation.InterfaceIncidences().items():
        if surf == surfaceIndex:
            interfaces.append(ID)
    return interfaces
        

def GetRegionsOnComponentSurface(tesselation, surfaceIndex):
    interfaces = GetInterfacesOnComponentSurface(tesselation,surfaceIndex)
    grains = []
    nRegions = tesselation.NumberOfRegions()
    for i in interfaces:
        if i[0] < nRegions:
            grains.append(i[0])
        elif i[1] < nRegions:
            grains.append(i[1])
        else:
            raise ValueError, "interfaces not on component surface"
    return grains



def GetUniformTesselation(ngrains):
    from DigitalMaterial.PX.Geometry.Models import qhullModels

    cube = qhullModels.UnitCube(3)
    source = PointSetSource.RboxGenerator()
    source.SetNumberOfPoints(ngrains)
    builder = VoronoiTesselationBuilder()
    builder.SetPointSource(source)
    builder.SetComponentGeometry(cube)
    tesselation = builder.GetTesselation()
    return tesselation

def GetNonuniformTesselation(infile):
    from DigitalMaterial.PX.Geometry.Models import qhullModels

    cube = qhullModels.UnitCube(3)
    source = PointSetSource.NonuniformPointSetGenerator()
    source.SetInputFile(infile)
    builder = VoronoiTesselationBuilder()
    builder.SetPointSource(source)
    builder.SetComponentGeometry(cube)
    tesselation = builder.GetTesselation()
    return tesselation


if __name__ == '__main__':
    tesselation = GetUniformTesselation(100)

