# Author: C.R. Myers, 2000-2001


class PolycrystalWriter:

    def SetFileName(self, filename):
        self.filename = filename


class PicklePolycrystalWriter(PolycrystalWriter):

    """PicklePolycrystalWriter writes out information about the
    polycrystal geometry and attributes using the Python pickle mechanism;
    __setstate__ and __getstate__ defined on Polycrystal define what
    constitutes the pickleable state"""

    def __init__(self, filename=None):
        self.filename = filename

    def write(self, polycrystal, binary=1):
        try:
            self._writePickleFile(polycrystal, binary)
        except:
            raise IOError, "failed to pickle %s to file %s" % \
		  (polycrystal, self.filename)

    def _writePickleFile(self, polycrystal, binary):
        f = open(self.filename, 'wb')
        import cPickle
        cPickle.dump(polycrystal, f, binary)
        f.close()
        return 0

DefaultPolycrystalWriter = PicklePolycrystalWriter


class VTKPolycrystalWriter(PolycrystalWriter):

    """VTKPolycrystalWriter writes out information about the
    polycrystal geometry in VTK format (suitable for reading by
    a VTKPolyDataReader).  VTK does not support polyhedral cells, so
    the geometry of the grain boundaries is written out instead.  A
    VTKFieldData array named grainIndices is also written to indicate
    which two grain ID numbers are associated with each grain boundary,
    following the specified orientation convention (right-hand rule
    around the boundary points to grainIndex 0"""

    def __init__(self, filename=None):
        self.filename = filename

    def write(self, polycrystal, binary=0):
        try:
            self._writeVTKFile(polycrystal, binary)
        except:
            raise ValueError, "failed to write polycrystal ( %s ) to VTK file" % polycrystal.ID
        
    def _writeVTKFile(self, polycrystal, binary):
        if binary == 1:
            raise ValueError, "binary VTK files not yet supported"

        f = open(self.filename, "w")
        f.write('# vtk DataFile Version 3.1\n')
        f.write(self.filename + '\n')
        f.write('ASCII\n')
        f.write('DATASET POLYDATA\n')

        f.write('POINTS %d float\n' % (polycrystal.NumberOfVertices()))
        for np in range(polycrystal.NumberOfVertices()):
            p = polycrystal.Vertex(np)
            f.write('%15.10f %15.10f %15.10f\n' % (p[0], p[1], p[2]))
        f.write('\n')

        npolys = polycrystal.NumberOfInterfaces()
	size = 0
	for inter in polycrystal.Interfaces():
            size = size + 1 + inter.NumberOfVertices()

        f.write('POLYGONS %d %d\n' % (npolys, size))
	for inter in polycrystal.Interfaces():
            f.write('%d ' % inter.NumberOfVertices())
            for v in inter.vertices:
                f.write('%d ' % v)
            f.write('\n')

        f.write('\n')            

        f.write('CELL_DATA %d\n' % npolys)
        f.write('FIELD FieldData 1\n')
        f.write('grainIndices 2 %d int\n' % npolys)
	for inter in polycrystal.Interfaces():
            f.write('%d %d\n' % inter.ID)
        f.write('\n')

        f.close()



class VertexInterfacePolycrystalWriter(PolycrystalWriter):

    """VertexInterfacePolycrystalWriter writes out information about the
    polycrystal geometry (tesselation) into two files, one containing
    coordinates of the tesselation vertices (*_v.dat), and one containing
    lists of which vertices participates in each of the tesselation
    interfaces/grain boundaries (*_i.dat)"""


    def __init__(self, filebase=None):
        self.filebase = filebase

    def write(self, polycrystal):
        try:
            self._write_vertex_interface_files(polycrystal)
        except:
            raise ValueError, \
                  "failed to write polycrystal ( %s ) to Sanjay files" % \
                  polycrystal.ID
        
    def _write_vertex_interface_files(self, polycrystal):

        ngrains = polycrystal.GetNumberOfGrains()
        v_filename = self.filebase + '_v.dat'
        f = open(v_filename, 'w')
        for nv in range(polycrystal.NumberOfVertices()):
            v = polycrystal.Vertex(nv)
            f.write('%20.16f %20.16f %20.16f\n' % (v[0], v[1], v[2]))
        f.close()

        i_filename = self.filebase + '_i.dat'
        f = open(i_filename, 'w')
        for i in polycrystal.Interfaces():
            for v in i.vertices:
                f.write('%d ' % v)
            f.write('\n')
        f.close()
        return


class InterfaceRegionPolycrystalWriter(PolycrystalWriter):

    def __init__(self, filebase=None):
        self.filebase = filebase

    def write(self, polycrystal):
        try:
            self._write_interface_region_files(polycrystal)
        except:
            raise ValueError, \
                  "failed to write polycrystal ( %s ) to Roland files" % \
                  polycrystal.ID
        
    def _write_interface_region_files(self, polycrystal):

        tesselation = polycrystal.geometry
        areas = {}
        for r in tesselation._regions.values():
            areas[r.ID] = []
            for fID in r.interfaces:
                for nID in fID:
                    if (r.ID != nID): ngID = nID
                areas[r.ID].append(ngID, tesselation.Interface(fID).area())
        
        filename = self.filebase + '.dat'
        f = open(filename, 'w')
        
        ngrains = polycrystal.GetNumberOfGrains()
        f.write('%d\n' % ngrains) 
        for i in range(ngrains):
            f.write('%d  ' % len(areas[i]))
            for ngID, area in areas[i]:
                f.write('%20.16f  ' % area)
            f.write('\n') 
        for i in range(ngrains):
            for ngID, area in areas[i]:
                f.write('%d  ' % (ngID+1)) #silly fortran offset
            f.write('\n') 
        for i in range(ngrains):
            ori = polycrystal.GetOrientation(i)
            f.write('%20.16f  %20.16f  %20.16f\n' % (ori[0], ori[1], ori[2]))
        f.close()
        return


# NetCDFPolycrystalWriter is incomplete in its implementation

class NetCDFPolycrystalWriter(PolycrystalWriter):

    def __init__(self, filename=None):
        self.filename = filename

    def _writeNetCDFFile(self, polycrystal, binary):
        from Scientific.IO.NetCDF import *
        ncfile = NetCDFFile(self.filename, 'w')
        ncfile.Conventions = 'DM/PX 0.2'
        ncfile.createDimension('numvertices',
                               polycrystal.GetNumberOfVertices())
        


# UNTESTED,
# and not consistent with the pickle-based PolycrystalReader defined in
# PolycrystalSource.py
#
# class VTKPolycrystalReader(PolycrystalReader):

#     def SetFileName(self, filename):
#         self.filename = filename

#     def show(self, polycrystal):

#         if self.filename is None:
#             import tempfile
#             self.filename = tempfile.mktemp()
#         self.write(polycrystal)

#         from DigitalMaterial.PX.Utils.Viz import VTKDisplay, ViewPolyData
#         self.display = VTKDisplay.VTKDisplay() 
#         self.viewer = ViewPolyData.PolycrystalViewer(self.filename)
#         actors = []
#         actors.extend(self.viewer.DrawInterfaces())
#         self.display.AddActors(actors)
        
#     def read(self):

#         import PolycrystalSource.PolycrystalReader
#         reader = PolycrystalSource.PolycrystalReader(self.filename)
#         return reader.GetPolycrystal()
