function [lens,gbs,gbtemp] = lgsize3d_ctest(verts,faces,nverts,lin,nseed)

rand('seed',nseed);  % seed the random number generator for use 
			%in computing line-plane intercepts 
tiny = 1e-10;        %We'll pretend this is the value of zero in 
			%a few cases                     	
gbcnt = 0;           %counters for the number of grain boundaries                     	
lencnt = 0;          %and lengths found.

h = waitbar(0,'');   %status bar command for MATLAB

%this is the big loop where the work is done.  The procedure is roughly
%as follows.  For each face in the tessellation, compute the 
%point of intersection for the plane of which the face is a subset
%and the specified line.  Then, check to see if this intersection
%point lies within the face, if so, consider it a GB intersection.

for i = 1:size(faces,1)
   waitbar(i/size(faces,1));      %MATLAB status bar command
   vert1 = verts(faces(i,1),:)';  %The first three vertices of the 
   vert2 = verts(faces(i,2),:)';  %face will be used to define the
   vert3 = verts(faces(i,3),:)';  %plane containing the face
   
   e1 = vert2-vert1;              %find a vector along two of the face edges
   e2 = vert3-vert2;              %to be used for defining the plane
   
   n = cross(e1,e2)/norm(cross(e1,e2));  %find a unit normal to the face
                                         %this normal and the first vert
                                         %from above define the plane
                                         
   if n(3) < 0      %these 'ifs' give a consistent orientation to the face
      n = -n;       %normals
   elseif n(3) == 0 & n(2) < 0
      n=-n;
   elseif n(3) == 0 & n(2)==0 & n(1) < 0
      n=-n;
   end
   
   
   intpt = linplanint(i,[vert1';n'],lin); %intpt computes the intersection
                                          %of the plane and line (see below)
                                          
%empty intpt means no intersection was found, i.e. the plane and line 
%are parallel. So, if an intersection was found, we proceed.  If the point
%is in the face then it should be consistently located to the same "side"
%of each of the face edges.  We go through and check this edge by edge.
%The counter suminout keeps track of how many of the edges the point
%lies on the correct side of.

	if ~isempty(intpt)      
      intpt = intpt';
   	suminout = 0;
   	for j = 1:nverts(i)
         tempvert1 = verts(faces(i,j),:)';   %grab the next two vertices of the face.
      	tempvert2 = verts(faces(i,j+1),:)';
                  	
         e = tempvert2-tempvert1;          %find a vector along this edge
         v = intpt - tempvert1;            %find a vector between the first 
                                           %of the verts and the intersection
                                           
         %We check three cases: 1) that the intersection is on a vertex,
         %2)that the intersection lies on the line containing the edge,
         %3)that neither of the first two is the case
         %if 1) then the intersection counts, if 2) it doesn't, if 3) we
         %need to keep checking
         if norm(intpt-tempvert2)<tiny|norm(v)<tiny
				suminout = nverts(i);
				break;
			elseif (norm(cross(e,v))) == 0
            inout = -1;
            suminout = suminout + inout;
         else
            %We're basically just using the sign of e x v to tell us which
            %side of the edge the intersection is on
            inout = sum((cross(e,v)/norm(cross(e,v))).*n);
            suminout = suminout + inout;
         end   
      end   
      %if the intersection is in the face, then it should have passed case 3) 
      %above for each of the edges, and suminout should be equal to the
      %total number of vertices in the face.  If so, then add the intersection
      %to the list of GB intersections, and increment the count of GB intersections
   	if round(abs(suminout)) == nverts(i)
         gbcnt = gbcnt + 1;
      	gbs(gbcnt,:) = intpt';
   	end   
   end
   
	   
end

%use nice MATLAB sorting to get the GB intersections in order along the
%line
gbs = sortrows(gbs,1);
gbs = sortrows(gbs,2);
gbs = sortrows(gbs,3);
gbtemp = gbs;
clear gbs;

%If you hit a vertex with your line, then that point will show up more than
%once in the list of GB intersections.  Here we go through and throw out
%duplicates, at the same time computing the line lengths.

gbs(1,:) = gbtemp(1,:);
gbcnt = 1;
for i = 1:size(gbtemp,1)-1
   lentemp = norm(gbtemp(i+1,:)-gbtemp(i,:));
   if lentemp > tiny
      gbs(gbcnt+1,:) = gbtemp(i+1,:);
      lens(gbcnt) = lentemp;
      gbcnt = gbcnt + 1;
   end
end
close(h);   %status bar command

%function intpt computes the point of intersection between a line and plane.
%Both the line and the plane are defined by a point and a vector.  The method
%for finding the intersection is as follows.  Generate two random vectors, 
%and cross each of them with the line direction vector.  These two new vectors
%each define, along with a point on the line, a plane.  The line is the 
%intersection of these two planes.  Thus, we can find the line-plane
%intersection by finding the intersection of the three planes we have now defined.

function [intpt] = linplanint(i,pl,lin)

x0=pl(1,1); %x0 is the point in the plane
y0=pl(1,2);
z0=pl(1,3);
a = pl(2,1); %[abc] is the plane normal vector
b = pl(2,2);
c = pl(2,3);

x0p = lin(1,1); %x0p is a point on the line
y0p = lin(1,2);
z0p = lin(1,3);
ap = lin(2,1); %[ap,bp,cp] is the line direction vector
bp = lin(2,2);
cp = lin(2,3);

v1 = [ap,bp,cp];          %line direction vector
v2 = cross(v1,rand(1,3)); %random vectors to define two planes
v3 = cross(v1,rand(1,3)); %which intersect at the line

v1 = v1/norm(v1);         %normalize all vectors
v2 = v2/norm(v2);
v3 = v3/norm(v3);

%We are solving a system of three equations (one for each plane)
%so we first need to check to see if the system is solvable by checking
%the determinant of the system matrix.  If it is not solvable then 
%the line and plane are parallel.  The expressions are obtained by
%Mathematica algebra.  If there is no intersection, return and empty
%vector.
mat = [a,b,c;v2;v3];
if det(mat) ~= 0
	xint = (a *(v2(3)* v3(2) - v2(2)* v3(3))* x0 + ...
         b *(-(v2(3)* v3(1)* x0p) + v2(1)* v3(3)* x0p + v2(3)* v3(2)* y0 - v2(2)* v3(3)* y0 - ...
            v2(3)* v3(2)* y0p + v2(2)* v3(3)* y0p) +...
         c* (v2(2)* v3(1)* x0p - v2(1)* v3(2)* x0p + v2(3)* v3(2)* z0 - v2(2)* v3(3)* z0 - ...
            v2(3)* v3(2)* z0p + v2(2)* v3(3)* z0p)) /...
       (c* v2(2)* v3(1) - b* v2(3)* v3(1) - c* v2(1)* v3(2) + a* v2(3)* v3(2) + b* v2(1)* v3(3) -...
         a* v2(2)* v3(3));


	yint = (b *(v2(3)* v3(1) - v2(1) *v3(3))* y0 +...
         a *(v3(3)* (-(v2(1)* x0) + v2(1)* x0p + v2(2) *y0p) +...
            v2(3)* (v3(1)* x0 - v3(1)* x0p - v3(2)* y0p)) +...
         c* (-(v2(2)* v3(1) *y0p) + v2(1)* v3(2)* y0p + v2(3)* v3(1)* z0 - v2(1)* v3(3)* z0 -...
            v2(3)* v3(1)* z0p + v2(1)* v3(3) *z0p)) /...
       (-(c* v2(2)* v3(1)) + b *v2(3)* v3(1) + c* v2(1)* v3(2) - a *v2(3)* v3(2) - b* v2(1)* v3(3) +...
         a* v2(2)* v3(3));

	zint =(c* (-(v2(2)* v3(1)) + v2(1)* v3(2))* z0 +...
         b *(-(v2(2)* v3(1)* y0) + v2(1)* v3(2) *y0 + v2(2)* v3(1)* y0p - v2(1)* v3(2)* y0p +...
            v2(3)* v3(1)* z0p - v2(1)* v3(3)* z0p) +...
         a *(v3(2)* (v2(1)* x0 - v2(1) *x0p - v2(3)* z0p) +...
            v2(2) *(-(v3(1)* x0) + v3(1)* x0p + v3(3)* z0p))) /...
       (-(c* v2(2)* v3(1)) + b* v2(3)* v3(1) + c* v2(1)* v3(2) - a* v2(3)* v3(2) - b* v2(1)* v3(3) +...
         a *v2(2)* v3(3));
intpt = [xint,yint,zint];
else
   intpt = [];
end

   













