import string
import VTK, VTKDisplay
from DigitalMaterial.PX.Structures.Polycrystals import VTKPolycrystal

class PolycrystalViewer:

    def __init__(self):
        self.filter = PolycrystalFilter()
        self.mapper = VTK.vtkPolyDataMapper()
        self.display = VTKDisplay.VTKDisplay()

    def _MapData(self):

        viewdata = self.filter.GetOutput()

        import types
        if type(viewdata) is types.TupleType:
            viewdata = list(viewdata)
        elif type(viewdata) is not types.ListType:
            viewdata = [viewdata]
        self.actors = []
        for d in viewdata:
            mapper = VTK.vtkPolyDataMapper()
            mapper.SetInput(d)
            mapper.SetLookupTable(self.filter.colorFunction.GetLookupTable())
            actor = VTK.vtkActor()
            actor.GetProperty().SetLineWidth(3.0)
            actor.SetMapper(mapper)
            actor.Update()
            self.actors.append(actor)


    def SetFileName(self, vtkfilename):
        pass

    def SetPolycrystal(self, polycrystal):
        self.filter.SetInput(polycrystal)
         
    def Display(self):
        
        self._MapData()
        self.display.AddActors(self.actors)
        self.display.Show()

    def __repr__(self):
        return 'PolycrystalViewer ( %s )' % (self.filter.polycrystal.ID)

    def __getattr__(self, attr):
        import string
        if string.find(attr, 'Get')==0 and hasattr(self.filter, attr):
            return getattr(self.filter, attr)
        else:
            raise AttributeError, 'attribute %s not found' % attr


class OrientationColoring:

    """OrientationColoring is a class implementing the standard RGB mapping
    of Bunge orientations.  An instance of OrientationColoring is callable,
    returning an index into the associated color table for a given
    orientation"""

    def __init__(self, rRange=[0,1.0], gRange=[0,1.0], bRange=[0,1.0],
                 resolution=8):
        self.rRange = rRange
        self.gRange = gRange
        self.bRange = bRange
        self.resolution = resolution
        self.maxIndex = resolution**3

    def GetLookupTable(self):
        lookupTable = VTK.vtkLookupTable()
        resolution = self.resolution
        lookupTable.SetNumberOfColors(resolution**3)
        lookupTable.Build()

        index = 0
        res = self.resolution
        rMin = self.rRange[0]
        rSlope = (self.rRange[1]-rMin)/(res-1)
        gMin = self.gRange[0]
        gSlope = (self.gRange[1]-gMin)/(res-1)
        bMin = self.bRange[0]
        bSlope = (self.bRange[1]-bMin)/(res-1)
        for red in xrange( res ):
            for green in xrange( res ):
                for blue in xrange( res ):
                    rVal = rMin+rSlope*red
                    gVal = gMin+gSlope*green
                    bVal = bMin+bSlope*blue
                    lookupTable.SetTableValue(index, rVal, gVal, bVal, 1.0)
                    index = index + 1

        #lookupTable.SetTableRange(0, index-1)
        return lookupTable

    def __call__(self, orientation):
        import math
        try:
            res = self.resolution
            red = int(res*orientation[0]/(2.*math.pi))
            green = int(res*(orientation[1]+math.pi)/(2*math.pi))
            blue = int(res*orientation[2]/(2*math.pi))
            val = (red*res + green)*res + blue
            return (float(val)/self.maxIndex)
        except ValueError:
            return 0.


class PolycrystalFilter:

    """PolycrystalFilter converts a polycrystal instance to vtkPolyData
    based on the specified filtering operation"""

    def __init__(self):

        self.polydata = None
        self.polycrystal = None
        self.vtkfilename = None
        self.viewdata = None

    def SetInput(self, polycrystal):

        self.polycrystal = polycrystal
        self.BuildPolyData()

    def GetOutput(self):
        return self.viewdata

    def BuildPolyData(self):

        if self.vtkfilename is not None:
            self.reader = VTK.vtkPolyDataReader()
            self.reader.SetFileName(self.vtkfilename)
            self.polydata = self.reader.GetOutput()
            self.reader.Update()

        elif self.polycrystal is not None:
            self.vtkbuilder = VTKPolycrystal.VTKPolycrystalBuilder()
            self.vtkbuilder.SetPolycrystal(self.polycrystal)
            self.vtkpolycrystal = self.vtkbuilder.GetOutput()
            self.polydata = self.vtkpolycrystal.GrainBoundaries
            self.polydata.Update()

        self.grainIndices = \
          self.polydata.GetCellData().GetFieldData().GetArray('grainIndices')
        
        if hasattr(self.polycrystal, 'orientations'):
            self.colorFunction = OrientationColoring()
        else:
            self.colorFunction = RandomColoring()


    def GetSurfaceGrains(self):

        from DigitalMaterial.PX.Geometry.Tesselation import Tesselation
        interfaces = []
        tess = self.polycrystal.geometry
        numSurfaces = tess.componentGeometry.NumberOfFacets()
        for i in range(numSurfaces):
            interfaces.extend(Tesselation.GetInterfacesOnComponentSurface(tess,
                                                                          i))
        surfacepolydata = self._buildpolydataFromGBIndexSet(interfaces)

        self.viewdata = self._colorpolydata(surfacepolydata)


    def GetGrains(self, indexSet=None):
        if indexSet is None:
            return self.GetSurfaceGrains()
        else:
            GBIndexSet = []
            for gID in indexSet:
                GBIndexSet.extend(\
                    self.polycrystal.GetGrain(gID).GrainBoundaries())
            return self.GetGrainBoundaries(GBIndexSet)


    def GetGrainBoundaries(self, indexSet=None):
        if indexSet is None:
            self.viewdata = self._colorpolydata(self.polydata)
        else:
            polydataSubset = self._buildpolydataFromGBIndexSet(indexSet)
            self.viewdata = self._colorpolydata(polydataSubset)
        
    def GetPlanarCut(self, knife):

        planeNormal, planeOffset = knife.planeNormal, knife.planeOffset

        self.clipPlane = VTK.vtkPlane()
        self.clipPlane.SetOrigin(self.polydata.GetCenter())
        apply(self.clipPlane.SetNormal, planeNormal)

        clipper = VTK.vtkClipPolyData()
        clipper.SetInput(self.polydata)
        clipper.SetValue(planeOffset)
        clipper.SetClipFunction(self.clipPlane)
        #        clipper.GenerateClippedOutputOn()
        clipData = clipper.GetOutput()
        clipData.Update()
        clipGBData = self._colorpolydata(clipData)
        clipGBData.Update()

        cutGBData = self._colorpolydata(knife.GetSurfaceGrains())
        cutGBData.Update()
        
        writer=VTK.vtkPolyDataWriter()
        writer.SetFileName('clipData.vtk')
        writer.SetInput(clipData)
        writer.Write()

        self.viewdata = (clipGBData, cutGBData)


    def _buildpolydataFromGBIndexSet(self, GBIndexSet):

        vtkcells = []
        gIs = VTK.vtkIntArray()
        gIs.SetNumberOfComponents(2)
#        gIs.SetNumberOfTuples(0)

        for i in GBIndexSet:
            vtkcells.append(self.vtkpolycrystal._VTKGBMap[i])

        sampledpolydata = VTK.vtkPolyData()
        sampledpolydata.SetPoints(self.polydata.GetPoints())
        sampledCells = VTK.vtkCellArray()

        for c in vtkcells:
            sampledCells.InsertNextCell(self.polydata.GetCell(c))
            gIs.InsertNextValue(self.grainIndices.GetComponent(c,0))
            gIs.InsertNextValue(self.grainIndices.GetComponent(c,1))

        gIField = VTK.vtkFieldData()
        gIField.SetArray(0, gIs)
        gIField.SetArrayName(0, 'grainIndices')
        sampledpolydata.SetPolys(sampledCells)
        sampledpolydata.GetCellData().SetFieldData(gIField)
        return sampledpolydata


    def _colorpolydata(self, gbpolydata):
        """attach scalar cell data suitable for coloring to input
        vtkPolyData instance; details of coloring are encapsulated in
        the AttributeMap instance passed to _colorpolydata"""

        GBFrontColors = VTK.vtkScalars()
        GBFrontColors.Allocate(gbpolydata.GetNumberOfPolys(), 10)

        grainIndices = \
            gbpolydata.GetCellData().GetFieldData().GetArray('grainIndices')

        for n in range(gbpolydata.GetNumberOfPolys()):
            i0 =  int(grainIndices.GetComponent(n,0))
            i1 =  int(grainIndices.GetComponent(n,1))
            try:
                color = self.colorFunction(self.polycrystal.orientations[i0])
            except:
                color = self.colorFunction(self.polycrystal.orientations[i1])
            GBFrontColors.InsertScalar(n, color)

        gbpolydata.GetCellData().SetScalars(GBFrontColors)
        return gbpolydata

def test(ngrains=10):
            
    from DigitalMaterial.PX.Structures.Polycrystals import PolycrystalSource
    poly = PolycrystalSource.test(ngrains)

    viewer = PolycrystalViewer()
    viewer.SetPolycrystal(poly)
    viewer.GetSurfaceGrains()
    viewer.Display()
    return viewer
    
if __name__ == '__main__':
    viewer = test(10)
    
