# PyVTK.PolyLines
import VTK
from Utils import *
#from Pipelines import *


class PolyLine:

    def __init__(self, points):

	"""Create a PyVTK.PolyLine object from a set of points in R^3;
	input argument "points" is assumed to be a list-like object 
	(i.e., supporting indexing by an integer) with a length-3 list
	named "coord" containing the coordinate data of each point."""

	self.points = points
	self.npoints = len(self.points)
	vtkpoints = VTK.vtkPoints()
	vtklines  = VTK.vtkCellArray()
	vtklines.InsertNextCell(self.npoints)
	
	for n in range(self.npoints):
	    pt = self.points[n]
	    vtkpoints.InsertPoint(n, pt.coord[0], pt.coord[1], pt.coord[2])
	    vtklines.InsertCellPoint(n)

	self.data = VTK.vtkPolyData()
	self.data.SetPoints(vtkpoints)
	self.data.SetLines(vtklines)
        self.data.Update()


class PolyLineCollection:

    def __init__(self, lines, withTube=0, withBox=1, withInteractor=0,
		 saveImages=0):
	
	self.lines = lines
	self.withTube = withTube
	self.withBox  = withBox
        self._build()
        self.currentImageNum = 0

    def _build(self):

	appendData = VTK.vtkAppendPolyData()
	for l in self.lines:
	    pl = PolyLine(l.points)
	    appendData.AddInput(pl.data)

        self.data = appendData.GetOutput()
        self.data.Update()

    def _map(self):

        self.mapper = VTK.vtkPolyDataMapper()
        self.actor = VTK.vtkActor()
        self.mapper.SetInput(self.data)
        self.actor.SetMapper(self.mapper)
        self.actor.GetProperty().SetColor(1.0000,0.7000,0.1600)
	bounds = self.data.GetBounds()
	maxbound = max(bounds)
	self.scalefactor = (1./maxbound)
	self.actor.SetScale(self.scalefactor, self.scalefactor, self.scalefactor)

	if self.withTube:
	    self.tube = VTK.vtkTubeFilter()
            self.tube.SetInput(self.data)
	    self.tube.SetNumberOfSides(8)
	    self.tube.SetRadius(maxbound/50.)
	    self.tubemapper = VTK.vtkDataSetMapper()
            self.tubemapper.SetInput(self.tube.GetOutput())
            self.tubeactor = VTK.vtkActor()
            self.tubeactor.SetMapper(self.tubemapper)
	    self.tubeactor.GetProperty().SetColor(1.0000,0.3882,0.2784)
	    self.tubeactor.SetScale(self.scalefactor, self.scalefactor, self.scalefactor)
            self.tubeactor.RotateY(-50.)
            
	if self.withBox:
	    self.box = VTK.vtkCubeSource()
	    self.box.SetBounds(bounds[0], bounds[1], bounds[2], 
                               bounds[3], bounds[4], bounds[5])
            self.boxmapper = VTK.vtkDataSetMapper()
            self.boxmapper.SetInput(self.box.GetOutput())
            self.boxactor = VTK.vtkActor()
            self.boxactor.SetMapper(self.boxmapper)
	    self.boxactor.GetProperty().SetColor(1.0,1.0,1.0)
	    self.boxactor.SetScale(self.scalefactor, self.scalefactor,
                                   self.scalefactor)
	    self.boxactor.GetProperty().SetOpacity(0.2)
            self.boxactor.RotateY(-50.)
	    #self.boxactor.center = ( (bounds[1]-bounds[0])/2., 
            #(bounds[3]-bounds[2])/2., 
            #			(bounds[5]-bounds[4])/2. )

        #self.display()


    def display(self):

	self.renderPipe.updateRenderer()
	cam = self.renderPipe.GetActiveCamera()
	cam.SetViewAngle(80.)
	cam.Dolly(0.5)
	cam.Azimuth(self.lines.numFrames * 3.0)

	self.renderPipe.AddActor(self.dataPipe.actor)
	if self.withTube:
	    self.renderPipe.AddActor(self.tube.actor)
	if self.withBox:
	    self.renderPipe.AddActor(self.box.actor)

	self.renderPipe.win.Render()
	if self.saveImages: self.saveAsPPM()
	if self.renderPipe.withInteractor:
	    self.vtkiren.Initialize()
	    self.vtkiren.Start()


    def saveAsPPM(self):

	filename = 'tmp/image.%04d.ppm' % (self.curImageNum)
	self.renderPipe.win.SetFileName(filename)
	self.renderPipe.win.SaveImageAsPPM()
	self.renderPipe.curImageNum = self.renderPipe.curImageNum + 1


