#include <stdio.h>
#include <math.h>
#include <assert.h>
#include "SquareGridField.h"
#include "SimplePeriodicBC.h"
#include "ConstantFieldInitializer.h"
#include "GnuPlotter.h"
#include "BallisticDeposition.h"
#include "ExplicitEuler.h"
#include "FieldInitializer.h"

int main(void);



class PhaseFieldInitializer : public virtual FieldInitializer {
	protected:
		Real slab_bottom, slab_surface, slab_peak;
	public:
		PhaseFieldInitializer(void);
		PhaseFieldInitializer(Real bottom, Real surface, Real peak);
		Real FieldValue(const Point &x) const {
			if (x(0)>0.5 && x(0)<0.6) 
				return((x(DIMENSION-1)>slab_bottom && x(DIMENSION-1)<slab_peak)?1.0:0.0); 
			else
				return((x(DIMENSION-1)>slab_bottom && x(DIMENSION-1)<slab_surface)?1.0:0.0); 
		}
		Real GetBottom(void) { return(slab_bottom); }
		Real GetSurface(void) { return(slab_surface); }
		Real GetPeak(void) { return(slab_peak); }

		bool SetBottom(Real bot) { slab_bottom=bot; return(true); }
		bool SetSurface(Real surf) { slab_surface=surf; return(true); }
		bool SetPeak(Real peak) { slab_peak=peak; return(true); }
};

PhaseFieldInitializer::PhaseFieldInitializer(void) : 
	slab_bottom(0.1), slab_surface(0.3), slab_peak(0.6)
{ ; } 

PhaseFieldInitializer::PhaseFieldInitializer(Real bottom, Real surface, Real peak) : 
	slab_bottom(bottom), slab_surface(surface), slab_peak(peak) 
{ ; } 


class IncidentFieldInitializer : public virtual FieldInitializer {
	protected:
		Real source_from, source_to, source_strength;
		Point velocity;
		Real slab_peak, slab_bottom;
	public:
		IncidentFieldInitializer(Real from, Real to, Real strength, Real peak, Real bottom, Point velo);
		Real FieldValue(const Point &r) const {
			Real y;
			if (r(DIMENSION-1)>source_from && r(DIMENSION-1)<source_to) 
				y=-source_strength*(source_to-r(DIMENSION-1))/velocity(DIMENSION-1);
			else if (r(DIMENSION-1)>slab_peak+slab_bottom && r(DIMENSION-1)<source_from) 
				y=-source_strength*(source_to-source_from)/velocity(DIMENSION-1);
			else 
				y=0.0;
			return(y);
		}
		Real GetFrom(void) { return(source_from); }
		Real GetTo(void) { return(source_to); }
		Real GetStrength(void) { return(source_strength); }
		Point GetVelocity(void) { return(velocity); }
		Real GetSlabPeak(void) { return(slab_peak); }
		Real GetSlabBottom(void) { return(slab_bottom); }

		bool SetFrom(Real from) { source_from=from; return(true); }
		bool SetTo(Real to) { source_to=to; return(true); }
		bool SetStrength(Real strength) { source_strength=strength; return(true); }
		bool SetVelocity(const Point &vel) { velocity=vel; return(true); }
		bool SetSlabPeak(Real peak) { slab_peak=peak; return(true); }
		bool SetSlabBottom(Real bottom) { slab_bottom=bottom; return(true); }
};

IncidentFieldInitializer::IncidentFieldInitializer(Real from, Real to, Real strength, Real peak, Real bottom, Point velo) : 
	source_from(from), source_to(to), source_strength(strength), velocity(velo), slab_peak(peak), slab_bottom(bottom) 
{ ; }

int main(void) 
{
	int i, n;
	Real source_from, source_to, source_strength;
	//char ccc;
	Index gridsize(48);
	Point velocity;
	gridsize[1]*=2;
	Point simulationbox(1.0);
	simulationbox[1]*=2.0;
	Real hh, wp;
	GnuSurfacePlot plotphasefield, plotincidentfield;
	GnuContourPlot contphasefield, contincidentfield;
	//GnuSurfacePlot checkworkspace;
	SquareGridField phasefield, incidentfield;
	FieldCollection fields;
	fields.AddField(&phasefield);
	fields.AddField(&incidentfield);
	//Field *fields[2]={&phasefield, &incidentfield};
	//Field *pworkspace;

	SimplePeriodicBC spbc(simulationbox);


	BallisticDeposition bd;
	cout << "Enter surface tension " << endl; cin >> wp;
	assert(bd.SetSurfaceTension(wp));
	cout << "Enter potential strength " << endl; cin >> wp;
	assert(bd.SetPotentialStrength(wp));
	cout << "Enter conversion rate " << endl; cin >> wp;
	assert(bd.SetConversionRate(wp));
	cout << "Enter source strength " << endl; cin >> source_strength;
	source_from=1.8;
	source_to=1.9;
	cout << "Enter incident velocity " << endl; cin >> velocity;
	assert(bd.SetIncidentVelocity(velocity));
	AddSource Source;
	assert(Source.SetStrength(source_strength));
	assert(Source.SetFrom(source_from));
	assert(Source.SetTo(source_to));
	assert(bd.SetSource(&Source));

	phasefield.Attach(&plotphasefield); 
	phasefield.Attach(&contphasefield); 
	plotphasefield.SetTitle("PhaseField(t)");
	plotphasefield.SetAngles(60.0, 210.0);
	contphasefield.SetTitle("PhaseField(t)");
	assert(phasefield.SetNumberOfFields(1));
	assert(phasefield.SetLogicalSize(gridsize));
	assert(phasefield.SetBoundaryCondition(&spbc));
	PhaseFieldInitializer phini;
	assert(phasefield.Initialize(phini));
	assert(bd.SetWorkSpace(fields));

	incidentfield.Attach(&plotincidentfield);
	plotincidentfield.SetTitle("IncidentField(t)");
	incidentfield.Attach(&contincidentfield);
	contincidentfield.SetTitle("IncidentField(t)");
	assert(incidentfield.SetNumberOfFields(1));
	assert(incidentfield.SetLogicalSize(gridsize));
	assert(incidentfield.SetBoundaryCondition(&spbc));
	IncidentFieldInitializer inini(source_from, source_to, source_strength, phini.GetPeak(), phini.GetBottom(), velocity);
	assert(incidentfield.Initialize(inini));


	ExplicitEuler euler;
	assert(euler.SetStepSize(hh=0.0));
	assert(euler.SetRightHandSide(&bd));
	assert(euler.SetWorkSpace(fields));
	//(pworkspace=euler.GetWorkSpace()[1])->Attach(&checkworkspace); checkworkspace.SetTitle("EulerWorkSpace IncidentField");


	phasefield.Notify(); incidentfield.Notify(); //pworkspace->Notify(); 
	fflush(stdin); getchar();
	
	for (n=1, hh=1e-9; hh>=0.0; cin >> hh) {
		euler.SetStepSize(hh);
		for (n=1; n>0; cin >> n) {
			for (i=0; i<n; i++) {
				euler.Move(fields);
				if (!(i%200)) { 
					phasefield.Notify(); incidentfield.Notify(); //pworkspace->Notify(); 
				}
			}
			phasefield.Notify(); incidentfield.Notify(); //pworkspace->Notify();
			//cout << phasefield.DumpToFile("muell.dat") << endl;
			cout << "Enter number of time steps (old=" << n << ") -- negative to exit" << endl;
		}
		cout << "Enter stepsize (old=" << hh << ") -- negative for exit" << endl;
	}

	phasefield.Detach(&plotphasefield);
	incidentfield.Detach(&plotincidentfield);
	phasefield.Detach(&contphasefield);
	incidentfield.Detach(&contincidentfield);
	//pworkspace->Detach(&checkworkspace);

	return(1);
}
