#ifdef _MSC_VER
#pragma warning(disable:4786)   // remove STL warnings
#endif //_MSC_VER

#include <stdlib.h>
#include <assert.h>
#include <iostream>
#include <algorithm>
#include <math.h>
#include <stdexcept>

#include "AdvancingFront.h"
using namespace std ;

const int AdvancingFront::NRandom = 3002;

AdvancingFront::AdvancingFront(set<Face>& front0, vector< Vector<int> >& pos0,
                               int cutoff, int nAtomic, int atomicCutoff,
                               ostream &outStream) :
    cutoff(cutoff), nAtomic(nAtomic), outStream(outStream)
{
    if(atomicCutoff == -1) this->atomicCutoff = cutoff;

    // watch out for integer overflow:
    const int nBits = 8 * sizeof(LongInt) - 1;
    const double max = pow(2.0, nBits);
    const double max0 = 72.0 * pow(cutoff, 5.0);
    const double max1 = 216.0 * pow(cutoff, 4.0) * NRandom;
    assert(max0 < max && max1 < max); // overflow error!
    if(max0 > max || max1 > max)
        throw runtime_error("overflow risk.");
    nAboveTests = 0;
    nAboveTests1 = 0;
    nInsideTests = 0;
    nInsideTests1 = 0;
    this->pos0 = pos0 ;
    
    pos = this->pos0 ;
    front.swap(front0) ;

    pbc[0] = pbc[1] = pbc[2] = false;

    CalculateEnclosingBox();

    faceNodes.resize(pos.size());
    for(int coord = 0; coord < 3; coord++)
    {
        assert(box[coord] > 2 * cutoff); // box too small!"
    }
    closestPair = pair<int, int>(-1, -1);
    MakeNeighborList();
    Randomize();
    // create initial front if no fronts are given:
    noFrontGiven = false ;
    if (front.size() == 0) {
        Face f = FindFirstFace();
        outStream << "Automatically generated the first face: " << f << endl;
        front.insert(f);
        f.Flip();
        front.insert(f);
        noFrontGiven = true ;
    }
    typedef set<Face>::const_iterator BI;
    for(BI i = front.begin(); i != front.end(); ++i)
        for(int n = 0; n < 3; n++)
            faceNodes[i->nodes[n]]++;
}


void AdvancingFront::CalculateEnclosingBox()
{
    if(pos.size() == 0) throw runtime_error("no points");
    Vector<int> max(pos[0]);
    Vector<int> min(pos[0]);
    for(int i=1; i<pos.size(); ++i)
    {
        for(int j=0; j<3; ++j)
        {
            if(pos[i][j] < min[j]) min[j] = pos[i][j];
            if(pos[i][j] > max[j]) max[j] = pos[i][j];
        }
    }
    for(int j=0; j<3; ++j) 
    {
        enclosingBox[j] = max[j] - min[j];
        box[j] = enclosingBox[j] + atomicCutoff/2; // so that we can always mesh 
    }
}

void AdvancingFront::SetBoxSize(const int box[3], bool wrap[3])
{
    for(int j=0; j<3; ++j) 
    {  
        this->pbc[j] = (wrap ? wrap[j] : true);
        if(pbc[j])
        {
            this->box[j] = box[j];
            if(box[j] < enclosingBox[j]) throw runtime_error("bad boundary conditions");
            if(box[j] > enclosingBox[j] + cutoff)
                throw runtime_error("I won't be able to mesh across the periodic boundaries");
        }
    }
}

Vector<int> AdvancingFront::Diff(int a1, int a2) const
{
    Vector<int> d = pos[a1] - pos[a2];
    for(int coord = 0; coord < 3; coord++)
    {
        int L = box[coord];
        int& x = d[coord];
        if(2 * x > L)
            x -= L;
        else if(2 * x <= -L)
            x += L;
    }
    return d;
}

int AdvancingFront::Distance2(int a1, int a2) const
{
    return Diff(a1, a2).Length2();
}

void AdvancingFront::Randomize()
{
    pos1.resize(pos.size());
    for(int a = 0; a < pos.size(); a++)
        for(int coord = 0; coord < 3; coord++)
        {
            pos1[a][coord] = rand() % (2 * NRandom) - NRandom;
            pos[a][coord] += rand() % 3 - 1;
        }
}

void AdvancingFront::MakeNeighborList()
{
    outStream << "Building Neighbor list ...";
    int cutoff2 = cutoff * cutoff;
    int atomicCutoff2 = atomicCutoff * atomicCutoff;
    neighbors.resize(pos.size());
    int dMin2 = cutoff2;
    int n1 = 0;
    for (; n1 < nAtomic; n1++)
    {
        int n2 = n1 + 1;
        for (; n2 < nAtomic; n2++)
        {
            int d2 = Distance2(n1, n2);
            if (d2 < atomicCutoff2)
            {
                neighbors[n1].insert(n2);
                neighbors[n2].insert(n1);
            }
            if (d2 < dMin2)
            {
                dMin2 = d2;
                closestPair = pair<int, int>(n1, n2);
            }
        }
        for(; n2 < pos.size(); n2++)
            if(Distance2(n1, n2) < cutoff2)
            {
                neighbors[n1].insert(n2);
                neighbors[n2].insert(n1);
            }
    }
    for(; n1 < pos.size(); n1++)
        for(int n2 = n1 + 1; n2 < pos.size(); n2++)
        {
            int d2 = Distance2(n1, n2);
            if (d2 < cutoff2)
            {
                neighbors[n1].insert(n2);
                neighbors[n2].insert(n1);
                if(nAtomic == 0 && d2 < dMin2)
                {
                    dMin2 = d2;
                    closestPair = pair<int,int>(n1, n2);
                }
            }
        }
    assert(closestPair.first != -1); // Baaaad!
    outStream << "done" << endl;
}


Face AdvancingFront::FindFirstFace()
{
    int n0 = closestPair.first;
    int n1 = closestPair.second;
    int n2 = -1;
    double dMin = 2 * cutoff;
    typedef set<int>::const_iterator SI;
    for (SI i = neighbors[n0].begin(); i != neighbors[n0].end(); ++i)
    {
        const int n = *i;
        if (neighbors[n1].find(n) != neighbors[n1].end())
        {
            double d = sqrt(Distance2(n0, n)) + sqrt(Distance2(n1, n));
            if (d < dMin)
            {
                dMin = d;
                n2 = n;
            }
        }
    }
    assert(n2 != -1); // not found!
    return Face(n0, n1, n2);
}
 
void AdvancingFront::Advance()
{
    outStream << "Advancing ..." << endl;
    while(front.size() > 0)
    {
        // we first try to grow the first face of the front
        set<Face>::iterator face = front.begin();
        bool done = false;
        while(!done) {
            int n0 = face->nodes[0];
            int n1 = face->nodes[1];
            int n2 = face->nodes[2];
            set<int> nb;
            // the candidate node is the reunion of the neighbor lists.
            typedef set<int>::const_iterator SI;
            nb = neighbors[n0] ;
            for(SI i = neighbors[n1].begin(); i != neighbors[n1].end(); ++i)
                nb.insert(*i);
            for(SI j = neighbors[n2].begin(); j != neighbors[n2].end(); ++j)
                nb.insert(*j);
			for(int node=0; node<3; ++node) nb.erase(face->nodes[node]);

            done = GrowFace(face, nb) ;
            if (!done) ++face ; // if we cannot grow this face, try the next one (we'll try this later).
            assert( face != front.end() || front.size() == 0); 
        }
        if (noFrontGiven)
            CheckFirstTetrahedron();
        noFrontGiven = false;
    }
    for(int n = 0; n < faceNodes.size(); n++)
        assert(0 == faceNodes[n]);
    outStream << "\"Above\" tests: " << nAboveTests
         << " (" << nAboveTests1 << ")\n";
    outStream << "\"Inside\" tests: " << nInsideTests
         << " (" << nInsideTests1 << ")\n";
    pos.swap(pos0) ;
    LongInt volume = WriteSummary(false);
    assert(volume == LongInt(box[0]) * box[1] * box[2] * 6);
    ApplyPeriodicBoundaryConditions();
    WriteSummary(true);
}

LongInt AdvancingFront::WriteSummary(bool print)
{
    LongInt volume = 0;
    int nSlivers = 0;
    typedef list<Tetrahedron>::const_iterator TI;
    for(TI t = tetrahedra.begin(); t != tetrahedra.end(); ++t)
    {
        const int* nodes = t->nodes;
        Vector<int> v1 = Diff(nodes[3], nodes[0]);
        Vector<int> v2 = Diff(nodes[1], nodes[0]).Cross(
                         Diff(nodes[2], nodes[0]));
        LongInt vol = 0;
        for(int coord = 0; coord < 3; coord++)
            vol += v1[coord] * v2[coord]; 

        volume += vol;
        // it is possible that volumes become negative if measured with
        // the non randomized positions, but this shoudl happen only with slivers.
        if(vol <= 0)   
            nSlivers++;
    }
    if(print)
    {
        outStream << "Elements: " << tetrahedra.size() << endl;
        outStream << "Volume: " << volume / 6.0 << endl;
        outStream << "Zero volume slivers: " << nSlivers << endl;
    }
    return volume;
}

void AdvancingFront::ApplyPeriodicBoundaryConditions()
{
    typedef list<Tetrahedron>::iterator TI;
    for (TI t = tetrahedra.begin(); t != tetrahedra.end();)
    {
        const int* nodes = t->nodes;
        int coord = 0;
        for (; coord < 3; coord++)
            if (!pbc[coord])
            {
                int xMax = pos[nodes[0]][coord];
                int xMin = xMax;
                for (int n = 1; n < 4; n++)
                {
                    int x = pos[nodes[n]][coord];
                    if (x > xMax)
                        xMax = x;
                    else if (x < xMin)
                        xMin = x;
                }
                if (xMax - xMin > box[coord] / 2)
                    break;
            }
        if (coord < 3)
            t = tetrahedra.erase(t);
        else
            ++t;
    }
}

void AdvancingFront::MakeTetrahedron(set<Face>::iterator f, int node)
{
    const int* nodes = f->nodes;
    tetrahedra.push_front(Tetrahedron(nodes[0], nodes[1], nodes[2], node));
//    cout << "T: " << Tetrahedron(nodes[0], nodes[1], nodes[2], node) << endl ;
    for(int n = 0; n < 3; n++)
    {
        Face newFace(nodes[(n + 1) % 3], nodes[(n + 2) % 3], node);
        Face oppositeFace = newFace;
        oppositeFace.Flip();
        set<Face>::iterator f2 = front.find(oppositeFace);
        if(f2 != front.end())
            DeleteFace(f2);
        else
        {
            front.insert(newFace);
            for(int n = 0; n < 3; n++)
                faceNodes[newFace.nodes[n]]++;
        }
    }
    DeleteFace(f);
}

inline LongInt iabs(LongInt x) {return x > 0 ? x : x * (-1);}

bool AdvancingFront::GrowFace(set<Face>::iterator f, set<int>& nb)
{
    /*! first time through, we check all the candidated neighbors that are 
    on right side of the face and build the new set of neighbors; we then
    perform Inside Sphere check on the new set */ 
    
    const int* nodes = f->nodes;
    const int node0 = nodes[0];
    const int node1 = nodes[1];
    const int node2 = nodes[2];
    const Vector<int> d10 = Diff(node1, node0);
    const Vector<int> d20 = Diff(node2, node0);
    Vector<int> d30;
    const Vector<int> dd10 = pos1[node1] - pos1[node0];
    const Vector<int> dd20 = pos1[node2] - pos1[node0];
    int node3;
    set<int>::iterator iter = nb.begin();
	set<int> tmp;
    for(; iter != nb.end(); ++iter)
    {
        node3 = *iter;
        d30 = Diff(node3, node0);
        bool tooBig = false;
        for(int coord = 0; coord < 3; coord++)
        {
            int L = box[coord]; // ??? > or >= ..., L > 2 * dMax ???
            if(iabs(d30[coord] - d10[coord]) > L / 2 ||
               iabs(d30[coord] - d20[coord]) > L / 2 ||
               iabs(d30[coord]) > L / 2)
                tooBig = true; // break; ???
        }
        if(tooBig){
            continue;
        }
        LongInt above = Above(d10, d20, d30) ;
        nAboveTests++;
        if (above > 0) 
            tmp.insert(*iter) ;
        else if (above == 0) {
            nAboveTests1++;
            Vector<int> dd30 = pos1[node3] - pos1[node0];
            LongInt above1 = Above1(d10, d20, d30, dd10, dd20, dd30);
            if (above1 == 0) {
                outStream << "Above1 failed: need 2nd order perturbation" << endl ;
            }

            else if (above1 > 0)
                tmp.insert(*iter) ;
        }
    }
	nb.swap(tmp);
    if(nb.size() == 0)
    {
        outStream << "Cannot grow face: " << node0 << ',' << node1 << ',' << node2 << endl;
        outStream << "will use next face in the front to grow" << endl;
        return false;
    }
    iter = nb.begin();
    int node4;
 
    node3 = *iter ;
    d30 = Diff(node3, node0);
    for(++iter; iter != nb.end(); ++iter)
    {
        node4 = *iter;
        Vector<int> d40 = Diff(node4, node0);
        bool tooBig = false;
        for(int coord = 0; coord < 3; coord++)
        {
            int L = box[coord];
            if(iabs(d40[coord] - d10[coord]) > L / 2 ||
               iabs(d40[coord] - d20[coord]) > L / 2 ||
               iabs(d40[coord]) > L / 2)
                tooBig = true;
        }
        if(tooBig)
        {
            continue;
        }
        LongInt inside = Inside(d10, d20, d30, d40);
        nInsideTests++;
        if(inside == 0)
        {
            // check first order correction:
            nInsideTests1++;
            Vector<int> dd30 = pos1[node3] - pos1[node0];
            Vector<int> dd40 = pos1[node4] - pos1[node0];
            inside = Inside1(d10, d20, d30, d40, dd10, dd20, dd30, dd40);
            if (inside == 0) outStream << "Inside1 fail" << endl ;
            assert (inside != 0);
        }
        if(inside > 0)
        {
            node3 = node4;
            d30 = Diff(node3, node0);
        }
    }
    MakeTetrahedron(f, node3);
    return true ;
}


void AdvancingFront::CheckFirstTetrahedron()
{
    // make sure that no points are inside the first tetrahedron. If
    // there is such a point, then our initial guess for the first face
    // was wrong!!
    int* nodes = tetrahedra.front().nodes;
    int node0 = nodes[0];
    int node1 = nodes[1];
    int node2 = nodes[2];
    int node3 = nodes[3];
    set<int> nb;
    typedef set<int>::const_iterator SI;
    int n ;
    for (n = 0; n < 4; n++)
        for(SI i = neighbors[nodes[n]].begin();
            i != neighbors[nodes[n]].end(); ++i)
            nb.insert(*i);
    for (n = 0; n < 4; n++)
        nb.erase(nodes[n]);
    const Vector<int> d10 = Diff(node1, node0);
    const Vector<int> d20 = Diff(node2, node0);
    const Vector<int> d30 = Diff(node3, node0);
    const Vector<int> dd10 = pos1[node1] - pos1[node0];
    const Vector<int> dd20 = pos1[node2] - pos1[node0];
    const Vector<int> dd30 = pos1[node3] - pos1[node0];
    typedef set<int>::const_iterator SI;
    int node4;
    for(SI iter = nb.begin(); iter != nb.end(); ++iter)
    {
        node4 = *iter;
        Vector<int> d40 = Diff(node4, node0);
        LongInt inside = Inside(d10, d20, d30, d40);
        if(inside == 0)
        {
            // check first order correction:
            Vector<int> dd40 = pos1[node4] - pos1[node0];
            inside = Inside1(d10, d20, d30, d40, dd10, dd20, dd30, dd40);
        }
        assert(inside != 0);
        assert(inside < 0); // oops!
    }
}

void AdvancingFront::DeleteFace(set<Face>::iterator f)
{
    for(int n = 0; n < 3; n++)
    {
        int node = f->nodes[n];
        if(--faceNodes[node] == 0)
        {
            set<int>& nb = neighbors[node];
            typedef set<int>::const_iterator SI;
            for(SI i = nb.begin(); i != nb.end(); ++i)
                neighbors[*i].erase(node);
            nb.erase(nb.begin(), nb.end());
        }
    }
    front.erase(f);
}

LongInt AdvancingFront::Above(const Vector<int>& d10,
                              const Vector<int>& d20,
                              const Vector<int>& d30)
{
    return (
        LongInt(d10[0]) * (d20[1] * d30[2] - d20[2] * d30[1]) -
        LongInt(d20[0]) * (d10[1] * d30[2] - d10[2] * d30[1]) +
        LongInt(d30[0]) * (d10[1] * d20[2] - d10[2] * d20[1]));
}

LongInt AdvancingFront::Above1(const Vector<int>& d10,
                               const Vector<int>& d20,
                               const Vector<int>& d30,
                               const Vector<int>& dd10,
                               const Vector<int>& dd20,
                               const Vector<int>& dd30)
{
    LongInt above1 =
        LongInt(dd10[0]) * (d20[1] * d30[2] - d20[2] * d30[1]) -
        LongInt(dd20[0]) * (d10[1] * d30[2] - d10[2] * d30[1]) +
        LongInt(dd30[0]) * (d10[1] * d20[2] - d10[2] * d20[1]) +
        LongInt(d10[0]) * ((dd20[1] * d30[2] - dd20[2] * d30[1]) +
                           (d20[1] * dd30[2] - d20[2] * dd30[1])) -
        LongInt(d20[0]) * ((dd10[1] * d30[2] - dd10[2] * d30[1]) +
                           (d10[1] * dd30[2] - d10[2] * dd30[1])) +
        LongInt(d30[0]) * ((dd10[1] * d20[2] - dd10[2] * d20[1]) +
                           (d10[1] * dd20[2] - d10[2] * dd20[1]));

  return (above1) ;
}

LongInt AdvancingFront::Inside(const Vector<int>& d10,
                               const Vector<int>& d20,
                               const Vector<int>& d30,
                               const Vector<int>& d40)
{
    return (
        d10.Length2() * Above(d20, d30, d40) -
        d20.Length2() * Above(d10, d30, d40) +
        d30.Length2() * Above(d10, d20, d40) -
        d40.Length2() * Above(d10, d20, d30));
}

LongInt AdvancingFront::Inside1(const Vector<int>& d10,
                                const Vector<int>& d20,
                                const Vector<int>& d30,
                                const Vector<int>& d40,
                                const Vector<int>& dd10,
                                const Vector<int>& dd20,
                                const Vector<int>& dd30,
                                const Vector<int>& dd40)
{
    return 2 * (
        (d10 * dd10) * Above(d20, d30, d40) -
        (d20 * dd20) * Above(d10, d30, d40) +
        (d30 * dd30) * Above(d10, d20, d40) -
        (d40 * dd40) * Above(d10, d20, d30)) +
        d10.Length2() * Above1(d20, d30, d40, dd20, dd30, dd40) -
        d20.Length2() * Above1(d10, d30, d40, dd10, dd30, dd40) +
        d30.Length2() * Above1(d10, d20, d40, dd10, dd20, dd40) -
        d40.Length2() * Above1(d10, d20, d30, dd10, dd20, dd30);
}

