#ifndef ADVANCINGFRONT_H
#define ADVANCINGFRONT_H

#include "Tetrahedron.h"
#include "Vector.h"
#include "Face.h"
#include <set>
#include <list>
#include <vector>
#include <utility>

#ifndef WIN32
typedef long long int LongInt;
#else
typedef __int64 LongInt;
#endif

//! A class to mesh a collection of points in 3D.
/*! Here is how this advancing front algorihm works. A front is a collection
of oriented faces that belong to the mesh. The goal is to advance this front
until the number of faces in the front reaches 0. To advance a face, the goal
is to identify a node that will form a new tetrahedron with it. Then the 3 
new faces (pointing outwards) are inserted in the front and the original 
face is removed. If a new face is already in the front with the opposite 
orientation, they cancel each other. 

One can start with a given front (for example if one wants to mesh a finite 
volume, one shoudl first mesh its surface, which is the front we start with)
or an empty front. In this latter case, a random face is generated starting 
from the closest pair of points. 

Identifying the correct node to grow a face is a difficult task in general: 
we have to make sure that we grow it in the right direction and (more difficult)
that the front never crosses itself. However, there is a special method that
makes  things in principle much simpler: this is the Delaunay triangulation.
A mesh is said to be a Delaunay triangulation if for each tetrahedron, the
circumscribing sphere does not enclose any other point of the mesh. If we grow
a face according to this criterion, then  there is no need to check that the 
front doesn't cross itself, because the Delaunay triangulation is unique.
(so if an initial front is provided, if must be compatible with the Delaunay 
criterion).

So the method to grow a face is to start with a candidate node for building
the next tetrahedron and make sure that no other point lies inside the 
circumscribing sphere of this potential tetrahedron.

However, it may happen that another point lies just on the sphere: this kind 
of degeneracy is likely to be present when we mesh a lattice.
In order to lift these degeneracies, one assign to each point a random 
perturbation vector. If a degeneracy occurs with a point, one checks what 
is the result on the first order of the perturbation. Given a perturbation for
all the points, the Delaunay mesh will be unique (the need to rely on the
2nd order perturbation is very unlikely, unless the corners of a face are aligned,
ie a face of zero area.).

Because this method involves many comparisons, we work with integer numbers. 
So prior to transforming the position of the points from double to integers, 
the user multiplies them by some arbitrary number. This number should be large
enough so that the relative distance between points is not too much modified
(and no points overlap), but it should not be too large because we would face
potential integer overflow problems (in the Inside function, we multiply 
up to 5 numbers).


In order the meshing be not too slow, we store a neighbor list for each point.
The building of this neighbor list is quite primitive at this time, but this
doesn't seem to be a big issue. A faster implementation could be done
with less calls of STL containers and functions.

The AdvancingFront internally uses periodic boundary conditions for the meshing process.
This means that we mesh a box which is slightly larger than the enclosing box, 
or a box provided by the user. In the end, the tetrahedra that cross the non periodic 
boundaries are removed.

Design Observation: in the near future we may want to have a different algorithm
than the Delaunay advancing front (for remeshing locally, for example). So this 
class should be renamed DelaunayAdvancingFront.

*/
class AdvancingFront
{
public:
    //! The constructor
    AdvancingFront(std::set< Face >& front, std::vector< Vector<int> >& pos0, 
                   int cuu, int nAtomic = 0, int atomicCutoff = -1, 
                   std::ostream &outStream = std::cerr);
    //! Specify a box size and tell along which direction we want to wrap (all by default)
    void SetBoxSize(const int box[3], bool wrap[3] = 0);
    //! The command to start meshing
    void Advance();
    //! Get the mesh
    const std::list<Tetrahedron>& GetTetrahedra() const {return tetrahedra;}
private:
    //! Make the neighbor list of the nodes. 
    /*! In order to identify the candidate nodes when we grow a face, 
    we use an internal neighbor list. It is built in a very crude way at this time.
    */
    void MakeNeighborList();
    //! Move the points randomly by one unit and assign them a virtual perturbation
    /*! Actually moving the points breaks the possible alignments, which are bad
    for the mesher.
    */
    void Randomize();
    //! Given a face, find the node on which we build the next tetrahedron
    bool GrowFace(std::set<Face>::iterator f, std::set<int>& nb); 
    //! Create the new tetrahedron and update the front
    void MakeTetrahedron(std::set<Face>::iterator f, int node);
    //! Return the separation between two points assuming periodic boundary conditions.
    Vector<int> Diff(int a1, int a2) const;
    //! Return the squared distance between 2 points, assuming periodic boundary cond.
    int Distance2(int a1, int a2) const;
    //! Remove a face from the front and update the neighbor list.
    void DeleteFace(std::set<Face>::iterator f);
    //! Remove all the tetrahedra that cross the box.
    void ApplyPeriodicBoundaryConditions();
    //! Compute the total volume and optionally print the number of elements and such. 
    LongInt WriteSummary(bool print);
    Face FindFirstFace();
    void CheckFirstTetrahedron();
    LongInt Above(const Vector<int>& d10,
                  const Vector<int>& d20,
                  const Vector<int>& d30);
    LongInt Above1(const Vector<int>& d10,
                   const Vector<int>& d20,
                   const Vector<int>& d30,
                   const Vector<int>& dd10,
                   const Vector<int>& dd20,
                   const Vector<int>& dd30);
    LongInt Inside(const Vector<int>& d10,
                   const Vector<int>& d20,
                   const Vector<int>& d30,
                   const Vector<int>& d40);
    LongInt Inside1(const Vector<int>& d10,
                    const Vector<int>& d20,
                    const Vector<int>& d30,
                    const Vector<int>& d40,
                    const Vector<int>& dd10,
                    const Vector<int>& dd20,
                    const Vector<int>& dd30,
                    const Vector<int>& dd40);
private:
    //! The current state of the front
    std::set<Face> front;

    std::vector< Vector<int> > pos0;
    std::vector< Vector<int> > pos;
    std::vector< Vector<int> > pos1;

    //! The box used for the peroidic boundary conditions
    /*! Either it is provided by the user, of it is slightly bigger
    than the enclosing box around the points
    */
    Vector<int> box;
    //! Enclosing box
    Vector<int> enclosingBox;
    void CalculateEnclosingBox();

    //! Tells which directions are periodic
    Vector<bool> pbc;
    //! The current state of the mesh
    std::list<Tetrahedron> tetrahedra;
    //! The neighbor list
    std::vector< std::set<int> > neighbors;
    std::vector<int> faceNodes;
    //! The cutoff distance for the internal neighbor list
    const int cutoff;
    //! The number nonlocal atoms (used it atomicCutoff is provided).
    const int nAtomic;
    //! The cutoff distance for pairs of non local atoms.
    /*! One may want to set it smaller than cutoff to prevent the number
    of neighbors of non local atoms to be too big.
    */
    int atomicCutoff;
    //! the index of the points that are the closest
    std::pair<int, int> closestPair;
    //! the amplitude of the perturbation
    const static int NRandom;

    //! The stream where we output the trace of the execution
    std::ostream &outStream;
    //! number of calls to Above()
    int nAboveTests;
    //! number of calls to Above1()
    int nAboveTests1;
    //! number of calls to Inside()
    int nInsideTests;
    //! number of calls to Inside1()
    int nInsideTests1;
    //! Tells whether we passed a front in the constructor
    bool noFrontGiven;
};
  
#endif //!ADVANCINGFRONT_H
