// MeshData.cpp: implementation of the MeshData class.
//
//////////////////////////////////////////////////////////////////////

#include "DMFirstHeader.h"
#include <iostream>
#include <math.h>
#include <algorithm>
#include <string>
#include <map>
#include <fstream>
#include "DigitalMaterial.h"
#include "DMArray.h"
#include "MeshData.h"
#include "QCListOfAtoms.h"
#include "ListOfSlavedAtoms.h"
#include "Element.h"
#include "BravaisLatticeWithBasis.h"
#include "BoundaryConditions.h"
#include "QCNeighborList.h"
//#include "QCNeighborList2.h"

#include "NeighborLocator.h"
#include "SerialBuilder.h"

#include "DMLastHeader.h"

using namespace std;

static SerialBuilder<MeshData> *_MeshDataCreator = 
        new SerialBuilder<MeshData>("MeshData");

MeshData::MeshData()
{
    volume = 0.0;
    elementsWeight = 0.0;
    nAtomicElements = 0;
    nAtomic = 0;
}

MeshData::~MeshData()
{
}

void MeshData::CreateElement(const vector<int>& ndList,
                             const QCListOfAtoms* masters,
                             const BravaisLatticeWithBasis* lattice)
{
    const DMArray<double, DIMENSION>& positions =
        *(masters->GetCartesianPositions());
    const BoundaryConditions* bc = masters->GetBoundaryConditions();
    int node, nodes[DIMENSION+1];
    ASSERT(DIMENSION + 1 == ndList.size());
    for(node = 0; node < DIMENSION + 1; ++node) nodes[node] = ndList[node];
    sort(nodes, nodes+DIMENSION+1);
    double diff[DIMENSION];
    TinyVector<int> nodesLocalCoord[DIMENSION];
    for(node = 0; node < DIMENSION; node++)
    {
        for(int coord = 0; coord < DIMENSION; coord++)
            diff[coord] = positions[nodes[node]][coord] -
                positions[nodes[DIMENSION]][coord];
        bc->DifferenceBoundaryConditions(&diff, 1);
        lattice->ClosestLatticeSiteIndices(nodesLocalCoord[node], diff);
    }
    nAtomic = masters->GetNonLocals()->GetNumber();
    Element element(nodes, nAtomic, nodesLocalCoord);
    double weight = element.GetWeight();
    // don't create element if it has only atomic corners and zero weight.
    // Note that the nodes have been sorted in ascending order.
    // Note also that we accept slivers for now (and delete them in PostProcessMesh).
    if(weight < 1e-9 && nodes[3] < nAtomic)
    {
        nAtomicElements++;
        return;
    }
    elements.push_back(element);
    lattices.push_back(lattice);
    elementsWeight += weight;
}



void MeshData::DecorateElement(Element& element,
                               const QCListOfAtoms* masters)
{
    const DMArray<double, DIMENSION>& positions =
        *(masters->GetCartesianPositions());
    const BoundaryConditions* bc = masters->GetBoundaryConditions();
    NeighborLocator* nl = masters->GetNonLocals()->GetNeighborLocator();
    const int length = 8000;
    int neighbors[length];
    double disp[length][DIMENSION];
    double dispSq[length];
    volume += element.CalculateVolume(masters); // ???
    bool newFace[DIMENSION + 1];
    bool newEdge[DIMENSION + 1][DIMENSION + 1];
    bool newVertex[DIMENSION + 1];
    for(int node0 = 0; node0 < DIMENSION + 1; node0++)
    {
        // is this node a new vertex?
        if (vertices.find(element.GetNode(node0)) != vertices.end())
            newVertex[node0] = false;
        else
        {
            newVertex[node0] = true;
            vertices.insert(element.GetNode(node0));
        }
        Face face;
        int n = 0;
        for(int node2 = 0; node2 < DIMENSION + 1; node2++)
            if(node2 != node0)
                face.nodes[n++] = element.GetNode(node2);
        // is this face a new face?
        set<Face>::iterator i = externalFaces.find(face);
        if(i != externalFaces.end()) // found
        {
            externalFaces.erase(face);
            newFace[node0] = false;
            for(int node1 = node0 + 1; node1 < DIMENSION + 1; node1++)
                newEdge[node0][node1] = false;
        }
        else
        {
            newFace[node0] = true;
            externalFaces.insert(face);
            for(int node1 = node0 + 1; node1 < DIMENSION + 1; node1++)
            {
                Edge edge;
                int n = 0;
                for(int node2 = 0; node2 < DIMENSION + 1; node2++)
                    if(node2 != node0 && node2 != node1)
                        edge.nodes[n++] = element.GetNode(node2);
                set<Edge>::iterator i = edges.find(edge);
                if(i != edges.end()) // found
                    newEdge[node0][node1] = false;
                else
                {
                    newEdge[node0][node1] = true;
                    edges.insert(edge);
                }
            }
        }
    }
    const int nAtomic = masters->GetNonLocals()->GetNumber();
    double rCut = nl->GetCutoffDistance();
    bool decorated = false;
    for(Element::latticePointsIterator it2 = element.begin();
                    it2 != element.end(); ++it2)
    {
        // find the real position of this slave candidate.
        double pos[DIMENSION];
        for(int coord = 0; coord < 3; coord++)
            pos[coord] = positions[element.GetNode(DIMENSION)][coord];
        for(int node = 0; node < DIMENSION; node++)
        {
            double R[DIMENSION];
            int coord;
            for(coord = 0; coord < DIMENSION; coord++)
                R[coord] = positions[element.GetNode(node)][coord] -
                    positions[element.GetNode(DIMENSION)][coord];
            bc->DifferenceBoundaryConditions(&R, 1);
            for(coord = 0; coord < 3; coord++)
                pos[coord] += it2->shape[node] * R[coord];
        }
        bc->EnforceBoundaryConditions(pos);
        int len = length;
        int numNeighbors;
        bool ok = nl->Neighbors(pos, neighbors, numNeighbors,
                                disp, dispSq, len, rCut);
//                                disp, dispSq, len, 2 * rCut);
        assert(ok);
        
        bool decorate = (numNeighbors != 0);
        if(decorate &&
           ((0 == it2->zeros) || // completely inside
            (1 == it2->zeros && newFace[it2->zero[0]]) || // on a new face
            (2 == it2->zeros && newEdge[it2->zero[0]][it2->zero[1]]) || // new edge
            (3 == it2->zeros && newVertex[it2->nonZero[0]] &&
                (element.GetNode(it2->nonZero[0]) >= nAtomic)))) // new vertex
        {
            for(int n = 0; n < numNeighbors; n++)
                slaveNeighbors[neighbors[n]].push_back(newIndices.size());
            newIndices.push_back(TinyVector<int>(it2->indices));
        }
        if(decorate)
            decorated = true;
    }
    if(decorated)
    {
        decoratedElementsVector.push_back(&element);
        numberOfSlaves.push_back(newIndices.size());
    }
}

void MeshData::InitializeSlaves(QCListOfAtoms* masters)
{
    ASSERT(lattices.size() == elements.size());

    if(masters->GetInitialized()) return;

    NeighborLocator* nl = masters->GetNonLocals()->GetNeighborLocator();
    nl->UpdatePosition();
    const BoundaryConditions* bc = masters->GetBoundaryConditions();

    while(PostProcessMesh(masters) != 0) ;  // get gid of slivers. 

    externalFaces.clear();
    edges.clear();
    vertices.clear();

    volume = 0;
    slaveNeighbors.clear();
    slaveNeighbors.resize(masters->GetNonLocals()->GetNumber());
    for(int e = 0; e < elements.size(); e++)
        DecorateElement(elements[e], masters);
    ListOfSlavedAtoms* slaves = masters->GetSlaves();
    slaves->SetShape(newIndices.size());
    slaves->SetElementsPointer(&elements[0]);
    int slave = 0;
    for(int i = 0; i < decoratedElementsVector.size(); i++)
    {
        int elementIndex = decoratedElementsVector[i] - &elements[0];
        for(; slave < numberOfSlaves[i]; slave++)
            slaves->SetElement(slave, elementIndex, newIndices[slave]);
    }
    slaves->AdjustPositions();
    masters->AdjustPositions();

//    QCNeighborList *qcNL = 
//        dynamic_cast<QCNeighborList *>(masters->GetAllAtoms()->GetNeighborLocator());
//    if(!qcNL) throw DMException("MeshData", "InitializeSlaves", "you must set a QCNerighborList on the master atoms");
//    qcNL->SetSlaveNeighborList(slaveNeighbors);
    masters->GetAllAtoms()->GetNeighborLocator()->UpdatePosition();
//    NeighborLocator* nl1 = new QCNeighborList(nl, this, masters,
//                                              slaveNeighbors);
//    masters->SetNeighborLocator(nl1);
//    masters->GetLocals()->SetNeighborLocator(
//        new QCNeighborList2(nl1, masters->GetNonLocals()->GetNumber())); // delete old one !!!
    
//     const double rCut =
//         masters->GetNonLocals()->GetNeighborLocator()->GetCutoffDistance();
//     masters->SetNeighborLocator(new CompositeNeighborList(0.0, masters));
//     cerr << rCut << endl;
//     masters->GetAllAtoms()->
//         SetNeighborLocator(new CompositeNeighborList(rCut,
//                                                      masters->GetAllAtoms()));
    cerr << elements.size() << "\telements\n";
    cerr << "volume of elements: " << volume << endl;
    cerr << elementsWeight << " local atoms" << endl;
    cerr << masters->GetNumber() << " masters atoms" << endl;
    cerr << masters->GetNonLocals()->GetNumber() << " non local masters" << endl;
    cerr << masters->GetLocals()->GetNumber() << " local masters" << endl;
    cerr << newIndices.size() << " slave atoms" << endl;
    cerr << "total number of atoms: " << elementsWeight +
        masters->GetNonLocals()->GetNumber() << endl;;
    newIndices.clear();
    decoratedElementsVector.clear();
    numberOfSlaves.clear();
    masters->SetInitialized(true);
}


int MeshData::PostProcessMesh(QCListOfAtoms* masters)
{    // Try to get rid of the slivers (a sliver is a zero volume element).
    BuildExternalFaces();

    int numberOfSliversDeleted = 0, numberOfSliversNotDeleted = 0;

    set<Face> extFaces = this->externalFaces; // make a copy.
    
    set<int> surfaceSlivers;

    while(!extFaces.empty())  // loop over all the external faces.
    {
        Face thisFace = *(extFaces.begin());
        ASSERT(thisFace.elemIds.size() == 1);
        const int elemId = *(thisFace.elemIds.begin());
        const Element &thisElem = elements[elemId];
        if(thisElem.GetDeterminant() == 0)  // found a sliver on the external surface
        {
            surfaceSlivers.insert(elemId); // schedule it for deletion
            for(int node1 = 0; node1 < 4; node1++)   // loop over its 4 faces
            {
                Face newFace;
                int n = 0;
                for(int node2 = 0; node2 < 4; node2++)
                    if(node2 != node1)
                        newFace.nodes[n++] = thisElem.GetNode(node2);
                if(newFace == thisFace) continue;
                vector<int> tmp = GetElementsSharingThisFace(newFace);
                ASSERT(tmp.size() == 1 || tmp.size() == 2);
                if(tmp.size() == 2)
                {
                    if(tmp[0] == elemId) newFace.elemIds.insert(tmp[1]);
                    else                 newFace.elemIds.insert(tmp[0]);
                    extFaces.insert(newFace);
                }
                else // means this face is also external
                {
                    extFaces.erase(newFace);
                }
            }
        }
        extFaces.erase(thisFace);
    }

    for(set<int>::reverse_iterator slivIt = surfaceSlivers.rbegin();
        slivIt != surfaceSlivers.rend(); ++slivIt)
    {
        elements.erase(elements.begin() + *slivIt);
        lattices.erase(lattices.begin() + *slivIt);
    }

    numberOfSliversDeleted = surfaceSlivers.size();

    cerr << numberOfSliversDeleted << " elements at surface deleted" << endl;

    set<int> elementsToRemove;
    vector<pair<vector<int>, int> > elementsToCreate;  // the nodes indices and the lattice index.

    // Now the strategy consists of trying to flip one of the edge of the slivers.
    const int numberOfElements = elements.size();

    for(int e1 = 0; e1 < numberOfElements; e1++)
        if(GetDeterminant(e1) == 0)
        {
            int neighborElements[4];  // the index of the 4 neighboring elements
            int oppositeNode[4]; // for each neighboring element, the node that the sliver doesn't have
            for(int node = 0; node < DIMENSION+1; ++node)
            {
                Face face;
                int n=0;
                for(int node2 = 0; node2 < DIMENSION+1; ++node2)
                    if(node2 != node) face.nodes[n++] = elements[e1].GetNode(node2);
                vector<int> tmp = GetElementsSharingThisFace(face);
                ASSERT(tmp.size() == 2);
                ASSERT(tmp[0] == e1 || tmp[1] == e1);
                int nbElem;
                if(tmp[0] == e1) nbElem = tmp[1];
                else             nbElem = tmp[0];
                ASSERT(elements[nbElem].GetDeterminant() != 0); // would be sliver on sliver.
                neighborElements[node] = nbElem;
                for(n = 0; n < 4; ++n)
                    if(find(face.nodes, face.nodes+3, elements[nbElem].GetNode(n)) == face.nodes+3)
                    {
                        oppositeNode[node] = elements[nbElem].GetNode(n);
                        break;
                    }
                ASSERT(n < 4); // assert we found the opposite node
            }

            // now check whether a pair of them have a common face (ie have to same opposite node)
            int nbElem1, nbElem2 = -1;
            for(int nbElem = 0; nbElem < 3; ++nbElem)
            {
                if(find(elementsToRemove.begin(), elementsToRemove.end(), neighborElements[nbElem])
                    != elementsToRemove.end()) continue; // this elem is scheduled to be removed.
                for(int otherEl=nbElem+1; otherEl <4; ++otherEl)
                {
                    if(find(elementsToRemove.begin(), elementsToRemove.end(), neighborElements[otherEl])
                       != elementsToRemove.end()) continue; // this elem is scheduled to be removed.
                    if(oppositeNode[nbElem] == oppositeNode[otherEl])
                    {
                        nbElem1 = nbElem;
                        nbElem2 = otherEl;
                        break;
                    }
                }
            }
            bool flip = (nbElem2 != -1);
            if(flip)
            {
                int nodes1[4];
                int nodes2[4];
                nodes1[0] = nodes2[0] = oppositeNode[nbElem1];
                // elements[e1].GetNode(i) is the sliver's node that doesn't belong to neighborElems[i]
                // this tells us two more nodes that the future new elements have in common.
                nodes1[1] = nodes2[1] = elements[e1].GetNode(nbElem1);
                nodes1[2] = nodes2[2] = elements[e1].GetNode(nbElem2);
                // determine the last nodes (one for each future new Element):
                int lastNode;
                int n=-1;
                do lastNode = elements[e1].GetNode(++n);
                while(lastNode == nodes1[1] || lastNode == nodes1[2]);
                nodes1[3] = lastNode;
                do lastNode = elements[e1].GetNode(++n);
                while(lastNode == nodes1[1] || lastNode == nodes1[2]);
                nodes2[3] = lastNode;
                sort(nodes1, nodes1+DIMENSION+1);
                sort(nodes2, nodes2+DIMENSION+1);
                // we cannot create the elements right now, because we cannot change 
                // the vector of elements (this would screw up calls to GetElementsSharingFace).
                elementsToCreate.push_back(make_pair(vector<int>(nodes1, nodes1+4), e1));
                elementsToCreate.push_back(make_pair(vector<int>(nodes2, nodes2+4), e1));
                elementsToRemove.insert(e1);
                elementsToRemove.insert(neighborElements[nbElem1]);
                elementsToRemove.insert(neighborElements[nbElem2]);
            }
            else
                numberOfSliversNotDeleted++;
        }
    numberOfSliversDeleted += elementsToRemove.size()/3;

    for(int newElem = 0; newElem < elementsToCreate.size(); ++newElem)
    {
        vector<int> &nodes = elementsToCreate[newElem].first;
        int lattIndex      = elementsToCreate[newElem].second;
        CreateElement(nodes, masters, lattices[lattIndex]);
    }
    set<int>::reverse_iterator elemRIter;
    for(elemRIter = elementsToRemove.rbegin();
        elemRIter != elementsToRemove.rend(); ++elemRIter)
    {
        elements.erase(elements.begin() + *elemRIter);
        lattices.erase(lattices.begin() + *elemRIter);
    }
    cerr << numberOfSliversDeleted << " slivers removed ( " << numberOfSliversNotDeleted << " remaining)" << endl;
    ASSERT(elements.size() == lattices.size());

    BuildExternalFaces();
    return numberOfSliversNotDeleted;
}


void MeshData::BuildExternalFaces()
{
    externalFaces.clear();
    vector<Element>::const_iterator elemIt;
    int elemId = 0;
    for(elemIt = elements.begin(); elemIt != elements.end(); elemIt++, elemId++)
    {
        for(int node0 = 0; node0 < DIMENSION + 1; node0++)
        {    // HERE WE ASSUME THAT THE NODES INDICES IN ELEMENTS ARE SORTED
            Face face;
            int n = 0;
            for(int node2 = 0; node2 < DIMENSION + 1; node2++)
                if(node2 != node0)
                    face.nodes[n++] = elemIt->GetNode(node2);
            set<Face>::iterator it = externalFaces.find(face);
            if(it != externalFaces.end()) // found
                externalFaces.erase(face);
            else
            {
                face.elemIds.insert(elemId);
                externalFaces.insert(face);
            }
        }
    }
}

vector<int> MeshData::GetElementsSharingThisFace(const Face &face) const
{
    vector<int> elemsHaveFace;
    vector<Element>::const_iterator elemIt;
    int elemId = 0;
    for(elemIt = elements.begin(); elemIt != elements.end(); elemIt++, elemId++)
    {
        int node;
        set<int> nodesInd;
        for(node = 0; node < DIMENSION + 1; node++)
            nodesInd.insert(elemIt->GetNode(node));
        bool good = true;
        for(node = 0; node < DIMENSION; ++node)
        {
            if (find(nodesInd.begin(), nodesInd.end(), face.nodes[node]) == nodesInd.end())
            {
                good = false;
                break;
            }
        }
        if(good) elemsHaveFace.push_back(elemId);
    }
    if (elemsHaveFace.size() > 2) 
        for(int i=0; i<elemsHaveFace.size(); ++i) cerr << elements[elemsHaveFace[i]] << endl;
    return elemsHaveFace;
}



void MeshData::Save(DMWriter *writer) const
{
    // First save the elements connectivity.
    const int nElements = elements.size();
    int elemId;
    int (*elemConnect)[DIMENSION+1] = new int[nElements][DIMENSION+1];
    for(elemId=0; elemId < nElements; ++elemId)
        for(int node=0; node<DIMENSION+1; ++node)
            elemConnect[elemId][node] = elements[elemId].GetNode(node);
    vector<int> dims(2, DIMENSION+1);
    dims[0] = nElements;
    writer->PutIntArray((int *)elemConnect, dims, "elemConnect");
    delete [] elemConnect;

    // Then store their nodesLocalCoord;
    int (*nodesLocalCoord)[DIMENSION][DIMENSION] =
        new int[nElements][DIMENSION][DIMENSION];
    for(elemId=0; elemId < nElements; ++elemId)
        for(int node=0; node<DIMENSION; ++node)
            memcpy(nodesLocalCoord[elemId][node], 
                elements[elemId].GetNodesCoord(node), DIMENSION*sizeof(int));
    dims[1] = DIMENSION;
    const int dimension = DIMENSION;
    dims.push_back(dimension);
    writer->PutIntArray((int *)nodesLocalCoord, dims, "nodesLocalCoord");
    delete [] nodesLocalCoord;

    // Store the total weight
//    writer->PutDouble(elementsWeight, "weight");
    writer->PutInt(nAtomic, "nAtomic");

    // Store the index of the Bravais lattice
    int *bravaisInd = new int[nElements];
    map<const BravaisLatticeWithBasis *, int> lat;
    int index=0;
    for(elemId=0; elemId<nElements; ++elemId)
    {
        const BravaisLatticeWithBasis *current = lattices[elemId];
        if(lat.count(current) == 0) lat[current] = index++;
        bravaisInd[elemId] = lat[current];
    }

    writer->PutInt(lat.size(), "numberOfLattices");
    map<const BravaisLatticeWithBasis *, int>::const_iterator lattIter;
    for(int latIndex=0; latIndex < lat.size(); ++latIndex)
    {
        SprintString name("lattice%i",latIndex);
        lattIter = lat.begin();
        while(lattIter->second != latIndex) lattIter++;
        writer->Put(lattIter->first, name);
    }

    dims.resize(1);
    writer->PutIntArray(bravaisInd, dims, "bravaisIndex");
    delete [] bravaisInd;

}

void MeshData::Load(DMReader *reader)
{
    // Get the data
    vector<int> dims;
    int (*elemConnect)[DIMENSION+1] = reinterpret_cast<int(*)[DIMENSION+1]>
                            (reader->GetIntArray(dims, "elemConnect"));
    if(dims.size() != 2 || dims[1] != DIMENSION+1)
        throw DMException("MeshData", "Load", "Corrupted elementConnectivities");

    int (*nodesLocalCoord)[DIMENSION][DIMENSION] = 
        reinterpret_cast<int (*)[DIMENSION][DIMENSION]>
            (reader->GetIntArray(dims, "nodesLocalCoord"));
    if(dims.size() != 3 || dims[1] != DIMENSION || dims[2] != DIMENSION)
        throw DMException("MeshData", "Load", "Corrupted localCoords");

//    this->elementsWeight = reader->GetDouble("weight");
    nAtomic = reader->GetInt("nAtomic");

    int *bravaisIndices = reader->GetIntArray(dims, "bravaisIndex");
    if(dims.size() != 1)
        throw DMException("MeshData", "Load", "Corrupted bravaisIndices");
    
    const int nLattices = reader->GetInt("numberOfLattices");
//    BravaisLattices
    vector<BravaisLatticeWithBasis*> bravaisLat;
    for(int latIndex=0; latIndex < nLattices; ++latIndex)
    {
        SprintString name("lattice%i",latIndex);
        bravaisLat.push_back((BravaisLatticeWithBasis *)(reader->Get(name)));
    }

    
    // Build the elements
    const int nElements = dims[0];
    int elemId;
    for(elemId=0; elemId < nElements; ++elemId)
    {
        TinyVector<int> tmp[DIMENSION];
        for(int node=0; node<DIMENSION; ++node)
            tmp[node] = nodesLocalCoord[elemId][node];
        Element newElem(elemConnect[elemId], nAtomic, tmp);
        const double weight = newElem.GetWeight();

        elementsWeight += weight;
        this->elements.push_back(newElem);
        lattices.push_back(bravaisLat[bravaisIndices[elemId]]);
    }

    delete [] bravaisIndices;
    delete [] nodesLocalCoord;
    delete [] elemConnect;
}

