// QCEMTPotential.cpp: implementation of the QCEMTPotential class.
#include "DigitalMaterial.h"
#include "QCEMTPotential.h"
#include "QCListOfAtoms.h"
#include "ListOfSlavedAtoms.h"
#include "DMArray.h"
#include "LeafToLeafIterator.h"
#include <assert.h>
#include <stdexcept>
#include "NeighborLocator.h"
#include "CellNeighborList.h"
#include "SphericalClusterInitializer.h"
#include "Element.h"
#include "MeshData.h"
#include "3DLattices.h"
#include "Timer.h"
#include "DMLastHeader.h"

#if DIMENSION == 3

#include "vectools.h"
#include "mass.h"

using namespace std;

class HalfSphereInitializer : public SphericalClusterInitializer
{
public:
    HalfSphereInitializer(double radius, BravaisLatticeWithBasis* lattice) :
        SphericalClusterInitializer(radius, lattice) {};
private:
    virtual bool Inside(double R[])
        {
            const double eps = 1e-9;
            return SphericalClusterInitializer::Inside(R) &&
                (R[2] > eps ||
                 (R[2] >= -eps && (R[1] > eps ||
                                   (R[1] >= -eps && R[0] > eps))));
        }
};

QCEMTPotential::QCEMTPotential(char *element) :
    EMTPotential(element), nMasters(0)
{
    PrimitiveFCCLattice lattice(1.0);
    ListOfAtoms atoms;
    double radius = sqrt(cut.shells[0] / 2.0 + 0.001);
    HalfSphereInitializer(radius, &lattice).Create(&atoms);
    nHalfNeighbors = atoms.GetNumber();
    ASSERT(nHalfNeighbors == 21);
    coef = new double[nHalfNeighbors][6];
    for (int n = 0; n < nHalfNeighbors; n++)
    {
        int indices[3];
        lattice.ClosestLatticeSiteIndices(indices,
                                          (*atoms.GetCartesianPositions())[n]);
        int k = 0;
        for (int i = 0; i < 3; i++)
        {
            for (int j = i; j < 3; j++)
            {
                coef[n][k] = double(indices[i] * indices[j]);
                if (j > i) coef[n][k] *= 2.;
                k++;
            }
        }
    }
}

QCEMTPotential::~QCEMTPotential()
{
    delete [] coef;
}

void QCEMTPotential::Allocate(ListOfAtoms *atoms, int nMasters)
{
    //cerr << "QCEMTPotential::Allocate" << endl;
    this->nMasters = nMasters;
    EMTPotential::Allocate(atoms);
    ASSERT(atoms->GetNumber() >= nHalfNeighbors);
    vdist = dTmp[0];
    wght = dTmp[1];
    dsigma1 = dTmp[2];
    dsigma2 = dTmp[3];
}

double QCEMTPotential::CalculateLocalEnergy(double sigma1, double sigma2)
{
    double tmp = sigma1 / (12.0 * emt->gamma1);
    double radiusMinusSeq = - log(tmp) / (BETA * emt->eta2);
    double ex1 = exp(-emt->kappa * radiusMinusSeq);
    tmp = emt->lambda * radiusMinusSeq;
    double Ec = emt->e0 * (tmp + 1.0) * exp(-tmp);
    double Eas = emt->V0 * (6.0 * ex1 - 0.5 * sigma2 / emt->gamma2);
    return Ec - emt->e0 + Eas;
}

void QCEMTPotential::CalculateLocalForce(const Element& element,
                                         double sigma1,
                                         double a[DIMENSION][DIMENSION],
                                         DMArray<double, DIMENSION>* forces,
                                         double stress[6])
{
    // vdist is still valid since the calculation of the sigmas
    vrec(vdist, vdist, &nHalfNeighbors);
    // wght  is still valid since the calculation of the sigmas
    vec_mul_scal_add_scal(wght, wght, cut.cutslope,
                          -cut.cutslope - emt->eta2, nHalfNeighbors); 
    vec_mul(dsigma1, dsigma1, wght, nHalfNeighbors);
    vec_mul(dsigma1, dsigma1, vdist, nHalfNeighbors);
    vec_mul_scal_add_scal(wght, wght, 1.0, emt->eta2 - emt->kappa / BETA,
                          nHalfNeighbors);
    vec_mul(dsigma2, dsigma2, wght, nHalfNeighbors);
    vec_mul(dsigma2, dsigma2, vdist, nHalfNeighbors);
    double dsigma1daa[6];
    double dsigma2daa[6];
    for (int k = 0; k < 6; k++)
    {
        double d1 = 0.0;
        double d2 = 0.0;
        for (int n = 0; n < nHalfNeighbors; n++)
        {
            d1 += coef[n][k] * dsigma1[n];
            d2 += coef[n][k] * dsigma2[n];
        }
        dsigma1daa[k] = d1;
        dsigma2daa[k] = d2;
    }
    double tmp = sigma1 / (12.0 * emt->gamma1);
    double radiusMinusSeq = - log(tmp) / (BETA * emt->eta2);
    double x = emt->lambda * radiusMinusSeq;
    double dEdsigma1 = (6.0 * emt->V0 * emt->kappa *
                        exp(-emt->kappa * radiusMinusSeq) +
                        emt->e0 * emt->lambda * x * exp(-x)) /
        (BETA * emt->eta2 * sigma1);
    double dEdsigma2 = -0.5 * emt->V0 / emt->gamma2;
    double weightOverDet = 1.0 / DIMENSION / (DIMENSION - 1);
    int m0 = element.GetNode(DIMENSION) - nAtomic;
    const DMArray<double, DIMENSION>& pos = *(masters->GetCartesianPositions());
    if (stress)
        for (int s = 0; s < 6; s++)
            stress[s] = 0.0;
    for (int node = 0; node < DIMENSION; node++)
    {
        int m = element.GetNode(node) - nAtomic;
        double R[DIMENSION];
        if (stress)
        {
            for (int coord = 0; coord < DIMENSION; coord++)
                R[coord] = pos[m + nAtomic][coord] - pos[m0 + nAtomic][coord];
            atoms->GetBoundaryConditions()->DifferenceBoundaryConditions(&R, 1);
        }
        int s = 0;
        for (int coord = 0; coord < 3; coord++)
        {
            double dsigma1dR = 0.0;
            double dsigma2dR = 0.0;
            int k = 0;
            for (int i = 0; i < 3; i++)
                for (int j = i; j < 3; j++)
                {
                    double daadR = element.GetCoef(node, i) * a[j][coord] +
                        element.GetCoef(node, j) * a[i][coord];
                    dsigma1dR += dsigma1daa[k] * daadR;
                    dsigma2dR += dsigma2daa[k] * daadR;
                    k++;
                }
            double f = -(dEdsigma1 * dsigma1dR + dEdsigma2 * dsigma2dR) *
                weightOverDet;
            if (stress)
                for (int coord2 = coord; coord2 < DIMENSION; coord2++)
                    stress[s++] -= f * R[coord2];
            else
            {
                if (m >= 0)  // means that this node is a local atom.
                    (*forces)[m][coord] += f;
                if (m0 >= 0)
                    (*forces)[m0][coord] -= f;
            }
        }
    }
    if (stress)
    {
        //double v = element.CalculateVolume(masters);
        for (int s = 0; s < 6; s++)
            //    stress[s] *= 2.0 / element.GetWeight();
            stress[s] *= 12.0 / element.GetDeterminant();
    }
}

void QCEMTPotential::CalculateLocalStuff(const Element& element,
                                         double latticeVectors[DIMENSION][DIMENSION],
                                         double dotProducts[6],
                                         double& sigma1, double& sigma2)
{
    element.CalculateLattice(masters, latticeVectors);
    int n, k = 0;
    for (int i = 0; i < 3; i++)
        for (int j = i; j < 3; j++)
        {
            double dot = 0.0;
            for (int coord = 0; coord < 3; coord++)
                dot += latticeVectors[i][coord] * latticeVectors[j][coord];
            dotProducts[k++] = dot;
        }
    // compute the real distance of the atoms in the lattice.
    for (n = 0; n < nHalfNeighbors; n++)
    {
        double r2 = 0.0;
        for (int k = 0; k < 6; k++)
            r2 += coef[n][k] * dotProducts[k];
        vdist[n] = r2;
    }
    // vdist = sqrt(vdist), distance between atoms. Will be used in LocalForces.
    vsqrt(vdist,vdist,&nHalfNeighbors);
    // dsigma1 = exp(-emeta2 * dist + emeta2BetaSeq)
    vec_mul_scal_add_scal(wght, vdist, -emt->eta2,
                          emt->eta2 * BETA * emt->seq, nHalfNeighbors);
    vexp(dsigma1, wght, &nHalfNeighbors);
    // dsigma2 = exp(-emKappaOverBeta * dist + emKappaSeq)
    vec_mul_scal_add_scal(wght, vdist, -emt->kappa/BETA,
                          emt->kappa * emt->seq, nHalfNeighbors);
    vexp(dsigma2, wght, &nHalfNeighbors);
    bool veryAccurate = true;
    if (veryAccurate)
    {
        // wght = exp(cutslope * dist - cutslopeCutdist)
        vec_mul_scal_add_scal(wght, vdist, cut.cutslope,
                              -cut.cutslope * cut.cutdist, nHalfNeighbors);
        vexp(wght, wght, &nHalfNeighbors);
        // wght = 1 / (wght + 1.0)
        vec_mul_scal_add_scal(wght, wght, 1.0, 1.0, nHalfNeighbors);
        vrec(wght, wght, &nHalfNeighbors);    // wght will be used in LocalForces.
        // dsigmai = wght * dsigmai
        vec_mul(dsigma1, wght, dsigma1, nHalfNeighbors);
        vec_mul(dsigma2, wght, dsigma2, nHalfNeighbors);
    }
    sigma1 = 0.0;
    sigma2 = 0.0;
    for (n = 0; n < nHalfNeighbors; n++)
    {
        sigma1 += dsigma1[n];
        sigma2 += dsigma2[n];
    }
    // since we considered half the neighbors
    sigma1 *= 2;
    sigma2 *= 2;
}

double QCEMTPotential::CalculateEnergy(ListOfAtoms *loa)
{
    masters = dynamic_cast<QCListOfAtoms*>(loa);
    if (masters == 0)
        throw DMException("QCEMTPotential", "CalculateEnergy", "not a QCListOfAtoms");
    nAtomic = masters->GetNonLocals()->GetNumber();
    ListOfAtoms* atoms = masters->GetAllAtoms(); // ????? EMTP::atoms ???
    ListOfSlavedAtoms *slaves = masters->GetSlaves();
    if (masters->GetInitialized() == false)
        meshData->InitializeSlaves(masters);
    if (nAtoms != atoms->GetNumber() || nMasters != masters->GetNumber())
        Allocate(atoms, masters->GetNumber());

    Timer t;
    // first calculate the sigmas and the energy of the non local atoms
//    CalculateSigmas(masters->GetNonLocals(), false);
    CalculateSigmas(masters->GetAllAtoms(), false);
    nAtoms = nAtomic;
//    EMTPotential::CalculateEnergyAfterSigmas(masters->GetNonLocals());
    EMTPotential::CalculateEnergyAfterSigmas(masters->GetAllAtoms());
    nAtoms = atoms->GetNumber();
    double energy = 0.0;
    for (int m = 0; m < nAtomic; m++)
        energy += (*Ec)[m] - emt->e0 + (*Eas)[m];
    Timer tmd;
    double a[DIMENSION][DIMENSION];
    double aa[6];
    elementEnergy = 0.0;
    for (int e = 0; e < meshData->GetNumberOfElements(); e++)
    {
        const Element& element = meshData->GetElement(e);
        double sigma1, sigma2;
        CalculateLocalStuff(element, a, aa, sigma1, sigma2);
        double weight = element.GetWeight();
        double en = CalculateLocalEnergy(sigma1, sigma2);
        elementEnergy += weight * en;
    }
    static bool firstTime = true;
    if (firstTime)
        cout << "Energy: " << 100.0 * tmd.Time() / t.Time() << '%' << endl;
    firstTime = false;
    return energy + elementEnergy;
}

void QCEMTPotential::CalculateStressesAndForces(ListOfAtoms* loa,
                                                DMArray<double, 3>* forces,
                                                DMArray<double, 6>* stresses)
{
    masters = dynamic_cast<QCListOfAtoms *>(loa);
    if (masters == 0)
        throw DMException("QCEMTPotential", "CalculateStressesAndForces", "not a QCListOfAtoms");
    nAtomic = masters->GetNonLocals()->GetNumber();
    ListOfAtoms* atoms = masters->GetAllAtoms();
    if (masters->GetInitialized() == false)
        meshData->InitializeSlaves(masters);
    if (nAtoms != atoms->GetNumber() || nMasters != masters->GetNumber())
        Allocate(atoms, masters->GetNumber());
    DMArray<double, 3>* nonLocalForces = forces->GetBranch(0);
    DMArray<double, 3>* localForces = forces->GetBranch(1);
    DMArray<double, 6>* nonLocalStresses = 0;
    DMArray<double, 6>* localStresses = 0;
    double* stress = 0;
    if (stresses)
    {
        nonLocalStresses = stresses->GetBranch(0);
        localStresses = stresses->GetBranch(1);
        localStresses->Zero();
        stress = new double[6];
    }
    Timer t;
    double a[DIMENSION][DIMENSION];
    double aa[6];
    localForces->Zero();
    vector<double> elementSigma1s(meshData->GetNumberOfElements());
    for (int e = 0; e < meshData->GetNumberOfElements(); e++)
    {
        const Element& element = meshData->GetElement(e);
        double sigma1, sigma2;
        CalculateLocalStuff(element, a, aa, sigma1, sigma2);
        elementSigma1s[e] = sigma1;
        CalculateLocalForce(element, sigma1, a, localForces, stress);
        if (stress)
        {
            double angles[DIMENSION + 1][DIMENSION + 1];
            throw DMException("QCEMTPotential", "CalculateForcesAndStresses", "function was commented out");
//            element.CalculateAngles(angles);
            for (int corner1 = 0; corner1 < DIMENSION + 1; corner1++)
            {
                int node = element.GetNode(corner1) - nAtomic;
                if (node >= 0)
                {
                    double weight = 0.0;
                    for (int corner2 = 0; corner2 < DIMENSION + 1; corner2++)
                        if (corner2 != corner1)
                            weight += angles[corner1][corner2];
                    weight = weight / 2.0 - 0.25;
                    for (int n = 0; n < DIMENSION * (DIMENSION - 1); n++)
                        (*localStresses)[node][n] += weight * stress[n];
                }
            }
        }
    }
    Timer tmd;
//    CalculateSigmas(masters->GetNonLocals(), true); // forcesOnly = true
    // with the hierachy of neighbor locators, we must pass all the atoms, instead of only the non locals.
    CalculateSigmas(masters->GetAllAtoms(), true); // forcesOnly = true
    nAtoms = atoms->GetNumber();
    DMArray<double>* slaveSigma1s = sigma1->GetBranch(1);
    masters->GetSlaves()->DistributeData(&elementSigma1s[0],
                                         &(*slaveSigma1s)[0]);
    CalculateForcesAfterSigmas(nonLocalForces, nonLocalStresses);
    static bool firstTime = true;
    if (firstTime)
    {
        cout << "Force: " << 100.0 * tmd.Time() / t.Time() << '%' << endl;
        firstTime = false;
    }
    if (stress)
        delete [] stress;
}

void QCEMTPotential::CalculateForces(ListOfAtoms *atoms,
                                     DMArray<double, DIMENSION> *forces)
{
    CalculateStressesAndForces(atoms, forces);
}

void QCEMTPotential::CalculateStressTensor(ListOfAtoms *atoms,
                                           DMArray<double, 6> *stresses)
{
    DMArray<double, 3> dummyForces(atoms->GetShape());
    CalculateStressesAndForces(atoms, &dummyForces, stresses);
}

double QCEMTPotential::CalculateEnergyAndForces(ListOfAtoms *atoms,
                                                DMArray<double, DIMENSION> *forces)
{
    throw std::runtime_error("not implemented!");
    return -1.0;
}

void QCEMTPotential::CalculateSigmas(ListOfAtoms *atoms, const bool forcesOnly)
{
    atoms->AdjustPositions(); // ??????????
    EMTPotential::CalculateSigmas(atoms, forcesOnly);
}

void QCEMTPotential::CalculateForcesAfterSigmas(DMArray<double, 3>* forces,
                                                DMArray<double, 6>* stresses)
{
    double rnb[ARRLIM][DIMENSION];
    double rnb2[ARRLIM];
    double cutslope = cut.cutslope;
    double cutslopecutdist = cut.cutslope * cut.cutdist;
    double emeta2 = emt[0].eta2;
    double emeta2betaseq = emt[0].eta2 * BETA * emt[0].seq;
    double eminv12gamma1 = 1.0 / (12.0 * emt->gamma1);
    double eminvbetaeta2 = 1.0 / (emt->eta2 * BETA);
    double emseq = emt->seq;
    double emkappa = emt->kappa;
    double emlambda = emt->lambda;
    double emlambdaseq = emt->lambda * emt->seq;
    double eme0lambdalambda = emt->e0 * emt->lambda * emt->lambda;
    double eme0lambdalambdaseq = emt->e0 * emt->lambda * emt->lambda * emt->seq;
    double emkappaoverbeta = emt[0].kappa / BETA;
    double em6V0kappa = 6.0 * emt->V0 * emt->kappa;
    double emkappaseq = emt[0].kappa * emt[0].seq;
    double emV0overgamma2 = emt[0].V0 / emt[0].gamma2;
    //    double df, dfx, dfy, dfz;
    double emneg_eta2 = -emeta2;
    double emneg_kob = -emkappaoverbeta;
    double vdist[ARRLIM], invr[ARRLIM]; //, sqdist[ARRLIM]
    double exarg[ARRLIM], vex1[ARRLIM], vex2[ARRLIM], vex3[ARRLIM];
    double invwght[ARRLIM], wght[ARRLIM], dwdr[ARRLIM];
    double dsg1dr[ARRLIM], dsg2dr[ARRLIM];
    // Allocate temporary arrays 
    double* tmp1 = dTmp[0];
    double* tmp2 = dTmp[1];
    double* ex1 = dTmp[2];
    double* ex2 = dTmp[3];
    ASSERT(tmp1 && tmp2 && ex1 && ex2);
    // Calculate gradients 
    // Prevent underflows
    // Prevent underflows 
    DMArrayIterator<double> sigIter(sigma1);
    for (sigIter.Init(); !sigIter.IsDone(); sigIter.Next())
    {
        if (sigIter.Current() < 1.0e-9)
            sigIter.Current() = 1.0e-9;
    }
    int is;

    LeafToLeafIterator<DMArray<double> > sigma1It(sigma1);
    LeafToLeafIterator<DMArray<double> > radiusIt(radius);
    LeafToLeafIterator<DMArray<double> > dE0dsIt(dE0ds);
    for (sigma1It.Init(), radiusIt.Init(), dE0dsIt.Init();
        !sigma1It.Done();
        sigma1It.Next(), radiusIt.Next(), dE0dsIt.Next())
    {
        const int size = sigma1It.Current()->Size();
        if (size == 0) continue;
        const double *pSig1 = &(*sigma1It.Current())[0];
        double *pRadi = &(*radiusIt.Current())[0];
        double *pdE0ds= &(*dE0dsIt.Current())[0];

        // radius = emseq - eminvbetaeta2 * log(sigma1[is] * eminv12gamma1); 
        vec_mul_scal_add_scal(tmp1, pSig1, eminv12gamma1, 0.0, size);
        vlog(tmp2, tmp1, &size);
        vec_mul_scal_add_scal(pRadi, tmp2, -eminvbetaeta2, emseq, size);
        // ex1 = exp(-emkappa * radius + emkappaseq); 
        // ex2 = exp(-emlambda * radius + emlambdaseq); 
        vec_mul_scal_add_scal(tmp1, pRadi, -emkappa, emkappaseq, size);
        vexp(ex1, tmp1, &size);  // tmp2 = ex1 
        vec_mul_scal_add_scal(tmp1, pRadi, -emlambda, emlambdaseq, size);
        vexp(ex2, tmp1, &size);  // tmp3 = ex2 
 
        // dEds = (- eme0lambdalambda * radius + eme0lambdalambdaseq) * ex2
        //  - em6V0kappa * ex1; 
        vec_mul_scal_add_scal(tmp1, pRadi, - eme0lambdalambda,
                              eme0lambdalambdaseq, size);
        vec_eq_self_mul_addscal_mul(tmp1, ex2, -em6V0kappa, ex1, size);
        // tmp1 = dEds
        // dEd0ds[is] =  - eminvbetaeta2 / sigma1[is] * dEds; 
        vrec(tmp2, pSig1, &size);
        vec_mul_mul_scal(pdE0ds, tmp1, tmp2, -eminvbetaeta2, size);

    }
    // Calculate forces 
    // Set forces to zero
    forces->Zero();
    if (stresses)
        stresses->Zero();
//    DMArray<double,DIMENSION>* forcesMaster = forces->GetBranch(0);
//    DMArray<double,DIMENSION>* forcesSlave = forces->GetBranch(1);
    int neighbors[ARRLIM];
    // Loop over all master-master pairs:
//    NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
    NeighborLocator *neighborLocator =
        masters->GetAllAtoms()->GetNeighborLocator(); // new hierarchy of neighbor locators.
//        masters->GetNonLocals()->GetNeighborLocator();
    int numnbr[ARRLIM];
//    for (is = 0; is < nMasters; )
    for (is = 0; is < nAtomic; )
    {
        
        
        int is1 = is;
        int slotsLeft = ARRLIM;
        bool accumulate = true;
        while (accumulate && (is < nAtomic) && (is-is1 < ARRLIM))
//        while (accumulate && (is < nMasters))
        {
            // convention for rnb: it gives position of j with respect to i
            accumulate = neighborLocator->
                HalfNeighbors(is, neighbors+(ARRLIM-slotsLeft), numnbr[is-is1],
                              rnb+(ARRLIM-slotsLeft), rnb2+(ARRLIM-slotsLeft),
                              slotsLeft);
            if (accumulate)
                is++;
        }
        int nbatch = ARRLIM - slotsLeft; // This is how many slots we have used
        //              vsqrt(vdist,sqdist,&nbatch);
        vsqrt(vdist,rnb2,&nbatch);
        // Calculate contributions to the sigmas
        // ex1 = exp(cutslope * dist - cutslopecutdist);
        vec_mul_scal_add_scal(exarg, vdist, cutslope, 
                              -cutslopecutdist, nbatch);
        vexp(vex1, exarg, &nbatch);
        // ex2 = exp(-emeta2 * dist + emeta2betaseq);
        vec_mul_scal_add_scal(exarg, vdist, emneg_eta2, emeta2betaseq, nbatch);
        vexp(vex2, exarg, &nbatch);
        // ex3 = exp(-emkappaoverbeta * dist + emkappaseq);
        vec_mul_scal_add_scal(exarg, vdist, emneg_kob, emkappaseq, nbatch);
        vexp(vex3, exarg, &nbatch);
        // wght = 1.0 / (1.0 + ex1);
        // dwdr = - cutslope * wght * (1 - wght);
        vec_mul_scal_add_scal(invwght, vex1, 1.0, 1.0, nbatch);
        vrec(wght, invwght, &nbatch);
        vec_diffweight(dwdr, wght, -cutslope, nbatch);
        // dsg1dr = (dwdr - wght * emeta2) * ex2;
        vec_mul_scal_add_mul(dsg1dr, wght, -emeta2, dwdr, vex2, nbatch);
        // dsg2dr = (dwdr - wght * emkappaoverbeta) * ex3;
        vec_mul_scal_add_mul(dsg2dr, wght, -emkappaoverbeta, 
                             dwdr, vex3, nbatch);
        // invr = 1.0 / vdist
        vrec(invr, vdist, &nbatch);
        // Distribute force contributions
        int numAtomsDone = is - is1;
        nbatch = 0;
        for ( ; is1 < is; is1++)
        {
//            bool atomic1 = (is1 < nAtomic);
            double *forces1 = (*forces)[is1];
            double* stress1;
            if (stresses)
                stress1 = (*stresses)[is1];
            for (int nb = 0; nb < numnbr[numAtomsDone-is+is1]; nb++)
            {
                int is2 = neighbors[nbatch];
//                 if (is2 >= nMasters)
//                 {
//                     assert(false);
//                     nbatch++;
//                     continue;
//                 }
                double df = ((*dE0ds)[is1] + (*dE0ds)[is2]) *
                    dsg1dr[nbatch] -
                    emV0overgamma2 * dsg2dr[nbatch];
                df *= invr[nbatch];
                double dfx = df * rnb[nbatch][0];
                double dfy = df * rnb[nbatch][1];
                double dfz = df * rnb[nbatch][2];
                forces1[0] += dfx;
                forces1[1] += dfy;
                forces1[2] += dfz;
                if (stresses)
                {
                    stress1[0] += dfx * rnb[nbatch][0];
                    stress1[1] += dfx * rnb[nbatch][1];
                    stress1[2] += dfx * rnb[nbatch][2];
                    stress1[3] += dfy * rnb[nbatch][1];
                    stress1[4] += dfy * rnb[nbatch][2];
                    stress1[5] += dfz * rnb[nbatch][2];
                }
                if (is2 < nAtomic)
                {
                    double *forces2 = (*forces)[is2];
                    forces2[0] -= dfx;
                    forces2[1] -= dfy;
                    forces2[2] -= dfz;
                    if (stresses)
                    {
                        double *stress2 = (*stresses)[is2];
                        stress2[0] += dfx * rnb[nbatch][0];
                        stress2[1] += dfx * rnb[nbatch][1];
                        stress2[2] += dfx * rnb[nbatch][2];
                        stress2[3] += dfy * rnb[nbatch][1];
                        stress2[4] += dfy * rnb[nbatch][2];
                        stress2[5] += dfz * rnb[nbatch][2];
                    }
                }
                nbatch++;
            }
        }
        assert(is == is1);
    }
}
#endif // DIMENSION == 3
