// QCEMTPotential.h: interface for the QCEMTPotential class.
//
//////////////////////////////////////////////////////////////////////

#ifndef QCEMTPotential_H
#define QCEMTPotential_H

#include "DigitalMaterial.h"
#include "EMTPotential.h"


//! Effective Medium Theory potential
/** This is a rewritten EMT potential that is designed to work in the QC frame.
    The energy is the weighted form: E_tot = \sum_{masters} W_m E_m. The force
    is defined as minus the gradient of the total energy w.r.t. the masters'
    positions
**/

#if DIMENSION == 3

class ListOfAtoms;
class QCListOfAtoms;
class ListOfSlavedAtoms;
class Element;
class MeshData;

class QCEMTPotential : public EMTPotential
{
public:
    QCEMTPotential(char *element);
    virtual ~QCEMTPotential() ;
    void SetMeshData(MeshData* md) {meshData = md;}
    const MeshData* GetMeshData() const {return meshData;}
    double CalculateEnergy(ListOfAtoms *atoms);
    double CalculateEnergyAndForces(ListOfAtoms *atoms,
				    DMArray<double, DIMENSION> *forces);
    double GetElementEnergy() const {return elementEnergy;}
#ifndef SWIG
    void CalculateStressesAndForces(ListOfAtoms *atoms,
                                    DMArray<double, 3> *forces,
                                    DMArray<double, 6> *stresses = 0);
    void CalculateForces(ListOfAtoms *atoms,
			 DMArray<double, DIMENSION> *forces);
    void CalculateStressTensor(ListOfAtoms *atoms,
                               DMArray<double, 6> *stresses);
private:
    void CalculateSigmas(ListOfAtoms *atoms, const bool forcesOnly=false);
    void CalculateForcesAfterSigmas(DMArray<double, DIMENSION> *forces,
                                    DMArray<double, 6>* stresses);
    void Allocate(ListOfAtoms *atoms, int nMasters);
    void CalculateLocalStuff(const Element&,
			     double a[DIMENSION][DIMENSION],
			     double aa[6],
			     double& sigma1, double& sigma2);
    double CalculateLocalEnergy(double sigma1, double sigma2);
    void CalculateLocalForce(const Element& element,
			     double sigma1,
			     double a[DIMENSION][DIMENSION],
			     DMArray<double, DIMENSION>* forces,
			     double stress[6] = 0);
    void AllocateForLocalElements();
    void DeAllocateForLocalElements();
    QCListOfAtoms* masters;
    int nMasters;
    int nHalfNeighbors;
    //! \a coef is used to calculate the real separation of atoms.
    /*! If the lattice coordinates of an atom are [n1, n2, n3], its
    real vector is n1 * v1 + n2 * v2 + n3 * v3, where the v's are
    the lattice vectors. So the distance  to the origin can be written as
    (n1*n1) * (v1*v1) + 2(n1*n2)*(v1*v2) + 2 (n1*n3)*(v1*v3) + (n2*n2)*(v2*v2) + etc...
    The dot products are calculated each time, since the strain of the lattice
    evolves, but the coefficients (n1*n1), ... are cached in \a coeff.
    We have 6 of these numbers per atom in the "half" lattice (which consists
    of 21 atoms). The 6 numbers are for the coeffs (11), (12), (13), (22), (23), (33).
    */
    double (*coef)[6];
    double* vdist;
    double* wght;
    double* dsigma1;
    double* dsigma2;
    int nAtomic;
    double elementEnergy;
    MeshData* meshData;
#endif // SWIG
};

#endif // DIMENSION == 3

#endif // !defined(QCEMTPotential_H)
