// QCPotential.cpp: implementation of the QCPotential class.
//
//////////////////////////////////////////////////////////////////////

#include "DigitalMaterial.h"

#include "QCPotential.h"

#include "BinaryWriter.h"
#include "BinaryReader.h"

#include "QCListOfAtoms.h"
#include "BravaisLatticeWithBasis.h"
#include "SphericalClusterInitializer.h"

#include "DMLastHeader.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

QCPotential::QCPotential(Potential *potential, BravaisLatticeWithBasis *lattice): potential(potential)
{
    // make a copy of the potential
    BinaryWriter out;
    out.Put(potential, "");
    BinaryReader in(out.GetBuffer());
    this->localPotential = dynamic_cast<Potential *>(in.Get(""));
    ASSERT(this->localPotential);

    this->atomicVolume = lattice->GetVolumeOfCell()/lattice->GetNAtomsPerCell();

    this->localLOA = new LocalListOfAtoms(lattice, potential->GetCutoffDistance());

}

QCPotential::~QCPotential()
{
    delete this->localPotential;
    delete this->localLOA;
}

double QCPotential::CalculateEnergy(ListOfAtoms *atoms)
{
    QCListOfAtoms *masters = dynamic_cast<QCListOfAtoms*>(atoms);
    if (0 == masters)
        throw DMException("QCPotential", "CalculateEnergy", "not a QCListOfAtoms");

    const int nAtomic = masters->GetNonLocals()->GetNumber();
    ASSERT (masters->GetInitialized());

    // here is the energy of the non local atoms. Thanks to the special neighbor locator 
    // we have attached on AllAtoms, this will correctly include the slave atoms.
    double nonLocalEnergy = 0.;
    for(int nonLocalAtom = 0; nonLocalAtom < nAtomic; ++nonLocalAtom)
        nonLocalEnergy += potential->CalculateAtomicEnergy(masters->GetAllAtoms(), nonLocalAtom);

    // now compute the energy of the elements.
    double lattice[DIMENSION][DIMENSION];
    double elementEnergy = 0.0;
    for (int e = 0; e < meshData->GetNumberOfElements(); e++)
    {
        const Element& element = meshData->GetElement(e);
        element.CalculateLattice(masters, lattice);
        localLOA->SetLattice(lattice);
        double weight = element.GetWeight();
        double en = localPotential->CalculateAtomicEnergy(localLOA, 0);
        elementEnergy += weight * en;
    }

    return nonLocalEnergy + elementEnergy;
}

void QCPotential::CalculateForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces)
{
    QCListOfAtoms *masters = dynamic_cast<QCListOfAtoms*>(atoms);
    if (0 == masters)
        throw DMException("QCPotential", "CalculateForces", "not a QCListOfAtoms");
    
    ListOfAtoms *allAtoms = masters->GetAllAtoms();
    // calculate the MD force on the non local atoms
    DMArray<double, DIMENSION> *extraForces = new DMArray<double, DIMENSION>(allAtoms->GetShape());
    potential->CalculateForces(allAtoms, extraForces);
    forces->GetBranch(0)->Copy(extraForces->GetBranch(0)->GetBranch(0));
    delete extraForces;

    forces->GetBranch(1)->Zero();
    const int nNonLocalAtoms = masters->GetNonLocals()->GetNumber();

    // compute the gradient of the energy of the elements.
    /* The method consists of determining the current lattice vectors starting 
    from the real position of the node. Then we compute the stess tensor in a 
    lattice with such lattice vectors. 
    The definition of the force on node alpha is the derivative of the element's
    energy w.r.t. node alpha's position (R_a)_j. (j is the coordinate index, and 
    we assume R_a is in fact R_a - R_DIMENSION) 
    We have \partial E/ \partial R_aj = Weight * atomicVolume * sigma_ik 
    * \partial \epsilon_ik / \partial R_aj.
    where \epsilon_ik is the deformation matrix, that can be written as
    (R1, R2, R3) (R10, R20, R30)^{-1}, where the index 0 refers to the reference 
    positions. Now (Ri0)j = N_ik v_kj with N_ik the nodes coord and v_kj the reference 
    lattice vectors. So (R10, R20, R30)^{-1} = (v)^{-1} N^{-1}. The inverse of N
    is given by the transpose of "coef" in the element class (divided by the determinant).
    so \partial \epsilon_ik / \partial R_aj = \delta_ij * (N^{-1})_al (v^{-1})_lk.

    Note that (N^{-1})_al (v^{-1})_lk can be obtained by the LatticeToCartesianCoordinates method.

    */
//    __asm int 3
    double lattice[DIMENSION][DIMENSION];
    double stress[DIMENSION*(DIMENSION+1)/2];
    for (int e = 0; e < meshData->GetNumberOfElements(); e++)
    {
        const Element& element = meshData->GetElement(e);
        element.CalculateLattice(masters, lattice);
        localLOA->SetLattice(lattice);

        double volume = element.GetWeight() * atomicVolume/element.GetDeterminant();
        localPotential->CalculateStressTensor(localLOA, 0, stress);
        const int n0 = element.GetNode(DIMENSION);
        for(int node = 0; node < DIMENSION; ++node)
        {
            const int n1 = element.GetNode(node);
            double tmp[DIMENSION];
            double nodesCoord[DIMENSION];
            for(int i=0; i<DIMENSION; ++i) nodesCoord[i] = element.GetCoef(node,i);
            meshData->GetLattice(e)->LatticeCoordinatesFromCartesianCoordinates(tmp, nodesCoord);

            for(int coord = 0; coord<DIMENSION; ++coord)
            {
                double f = 0.;
                for(int i = 0; i < DIMENSION; ++i)
                {
                    int comp = stresscomp[coord][i];
                    f += stress[comp] * tmp[i];
                }
                f *= volume;
                if(n1 >= nNonLocalAtoms) (*forces)[n1][coord] -= f;
                if(n0 >= nNonLocalAtoms) (*forces)[n0][coord] += f;
            }
        }
    }
    

}

////////////////////////////////////
// LocalNeighborLocator
////////////////////////////////////

QCPotential::LocalListOfAtoms::LocalListOfAtoms(BravaisLatticeWithBasis *lattice, 
                                                double cutoff,
                                                double maxStrain):
                                                cutoff(cutoff), 
                                                cutoff2(cutoff*cutoff),
                                                maxStrain(maxStrain)
{
    if(lattice->GetNAtomsPerCell() > 1)
        throw DMException("QCPotential::LocalListOfAtoms", "LocalListOfAtoms",
            "at this time, QC accepts only one atom per unit cell");
    // we want to have an atom at the origin:
    for(int coord=0; coord<DIMENSION; ++coord) 
        ASSERT(lattice->GetOrigin()[coord] == 0);

    double radiusExt = cutoff * (1. + maxStrain);
    double radiusInt = cutoff / (1. + maxStrain);
    SphericalClusterInitializer sphere(radiusExt, lattice);
    sphere.Create(this);
    const int nExt = this->GetNumber();
    sphere.SetRadius(radiusInt);
    sphere.Create(this);
    const int nInt = this->GetNumber();
    sphere.SetRadius(cutoff);
    sphere.Create(this);
    if(nExt != nInt)
        throw DMException("QCPotential::LocalListOfAtoms", "LocalListOfAtoms", 
            SprintString("such a large maximum strain could pose a problem since the number of atoms could vary between %i and %i",
            nInt, nExt));

    maxNeighbors       = this->GetNumber() - 1;
    neighbors          = new int[maxNeighbors];
    latticeCoordinates = new double[maxNeighbors][DIMENSION];
    int curr = 0;
    // we store the lattice coordinates of the atoms in displacements, because in SetLattice
    // we get the new lattice vectors.
    for(int at=0; at<maxNeighbors+1; ++at)
    {
        TinyVector<double> dist;
        lattice->LatticeCoordinatesFromCartesianCoordinates(dist, this->positions[at]);
        if(dist.norm2() < 1.0e-10) continue;
        neighbors[curr] = at;
        for(int coord=0; coord<DIMENSION; ++coord) 
            latticeCoordinates[curr][coord] = dist[coord];
        curr++;
    }
    ASSERT(curr == maxNeighbors);
    this->localNL = new LocalNeighborLocator(maxNeighbors);
    SetNeighborLocator(this->localNL);
}

QCPotential::LocalListOfAtoms::~LocalListOfAtoms()
{
    SetNeighborLocator(0);
    delete this->localNL;
    delete [] neighbors;
    delete [] latticeCoordinates;
}

void QCPotential::LocalListOfAtoms::SetLattice(const double latticeVectors[DIMENSION][DIMENSION])
{
    int numNeighbors = 0;
    for(int nb = 0; nb < maxNeighbors; ++nb)
    {
        double tmp[DIMENSION];
        for(int coord = 0; coord<DIMENSION; ++coord) tmp[coord] = 0.0;
        for(int i=0; i<DIMENSION; ++i)
        {
            for(int coord=0; coord<DIMENSION; ++coord)
                tmp[coord] += latticeCoordinates[nb][i] * latticeVectors[i][coord];
        }
        double norm2 = 0;
        for(int coord2=0; coord2<DIMENSION; ++coord2)
            norm2 += tmp[coord2]*tmp[coord2];
        if(norm2 < cutoff2) 
        {
            ASSERT(norm2 > 0.001) ;
            localNL->neighbors[numNeighbors] = neighbors[nb];
            for(int coord = 0; coord<DIMENSION; ++coord)
                localNL->displacements[numNeighbors][coord] = tmp[coord];
            localNL->displacementsSq[numNeighbors] = norm2;
            ++numNeighbors;
        }
    }
    localNL->numNeighbors = numNeighbors;
}


QCPotential::LocalNeighborLocator::LocalNeighborLocator(int maxNeighbors)
{
    this->neighbors       = new int[maxNeighbors];
    this->displacements   = new double[maxNeighbors][DIMENSION];
    this->displacementsSq = new double[maxNeighbors];
}

QCPotential::LocalNeighborLocator::~LocalNeighborLocator()
{
    delete [] neighbors;
    delete [] displacements;
    delete [] displacementsSq;
}

bool QCPotential::LocalNeighborLocator::Neighbors(int atomNumber, int *neighbors, int &numNeighbors, 
                                             double (*displacements)[DIMENSION], double *displacementsSquared,
                                             int &spacesLeftInArrays)
{
    if(0 != atomNumber) 
        throw DMException("QCPotential::LocalNeighborLocator", "Neighbors", "bad atom number");
    if(spacesLeftInArrays < this->numNeighbors) return false;
    numNeighbors = this->numNeighbors;
    memcpy(neighbors, this->neighbors, numNeighbors * sizeof(int));
    memcpy(displacements, this->displacements, numNeighbors*DIMENSION*sizeof(double));
    memcpy(displacementsSquared, this->displacementsSq, numNeighbors*sizeof(double));
    spacesLeftInArrays -= numNeighbors;
    return true;
}
