// QCPotential.h: interface for the QCPotential class.
//
//////////////////////////////////////////////////////////////////////

#ifndef QCPOTENTIAL_H
#define QCPOTENTIAL_H

#include "DMFirstHeader.h"

#include "DigitalMaterial.h"

#include "Potential.h"
#include "ListOfAtoms.h"
#include "NeighborLocator.h"
#include "BravaisLatticeWithBasis.h"
#include "MeshData.h"

//! This class computes forces and energies for the quasi-continuum method.
/*!
    It decorates any potential and computes the forces and energy of the 
    non local atoms (if a suitable neighbor locator was set on the masters
    so the slaves are correctly taken into account) and uses an internal 
    list of atoms to compute the energy of the elements (uniformly strained).
    To get the forces on the local atoms, it uses to CalculateStress function
    of the decorated potential.
    WARNING: the decorated potential must implement the 
    CalculateAtomicEnergy(ListOfAtoms *, int i)
    and CalculateStress(LOA *, int i, stress[]) functions and these functions
    should use only the neighbors of atoms i to compute energies and stresses.
    Thus we CANNOT USE EMTPOTENTIAL as it is implemented now !
*/
class QCPotential : public Potential  
{
public:
	QCPotential(Potential *potential, BravaisLatticeWithBasis *lattice);
	virtual ~QCPotential();
    double CalculateEnergy(ListOfAtoms *atoms);
    void CalculateForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);

    double GetCutoffDistance() const {return potential->GetCutoffDistance();}
    void SetMeshData(const MeshData *meshData){ this->meshData = meshData;}
private:
    //! The decorated potential
    Potential *potential;
    //! A copy of the decorated potential used for calculating energies of elements.
    Potential *localPotential;
    //! We keep the atomic volume for the force calculation
    double atomicVolume;
    //! The mesh that contains the elements
    const MeshData *meshData;
    //! An internally used list of atoms
    /* 
    We use this internal list of atoms in order to compute the energy and stress of
    a uniformly deformed crystal. We can efficiently use the potential itself (or
    a copy of it) if we provide a special neighbor locator.
    */
    class LocalNeighborLocator;
    class LocalListOfAtoms: public ListOfAtoms
    {
    public:
        LocalListOfAtoms(BravaisLatticeWithBasis *lattice, 
                         double cutoff,
                         double maxStrain = 0.05);
        ~LocalListOfAtoms();
        void SetLattice(const double lattice[DIMENSION][DIMENSION]);
    private:
        LocalNeighborLocator *localNL;
        BravaisLatticeWithBasis *lattice;
        const double cutoff;
        const double cutoff2;
        const double maxStrain;
        int maxNeighbors;
        int *neighbors;
        double (*latticeCoordinates)[DIMENSION];
    } *localLOA;

    //! The neighbor locator we attach to \a localLOA.
    class LocalNeighborLocator: public NeighborLocator
    {
    public:
        LocalNeighborLocator(int maxNeighbors);
        ~LocalNeighborLocator();
        bool Neighbors(int atomNumber, int * neighbors, int& numNeighbors,
                double (*displacements)[DIMENSION], double* displacementsSquared,
                int& spacesLeftInArrays);

	    bool HalfNeighbors(int atomNumber, int * neighbors, int& numNeighbors,
    	    		double (*displacements)[DIMENSION],	double* displacementsSquared,
			    	int& spacesLeftInArrays)
        {
            throw DMException("QCPotential::LocalNeighborLocator","HalfNeighbors",
                "this function cannot be called");
        }
        bool Neighbors(const double *center, int* neighbors, int& numNeighbors,
                     double (*displacements)[DIMENSION], double* displacementsSquared,
                     int& spacesLeftInArrays, double cutoff=-1.)
        {
            throw DMException("QCPotential::LocalNeighborLocator","Neighbors",
                "this function cannot be called");
        }
        void UpdatePosition() {}
        NeighborLocator *Copy() const {ASSERT(false); return 0;}
        int numNeighbors;
        int *neighbors;
        double (*displacements)[DIMENSION];
        //! The distances to the origin
        double *displacementsSq;
    };

};

#endif // !defined QCPOTENTIAL_H
