// Refiner.cpp: implementation of the Refiner class.
//
//////////////////////////////////////////////////////////////////////

#include "DMFirstHeader.h"
#include <iostream>
#include <math.h>
#include <algorithm>
#include "DigitalMaterial.h"
#include "Element.h"
#include "Refiner.h"
#include "MeshData.h"
#include "Face.h"
#include "PolyClusterInitializer.h"
#include "BravaisLatticeWithBasis.h"
#include "DMLastHeader.h"

using namespace std ;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
Refiner::Refiner(MeshData& meshData):meshData(meshData)
{
	
}

Refiner::~Refiner()
{
}

bool Refiner::Cluster::Master::operator==(const Master& master)const
{
    bool equal = (id == master.id);
    for(int coord = 0; coord < DIMENSION; ++coord)
	equal &= (index[coord] == master.index[coord]);
    return equal;
}

Refiner::Cluster::Cluster(Element *elem)
{
    elems.insert(elem);
    int numberOfNodes = elem->GetNumberOfNodes();
    Master *newMasters = new Master[numberOfNodes];
    for(int node=0; node<numberOfNodes;++node)
    {
	newMasters[node].id = elem->GetNode(node);
	memcpy(newMasters[node].index, elem->GetNodeRelativeLatticeIndices(node),
	       DIMENSION*sizeof(int));
    }
    masters.insert(newMasters, newMasters+numberOfNodes); 
}


void Refiner::Cluster::Merge(const Cluster &cluster)
{
    // Look in the other cluster for a common Master id.
    bool found=false;
    set<Master>::const_iterator myMaster, master = cluster.masters.begin(), masterIt;
    while(!found && master != cluster.masters.end())
    {
	myMaster=masters.begin();
	while(!found && myMaster!=masters.end())
	    if(myMaster->id == master->id) found = true;
	    else ++myMaster;
	if(!found) ++master;
    }
    ASSERT(found);

    int delta[DIMENSION];
    // Having found a common Master, this tells us how much we have to shift 
    // clusters'Masters indices.
    for(int coord=0; coord<DIMENSION; ++coord)
	delta[coord] = myMaster->index[coord] - master->index[coord];

// Now we can shift the cluster (we actually copy the shifted version in newMasters)
    set<Master> newMasters;
    for(masterIt=cluster.masters.begin(); masterIt != cluster.masters.end(); ++masterIt)
    {
	Master tmp;
	tmp.id = masterIt->id;
	for(int coord=0; coord<DIMENSION; ++coord)
	    tmp.index[coord] = masterIt->index[coord] + delta[coord];
	newMasters.insert(tmp);
    }
    // Compatibility test: 
    for(masterIt=newMasters.begin(); masterIt != newMasters.end(); ++masterIt)
    {
	found = false;
	myMaster=masters.begin();
	while (!found && myMaster!=masters.end())
	    if(myMaster->id == masterIt->id) found = true;
	    else ++myMaster;
	if(found && !(*myMaster == *masterIt))
	{
	    cerr << "Incompatibilty detected between " << myMaster->id << " and "<<masterIt->id << endl ;
	    for(int i=0; i<DIMENSION; ++i ) 
		cerr << myMaster->index[i] << " / " << masterIt->index[i] << endl;
	    throw DMException("MeshData::Cluster", "Merge", "Incompatibility detected");
	}
    }
    //! Finaly merge data (Microsoft doesn't like the condensed version of insert
    for(set<Element*>::const_iterator elem = cluster.elems.begin();
	elem!= cluster.elems.end(); ++elem) elems.insert(*elem );
    for(master = newMasters.begin(); master!=newMasters.end();++master)
	masters.insert(*master);
}

void Refiner::CreateClustersOfElementsToRefine()
{
    // We first clear all the clusters, if the vector isn't empty
    clusters.clear();
    const set<int> &indexOfFacesToRefine = meshData.GetIndexOfFacesToRefine();
    // Then for each face we want to refine, we create a cluster containing the elements 
    // sharing this face (WE MAY ALSO CONSIDER ADDING THE NEIGHBORING ELEMENTS, IN CASE 
    // THE CLUSTERS ARE TOO SMALL, ie if the face cannot be refined) 
    for(set<int>::const_iterator faceIter=indexOfFacesToRefine.begin();
	faceIter != indexOfFacesToRefine.end(); ++faceIter)
    {
	const Face& faceToRefine = meshData.GetFace(*faceIter);
	ASSERT(faceToRefine.elemIds.size() > 0);
	set<int>::const_iterator elemId=faceToRefine.elemIds.begin(); 
	clusters.push_back(Cluster(meshData.GetElement(*elemId)));
	while(++elemId != faceToRefine.elemIds.end()) 
	    clusters.back().Merge(Cluster(meshData.GetElement(*elemId)));
    }
    // For each existing cluster, check whether its elements are shared 
    // with other clusters. If yes, merge them.
    // The strategy consists of iterating over the clusters, looking whether a cluster 
    // with a higher index shares an Element and if yes, send the current cluster to the 
    // found one and erase the current. 
    list<Cluster>::iterator cluster = clusters.begin();
    while(cluster != clusters.end())
    {		
	bool toMerge = false;
	list<Cluster>::iterator tmpCluster = cluster;
	++tmpCluster;
	// Loop over the clusters with higher index...
	while(!toMerge && (tmpCluster != clusters.end()))
	{
	    // ... and check whether one of cluster's elements is in this cluster 
	    for(set<Element*>::iterator elemIter = cluster->elems.begin();
		elemIter != cluster->elems.end(); ++elemIter)
	    {
		set<Element*>::iterator found = 
		    find(tmpCluster->elems.begin(),tmpCluster->elems.end(), *elemIter);
		if(found != tmpCluster->elems.end()) toMerge = true;
	    }
	    if(!toMerge) ++tmpCluster;
	}
	// Merge them if we have to and delete this cluster 
	if(toMerge)
	{
	    tmpCluster->Merge(*cluster);
	    clusters.erase(cluster++);
	}
	else ++cluster;	
    }
    cout << "We have to refine " << clusters.size() << " clusters" << endl;
}



void Refiner::RefineCluster(Cluster &cluster)
{	
    if(cluster.elems.size() == 0) return;
    const vector<set<int> > extFaces = cluster.GetExternalFaces();
    vector<set<int> >::const_iterator face;
    set<Cluster::Master>::iterator master;
    cout << "Here is a cluster made of "<< cluster.elems.size() << " elements" << endl;
    cout << "Atoms in this cluster:" << endl;

    for(master=cluster.masters.begin(); master!=cluster.masters.end(); ++master)
    {
	cout << "atom: " <<master->id << " at pos";
	for(int coord=0; coord<DIMENSION; ++coord) cout<< " "<<master->index[coord];
	cout << endl;
    }
    cout << "Elements in this cluster: ";
    for(set<Element*>::iterator e=cluster.elems.begin(); e!=cluster.elems.end(); ++e) 
	cout <<  (*e)->GetElementNumber()<< " " ;
    cout << endl;
    cout << "External faces are given by the following nodes index:" << endl;
    for(face=extFaces.begin(); face!=extFaces.end(); ++face)
    {
	for(set<int>::const_iterator node=face->begin(); node!=face->end(); ++node) 
	    cout << *node << " " ;
	cout << endl;
    }
    // Instanciate a polycluster initializer that will help us make sure the new nodes
    // are inside the cluster: we need to set arrays for the corners and their connectivity.
    // We have to translate the indices in the external faces into the connectivity 
    // wrt the indices of the corners we pass to the polyCluster.
    DMArray<int, DIMENSION> extFaceConnectivity(extFaces.size());
    DMArrayIterator<int, DIMENSION> connIt(&extFaceConnectivity);
    face = extFaces.begin();
    for(connIt.Init(); !connIt.IsDone(); connIt.Next(), ++face)
    {
	ASSERT(face->size() == DIMENSION); // the number of points on the face
	set<int>::const_iterator node = face->begin();
	for(int coord=0; coord<DIMENSION; ++coord, ++node) 
	    connIt.Current()[coord] = *node;  //  note that this is not the right index (yet)
    }
    DMArray<double, DIMENSION> corners(cluster.masters.size());
    DMArrayIterator<double, DIMENSION> cornerIt(&corners);
    master = cluster.masters.begin();
    int currentCorner=0;
    for(cornerIt.Init(); !cornerIt.IsDone(); cornerIt.Next(), ++master,++currentCorner)
    {
	for(int coord=0; coord<DIMENSION; ++coord) 
	    cornerIt.Current()[coord] = (double) master->index[coord];
	int target = master->id;
	for(connIt.Init(); !connIt.IsDone(); connIt.Next())
	    for(int coord=0; coord<DIMENSION; ++coord) 
		if(connIt.Current()[coord] == target) connIt.Current()[coord]=currentCorner;
    }
	
// we temporarily need a lattice (that's not going to be used anyway)
	GeneralLattice lattice(1.0);
	PolyClusterInitializer polyCluster(&corners, &lattice, &extFaceConnectivity);

// Let's find an atomic position in the middle of the faces we want to refine
	const set<int> &indexOfFacesToRefine = meshData.GetIndexOfFacesToRefine();
	for(set<int>::const_iterator faceIndex=indexOfFacesToRefine.begin(); 
		faceIndex!=indexOfFacesToRefine.end(); ++faceIndex)
	{
		const Face &face = meshData.GetFace(*faceIndex);
// check if this face belongs to the cluster we are refining 
// (note that we could have cached this information)
		Element *firstElemInFace = meshData.GetElement(*(face.elemIds.begin()));
		if(find(cluster.elems.begin(), cluster.elems.end(), firstElemInFace)
		   != cluster.elems.end())
		{
			int newPoint[DIMENSION];
			memset(newPoint, 0, DIMENSION*sizeof(int));
			for(set<int>::const_iterator node=face.nodes.begin(); node!=face.nodes.end(); ++node)
			{
				set<Cluster::Master>::iterator master=cluster.masters.begin();
				while(master->id != *node) if(master++ == cluster.masters.end())
					throw DMException("MeshData","RefineCluster", 
									  "couldn't find a vertex of the face in the cluster's masters"); 
				for(int coord=0; coord<DIMENSION; ++coord) newPoint[coord] += master->index[coord];	
			}
// We look for the closest lattice site.
			double dNewPoint[DIMENSION];
			for(int coord=0; coord<DIMENSION; ++coord)
				dNewPoint[coord] = newPoint[coord]/(double)DIMENSION;
			if(FindClosestLatticeSiteInCluster(dNewPoint, polyCluster)) 
// this looks for the closest lattice site that's in the cluster.
			{
				// let's add a new Master in the cluster with id -1;
		Cluster::Master newMaster;
		newMaster.id = -1;
		for(int coord=0; coord<DIMENSION; ++coord) 
		    newMaster.index[coord] = (int) (dNewPoint[coord]+0.5);			
	    }
	}
    }

    /*	ConstrainedMesher *mesher = dynamic_cast<ConstrainedMesher*>(meshData.GetMesher());
	if(!mesher) throw DMException("Refiner", "RefineCluster", "No constrained mesher passed");

	mesher->SetPoints();
	mesher->SetConnectivity();

	mesher->GenerateMesh();
	mesher->GetElementsConnectivity();
    */
	
    //! Then we are going to delete the elements in the cluster

    //! ...

}

bool Refiner::FindClosestLatticeSiteInCluster(double point[DIMENSION],
					      PolyClusterInitializer &poly)
{
    double center[DIMENSION];
    int coord;
    for(coord=0; coord<DIMENSION; ++coord)
	center[coord] = floor(point[coord]+0.5);
    if(poly.StrictlyInside(center))
    {
	memcpy(point, center, DIMENSION*sizeof(double));
	return true; // ok
    }
    else// look around for a suitable lattice site
    {
	int epsilon[DIMENSION]; // epsilon is a counter: goes from -1 to 1 in each direction
	for(coord=0; coord<DIMENSION; ++coord) epsilon[coord] = -1;
	while(epsilon[DIMENSION-1] < 2)
	{
	    for(coord=0; coord<DIMENSION; ++coord) point[coord] = center[coord] + epsilon[coord];
	    if(poly.StrictlyInside(point)) return true;
	    epsilon[0]++;
	    for(coord=0; coord<DIMENSION-1;++coord) if(epsilon[coord] > 1)
	    {
		epsilon[coord+1]++;
		epsilon[coord] = -1;
	    }
	}
	return false;
    }
}

vector<set<int> > Refiner::Cluster::GetExternalFaces() const
{
    vector<set<int> > extFaces;
    for(set<Element*>::const_iterator elem=elems.begin(); 
	elem != elems.end(); ++elem)
    {
	vector<int> a;
	int i, size = (*elem)->GetNumberOfNodes();
	for(i=0; i<size; ++i) a.push_back((*elem)->GetNode(i));
	for(i=0; i<size; ++i)
	{
	    set<int> b;
	    for(int j=0; j<size; ++j) if(i!=j) b.insert(a[j]);
	    bool newFace = true;
	    vector<set<int> >::iterator faceIter = extFaces.begin();
	    while(newFace && (faceIter != extFaces.end()))
	    {
		if(b==*faceIter) newFace = false;
		else ++faceIter;
	    }
	    if(!newFace) extFaces.erase(faceIter);
	    else extFaces.push_back(b);
	}
    }
    return extFaces;
}

void Refiner::Refine()
{
    CreateClustersOfElementsToRefine();
    for(list<Cluster>::iterator cluster=clusters.begin(); cluster!=clusters.end();++cluster)
    {
	RefineCluster(*cluster);
    }
}
